; -*- Scheme -*-

	;;; counters

(define 14.1-counter-points 0)
(define 14.1-counter-faul 1)

	;;; table things...

(define (14.1-table-select id)
  (case id
    ((0) table/pockets/8ft)
    ((1) table/pockets/9ft)))

(define (14.1-rack table ball-list)
  (let* ((has-1    (memq 1 ball-list))
	 (has-5    (memq 5 ball-list))
	 (ball-num (length ball-list))
	 
	 (pos-list (append
		    '(1 2 3 4 5 6 7 8 9 11 12 13)
		    (case ball-num
		      ((14) '())
		      ((15) '(0))
		      (else (error "Bad ball-list")))
		    (if has-5 '() '(10))
		    (if has-1 '() '(14))))
	 (foot-spot (table-foot-spot table))
	 (mult	    (+ pocket-ball-radius (mm 0.1))))
	
    (define (place bid pyr-pos)
      (let ((pos (ball-in-pyramid pyr-pos)))
	(array-map! pos (lambda (x) (* x mult)) pos)
	(array-map! pos + pos foot-spot)
	(place-ball/rr bid pos)))

    (and has-1 (place 1 14))
    (and has-5 (place 5 10))

    (for-each place
       (list-remove-member (list-remove-member ball-list 5) 1)
       (unsort pos-list))))

	;;; 14.1 GAME state:
	  ;   0	table
	  ;   1 nof points to win a game
	  ;   2 whether to ask for ball/pocket
	  ;   3 list of pocketed balls [per one shot]
	  ;   4 list of balls on the table
	  ;   5 cue ball in the kitchen
	  ;   6 num of ball x cushion colls after first ball x ball col
	  ;   7 cue ball collided with any cusshion [openning break]
	  ;   8 cue ball collided with foot rail [playing from kitchen]

(define (14.1-gs-new table pts ask)
  (vector
	  ;   0	table
     table
	  ;   1 nof points to win a game
     pts
	  ;   2 whether to ask for ball/pocket
     ask
	  ;   3 list of pocketed balls [during one shot]
     '()
	  ;   4 list of balls on the table
     (make-interval 1 15)
	  ;   5 cue ball in the kitchen
     #t
	  ;   6 num of ball x cushion colls after first ball x ball col
     2
	  ;   7 cue ball collided with any cusshion [openning break]
     #f
	  ;   8 cue ball collided with foot rail [playing from kitchen]
     #f))


(define (14.1-gs-copy x)
  (let ((result (make-vector 9)))
    (array-copy! x result)
	  ;   3 list of pocketed balls [per one shot]
    (vector-set! result 3 (list-copy (vector-ref x 3)))
	  ;   4 list of balls on the table
    (vector-set! result 4 (list-copy (vector-ref x 4)))
    result))

(define (14.1-gs-max-points x)
  (vector-ref x 1))

(define (14.1-gs-ask? x)
  (vector-ref x 2))

(define (14.1-gs-pocketed-balls x)
  (vector-ref x 3))

(define (14.1-gs-pocketed-balls-reset! x)
  (vector-set! x 3 '()))

(define (14.1-gs-add-pocketed-ball! x bid)
  (vector-set! x 3 (cons bid (vector-ref x 3))))

(define (14.1-gs-balls x )
  (vector-ref x 4))

(define (14.1-gs-balls-remove-ball! x bid)
  (vector-set! x 4 (list-remove-member (vecto-ref x 4) bid)))

(define (14.1-gs-in-the-kitchen? x)
  (vector-ref x 5))

(define (14.1-gs-from-kitchen! x i)
  (vector-set! x 5 i)
  (vector-set! x 8 #f))

(define (14.1-gs-ball-cusshion-kisses x)
  (vector-ref x 6))

(define (14.1-gs-ball-cushion-kisses-reset! x)
  (vector-set! x 6 1))

(define (14.1-gs-ball-cusshion-kisses-dec! x)
  (vector-set! x 6 (-1+ (vector-ref x 6))))

(define (14.1-gs-cue-ball-cushion-kisses x)
  (vector-ref x 7))

(define (14.1-gs-cue-ball-cushion-kiss! x)
  (vector-set! x 7 #t))

(define (14.1-gs-cue-ball-foot-rail-kiss? x)
  (vector-ref x 8))

(define (14.1-gs-faul-reset! x pid)
  (vector-set! x (+ 9 pid) 0))

(define (14.1-add-faul! x pid faul)
  (cond
   (faul
    (game-add-point 14.1-counter-faul)
    (game-add-point 14.1-counter-points -1))
   (else
    (game-reset-score  pid 14.1-counter-faul)))

  (cond 
   ((<= (get-player-points pid 14.1-counter-faul))
    (game-add-point 14.1-counter-points -14)
    (game-reset-score pid 14.1-counter-faul)
    #t)
   (else
    #f)))

(define (14.1-game-reward state)
  (or (eq? state 'faul) (game-reward)))

(define (14.1-game-faul)
  (game-faul)
  'faul)

(pool-rules "14.1 continuous"
   'number-of-balls		16
   'number-of-players		2      

   'help-file			"game_14-1.html"

   'game-state-deep-copy	14.1-gs-copy
   'safe-to-undo		#t

   'score
     '((counter		"Points")
       (faul-counter	"Fauls"))
   
   'options
     '((group "14.1 options"
         (integer "Number of points to win a game" 150 1 999)
	 (yes/no  "Select ball and hole" #t))
       (group "Table size"
	 (one-of 1
		 "8ft table"
		 "9ft table")))

   'queries
     `((ball  "Called ball"
	      "Please designate the called ball"
	     ,(lambda (round player-id game-state)
		(14.1-gs-ask? game-state))
	     ,(lambda (round player-id game-state ball-id)
		(member ball-id (14.1-gs-balls game-state))))
       (pocket "Called pocket"
	       "Please designale the called pocket"
	       ,(lambda (round player-id game-state)
		  (14.1-gs-ask? game-state))
	       #t))

   'table-setup
     (lambda (points ask table-size-id)
       (8ball-skins #t)
       (let ((table (14.1-table-select table-size-id)))
	 (table-new table 'foot-spot 'head-spot 'kitchen)
	 (cushion-height (+ pocket-ball-radius (cm 0.2)))))

   'new-game
     (lambda (points ask table-size-id)
       (let ((table (14.1-table-select table-size-id)))

	 (game-state-set! (14.1-gs-new table points ask))

	 (place-ball 8ball-cue-ball (table-head-spot table))
	 (place-cue 0 #(-90 0))
	 (14.1-rack table (make-interval 1 15))
	 (ball-can-move 8ball-cue-ball
			(table-in-the-kitchen-pred table))))

   'next-shot
     (lambda (player-id game-state)
       #t)

   'end-of-shot
     (lambda (last-state game-state balls-out ball hole)
       #t)

   'event-handler
     (lambda (event-type info1 info2 state game-state called-ball called-hole)
       (DebugDisplayEvent event-type info1 info2 state game-state
			  "Ball on" ball "Hole" hole)
       (case event-type
	 ((cue-ball-collision)
	  #t)

	 ((ball-jump-off)
	  (game-faul)
	  (game-ball-out info1)
	  'faul)

	 ((ball-ball-collision)
	  #t)

	 ((ball-mantinel-collision)
	  #t)

	 ((ball-pocketed)
	  (cond
	   ((eq? info1 8ball-cue-ball)
	    (game-faul)
	    (game-ball-out info1)
	    'faul)
	   ((eq? info1 called-ball)
	    (cond
	     ((eq? info2 called-hole)
	      (14.1-add-pocketed-ball! game-state info1)
	      (14.1-game-reward state)
	      state)
	     (else
	      (14.1-game-faul))))
	   (else  ;;; additionally pocketted balls...
	    (14.1-add-pocketed-ball! game-state info1)
	    (14.1-game-reward state)
	    state)))


	  #t)))
