; -*- Scheme -*-

(define (8ball-setup-rack table)
  (let* ((stripes	(unsort 8ball-stripes-list))
	 (solids	(unsort 8ball-solids-list))

	 (one-stripe	(car stripes))
	 (one-solid	(car stripes))

	 (foot-spot	(table-foot-spot table)))

    (define (*r x)
      (* x (+ pocket-ball-radius (mm 0.01))))

    (define (place-ball-in-pyramid bid pos)
      (let ((pos	(ball-in-pyramid pos)))
	(array-map! pos *r pos)
	(array-map! pos + pos foot-spot)
	(place-ball/rr bid pos)))
    
    (place-ball-in-pyramid 8ball-8ball 4)
    (place-ball-in-pyramid one-solid 10)
    (place-ball-in-pyramid one-stripe 14)

    (for-each place-ball-in-pyramid
	(append (cdr stripes) (cdr solids))
	(unsort '(0 1 2 3 5 6 7 8 9 11 12 13)))))

(define (8ball-table-select id)
  (case id
    ((0) table/pockets/8ft)
    ((1) table/pockets/9ft)))

(pool-rules "8-ball"
     'number-of-balls		16
     'number-of-players		2

     'help-file			"game_8ball.html"


     ; 'game-state-deep-copy
     'safe-to-undo		#t

     'score			'((balls	""))
     'options			'((group "Table size"
				   (one-of 1 
					   "8ft table"
					   "9ft table")))
     'table-setup
       (lambda (table-size-id)
	 (8ball-skins 'all)
	 (table-new (8ball-table-select table-size-id) 
		    'center-spot 'foot-spot 'head-spot 'kitchen)
	 (cushion-height (+ pocket-ball-radius (mm 2))))

      'new-game
        (lambda (table-size-id)
	  (let ((table (8ball-table-select table-size-id)))

	    (8ball-setup-rack table)
	    (place-ball 0 (table-head-spot table))


	    (place-cue 0 #(-90 0))
	    (ball-can-move 9ball-cue-ball
			   (table-in-the-kitchen-pred table))))

      'end-of-shot 
        (lambda (last-state game-state balls-out)
	  #t)

      'next-shot
        (lambda (player-id game-state)
	  #t)

      'event-handler
        (lambda (event-type info1 info2 state game-state)
	  (DebugDisplayEvent event-type info1 info2 state game-state)))
