/* animation3 1999/11/27 16:10:104 1999/11/27 18:45:44
   $Id: animation.c,v 1.29 2000/02/29 15:23:32 rodo Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>
                      Daniel Skarda <0rfelyus@suse.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <gtk/gtk.h>

#include <stdio.h>
#include <string.h>

#include <math.h>
#include "mathutils.h"

#include "pool.h"
#include "game.h"
#include "gtulpas.h"
#include "poolview.h"
#include "poolfeedgl.h"
#include "fyjik.h"
#include "guileglue.h"

static GMemChunk	* event_mem_chunk	 = NULL;

static void anim_event_process (Animation *, AnimEvent *event);
static void pool_anim_skip_event_till_time (Animation *anim, int itime);
static void pool_animate_do (void);

Animation *
animation_new (gint nob, gint fps, gint init_size)
{
  Animation * anim;

  if (event_mem_chunk == NULL)
    event_mem_chunk = g_mem_chunk_create (AnimEvent, 128, G_ALLOC_AND_FREE); 

  anim = g_new (Animation, 1);

  anim -> game = gtulpas -> game;
  anim -> length = 1;
  anim -> fps    = fps;
  anim -> current_frame = 0;
  anim -> nof_balls     = nob;

  anim -> balls = g_new (PoolBallSpaceInfo, init_size * nob); 
  anim -> ball_array_size = init_size;

  anim -> finished = TRUE;

  anim -> events = NULL;
  anim -> last_event = NULL;

  anim -> ball_ball_num = anim -> game -> anim_event_ball_ball_num;
  anim -> ball_mantinel_num = anim -> game -> anim_event_ball_mantinel_num;

  anim -> ugly_ptr = NULL;

  return anim;
}

static void 
free_event_callback (AnimEvent *event, GMemChunk *memchunk)
{
  g_chunk_free (event, memchunk);
}

void
animation_destroy (Animation * anim)
{
  g_return_if_fail (anim);

  g_free (anim->balls);

  g_list_foreach (anim->events, (GFunc) free_event_callback, event_mem_chunk);
  g_list_free (anim->events);

  g_free (anim);
}

static int
find_event_callback (AnimEvent *event, guint *time)
{
  return event->time_since_bang <= *time;
}

AnimEvent*
animation_add_event (Animation *anim, guint time, AnimEventType type, gint info1, gint info2)
{
  AnimEvent	*event;

  event = g_chunk_new (AnimEvent, event_mem_chunk);

  event -> processed = 0;
  event -> time_since_bang = time;
  event -> type = type;

  event -> info1 = info1;
  event -> info2 = info2;

  glue_process_anim_event (anim, event);

  if (anim -> last_event == NULL)
    anim->last_event = anim->events = g_list_append (NULL, event);
  else
    {
      g_list_append (anim->last_event, event);
      anim->last_event = anim->last_event->next;
    }

  return event;
}

AnimEvent*
animation_insert_event (Animation *anim, guint time, AnimEventType type, gint info1, gint info2)
{
  AnimEvent	*event;
  GList		*found;
  
  event = g_chunk_new (AnimEvent, event_mem_chunk);

  event -> processed = 0;
  event -> time_since_bang = time;
  event -> type = type;

  event -> info1 = info1;
  event -> info2 = info2;

  found = g_list_find_custom (anim->events, &time, (GCompareFunc) find_event_callback);

  if (found == NULL)
    if (anim->last_event == NULL)
      anim->last_event = anim->events = g_list_append (NULL, event);
    else
      {
	g_list_append (anim->last_event, event);
	anim->last_event = anim->last_event->next;	
      }
  else
    if (found == anim->events)
      anim->events = g_list_prepend (found, event);
    else
      g_list_prepend (found, event);

  glue_process_anim_event (anim, event);

  return event;
}

/**** old ****/

GTimer		*pool_anim_timer;
GTimer		*pool_anim_spin_timer;
static GTimer	*pool_anim_blink_timer;
static GTimer   *pool_anim_gl_speed;

static int	 anim_timeout_id = -1;
static int	 anim_required_fps = -1;

static int	 anim_play = 0;
static int	 anim_is_replay = 0;
static int	 anim_end = 1;

static int	 anim_grab_id = 0;
static int	 anim_grab_interleave = 0;
static int	 anim_grab_nof_frames = 0;
static int	 anim_grab_semaphor = TRUE;
static dbl	 anim_not_stopped = 1.0;
static dbl	 anim_speed = 1.0;

int animation_finished_p (void)
{
  Animation *anim;

  anim = gtulpas->game->animation;

  if (anim_end == 1 && (anim -> current_frame == (dbl) (anim -> length - 1)))
    return TRUE;
  else
    return FALSE;
}

void
animation_undo (Animation *anim)
{
  anim -> balls = g_renew (PoolBallSpaceInfo, anim -> balls, anim -> nof_balls);
  anim -> current_frame = 0;
  anim -> length = 1;
  anim -> ball_array_size = anim -> nof_balls;

  g_list_foreach (anim->events, (GFunc) free_event_callback, event_mem_chunk);
  g_list_free (anim->events);
  anim -> events = NULL;
  anim -> cur_event = &(anim -> events);
  anim -> last_event = NULL;

  if (anim -> events)
    { g_list_free (anim -> events); anim -> events = NULL; }

  anim -> finished = TRUE;
  anim_end = 1;

  pool_view_redraw_mark_all ();

  ANIM_PANEL_SET_POS (0);
  ANIM_PANEL_SET_MAX (1);
}

void
animation_force_complete (void)
{
  Animation	*anim;

  anim = gtulpas -> game -> animation;

  while (anim->finished == FALSE)
    {
#ifdef DEBUG
      g_message ("Dopocitavam (forced)...");
#endif
      pool_fyjik (anim, 1000, 0.5 /* FIXME */);

      ANIM_PANEL_SET_MAX (anim->length - 1);
    }
}

void 
animation_skip_till_end (void)
{
  Animation	*anim;

  anim = gtulpas -> game -> animation;

  if (animation_finished_p ())
    return;

#ifdef DEBUG
  g_message ("Skip till end ...");
#endif

  animation_force_complete ();
  anim -> current_frame = (dbl) (anim -> length - 1);
}

void
animation_reset_fyjik (int frame)
{
  Animation	*anim;

  anim = gtulpas -> game -> animation;

  anim -> length = (frame == -1) 
    ? (int) ((floor) (anim -> current_frame))
    : frame;

  anim -> finished = 0;
}

static char animation_file_id[] = "PoolAnim";

void
animation_save (char *filename)
{
  Animation *anim;
  FILE	    *file;

  if ((file = fopen (filename, "wb")) == NULL)
    return;

  anim = gtulpas -> game -> animation;

  fwrite (animation_file_id, 1, sizeof (animation_file_id), file);
  fwrite (anim, 1, sizeof (Animation), file);
  fwrite (anim -> balls, 1, sizeof (PoolBallSpaceInfo) * 
	  anim -> length * anim -> nof_balls, file);
  fclose (file);

  gdk_beep ();
}

void animation_load (char *filename)
{
  PoolBallSpaceInfo *ball;
  PoolBall	    *bi;
  Animation aaa;
  Animation *anim;
  FILE	    *file;
  char	     tmp [16];
  int	     ck, i,j;

  if ((file = fopen (filename, "rb")) == NULL)
    return;

  fread (tmp, 1, sizeof (animation_file_id), file);
  ck = fread (&aaa, 1, sizeof (Animation), file);

  if ((ck != sizeof (Animation)) ||
       strncmp (animation_file_id, tmp, sizeof (animation_file_id)) ||
      (gtulpas -> game -> balls_nof != aaa.nof_balls))
    { fclose (file); return; }

  anim = animation_new (aaa.nof_balls, aaa.fps, aaa.length);
  i = sizeof (PoolBallSpaceInfo) * aaa.length * aaa.nof_balls;
  ck = fread (anim -> balls, 1, i, file);
  fclose (file);

  if (i != ck)
    { animation_destroy (anim); return; }

  animation_destroy (gtulpas -> game -> animation);
  gtulpas -> game -> animation = anim;

  ball = anim -> balls;
  *anim = aaa;
  anim -> balls = ball;

  anim -> game = gtulpas-> game;
  anim -> events = NULL;
  anim -> cur_event = &(anim -> events);
  anim -> last_event = NULL;
  anim -> ugly_ptr = NULL;
  anim -> current_frame = aaa.current_frame;

  bi = gtulpas -> game -> balls;
  ball = anim -> balls;
  for (i = 0; i < anim -> length; i++)
    for (j = 0; j < anim -> nof_balls; j++, ball++)
      ball -> ball = bi + j;

  ANIM_PANEL_SET_MAX (anim -> length - 1);
  ANIM_PANEL_SET_POS (anim -> current_frame);

  animation_stop ();
  gdk_beep ();
}

void animation_play (void)
{
  Animation *anim;

  anim_not_stopped = 1.0;

  if (anim_end == 1)
    animation_set_position (gtulpas -> game -> animation -> current_frame);
}

void animation_stop (void)
{
  anim_not_stopped = 0.0;
}

void animation_set_speed (dbl newspeed)
{
  anim_speed = newspeed;
}

void anim_grab_start_grab (int nof_frames, int interleave)
{
  anim_grab_id = 1;
  anim_grab_nof_frames = nof_frames;
  anim_grab_interleave = interleave ? interleave : 1;
  anim_grab_semaphor = TRUE;
  gtulpas -> game ->animation -> current_frame -= interleave;
  animation_play ();
}

void animation_set_position (dbl newposition)
{
  Animation	*anim;
  dbl		 ns;

  anim = gtulpas -> game -> animation;

  if (newposition < 0.0)
    newposition = 0.0;

  if (newposition > (dbl) (anim -> length - 1))
    newposition = (anim -> length - 1);

  ns = (anim_end) ? 0 : anim_not_stopped;
  pool_replay_current_animation ();
  anim_not_stopped = ns;

  anim -> current_frame = newposition;
  ANIM_PANEL_SET_POS (newposition);

  pool_anim_skip_event_till_time (anim, (int) (newposition * 1000.0));
  pool_view_redraw_mark_all ();
}

void animation_kill (void)
{
  anim_play = 0;
  anim_end = 1;
}

static void
pool_mark_to_redraw (PoolViewGL	*view)
{
  view -> needs_redraw = 1;
}

void pool_view_redraw_mark_all (void)
{
  g_list_foreach (pool_view_list, 
		  (GFunc) pool_mark_to_redraw,
		  NULL);
}

static void
animation_mark_to_blink (PoolViewGL *view, gpointer *blink)
{
  view_blink_start (view, (ViewBlinkType) blink);
}

static void 
animation_blink_all (ViewBlinkType blink)
{
  g_list_foreach (pool_view_list,
		  (GFunc)  animation_mark_to_blink, 
		  (gpointer) blink);
}

void pool_play_current_animation (void)
{
  anim_play = 1;
  anim_is_replay = 0;
  anim_not_stopped = 1.0;
}

void pool_replay_current_animation (void)
{
  Animation		*anim;

  anim = gtulpas->game->animation;

  if (anim_end == 0)     /* already playing ? */
    anim_play = 1;       /* anim_is_replay flag is retained */
  else
    {
      anim_play = 1;
      anim_is_replay = 1;
    }

  anim_not_stopped = 1.0;
  anim -> current_frame = 0.0;
  ANIM_PANEL_SET_POS (0.0);
}

static void
pool_anim_process_event (Animation *anim, AnimEvent *event)
{
  switch (event->type)
    {
    case S_Cue_Ball_Collision:
    case S_Ball_Ball_Collision:
    case S_Ball_Pocketed:
    case S_Ball_Jump_Off:
    case S_Ball_Mantinel_Collision:
      /* FIXME: launch sound events ... */
      break;

    case S_End_Of_Shoot:
      if (! event->processed)
	game_add_event_type (anim->game, SHOT_END);
      event->processed = TRUE;
      break;

    case S_Game_Point:
      if (! event->processed)
	game_add_event_type_data (anim->game, POINT, (gpointer) event_score_new (event->info1, event->info2));
      event->processed = TRUE;
      break;

    case S_Game_Penalty:
      if (! event->processed)
	game_add_event_type_data (anim->game, PENALTY, (gpointer) event_score_new (event->info1, event->info2));
      event->processed = TRUE;
      break;

    case S_Game_Pocketed:
      if (! event->processed)
	game_add_event_type_data (anim->game, POCKETED, (gpointer) event_score_new (event->info1, event->info2));
      event->processed = TRUE;
      break;

    case S_Game_Reward:
      animation_blink_all (Blink_Good);
      event -> processed = TRUE;
      break;

    case S_Game_Faul:
      if (! event->processed)
	game_add_event_type (anim->game, FAUL);
      animation_blink_all (Blink_Bad);
      event->processed = TRUE;
      break;
                            
    }
}

static void
pool_anim_skip_event_till_time (Animation *anim, int itime)
{
  GList	       **levent;
  AnimEvent	*event;

  levent = & (anim -> events);
  while ((*levent) != NULL)
    {
      event = (AnimEvent *) ((*levent) -> data);
      if ((itime != -1) && (event->time_since_bang > (guint) itime))
	break;

      levent = & ((*levent) -> next);
    }

  anim -> cur_event = levent;
}

static void
pool_anim_process_event_till_time (Animation *anim, int itime)
{
  GList	       **levent;
  AnimEvent	*event;

  levent = anim -> cur_event;
  while ((*levent) != NULL)
    {
      event = (AnimEvent *) ((*levent) -> data);
      if ((itime != -1) && (event->time_since_bang > (guint) itime))
	break;

      pool_anim_process_event (anim, event);
      levent = & ((*levent) -> next);
    }

  anim -> cur_event = levent;
}

static void
pool_animate_do (void)
{
  Animation		*anim;

  anim = gtulpas->game->animation;

  if (anim_play == 1)
    {
      anim_play = 0;
      anim_end = 0;

      anim -> cur_event = &(anim -> events);
      g_timer_start (pool_anim_timer);
    }

  if (anim_end != 1)
    {
      dbl		te;
      int		ite;

      dbl		frame;

      te = g_timer_elapsed (pool_anim_timer, NULL);
      g_timer_reset (pool_anim_timer);

#ifdef DEBUG
      if  (anim_grab_nof_frames > 0)
	{
	  frame = anim -> current_frame + anim_grab_interleave * anim_grab_semaphor;
	  anim_grab_semaphor = FALSE;
	}
      else
#endif
      frame =  (anim->current_frame) + te * (dbl) (anim->fps) 
#ifdef DEBUG
	* anim_speed * anim_not_stopped
#endif
	;
      
      if (frame > (dbl) (anim->length - 1))
	{
	  if (anim->finished == FALSE)
	    {
	      int   to_compute;

	      to_compute = ((gint) ceil (frame)) - (anim->length - 1);
	      to_compute += ANIMATION_DEFAULT_STEP - to_compute % ANIMATION_DEFAULT_STEP;

#ifdef DEBUG
	      g_message ("Dopocitavam...%d", to_compute);
#endif
	      pool_fyjik (anim, to_compute, 0.5 /* FIXME */);
	      ANIM_PANEL_SET_MAX (anim->length - 1);
	    }

	  /* FIXME: should call some function to update game events... */

	  if (frame > (dbl) (anim->length - 1)) 
	    {
	      frame = anim->length - 1;
	      anim_end = 1;
	      anim_grab_nof_frames = 0;
	      
	      anim -> current_frame = frame;
	      pool_anim_process_event_till_time (anim, -1);
	      anim -> current_frame = frame - 1;
	    }
	}

      ite = (int) (frame * 1000.0 / (dbl) (anim->fps));
      pool_anim_process_event_till_time (anim, ite);

      if (frame != anim -> current_frame)
	{
	  anim -> current_frame = frame;
	  ANIM_PANEL_SET_POS (frame);
	  pool_view_redraw_mark_all ();
	}
    }
}

static PoolViewGL*
pool_find_next_to_redraw (void)
{
  GList		*lptr;
  PoolViewGL	*pv;

  for (lptr = pool_view_list; lptr; lptr = lptr->next)
    {
      pv = (PoolViewGL *) lptr -> data;

      if ((pv -> needs_redraw == 1) && (pv -> redrawed == 0))
	return pv;
    }

  return NULL;
}

gint
poolanim_timeout_callback (void)
{
  static int	 lock = 1;
  
  if (lock)
    {
      GList		*lptr;
      PoolViewGL	*pv;

      lock = 0;

      g_list_foreach (pool_view_list, (GFunc) view_blink_process, NULL);
      pool_views_mover ();
      pool_animate_do ();

      while ((pv = pool_find_next_to_redraw ()) != NULL)
	{
	  pv -> needs_redraw = 0;
	  pv -> redrawed = 1;

	  if (pv -> show_fps)
	    {
	      g_timer_reset (pool_anim_gl_speed);
	      pool_view_do_redraw (pv);
	      pool_view_GL_finish (pv);
	      pool_view_draw_fps (pv, 
		  g_timer_elapsed (pool_anim_gl_speed, NULL));
	    }
	  else
	    pool_view_do_redraw (pv);

	  while (gtk_events_pending ()) 
	    gtk_main_iteration ();
	}

      if (anim_grab_semaphor == FALSE)
	{
	  anim_grab_do (anim_grab_id++);
	  if (--anim_grab_nof_frames)
	    anim_grab_semaphor = TRUE;
	}

      for (lptr = pool_view_list; lptr ; lptr = lptr->next)
	((PoolViewGL *)(lptr->data)) -> redrawed = 0;
      
      lock = 1;
    }

  return TRUE;
}

void
poolanim_fps_set (int newfps)
{
  if (newfps != anim_required_fps)
    {
      if (anim_timeout_id != -1)
	gtk_timeout_remove (anim_timeout_id);

      anim_required_fps = newfps;
      anim_timeout_id = 
	gtk_timeout_add (1000 / newfps, (GtkFunction) poolanim_timeout_callback, NULL);
    }
}

void
poolanim_setup ()
{
  poolanim_fps_set (POOL_DEFAULT_REQUIRED_FPS);
  
  pool_anim_timer = g_timer_new ();
  pool_anim_spin_timer = g_timer_new ();
  pool_anim_blink_timer = g_timer_new ();
  pool_anim_gl_speed = g_timer_new ();

  g_timer_start (pool_anim_spin_timer);
}

void
poolanim_shutdown ()
{
  if (anim_timeout_id != -1)
    gtk_timeout_remove (anim_timeout_id);

  anim_timeout_id = -1;

  g_timer_destroy (pool_anim_timer);
  g_timer_destroy (pool_anim_spin_timer);
  g_timer_destroy (pool_anim_blink_timer);
  g_timer_destroy (pool_anim_gl_speed);
}
