/* animation3 1999/11/27 16:10:114 1999/11/27 18:45:447 1999/12/04 13:34:23
   $Id: animation.h,v 1.21 2000/02/26 21:15:18 0rfelyus Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>
                      Daniel Skarda <0rfelyus@suse.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef __ANIMATION_H__
#define __ANIMATION_H__

#include "pool.h"
#include "mathutils.h"
#include <glibconfig.h>

#define ANIMATION_LAST_FRAME(A)	((A)->balls + ((A)->length - 1) * ((A)->nof_balls))

#define ANIMATION_DEFAULT_STEP	         5

struct _Game;

typedef struct _Animation      Animation;
typedef struct _AnimEvent      AnimEvent;
typedef enum   _AnimEventType  AnimEventType;

enum _AnimEventType
{
  S_Cue_Ball_Collision      = 1 <<  0,
  S_Ball_Ball_Collision     = 1 <<  1,
  S_Ball_Pocketed           = 1 <<  2,
  S_Ball_Jump_Off	    = 1 <<  3,
  S_Ball_Mantinel_Collision = 1 <<  4,
  S_End_Of_Shoot	    = 1 <<  5,
				    
  S_Game_Point		    = 1 <<  7,
  S_Game_Penalty	    = 1 <<  8,
  S_Game_Pocketed	    = 1 <<  9,
  S_Game_Faul		    = 1 << 10,
  S_Game_Reward		    = 1 << 11
};

struct  _AnimEvent
{
  guint			 time_since_bang; /* ms */
  
  AnimEventType		 type;
  
  int			 info1;		/* usually - an active  ball */
  int			 info2;		/* secondary info - pocket, second ball (collision) */

  guint			 processed : 1; /* some type of events should not be processed more than once ... */
};

struct _Animation
{
  struct _Game	*game;

  gint		length;
  gint		fps;			/* precision of generated animation */
  gint		nof_balls;
  gint		ball_array_size;	/* actual allocated balls */

  PoolBallSpaceInfo * balls;

  /* events */

  GList	      * events;			/* list of events */
  GList	     ** cur_event;		/* event to be processed next */
  GList	      * last_event;		/* pointer to last event */

  dbl		current_frame;		/* filled by pool_animate_do */

  /* FIXME: move to Game (when Game is splitted to Game & Rules */

  gint		ball_ball_num;		/* number of ball-ball events to process by guile */
  gint		ball_mantinel_num;	/* number of ball-mantinel events to process by guile */

  /* various ugly things to speed-up for guile bindings */
  
  void		*ugly_ptr;

  /* `finished' determines wheter computation of animation was finished */ 

  guint		finished : 1;
};

Animation * animation_new (gint nob, gint fps, gint init_size);
void        animation_destroy (Animation * anim);
void	    animation_kill (void);
void	    animation_undo (Animation *anim);

	/* returns true if no animation is running */
int	    animation_finished_p (void);


	/* appends new event to animation's event list */
	/* presumes that time is greater than time of already added events */
AnimEvent* animation_add_event (Animation *anim, guint time, AnimEventType event, gint info1, gint info2);

	/* inserts event in animation's event list */

AnimEvent* animation_insert_event (Animation *anim, guint time, AnimEventType event, gint info1, gint info2);

	/* animation debugging facility ... */

void	   animation_set_position (dbl);
void	   animation_set_speed (dbl);
void	   animation_play (void);
void	   animation_stop (void);
void	   animation_force_complete (void);
void	   animation_skip_till_end (void);
void	   animation_reset_fyjik (int frame);

void	   anim_panel_show (void);
void	   anim_panel_set_position (dbl);
void	   anim_panel_position_max (dbl);

void	   animation_save (char *);
void	   animation_load (char *);

void	   anim_grab_panel_show (void);
void	   anim_grab_do (int frame);
void	   anim_grab_start_grab (int nof_frames, int interleave);

#ifdef DEBUG

# define ANIM_PANEL_SET_MAX(MAX)		anim_panel_position_max (MAX)
# define ANIM_PANEL_SET_POS(POS)		anim_panel_set_position(POS)

#else	/* DEBUG */

# define ANIM_PANEL_SET_MAX(MAX)		/* NOP */
# define ANIM_PANEL_SET_POS(POS)		/* NOP */

#endif  /* DEBUG */


#endif
