#include <math.h>
#include "mathutils.h"

#include "animation.h"
#include "pool.h"
#include "cue.h"
#include "fyjik.h"
#include "poolview.h"
#include "gtulpas.h"

GList		 *pool_cue_type_list = NULL;
PoolCue		  pool_cue = {NULL,  NULL, 
			      /* ball */0, /* alfa */ 0, /* beta */ 0, /* ccd */ 20,
			      /* cor */ {0,0,0}, /* top */ {0,0,0}, /* up */ {0,0,1},
			      /* right */ { 0,1,0}, /* orient */ { -1, 0,0}, 
			      /* adjust */ {0,0}, /* atop */ {0,0,0}, /* abot */{0,0,0},
			      /* rebuild_gl_list */ TRUE};
dbl		  pool_cue_mouse_speed = 0.1;

#define SWAP_YZ(V) { dbl tmp; tmp = (V)[Y]; (V)[Y] = (V)[Z]; (V)[Z] = tmp; }

void
cue_update_orientation (PoolCue *cue)
{
  angles_to_vectors (cue -> alfa, cue -> beta,
		     cue -> orientation, cue -> right, cue -> up);
}

void
cue_update_position (PoolCue *cue)
{
  cue_get_new_position (cue, cue -> cue_center_distance,
		    cue -> top, cue -> adjusted_top, cue -> adjusted_bottom);
}

void
cue_get_new_position (PoolCue *cue, dbl new_cue_center_distance,
		      dbl *top, dbl *adjusted_top, dbl *adjusted_bottom)
{
  dbl		hv [3];
  dbl		vv [3];
  dbl		bottom [3];
  PoolBall     *ball;
  dbl		radius;

  g_assert (top != NULL && adjusted_top != NULL);

  ball = gtulpas -> game -> balls  + cue -> ball;
  radius = ball -> ball_type -> radius;
 
  vec_mul_c (top, cue -> orientation, - new_cue_center_distance);
  vec_add (top, top, cue -> center_of_rotation);

  vec_mul_c (hv, cue -> right, cue -> adjust [0] * radius);
  vec_mul_c (vv, cue -> up, cue -> adjust [1] * radius);

  vec_add (hv, hv, vv);
  vec_add (adjusted_top, hv, top);

  if (adjusted_bottom != NULL)
    {
      vec_mul_c (adjusted_bottom, cue -> orientation, (cue -> cue_type -> length));
      vec_sub (adjusted_bottom, adjusted_top, adjusted_bottom);
    }
}

void
cue_rotate (PoolCue *cue, dbl h_change, dbl v_change)
{
  if (! animation_finished_p ())
    return;

  /* FIXME : should check for cue x ball (table) collision  !!! */
  /* FIXME : remove SWAP_YZ....  */

  cue -> alfa = fmod (cue -> alfa - h_change, 2.0 * PI);
  cue -> beta = fmod (cue -> beta - v_change + PI, 2.0 * PI) - PI;

  if (cue -> beta < 0.0)
    cue -> beta = 0.0;

  if (cue -> beta > (PI / 2.))
    cue -> beta = PI / 2.;

  cue_update_orientation (cue);
  cue_update_position (cue);

  pool_view_redraw_mark_all ();
  cue_recalc_hints (NULL);
}

void
cue_recalc_hints (void *ptr)
{
  GList		* lview;
  PoolViewGL	*  view;
  Animation	* anim;
  int		 c;

  if (! animation_finished_p ())
    return;

  for (c = 0, lview = pool_view_list; lview; lview = lview -> next)
    {
      view = (PoolViewGL *) (lview -> data);
      if ((view -> show_hints) && ((view == ptr) || (ptr == NULL)))
	{ c++; view -> needs_redraw = TRUE; }
    }

  if (c == 0)
    return;
 
  anim = gtulpas -> game -> animation;
  pool_pythie (ANIMATION_LAST_FRAME (anim), 
	       anim -> nof_balls,
	       pool_cue.ball,
	       pool_cue.orientation);
}

void 
cue_adjust (PoolCue *cue, dbl x, dbl y)
{
  if (x == cue -> adjust [X] && y == cue->adjust [Y])
    return;

  cue -> adjust [X] = x;
  cue -> adjust [Y] = - y;

  cue_update_position (cue);
  pool_view_redraw_mark_all ();
}

void
cue_set_ball (PoolCue *cue, int ball_id)
{
  PoolBallSpaceInfo	*ball;

  ball = ANIMATION_LAST_FRAME (gtulpas -> game -> animation) + ball_id;

  vec_cpy (cue -> center_of_rotation, ball -> p);
  cue -> cue_center_distance = 4.0 * ball -> ball -> ball_type -> radius;

  cue -> ball = ball_id;
  cue_update_position (cue);

  views_rebind_to_cue ();
}

void
cue_set_orientation (PoolCue *cue, dbl alfa, dbl beta)
{
  dbl	p;

  cue -> alfa = alfa;
  cue -> beta = beta;

  cue_update_orientation (cue);
  cue_update_position (cue);
}

void
cue_reset_position (PoolCue *cue)
{
  cue_set_ball (cue, cue -> ball);
}

void
cue_boom (dbl *oldpos, dbl *newpos, dbl speed)
{
  Animation		*anim = gtulpas->game->animation;
  Animation		*newanim;
  PoolBallSpaceInfo	*ball;
  AnimEvent		*events;

  ball = ANIMATION_LAST_FRAME (anim);

  pool_big_mover (ball, anim->nof_balls, ANIMATION_DEFAULT_STEP,
		  oldpos, newpos, speed, 0.5 /* FIXME : pool_cur_cue->roughness */,
		  &newanim);

  animation_destroy (anim);

  gtulpas->game->animation = newanim;

  ANIM_PANEL_SET_POS (0);
  ANIM_PANEL_SET_MAX (newanim->length);

  pool_play_current_animation ();
}

void
cue_motion (PoolCue *cue, dbl delta_y, dbl speed)
{
  PoolBallSpaceInfo	*balls;
  Animation             *anim = gtulpas->game->animation;

  dbl			___tmp [3];
  dbl			new_adj_top [3];
  dbl			new_distance;


  if (game_cue_active (gtulpas->game))
    {
      new_distance = cue -> cue_center_distance + delta_y * pool_cue_mouse_speed;
      cue_get_new_position (cue, new_distance,
			    ___tmp, new_adj_top, NULL);

      balls = ANIMATION_LAST_FRAME (anim);

      switch  (pool_cue_ball (balls, 
			      anim->nof_balls,
			      cue->adjusted_top,
			      new_adj_top))
	{
	case 1:
#ifdef DEBUG
	  g_message ("BZUK: (%.6f %.6f)", delta_y, speed);
#endif
	  game_add_event_type_data (gtulpas->game, CUE_X_BALL, 
				    (gpointer) event_cue_x_ball_new (cue->adjusted_top, 
								     new_adj_top, 
								     pool_cue_mouse_speed * speed));
	  break;

	case 0:
	  break;

	case -1:
#ifdef DEBUG
	  g_message ("Table was hit");
#endif
	  break;
	default:
	  g_assert_not_reached ();
	}

      cue -> cue_center_distance = new_distance;
      cue_update_position (cue);

      pool_view_redraw_mark_all ();
    }
}

