/* debug.h -- debug messages
   $Id: debug.h,v 1.2 1999/11/27 00:41:16 tomby Exp $

   Copyright (C) 1996 - 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef __DEBUG_H__
#define __DEBUG_H__

#define DSTR(s) (s) ? s : "(null)"

/*
  Tohle jsou debugovaci rutiny

  zapnout debugovaci mod muzete nastavenim env. promenne DEBUG na Yes
  treba takhle:

  tcsh:
                    setenv DEBUG Yes

  sh, bash, ksh:
                    DEBUG=Yes ; export DEBUG

  predstavuju si, ze pokud se kompiluje bez debugovani, neni nic navic
  a v opacnem pripade si muzeme nechat vypisovat potrebne hlasky

  rutiny jsou inlinovane - tzn. pokud mate jen cast zkompilovanou
  s debugovanim, dostavate hlasky jen z teto casti

  debug (..)       - vypise debugovaci msg identifikovanou DSECTION
  udebug (..)      - vypise neidentifkovanou msg (treba kdyz neni na zacatku radky)


  Pozn: (Petr Danecek)  
    - Aby to fungovalo, je zrejme treba nejdrive definovat DSECTION a pak teprve
     inkludovat debug:
        ..
        #define DSECTION "name of my section"
         ..
        #include "../debug.h"

  nepouzivejte v .h souborech (ostatni si je includnou a predefinuje se jim
  jejich DSECTION)
 
*/
#include <stdio.h>

#ifndef DSECTION
#define DSECTION ""
#endif

#ifdef DEBUG
#define debug(args...) { \
	fprintf (stderr, "%15s(%6d)\t", __FILE__, __LINE__); fprintf (stderr, ## args); \
}

#define udebug(args...) fprintf (stderr, ## args);

#else

#define debug(args...)
#define udebug(args...)

#endif

#define FIXME(S)  g_message ("%s:%d FIXME - %s - write me, plz - %s", __FILE__, __LINE__, __PRETTY_FUNCTION__, (S))

#define exit_if_fail(COND, MESSAGE)	do { if (! (COND)) { g_error (MESSAGE); exit (0); }} while (0)

#ifdef DEBUG
#define DO_IF_TRUE( X, Y ) if ( X ) Y
#else
#define DO_IF_TRUE( X, Y ) Y
#endif

#endif
