/* player
   $Id: dialog_players.c,v 1.3 1999/12/07 15:27:09 rodo Exp $

   Copyright (C) 1999 Daniel Skarda <0rfelyus@suse.cz>
   Copyright (C) 1999 Radek Doulik <rodo@suse.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "dialog_players.h"
#include "debug.h"
#include <config.h>
#include <gnome.h>

static gchar * text [2];

static void
set_text_cb (gchar * text, gchar **t)
{
	debug ("player name %s\n", text);

	/* fix me -- memory leak */
	if (*t)
		g_free (*t);
	(*t) = g_strdup (text);
}

static void
rename_cb (GtkWidget *w, DialogPlayers *dp)
{
	GnomeDialog *dialog;
	gchar *new_name = NULL;
	debug ("rename_cb\n");

	if (dp->selected < 0)
		return;

	dialog = GNOME_DIALOG (gnome_request_dialog (FALSE, _("Enter new name"),
						     dp->gt->game->player [dp->selected]->name, 64,
						     (GnomeStringCallback) set_text_cb, (gpointer) &new_name,
						     GTK_WINDOW (dp->win)));

	if (!gnome_dialog_run_and_close (dialog)) {
		gtk_clist_set_text (GTK_CLIST (dp->clist), dp->selected, 0, new_name);
		gtk_clist_columns_autosize (GTK_CLIST (dp->clist));
		dp->gt->game->player [dp->selected]->name = new_name;
		gnome_tulpas_score_refresh (dp->gt->score);
	} else if (new_name)
		g_free (new_name);
}

static void
add_player_cb (GtkWidget *w, DialogPlayers *dp)
{
	debug ("add_player_cb\n");

	text [0] = _("Unknown");

	if (!game_add_player (dp->gt->game, text [0]))
		return;

	gtk_clist_insert (GTK_CLIST (dp->clist), -1, text);
	gtk_clist_select_row (GTK_CLIST (dp->clist), GTK_CLIST (dp->clist)->rows-1, 0);

	while (gtk_events_pending ()) gtk_main_iteration_do (FALSE);

	rename_cb (w, dp);

	if (dp->gt->game->nof_players >= dp->gt->game->max_players_nof)
		gnome_dialog_set_sensitive (GNOME_DIALOG (dp->win), 0, FALSE);
}

static void
select_row_cb (GtkWidget *w, gint row, gint col, GdkEvent *e, DialogPlayers *dp)
{
	debug ("select_row_cb\n");
	dp->selected = row;
	gnome_dialog_set_sensitive (GNOME_DIALOG (dp->win), 1, TRUE);
}

static void
unselect_row_cb (GtkWidget *w, gint row, gint col, GdkEvent *e, DialogPlayers *dp)
{
	debug ("unselect_row_cb\n");
	dp->selected = -1;
	gnome_dialog_set_sensitive (GNOME_DIALOG (dp->win), 1, FALSE);
}

static void
close_cb (GtkWidget *w, DialogPlayers *dp)
{
	debug ("close\n");
	gtk_widget_destroy (dp->win);
}

static void
destroy_cb (GtkWidget *w, DialogPlayers *dp)
{
	debug ("destroy\n");
	*(dp->ref) = NULL;
}

DialogPlayers *
dialog_players_new (GTulpas *gt, DialogPlayers **ref)
{
	DialogPlayers * ndp;
	gint            i;
	gchar         * titles [2];
	guchar        * acc_add = _("a - add accelerator");
	guchar        * acc_ren = _("r - rename accelerator");

	ndp = g_new (DialogPlayers, 1);

	ndp->selected = -1;

	ndp->gt  = gt;
	ndp->ref = ref;
	ndp->win = gnome_dialog_new (_("Players"),
				     _("Add player"),
				     _("Rename"),
				     GNOME_STOCK_BUTTON_CLOSE,
				     NULL);

	gtk_widget_set_usize (ndp->win, -1, 240);

	gnome_dialog_set_default (GNOME_DIALOG (ndp->win), 2);
	gnome_dialog_button_connect (GNOME_DIALOG (ndp->win),
				     0, (GtkSignalFunc) add_player_cb, ndp);
	gnome_dialog_button_connect (GNOME_DIALOG (ndp->win),
				     1, (GtkSignalFunc) rename_cb, ndp);
	gnome_dialog_button_connect (GNOME_DIALOG (ndp->win),
				     2, (GtkSignalFunc) close_cb, ndp);
	gnome_dialog_set_accelerator (GNOME_DIALOG (ndp->win), 0, acc_add [0], 0);
	gnome_dialog_set_accelerator (GNOME_DIALOG (ndp->win), 1, acc_ren [0], 0);

	if (gt->game->nof_players >= gt->game->max_players_nof)
		gnome_dialog_set_sensitive (GNOME_DIALOG (ndp->win), 0, FALSE);

	titles [0] = _("Name");
	titles [1] = _("Location");

	ndp->clist = gtk_clist_new_with_titles (2, titles);

	gtk_signal_connect (GTK_OBJECT (ndp->win), "destroy", destroy_cb, ndp);
	gtk_signal_connect (GTK_OBJECT (ndp->clist), "select_row", select_row_cb, ndp);
	gtk_signal_connect (GTK_OBJECT (ndp->clist), "unselect_row", unselect_row_cb, ndp);

	for (i=0; i<gt->game->nof_players; i++) {
		text [0] = gt->game->player [i]->name;
		text [1] = _("local");
		gtk_clist_insert (GTK_CLIST (ndp->clist), i, text);
	}

	gtk_clist_select_row (GTK_CLIST (ndp->clist), 0, 0);

	gtk_clist_columns_autosize (GTK_CLIST (ndp->clist));
	gtk_box_pack_start_defaults (GTK_BOX (GNOME_DIALOG (ndp->win)->vbox), ndp->clist);
	gtk_widget_show_all (ndp->win);

	return ndp;
}


