/* (c) 1999 Tomby */

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <math.h>
#include <stdlib.h>

#include "mathutils.h"
#include "pool.h"
#include "game.h"
#include "animation.h"
#include "fyjik.h"
#include "fyjpriv.h"
#include "gtulpas.h"

#include "cue.h"
#include "table.h"

#include "bbfyjik.h"

/* !!!!!!!!!!!!!!!!!!!! POZNAMKA: VSE JE V CGS !!!!!!!!!!!!!!!!!!!! */

static dbl		null[3] = {0,0,0};
static bang_info	bum;

static PernikovaChaloupka chanull = {-1, {0.0,0.0,0.0}, {0.0,0.0,0.0}, 0.0, 10.0};

static void (*sphere_sphere_handler)() = NULL;
static void (*sphere_plane_handler)() = NULL;
static void (*sphere_hole_handler)() = NULL;
static void (*ball_out_handler)() = NULL;
static int fyjik_state = -1;

static dbl		pythie [ 3 * MAX_PYTHIE_COLS * MAX_BALLS ];
static dbl	       *ppythie [ MAX_BALLS ];
static int		pythie_count [ MAX_BALLS ];

static dbl		prec_back_shock [ MAX_BALLS ];
static int	       *pocketed = NULL;

static dbl	       *bb_balls = NULL;
static dbl	       *bb_holes = NULL;
static dbl	       *bb_edges = NULL;
static dbl		bb_cue [6];

static Snehurka	       *problemal_cols = NULL;
static int	       *isected = NULL;
static dbl	       *vs = NULL;

static Bajaja	       *BA = NULL;
static dbl	        T_INTS;
static dbl		cum_tr, cum_scl;

static dbl		*state = NULL, ts;
static DedecekHribecek	*hrib = NULL;
static int		*hrib_qu = NULL;
static int		*hrib_lens = NULL;
static int		*crushed = NULL;
static int		nof_crush = 0;

static PoolCue	        my_pool_cue;

#ifdef DEBUG
static int		fyjik_inited = 0;
#endif

dbl ** pool_pythie_positions = ppythie;
int  * pool_pythie_counts = pythie_count;

dbl	*table_d = NULL, *table_h = NULL;
dbl	**table_sides = NULL;
int	*counts = NULL;
int	nof_edges = 0;
dbl	*hole_cuts;
int	*map_edges;

/*
void    (*fyjik_one_step) (Sphere *ball, int ids, dbl time);
void   	(*fyjik_end_proc) (Sphere *ball, int ids, dbl whole_interval_time, dbl time_since_last_col);
void	(*fyjik_cue_dyn) (dbl *c_pos, dbl *c_dir, dbl power, dbl roughnes, Sphere *bal, int id_s);
void	(*fyjik_s2s_col_dyn) (Sphere *ball, int id_s1, int id_s2, Babovicka *);
void*   (*fyjik_s2p_alc) (Sphere *ball);
void	(*fyjik_s2p_col_dyn_preproc) (Sphere *ball, PernikovaChaloupka *, void *data, int id_s, int id_p);
void	(*fyjik_s2p_col_dyn_proc) (Sphere *ball, void *data, int flags, dbl dt);
*/

void    (*fyjik_one_step) (Sphere *, int, dbl);
void   	(*fyjik_end_proc) (Sphere *, int, dbl, dbl);
void	(*fyjik_cue_dyn) (dbl *, dbl *, dbl , dbl , Sphere *, int );
void	(*fyjik_s2s_col_dyn) (Sphere *, int , int , Babovicka *);
void*   (*fyjik_s2p_alc) (Sphere *);
void	(*fyjik_s2p_col_dyn_preproc) (Sphere *, PernikovaChaloupka *, void *, int , int );
void	(*fyjik_s2p_col_dyn_proc) (Sphere *, void *, int , dbl );

void 	babovicka_on_pisecek( DedecekHribecek *hrib, int ns, int np, int nh, int k );

/***********
 *   BBs   *
 ***********/

static inline void make_bb_from_ball_path( dbl *bb, dbl *p1, dbl *p2, dbl r )
{
  if ( p1 [X] < p2 [X] )
    {
      bb [LLX] = p1 [X];
      bb [URX] = p2 [X];
    }
  else
    {
      bb [LLX] = p2 [X];
      bb [URX] = p1 [X];
    }

  if ( p1 [Y] < p2 [Y] )
    {
      bb [LLY] = p1 [Y];
      bb [URY] = p2 [Y];
    }
  else
    {
      bb [LLY] = p2 [Y];
      bb [URY] = p1 [Y];
    }

  if ( p1 [Z] < p2 [Z] )
    {
      bb [LLZ] = p1 [Z];
      bb [URZ] = p2 [Z];
    }
  else
    {
      bb [LLZ] = p2 [Z];
      bb [URZ] = p1 [Z];
    }

  bb [LLX] -= r;
  bb [LLY] -= r;
  bb [LLZ] -= r;

  bb [URX] += r;
  bb [URY] += r;
  bb [URZ] += r;
}

static inline void make_bb_from_poly( dbl *bb, dbl **p, int n )
{
  int	i;

  bb [URX] = bb [LLX] = (*p) [X];
  bb [URY] = bb [LLY] = (*p) [Y];
  bb [URZ] = bb [LLZ] = (*p) [Z];

  for (i = 0; i < n; i ++)
    {
      if (bb [LLX] > p [i] [X])
	bb [LLX] = p [i] [X];
      else
	if (bb [URX] < p [i] [X])
	  bb [URX] = p [i] [X];

      if (bb [LLY] > p [i] [Y])
	bb [LLY] = p [i] [Y];
      else
	if (bb [URY] < p [i] [Y])
	  bb [URY] = p [i] [Y];

      if (bb [LLZ] > p [i] [Z])
	bb [LLZ] = p [i] [Z];
      else
	if (bb [URZ] < p [i] [Z])
	  bb [URZ] = p [i] [Z];
    }
}

static inline void make_bb_from_hole( dbl *bb, dbl *center, dbl r, dbl hl, dbl ht )
{
  bb [LLX] = center [X] - r;
  bb [LLY] = center [Y] - r;
  bb [LLZ] = hl;
  bb [URX] = center [X] + r;
  bb [URY] = center [Y] + r;
  bb [URZ] = ht;
}

static inline int bb_has_ints( dbl *bb1, dbl * bb2 )
{
  if ( bb1 [URX] < bb2 [LLX] ||
       bb1 [LLX] > bb2 [URX] ||
       bb1 [URY] < bb2 [LLY] ||
       bb1 [LLY] > bb2 [URY] ||
       bb1 [URZ] < bb2 [LLZ] ||
       bb1 [LLZ] > bb2 [URZ] )
    return 0;

  return 1;
}

/***********/

static void print_sphere( int id, Sphere *s, dbl *p )
{
  if ( p )
    printf( "ball %d:\npos: %f %f %f\nv: %f %f %f\nomega: %f %f %f\n",
	    id, p[0], p[1], p[2], s->v[0], s->v[1], s->v[2],
	    s->omega[0], s->omega[1], s->omega[2] );
  else
    printf( "ball %d:\npos: %f %f %f\nv: %f %f %f\nomega: %f %f %f\n",
	    id, s->p[0], s->p[1], s->p[2], s->v[0], s->v[1], s->v[2],
	    s->omega[0], s->omega[1], s->omega[2] );
}

/************ 
 * HANDLERS *
 ************/

static void spherespherehandler( int id1, int id2, dbl *p1, dbl *p2, bang_info *info )
{
  AnimEvent *ev;

  animation_add_event( info->ani, info->time * 1000.0, S_Ball_Ball_Collision, id1, id2 );

#ifdef DEBUG
  printf("------------\nS2S s1:%d s2:%d t:%f\n", id1, id2, info->time );
  print_sphere( id1, info->act_sphs + id1, (dbl*)p1 );
  print_sphere( id2, info->act_sphs + id2, (dbl*)p2 );
#endif
}

static void sphereplanehandler( int ids, int idp, dbl *p1, void *p )
{
  bang_info *bb = (bang_info*)p;

  animation_add_event( bb->ani, bb->time * 1000.0, S_Ball_Mantinel_Collision, ids, idp );

#ifdef DEBUG
  printf("-----------\nS2P s:%d p:%d t:%f\n", ids, idp,  bb->time);
  print_sphere( ids, bb->act_sphs + ids, (dbl*)p1 );
#endif
}

static void sphereholehandler( int ids, int idh, dbl *p1, void *p )
{
  bang_info *bb = (bang_info*)p;

  animation_add_event( bb->ani, bb->time * 1000.0, S_Ball_Pocketed, ids, idh );

#ifdef DEBUG
  printf("----------\nS2H s:%d h:%d t:%f\n", ids, idh, bb->time );
  print_sphere( ids, bb->act_sphs + ids, (dbl*)p1 );
#endif
}

static void holeouthandler( int ids, void *p )
{
  bang_info *bb = (bang_info*)p;

  animation_add_event( bb->ani, bb->time * 1000.0, S_Ball_Jump_Off, ids, 0 );

#ifdef DEBUG
  printf("-------------\nBalll out s:%d t:%f\n", ids, bb->time );
#endif
}

/**********/

static void sphere_sphere_pythie( int id1, int id2, dbl *p1, dbl *p2, bang_info *info )
{

#ifdef DEBUG
  printf("pythie s2s s1:%d s2:%d at %f %f %f - %f %f %f t:%f\n", id1, id2,
	 p1[0], p1[1], p1[2],
	 p2[0], p2[1], p2[2],
	 info->time
	 );
#endif

  if ( pythie_count [id1] < MAX_PYTHIE_COLS )
    {
      cvec( ppythie [id1] + 3 * pythie_count [id1], p1 );
      ppythie [id1][3 * pythie_count [id1] + Z] = 2 * SLIGHTLY_ABOVE_TABLE;
      pythie_count [id1] ++;
    }
  else
    cvec( info->act_sphs [id1].v, null );

  if ( pythie_count [id2] < MAX_PYTHIE_COLS )
    {
      cvec( ppythie [id2] + 3 * pythie_count [id2], p2 );
      ppythie [id2][3 * pythie_count [id2] + Z] = 2 * SLIGHTLY_ABOVE_TABLE;
      pythie_count [id2] ++;
    }
  else
    cvec( info->act_sphs [id2].v, null );
}

static void sphere_plane_pythie( int ids, int idp, dbl *p1, void *p )
{
  bang_info *info = (bang_info*)p;


#ifdef DEBUG
  printf("pythie s2p s:%d p:%d at %f %f %f t:%f - v: %f %f %f\n", ids, idp, 
	 p1[0], p1[1], p1[2], info->time,
	 info->act_sphs[ids].v[0], info->act_sphs[ids].v[1], info->act_sphs[ids].v[2]);
#endif

  if ( pythie_count [ids] < MAX_PYTHIE_COLS )
    {
      cvec( ppythie [ids] + 3 * pythie_count [ids], p1 );
      ppythie [ids][3 * pythie_count [ids] + Z] = 2 * SLIGHTLY_ABOVE_TABLE;
      pythie_count [ids] ++;
    }
  else
    cvec( info->act_sphs [ids].v, null );

}

static void sphere_hole_pythie( int ids, int idh, dbl *p1, void *p )
{
  bang_info *bb = (bang_info*)p;

#ifdef DEBUG
  printf("s2h s:%d h:%d t:%f\n", ids, idh, bb->time );
#endif
}

static void hole_out_pythie( int ids, void *p )
{
  bang_info *bb = (bang_info*)p;

#ifdef DEBUG
  printf("balll out s:%d t:%f\n", ids, bb->time );
#endif
}


/***********/

static inline int has_ints_with( int i, int j, int hrib_line_len )
{
  int	k;
  
  for (k = 0; k < hrib_lens [i]; k ++)
    if ( hrib_qu [i * hrib_line_len + k] == j )
      return 1;

  return 0;
}

/***********/

static inline void update_BA( int i, dbl t, int hrib_line_len, int j )
{
  if ( BA [i].t > t )
    {
      BA [i].t = t;
      BA [i].idx = hrib_lens [i];
    }
  
  hrib_qu [hrib_line_len * i + hrib_lens [i]] = j;
  hrib_lens [i] ++;
}

/***********/

dbl getPlaneResult( dbl *n, dbl *v, dbl *s )
{
  dbl d;

  d = (n)[0]*( (s)[0] - (v)[0] ) + 
      (n)[1]*( (s)[1] - (v)[1] ) + 
      (n)[2]*( (s)[2] - (v)[2] );
  
  return d;
}

/***********/

void int_vec_vec( dbl *v1, dbl *s1, dbl *v2, dbl *s2, dbl *u, dbl *v )
{
  dbl a[2][2], b[2], d;
  int i, j, k[2];

  for( j = i = 0; i < 3; i++ )
    {
      k[j] = 0;
      k[j] |= (fabs( v1[i] ) > EPSILON) | 2*(fabs( v2[i] ) > EPSILON);
      if ( ((!j) && k[0]) || (j && ((k[0]|k[1]) == 3)) )
	{
	  a[j][0] = v1[i];
	  a[j][1] = -v2[i];
	  b[j++] = s2[i] - s1[i];
	  if ( j == 2 )
	    break;
      }
    }

  d = det2( a[0][0], a[0][1], a[1][0], a[1][1] );

  *u = det2( s1[0], a[0][1], s1[1], a[1][1] ) / d;
  *v = det2( a[0][0], s2[0], a[1][0], s2[1] ) / d;
}

/***********/

int cut_absc( dbl *n, dbl *v, dbl *s, dbl *dir, dbl *t1, dbl *t2 )
{
  dbl	tmp[3];
  dbl	r1, r2, t;
  
  get_point( tmp, s, dir, *t1 );
  r1 = getPlaneResult( n, v, tmp );

  get_point( tmp, s, dir, *t2 );
  r2 = getPlaneResult( n, v, tmp );

  if ( r1*r2 < 0.0 )
    {
      t = -r1 / (r2 - r1);
      t = *t1 + t*(*t2-*t1);
      
      if ( r1 > 0.0 )
	*t2 = t;
      else
	*t1 = t;
      
      return 0;
    }

  return (r1 < 0.0)?(-1):(1);
}

/***********/

int cut_to_normal_region( dbl *t1, dbl *t2, dbl *s, dbl *dir, dbl *from, dbl *to )
{
  dbl	n[3];

  vec_sub( n, to, from );
  if ( cut_absc( n, from, s, dir, t1, t2 ) == -1 )
    return 0;

  vec_neg( n );
  if ( cut_absc( n, to, s, dir, t1, t2 ) == -1 )
    return 0;

  return 1;
}

/***********/

int edS_path_col( dbl *es, dbl *edir, dbl *ps, dbl *pdir, dbl t1, dbl t2, dbl *t, dbl r, dbl *pcol )
{
  dbl 	cc, cacc, cbcdcc;
  dbl	tmp[3], ints[2];
  dbl	a, b, c;
  int	i;
  
  cc = vec_dot( edir, edir );
  cacc = vec_dot( edir, pdir ) / cc;
  vec_sub( tmp, ps, es );
  /*  cbcdcc = (vec_dot( edir, ps ) - vec_dot( edir, tmp )) / cc; */
  cbcdcc = vec_dot( edir, tmp ) / cc;

  a = b = c = 0.0;

  for( i = 0; i < 3; i++ )
    {
      a += sqr( pdir[i] - edir[i]*cacc );
      b += (pdir[i] - edir[i]*cacc)*(tmp[i] - edir[i]*cbcdcc);
      c += sqr(tmp[i] - edir[i]*cbcdcc);
    }
  
  b *= 2.0;
  c -= sqr(r);
  
  if ( !solve_quadr_eq( a, b, c, ints, ints + 1 ) )
    return 0;

  if ( ints[0] > ints[1] )
    {
      *t = ints[1];
      ints[1] = ints[0];
    }
  else
    *t = ints[0];

  if ( t1 <= *t && *t <= t2 )
    {
      get_point( pcol, es, edir, cacc*(*t) + cbcdcc );
      return 1;
    }

  *t = ints[1];

  if ( t1 <= *t && *t <= t2 )
    {
      get_point( pcol, es, edir, cacc*(*t) + cbcdcc );
      return 1;
    }
  
  return 0;
}

/***********/

/*
                  P 
  x---------------x
                   } dist - for intersection with sphere must be dist gt or eq then sph radius
                  x
                   \ A - vector
                    \  
                     \
                      x B - begining point

    X = A*t + B
    
    r^2 = dist( X, P )^2 = sum( 0..2, ( A[i]*t + B[i] - P[i] )^2 )
    a = A[i]^2
    b = 2*A[i]*(B[i]-P[i])
    c = (B[i]-P[i])^2 - r^2 
*/

int point_path_col( dbl *p, dbl *pas, dbl *padir, dbl r, dbl *t, dbl *p_ints )
{
  dbl	a, b, c;
  dbl	ints[2];
  int	i;

  a = b = c = 0.0;
  
  for( i = 0; i < 3; i++ )
    {
      a += sqr(padir[i]);
      b += padir[i]*(pas[i] - p[i]);
      c += sqr(pas[i] - p[i]);
    }

  b *= 2.0;
  c -= sqr(r);

  if ( !solve_quadr_eq( a, b, c, ints, ints + 1 ) )
    return 0;

  if ( ints[0] > ints[1] )
    {
      *t = ints[1];
      ints[1] = ints[0];
    }
  else
    *t = ints[0];

  /* FIXME: pridat t1 a t2 kde se to ma kontrolovat nebo staci mezi 0,1 ? */

  if ( 0.0 <= *t && *t <= 1.0 )
    {
      cvec( p_ints, p );
      return 1;
    }

  *t = ints[1];

  if ( 0.0 <= *t && *t <= 1.0 )
    {
      cvec( p_ints, p );
      return 1;
    }
  
  return 0;
  
}

/***********/

/* normal vector must be unit vector */
int colidePlaneSphere( dbl **v, int num, dbl *n, dbl *centr, dbl *old_pos, dbl r, dbl *ints, dbl *tints )
{
  int    i, j, left, right, si = 1;
  dbl  	 d, dd, ee;
  dbl    t, t1, t2;
  dbl    nn[3], ss[3], tmp[3];
  dbl	 s[3];

  /* compute moving vector */
  vec_sub( s, centr, old_pos );
  cvec( ss, s );

  /* if ball is not moving don't check colision */
  if ( vec_size2(s) == 0.0 )
    return 0;

  /* check on which side is ball */
  ee = getPlaneResult( n, v[0], old_pos );
  t1 = getPlaneResult( n, v[0], centr );

  if ( fabs( ee ) < r )
    si = 0;

  if ( ee < 0.0 )
    dd = -r;
  else
    dd = r;

  /* normal vector of poly * moving vector of ball */
  d = vec_dot( n, s );

  /* moving is along polygon */
  if ( d == 0.0 )
    return 0;

  /* compute intersection t */
  vec_sub( tmp, old_pos, v[0] );
  t = ( dd - vec_dot( n, tmp ) ) / d;

  if ( (t < 0.0 || t > 1.0) && si )
    return 0;

  /* intersection is not lie on path */
  if ( t >= 0.0 && t <= 1.0 )
    {
      if ( ee*t1 >= 0.0 && (fabs( ee ) - fabs( t1 )) < 0.0 )
	return 0;

      /* and compute intersection point */
      vec_mul_c( s, s, t );
      vec_add( s, s, old_pos );

      j = num - 1;

      left = right = 0;
  
      for( i = 0; i < num; i++ )
	{
	  vec_sub( tmp, v[i], v[j] );
	  vec_cross( nn, n, tmp );
	  d = getPlaneResult( nn, v[i], s );
	  if ( d >= 0 )
	    left++;
	  else
	    if ( d < 0 )
	      right++;
	  j = i;
	}
  
      if ( (left && !right) || (right && !left) )
	{
	  ints[0] = -dd*n[0]+s[0];
	  ints[1] = -dd*n[1]+s[1];
	  ints[2] = -dd*n[2]+s[2];
	  
	  /* UGLY HACK */

	  t -= 2 * EPS / vec_size (ss);
	  if ( t < 0.0 )
	    t = 0.0;

	  /*************/

	  *tints = t;
	  
	  return 1;
	}
    }

  j = num - 1;
  
  for( i = 0; i < num; j = i++ )
    {
      t1 = 0.0; 
      t2 = 1.0;
      t = 10.0;
#ifdef DEBUG
      if ( (sqr(old_pos [0] - v[j][0]) + sqr(old_pos [1] - v[j][1])) < sqr(r) )
	printf ("Oh may shit - there must be somfing f**ked\n");
#endif
      if ( cut_to_normal_region( &t1, &t2, old_pos, ss, v[j], v[i] ) )
	{
	  vec_sub( tmp, v[i], v[j] );
	  if ( edS_path_col( v[j], tmp, old_pos, ss, t1, t2, &d, r, ints ) )
	    t = d;
	}
      
      if ( point_path_col( v[i], old_pos, ss, r, &d, nn ) )
	if ( d < t )
	  {
	    t = d;
	    cvec( ints, nn );
	  }

      if ( point_path_col( v[j], old_pos, ss, r, &d, nn ) )
	if ( d < t )
	  {
	    t = d;
	    cvec( ints, nn );
	  }
      
      if ( 0.0 <= t && t <= 1.0 )
	{
	  /* UGLY HACK */

	  t -= 2 * EPS / vec_size (ss);
	  if ( t < 0.0 )
	    t = 0.0;
	  /************/

	  *tints = t;
	  return 2;
	}
    }

  return 0;
}

/*
 Old - edge collision with only position not path

  j = num - 1;

  for( i = 0; i < num; i++ )
    {
      vec_sub( tmp, v[i], v[j] );
      vec_sub( ss, v[i], s );
      if ( !vec_dot( tmp, ss ) )
	{
	  vec_cross( nn, tmp, ss );
	  vec_cross( ss, nn, tmp );
	  vec_norm( ss, ss );
	  int_vec_vec( tmp, v[j], ss, s, &d, &dd );

	  if ( dd > r )
	    return 0;

	  if ( d >= 0 && d <= 1 )
	    {
	      ints[0] = dd*ss[0]+s[0];
	      ints[1] = dd*ss[1]+s[1];
	      ints[2] = dd*ss[2]+s[2];
	      
	      return 2;
	    }
	}
      if ( euclid( s, v[i] ) < r )
	{
	  vec_sub( tmp, v[i], s );

	  ints[0] = r*tmp[0]+s[0];
	  ints[1] = r*tmp[1]+s[1];
	  ints[2] = r*tmp[2]+s[2];

	  return 2;
	}

      if ( euclid( s, v[j] ) < r )
	{
	  vec_sub( tmp, v[j], s );

	  ints[0] = r*tmp[0]+s[0];
	  ints[1] = r*tmp[1]+s[1];
	  ints[2] = r*tmp[2]+s[2];

	  return 2;
	}
    }
*/

/***********/

/*
  dbl r = euclid( s1, s2 );
  
  if ( r < ( r1 + r2 ) )
    {
      (norm)[0] = (s1)[0] - (s2)[0];
      (norm)[1] = (s1)[1] - (s2)[1];
      (norm)[2] = (s1)[2] - (s2)[2];
      vec_norm( norm, norm );
      return 1;
    }

  return 0;
*/

int colideSphereSphere( dbl *s1, dbl r1, dbl *s2, dbl r2, dbl *old_pos1, dbl *old_pos2, dbl *norm, dbl *t )
{
  dbl 	a,b,c,t1,t2;
  dbl 	A[3],C[3], tmp [3], tmp1 [3];
  int	i;

  vec_sub( A, s1, old_pos1 );
  vec_sub( C, s2, old_pos2 );

  if ( euclid2 (old_pos1, old_pos2) <= sqr (r1 + r2) )
    {
#ifdef DEBUG
      printf ("UAUAUAUAUAUHA - hejhola pachola se tresou\n");
#endif
      t1 = 0;
      get_point( tmp, old_pos1, A, t1 );
      get_point( tmp1, old_pos2, C, t1 );
      *t = t1;

      (norm)[0] = (tmp)[0] - (tmp1)[0];
      (norm)[1] = (tmp)[1] - (tmp1)[1];
      (norm)[2] = (tmp)[2] - (tmp1)[2];

      return 1;      
    }

  a = b = c = 0;

  for( i = 0; i < 3; i++ )
    {
      a += sqr(A[i]-C[i]);
      b += (A[i]-C[i])*(old_pos1[i]-old_pos2[i]);
      c += sqr(old_pos1[i]-old_pos2[i]);
    }

  b *= 2.0;
  c -= sqr( r1+r2 );

  if ( !solve_quadr_eq( a, b, c, &t1, &t2 ) )
    return 0;

  if ( t1 > t2 )
    {
      SWAP(t1, t2, a);
    }

  if ( t1 > 1.0 || ( t1 < 0.0 && (t2 < 0.0 || t2 > 1.0) ) )
    return 0;

  if ( t1 < 0.0 && t2 >= 0.0 && t2 < 1.0 )
    t2 = t1;
 
  get_point( tmp, old_pos1, A, t1 );
  get_point( tmp1, old_pos2, C, t1 );

  /* UGLY HACK */
  
  a = vec_size2 (A);
  b = vec_size2 (C);

  if  ( a > b )
    t1 -= 2 * EPS / sqrt (a);
  else
    t1 -= 2 * EPS / sqrt (b);

  if ( t1 < 0.0 )
    t1 = 0.0;

  /********/
  
  *t = t1;

  (norm)[0] = (tmp)[0] - (tmp1)[0];
  (norm)[1] = (tmp)[1] - (tmp1)[1];
  (norm)[2] = (tmp)[2] - (tmp1)[2];

  return 1;
}

/***********/

/* ints - array dbl [2] because ray and sphere could have two intersection */
int ray2sphere( dbl *p, dbl *v, dbl *s, dbl r, dbl *ints )
{
  dbl	a,b,c;
  int	i;
  
  a = b = c = 0.0;

  /*
    a*x^2 + b*x + c = 0	[1]

    D = b^2 - 4*a*c

    roots of equation [1] are

    x_1 = (-b + sqrt( D ))/(2*a)
    x_2 = (-b - sqrt( D ))/(2*a)

    ----------
    
    ray: 	x[i] = v[i]*t + p[i]
    sphere:	sum( i = 0..2, (x[i]-s[i])^2 ) = r^2

    =>
    
    sphere ray intersection

    sum( i = 0..2, (v[i]*t+p[i]-s[i])^2 ) = r^2

    =>

    a = sum( i = 0..2, v[i]^2 )
    b = 2*sum( i = 0..2, v[i]*(p[i]-s[i]) )
    c = sum( i = 0..2, (p[i]-s[i])^2 ) - r^2
  */

  for( i = 0; i < 3; i++ )
    {
      a += sqr(v[i]);
      b += v[i]*(p[i]-s[i]);
      c += sqr(p[i]-s[i]);
    }

  b *= 2.0;
  c -= sqr(r);

  return solve_quadr_eq( a, b, c, ints, ints + 1 );
}

/***********/

static dbl circle( dbl *center, dbl r, dbl *p )
{
  return ( sqr(p[0] - center[0]) + sqr(p[1] - center[1]) - sqr(r) );
}

/***********/

/*

  X = A*t + B

  (x-x0)^2 + (y-y0)^2 - r^2 = 0

  ((Ax*t+Bx) - x0)^2 - ((Ay*t+By) - y0)^2 - r^2 = 0

  (Ax*t + (Bx - x0))^2 - (Ay*t + (By-y0))^2 - r^2 = 0
*/

int ray2circle( dbl *pos, dbl *dir, dbl *center, dbl r, dbl *ints )
{
  dbl	a, b, c;
  int	i;

  a = b = c = 0.0;

  for( i = 0; i < 2; i ++ )
    {
      a += sqr(dir[i]);
      b += dir[i] * (pos [i] - center [i]);
      c += sqr (pos [i] - center [i]);
    }

  b *= 2.0;
  c -= sqr(r);

  return solve_quadr_eq( a, b, c, ints, ints + 1 );
}

/***********/

static dbl cyl2sphere_rpos( dbl *p, dbl *dir, dbl t, dbl rs, dbl *center, dbl rh, dbl h )
{
  dbl	tmp [3];
  dbl	a, x;

  get_point( tmp, p, dir, t );

  a = (rs + tmp [Z] - h) * (rs - tmp [Z] + h);
  x = sqr( tmp [X] - center [X] );
  x += sqr( tmp [Y] - center [Y] );
  x -= sqr( rh ) + a;
  x = sqr( x );
  x -= 4 * sqr( rh ) * a;

  return x;
}

/***********/

static int cyle2sph( dbl *p, dbl *dir, dbl t1, dbl t2, dbl rs, dbl *center, dbl rh, dbl h,
		     dbl *tints, dbl *norm, dbl *ints )
{
  dbl 	t, r1, r2, tt;
  dbl	its [2];

  t = (t2 - t1) / 2.0;

  if ( t < EPSILON )
    {
      t += t1;

      if ( (dir [Z] * t + p[Z]) <= h )
	return 0;

      norm [Z] = 0.0;
      norm [X] = (center [X] - (dir [X] * t + p[X]));
      norm [Y] = (center [Y] - (dir [Y] * t + p[Y]));

      ray2circle( center, norm, center, rh, its );
      if ( its [0] < 0.0 )
	tt = its [0];
      else
	tt = its [1];
      ints [X] = norm [X] * tt + center [X];
      ints [Y] = norm [Y] * tt + center [Y];
      ints [Z] = h;

      vec_norm( norm, norm );
      /*      norm [Z] = 1.0; */
      
      /* UGLY HACK */
      
      t -= 2 * EPS / vec_size (dir);

      if ( t < 0.0 )
	t = 0.0;

      /*************/

      *tints = t;
      return 1;
    }
  
  t += t1;

  r1 = cyl2sphere_rpos( p, dir, t1, rs, center, rh, h );
  r2 = cyl2sphere_rpos( p, dir, t2, rs, center, rh, h );

  if ( r1 * r2 > 0 )
    return 0;

  if ( cyle2sph( p, dir, t1, t, rs, center, rh, h, tints, norm, ints ) )
    return 1;

  return cyle2sph( p, dir, t, t2, rs, center, rh, h, tints, norm, ints );
}

/***********/

int colideSphereHole( dbl *pos, dbl *old_pos, dbl sr, dbl *center, dbl hr, dbl a1, dbl a2, dbl h,
		      dbl *norm, dbl *tints, dbl *ints )
{
  dbl	tmp[3], roots[2], tmp1[3], tmp2[3];
  dbl	t, a, d = -10.0;
  int	i, j;
  
  roots[0] = -10;
  roots[1] = -10;

  vec_sub( tmp, pos, old_pos );

  ray2circle( old_pos, tmp, center, hr - sr, roots );

  if ( roots [0] > roots [1] )
    SWAP( roots [0], roots [1], a );

  for( i = 0; i < 2; i ++ )
    if ( roots [i] > -EPSILON && roots [i] < 1.0 + EPSILON )
      break;
    
  if ( !i && roots [1] < 1.0 + EPSILON )
    t = roots [1];
  else
    t = roots [i];

  if ( (i == 1) || ( !i && ( t <= 1.0 + EPSILON ) ) )
    {
      vec_mul_c( tmp2, tmp, t );
      vec_add( tmp2, tmp2, old_pos );
      norm [Z] = 0.0;
      norm [X] = center [X] - tmp2 [X];
      norm [Y] = center [Y] - tmp2 [Y];

      if ( fabs(norm [X]) < EPSILON )
	if ( norm [Y] < 0.0 )
	  a = PI / 2;
        else
	  a = -PI / 2;
      else
	a = atan( norm [Y] / norm [X] );

      if ( norm [X] >= 0.0 )
	a += PI;
      
      if ( a < 0.0 )
	a += 2 * PI;
      
      if ( a1 < a2 )
	if ( a >= a1 && a <= a2 )
	  d = h;
	else
	  d = 0;
      else
	if ( a >= a1 || a <= a2 )
	  d = h;
        else
	  d = 0;

      if ( tmp2 [Z] < d )
	{
	  /* UGLY HACK */

	  t -= 2 * EPS / vec_size (tmp);

	  if ( t < 0.0 )
	    t = 0.0;

	  /*************/

	  *tints = t;
	  return 1;
	}
    }
  /* TODO: may be someday would be (if i don't be lazy), simplyfy horrible equation of degree 4, this code  */
  /*       faster */

  if ( d == -10.0 )
    d = 0.0;
  
  if (cyle2sph( old_pos, tmp, 0.0, 1.0, sr, center, hr, d, tints, norm, ints ))
    return 2;

  return 0;
}

/***********/

/*
int sphere2holes( Sphere *sph, int id, dbl *old_pos, DedecekHribecek *hrib, int hlen, int nofs, int nofp, int nofh, 
		  TableHole *holes, dbl *cutouts )
{
  int	i, j, k = 0;
  dbl	t, ints [3], norm [3];
  dbl	tmp1[3], tmp2[3];

  PernikovaChaloupka *ch;

  for( i = 0, j = id * hlen + nofs + nofp; i < nofh; i++, j++ )
    {
      ch = &hrib [j].ch;
      switch ( colideSphereHole( sph->p, old_pos, sph->ball->ball_type->radius, holes [i].p, holes [i].radius,
				 cutouts [2*i], cutouts [2*i + 1], TABLE_H, norm, &t, ints ) )
	{
	case 1:
#ifdef DEBUG
	  printf("holes 1\n");
#endif
	  cvec( hrib [j].ch.karkulka, norm );
	  ch->t = t;
	  ch->type = 1;
	  k = 1;
	  break;

	case 2:
#ifdef DEBUG
	  printf("holes 2 - s:%d h:%d\n", id, i);
	  if ( t < 0.0 )
	    printf( "Sakraa co to s:%d - time: %f , t = %f!\n", id, bum.time, t );
#endif
	  vec_sub( tmp2, sph->p, old_pos );
	  get_point( tmp1, old_pos, tmp2, t );
	  vec_sub( tmp1, ints, tmp1 );
	  if ( vec_size2( sph->v ) > EPSILON && vec_size2( tmp1 ) > EPSILON )
	    ch->coef = vec_angle( sph->v, tmp1 );
	  else
	    ch->coef = 0;
	  vec_cross( ch->vochomurka, sph->v, tmp1 );
	  vec_neg( tmp1 );
	  cvec( ch->kremilek, tmp1 );
	  ch->t = t;
	  ch->type = 2;
	  k = 2;
	  break;

	default:
	}
    }

  return k;
}
*/

int sphere2holes( Sphere *sph, int id, dbl *old_pos, DedecekHribecek *hrib, int hlen, int nofs, 
		  int nofp, int nofh, TableHole *holes, dbl *cutouts )
{
  int	i, j, k = 0;
  dbl	t, ints [3], norm [3];
  dbl	tmp1[3], tmp2[3], a;

  PernikovaChaloupka *ch;

  for( i = 0, j = nofs + nofp; i < nofh; i++, j++ )
    {
      ch = &hrib [id * hlen + hrib_lens [id]].ch;
      if ( bb_has_ints (bb_balls + 6*id, bb_holes + 6*i) )
	switch ( colideSphereHole( sph->p, old_pos, sph->ball->ball_type->radius, holes [i].p, 
				   holes [i].radius,
				   cutouts [2*i], cutouts [2*i + 1], TABLE_H, norm, &t, ints ) )
	  {	
	  case 1:
#ifdef DEBUG
	    printf("holes 1\n");
#endif
	    cvec( ch->n, norm );
	    ch->t = CT(t);
	    ch->type = 4;
	    k = 4;
	    update_BA (id, CT(t), hlen, j);
	    break;

	  case 2:
	    vec_sub( tmp2, sph->p, old_pos );
	    get_point( tmp1, old_pos, tmp2, t );
	    /*	    a = circle( holes [i].p, holes [i].radius, old_pos ); */
	    if ( /*a >= (sqr (holes [i].radius - sph->ball->ball_type->radius) - sqr (holes [i].radius)) &&*/
		 vec_dot (sph->v, norm) >= 0.0 )
	      {
#ifdef DEBUG
		printf("holes 3 - s:%d h:%d\n", id, i);
		if ( t < 0.0 )
		  printf( "Sakraa co to s:%d - time: %f , t = %f!\n", id, bum.time, t );
#endif
		cvec( ch->r, tmp1 );
		ch->type = 3;
		k = 3;
	      }
	    else
	      {
#ifdef DEBUG
		printf("holes 2 - s:%d h:%d\n", id, i);
		if ( t < 0.0 )
		  printf( "Sakraa co to s:%d - time: %f , t = %f!\n", id, bum.time, t );
#endif
		vec_sub( ch->r, ints, tmp1 );
		cvec( ch->n, norm );
		ch->type = 2;
		k = 2;
	      }
	    ch->t = CT(t);
	    update_BA (id, CT(t), hlen, j);
	    break;

	  default:
	  }
    } 

  if ( k )
    isected [id] = 1;

  return k;
}

/***********/

static int inHole( dbl *old_pos, dbl *center, dbl r )
{
  return ( circle( center, r, old_pos ) < 0 );
}

/***********/

static inline int chck_in_holes( Sphere *sph, int id )
{
  int	i;

  for( i = 0; i < pool_table.nof_holes; i++ )
    if ( bb_has_ints (bb_balls + 6*id, bb_holes + 6*i) )
      if ( inHole( sph->p, pool_table.holes[i].p, pool_table.holes[i].radius ) )
	return 1;
  
  return 0;
}

/**********/

int ball_in_hole( Sphere *sph )
{
  int	i;

  for( i = 0; i < pool_table.nof_holes; i++ )
    if ( inHole( sph->p, pool_table.holes[i].p, pool_table.holes[i].radius ) )
      return TRUE;
  
  return FALSE;
}

/***********/

static inline int chck_on_table( Sphere *sph, int id )
{
  if ( chck_in_holes( sph, id ) )
    return 0;

  return ( ( sph->p[Z] - sph->ball->ball_type->radius ) < EPSILON );
}

/***********/

static inline int chck_near_table( Sphere *sph, int id )
{
  if ( chck_in_holes( sph, id ) )
    return 0;

  return ( ( sph->p[Z] - sph->ball->ball_type->radius ) < NEAR_TABLE );
}

/***********/

static inline int chck_hole_bottom( Sphere *sph, int ids, int send )
{
  int	i;
  dbl	v = vs [ids]; /*vec_size( sph->v ); */
  dbl	r = sph->ball->ball_type->radius;

  for( i = 0; i < pool_table.nof_holes; i++ )
    if ( bb_has_ints (bb_balls + 6*ids, bb_holes + 6*i) )
      if ( inHole( sph->p, pool_table.holes[i].p, pool_table.holes[i].radius ) )
	{
	  if ( send && (( sph->p[Z] <= 0.0 && v > ENDER )  || (v < ENDER && sph->p[Z] > 0.0)) )
	    {
	      pocketed [ids] = 1;
	      cvec( sph->v, null );
	      cvec( sph->omega, null );
	      sph->p[Z] = HOLE_DEPTH;
	      if ( send )
		sphere_hole_handler( ids, pool_table.holes[i].hid, NULL, &bum );
	      return 1;
	    }
	  else
	    if ( sph->p[Z] <= 0.0 || v < ENDER )
	      return 1;
	}
  
  return 0;
}

/***********/

static void parametric_plane( dbl *res, dbl u, dbl v, dbl *up, dbl *right, dbl *pos )
{
  res [0] = right [0] * u + up [0] * v + pos [0];
  res [1] = right [1] * u + up [1] * v + pos [1];
  res [2] = right [2] * u + up [2] * v + pos [2];
}

/***********/

static dbl	begin_surf [4 * 3];
static dbl	end_surf [4 * 3];

static dbl	*surfs[] = {begin_surf,
			    begin_surf + 3,
			    begin_surf + 3 * 2,
			    begin_surf + 3 * 3,
			    
			    begin_surf,
			    begin_surf + 3,
			    end_surf + 3,
			    end_surf,
			    
			    begin_surf + 3,
			    begin_surf + 3 * 2,
			    end_surf + 3 * 2,
			    end_surf + 3,
			    
			    begin_surf + 3 * 2,
			    begin_surf + 3 * 3,
			    end_surf + 3 * 3,
			    end_surf + 3 * 2,
			    
			    begin_surf + 3 * 3,
			    begin_surf + 3,
			    end_surf + 3,
			    end_surf + 3 * 3,
			    
			    end_surf,
			    end_surf + 3,
			    end_surf + 3 * 2,
			    end_surf + 3 * 3
			    
};

static	int	scounts[6] = {4, 4, 4, 4, 4, 4};

void col_cue2sphere( dbl *begin, dbl *end, dbl r_begin, dbl r_end, dbl *up, dbl *right,
		     Sphere *sphs, int i, int n, dbl *old_pos, int nofp, DedecekHribecek *hrib )
{
  int 	k, j;

  if ( bum.time >= 0.0 || (fyjik_state == STATE_PYTHIE) )
    return;

  parametric_plane( begin_surf + 3 * 0,  r_begin,  r_begin, up, right, begin );
  parametric_plane( begin_surf + 3 * 1, -r_begin,  r_begin, up, right, begin );
  parametric_plane( begin_surf + 3 * 2, -r_begin, -r_begin, up, right, begin );
  parametric_plane( begin_surf + 3 * 3,  r_begin, -r_begin, up, right, begin );

  parametric_plane( end_surf + 3 * 0,  r_end,  r_end, up, right, end );
  parametric_plane( end_surf + 3 * 1, -r_end,  r_end, up, right, end );
  parametric_plane( end_surf + 3 * 2, -r_end, -r_end, up, right, end );
  parametric_plane( end_surf + 3 * 3,  r_end, -r_end, up, right, end );

  for( k = j = 0; j < 6; k += scounts[j++] )
    sphere2plane( sphs + i, old_pos + 3*i, hrib, n + nofp + DED_HRIB, n, i, surfs + k, 
		  scounts[j], j + 1 + nofp );

}

/***********/

void rebound( dbl *n, dbl *v )
{
  dbl  dot;
  dbl  nn[3], vv[3];

  cvec( nn, n );

  vv[0] = -(v)[0];
  vv[1] = -(v)[1];
  vv[2] = -(v)[2];

  dot = 2*vec_dot( nn, vv );
  vec_scl( dot, nn );
  vec_sub( v, nn, vv );

  vec_norm( v, v );
}

/***********/

static inline dbl sim( dbl m, dbl r )
{
  return 2 * m * sqr(r) / 5;
}

/***********/

void rigid_collision( dbl *v1, dbl *v2, dbl *omega1, dbl *omega2, dbl *col_n, dbl *r_to_col,
		      dbl om1, dbl om2, dbl oi1, dbl oi2, dbl r1, dbl r2,
		      dbl e, dbl roughnes, int flaf )
{
  dbl	j;
  dbl	n [3], v [3], rn [3], tmp [3], rr [3], col_r [3], do1 [3], do2 [3], tmp1 [3];
  dbl	oi, jj, V1, V2, O1, O2, d1, d2;

#ifdef DEBUG
  if ( isnan( omega1[0] ) || isnan( omega2[0] ) )
    {
      printf(" 1 - rigid body!\n" );
      exit(1);
    }
#endif  

  V1 = vec_size( v1 ); /* TIDIT */
  V2 = vec_size( v2 );
  O1 = vec_size( omega1 );
  O2 = vec_size( omega2 );

  rr [0] = 0.0;
  rr [1] = 0.0;
  rr [2] = 1.0;

  if ( O1 < EPSILON )
    O1 = 0.0;

  if ( O2 < EPSILON )
    O2 = 0.0;

  vec_norm( n, col_n );
  vec_norm( col_r, r_to_col );

  vec_sub( v, v1, v2 );

  vec_mul_c( v, v, -(1 + e) );
  
  if ( om1 > 0.0 && om2 > 0.0 )
    {
      cvec( rn, null );
    }
  else
    {
      vec_cross( rn, col_r, n );
      vec_mul_c( rn, rn, ( oi1 > 0.0 ) ? ( oi1 ) : ( oi2 ) );
      vec_cross( rn, rn, col_r );
    }

  /*
  if ( oi1 > 0.0 && flaf )
    {
      vec_cross( tmp, rr, n );
      vec_mul_c( tmp, tmp, oi1 );
      vec_cross( tmp, tmp, rr );
      vec_add( rn, rn, tmp );
    }

  if ( oi2 > 0.0 && flaf )
    {
      vec_cross( tmp, rr, n );
      vec_mul_c( tmp, tmp, oi2 );
      vec_cross( tmp, tmp, rr );
      vec_add( rn, rn, tmp );
    }
  */
  /*
  if ( oi1 > 0.0 && fabs ( v1 [2] ) < EPSILON && V1 > EPSILON && O1 > EPSILON )
    {
      do1 [0] = - v1 [1] / r1;
      do1 [1] = v1 [0] / r1;
      do1 [2] = 0.0;
      cvec( tmp, omega1 );
      d1 = vec_dot( do1, tmp );
      d1 /= V1*O1;
      vec_sub( do1, do1, tmp );
    }
  
  if ( oi2 > 0.0 && fabs ( v2 [2] ) < EPSILON && V2 > EPSILON && O2 > EPSILON )
    {	
      do2 [0] = - v2 [1] / r2;
      do2 [1] = v2 [0] / r2;
      do2 [2] = 0.0;
      cvec( tmp, omega2 );
      d2 = vec_dot( do2, tmp );
      d2 /= O2 * V2;
      vec_sub( do2, do2, tmp );
    }
  */

  j = vec_dot( v, n );
  
  if ( fabs( j ) < 0.1 )
    {
#ifdef DEBUG
      printf("ola ola ha!\n");
#endif
      j *= 16.0;
    }

  j /= ( om1 + om2 + vec_dot( rn, n ) );
  
#ifdef DEBUG
  if ( isnan( j ) )
    {
      printf("nana!\n");
      exit(1);
    }
#endif

  if ( om1 > 0.0 )
    {
      vec_mul_c( tmp, n, j * om1 );
      vec_add( v1, v1, tmp );
    }

  if ( om2 > 0.0 )
    {
      vec_mul_c( tmp, n, j * om2 );
      vec_sub( v2, v2, tmp );
    }

  vec_cross( tmp1, col_r, n );
  vec_cross( rr, rr, n );
  if ( oi1 > 0.0 )
    {
      if ( flaf && V1 > EPSILON && O1 > EPSILON && fabs (v1 [2]) < EPSILON )
	{
	  tmp [0] = - v1 [1] / r1;
	  tmp [1] = v1 [0] / r1;
	  tmp [2] = 0.0;
	  /*	  vec_sub( tmp, tmp, do1 ); */
	  cvec( omega1, tmp );
	}
      vec_mul_c( tmp1, tmp1, j * oi1 );
      vec_add( omega1, omega1, tmp1 );
    }

  if ( oi2 > 0.0 )
    {
      if ( flaf && V2 > EPSILON && O2 > EPSILON && fabs (v2 [2]) < EPSILON )
	{
	  tmp [0] = - v2 [1] / r2;
	  tmp [1] = v2 [0] / r2;
	  tmp [2] = 0.0;
	  /*	  vec_sub( tmp, tmp, do2 ); */
	  cvec( omega2, tmp );
	}
      vec_mul_c( tmp1, tmp1, j * oi2 );
      vec_sub( omega2, omega2, tmp1 );
    }

#ifdef DEBUG
  if ( isnan( omega1[0] ) || isnan( omega2[0] ) )
    {
      printf(" 2 - rigid body!\n" );
      exit(1);
    }
#endif
}

/***********/

/* TODO: dodelat ruzna hmotnost */
/*
       __
      /  \
   __ |S1|
  /  \\__/ 
  |S2|
  \__/
    

  interaction S1->S2 ( with rebounds to S1 )

  vv1 - vector of speed S1
  v1 - size of speed vector S1 
  v2 - size of speed vector S2
  tmp - normal of colision ( normal of S2 )
  vv2 - vector of speed S2
  vv3 - vector of angular speed S2
  f1 - -\
  f2 -   \_ same as f1-f4 constant in PoolBallType
  f3 -   /
  f4 - -/
*/

/* convert tmp to -tmp and unite it*/

void proc_sphere_inter( dbl *vv1, dbl v1, dbl v2, dbl *tmp, dbl *vv2, dbl *vv3, 
			dbl f1, dbl f2, dbl f3, dbl f4 )
{
  dbl 	tmp3[3], tmp1[3];
  dbl  	v, vv, a;

  if ( v1 < EPSILON )
    {
      cvec( vv2, null );
      cvec( vv3, null );
      vec_neg( tmp );
      return;
    }

  cvec( tmp1, vv1 );

  vec_norm( tmp, tmp );

  /* rebound */
  rebound( tmp, tmp1 );
  /* compute normal component of colision force  */
  if ( vec_size2( tmp ) > EPSILON && vec_size2( tmp1 ) > EPSILON )
    a = vec_angle( tmp, tmp1 );
  else
    a = 0.0;
  vec_neg( tmp );
  /* size of normal component */
  v = (v1*fabs(cos( a ))*f1 - ((v2 == 0)?(1):(0))*f2);
  vec_mul_c( tmp3, tmp, v );
  cvec( vv2, tmp3 );
  /* size of angular component */
  /* v += (vv=v1*sin( a )*f3); - ha ??? */
  vv=v1*fabs(sin( a ))*f3;
  vec_mul_c( tmp3, tmp1, (v1-v) );
  cvec( vv1, tmp3 );

  vec_cross( tmp3, tmp1, tmp );
  /*  vec_cross( tmp3, tmp3, tmp ); */
  vec_norm( tmp3, tmp3 );
  vec_mul_c( tmp3, tmp3, vv );
  vec_mul_c( vv3, vv3, (1-f4) );
  vec_add( vv3, vv3, tmp3 );
}

/***********/

void sphere2sphere( Sphere *sphs, dbl *old_pos, int ns, int sph, 
		    DedecekHribecek *hrib, int hrib_line_len )
		   
{
  int	i, j;
  dbl   tmp[3];
  dbl	t;
  
  for( i = 0; i < ns; i++, t = 10.0 )
    if ( i != sph && (vs [sph] > EPSILON || vs [i] > EPSILON) && 
	 bb_has_ints (bb_balls + 6*sph, bb_balls + 6*i) )
      if ( !has_ints_with (sph, i, hrib_line_len) && 
	   colideSphereSphere( sphs[sph].p, 
			       sphs[sph].ball->ball_type->radius, sphs[i].p, 
			       sphs[i].ball->ball_type->radius, 
			       old_pos+3*sph, old_pos+3*i, tmp, &t ) )
	{
	  hrib[hrib_line_len * sph + hrib_lens [sph]].b.t = CT(t);
	  cvec(hrib[hrib_line_len * sph + hrib_lens [sph]].b.n, tmp);
	  hrib[hrib_line_len * i + hrib_lens [i]].b.t = CT(t);
	  cvec(hrib[hrib_line_len * i + hrib_lens [i]].b.n, tmp);
	  isected [i] = isected [sph] = 1;
	  update_BA (i, CT(t), hrib_line_len, sph);
	  update_BA (sph, CT(t), hrib_line_len, i);
	}
}

/***********/

/*
int sphere2sphere_endproc( Sphere *sphs, int sph, int i,
			   Babovicka *bab )
{
  dbl   tmp[3], tmp1[3], tmp2[3], tmp3[3], tmp4[3];
  dbl	v, v1, v2;
  int	k;

  cvec( tmp, bab->n );

  v1 = vec_size( sphs[sph].v );
  v2 = vec_size( sphs[i].v );
  cvec( tmp3, sphs[i].omega );
  cvec( tmp2, sphs[i].v );
  proc_sphere_inter( sphs[sph].v, v1, v2, tmp, 
		     sphs[i].v, sphs[i].omega,
		     sphs[sph].ball->ball_type->f1, sphs[sph].ball->ball_type->f2, 
		     sphs[sph].ball->ball_type->f3, 
		     sphs[i].ball->ball_type->f4 );
  cvec( tmp1, sphs[sph].omega );
  vec_mul_c( tmp1, tmp1, sphs[sph].ball->ball_type->f4 );
  vec_add( sphs[i].omega, sphs[i].omega, tmp1 );
  vec_sub( sphs[sph].omega, sphs[sph].omega, tmp1 );
      

  cvec( tmp4, sphs[sph].v );

  proc_sphere_inter( tmp2, v2, v1, tmp, 
		     sphs[sph].v, sphs[sph].omega,
		     sphs[i].ball->ball_type->f1, sphs[i].ball->ball_type->f2, 
		     sphs[i].ball->ball_type->f3,
		     sphs[sph].ball->ball_type->f4 );

  vec_add( sphs[sph].v, sphs[sph].v, tmp4 );
  vec_add( sphs[i].v, sphs[i].v, tmp2 );

  vec_mul_c( tmp3, tmp3, sphs[i].ball->ball_type->f4 );
  vec_add( sphs[sph].omega, sphs[sph].omega, tmp3 );
  vec_sub( sphs[i].omega, sphs[i].omega, tmp3 );

  return 1;
}
*/

int sphere2sphere_endproc( Sphere *sphs, int sph, int i,
			   Babovicka *bab )
{

  rigid_collision( sphs[i].v, sphs[sph].v, sphs[i].omega, sphs[sph].omega, bab->n, bab->n, 
		   1.0 / sphs[i].ball->ball_type->weight, 1.0 / sphs[sph].ball->ball_type->weight,
		   1.0 / sim( sphs[i].ball->ball_type->weight, sphs[i].ball->ball_type->radius ),
		   1.0 / sim( sphs[sph].ball->ball_type->weight, sphs[sph].ball->ball_type->radius ),
		   sphs[i].ball->ball_type->radius, sphs[sph].ball->ball_type->radius,
		   sphs[sph].ball->ball_type->f1, sphs[sph].ball->ball_type->f3, 1 );

  prec_back_shock [i] = PREC_BACK_SHOCK;
  prec_back_shock [sph] = PREC_BACK_SHOCK;

  return 1;
}

/***********/

void planeNormal( dbl **v, dbl *n )
{
  dbl v1[3], v2[3];

  v2[0] = (v[3])[0] - (v[0])[0];
  v2[1] = (v[3])[1] - (v[0])[1];
  v2[2] = (v[3])[2] - (v[0])[2];

  v1[0] = (v[1])[0] - (v[0])[0];
  v1[1] = (v[1])[1] - (v[0])[1];
  v1[2] = (v[1])[2] - (v[0])[2];
  
  vec_cross( n, v2, v1 );

  vec_norm( n, n );
}

/***********/
/*
void sphere2plane( Sphere *sph, dbl *old_pos, DedecekHribecek *hrib, int hrib_line_len, int ns,
		   int ids, dbl **v, int nv, int idp )
{
  dbl 	tmp[3], tmp1[3], n[3], tmp2[3];
  dbl	t = 10.0;

  // hrib_line_len * ids -> beginig of DH sphere ids, + ids -> bigining of polys, + idp -> actual poly
  PernikovaChaloupka *ch = &hrib[hrib_line_len * ids + idp + ns].ch;

  planeNormal( v, n );

  switch( colidePlaneSphere( v, nv, n, sph->p, old_pos, sph->ball->ball_type->radius, tmp, &t ) )
    {
    case 1:
      cvec( ch->karkulka, n );
      ch->t = t;
      ch->type = 1;
      break;

    case 2:
    // now only one place colision
      vec_sub( tmp2, sph->p, old_pos );
      get_point( tmp1, old_pos, tmp2, t );
      vec_sub( tmp1, tmp, tmp1 );
      if ( vec_size2( sph->v ) > EPSILON && vec_size2( tmp1 ) > EPSILON )
	ch->coef = vec_angle( sph->v, tmp1 );
      else
	ch->coef = 0.0;
      vec_cross( ch->vochomurka, sph->v, tmp1 );
      // cvec( ch->vochomurka, tmp );
      // vec_cross( ch->vochomurka, tmp, tmp1 );
      // cvec( tmp, sph->v );
      // rebound( n, tmp );
      // vec_add( ch->kremilek, ch->kremilek, tmp );
      vec_neg( tmp1 );
      cvec( ch->kremilek, tmp1 );
      ch->t = t;
      ch->type = 2;
      break;

    }
}
*/

void sphere2plane( Sphere *sph, dbl *old_pos, DedecekHribecek *hrib, int hrib_line_len, int ns,
		   int ids, dbl **v, int nv, int idp )
{
  dbl 	tmp[3], tmp1[3], n[3], tmp2[3];
  dbl	t = 10.0;

  /* hrib_line_len * ids -> beginig of DH sphere ids, + ids -> bigining of polys, + idp -> actual poly */
  PernikovaChaloupka *ch = &hrib[hrib_line_len * ids + hrib_lens [ids]].ch;

  planeNormal( v, n );

  switch( colidePlaneSphere( v, nv, n, sph->p, old_pos, sph->ball->ball_type->radius, tmp, &t ) )
    {
    case 1:
      cvec( ch->n, n );
      ch->t = CT(t);
      ch->type = 1;
      isected [ids] = 1;
      update_BA (ids, CT(t), hrib_line_len, ns + idp);
      break;

    case 2:
      vec_sub( tmp2, sph->p, old_pos );
      get_point( tmp1, old_pos, tmp2, t );
      vec_sub( ch->r, tmp, tmp1 );
      cvec( ch->n, n );
      ch->t = CT(t);
      ch->type = 2;
      isected [ids] = 1;
      update_BA (ids, CT(t), hrib_line_len, ns + idp);
      break;

    }
}

/***********/

/*
void sphere2plane_end_preproceed( Sphere *sph, PernikovaChaloupka *ch, dbl **datas, int ids, int idp )
{
  dbl	tmp[3], tmp1[3];
  dbl	v;
  dbl	*n, *nn;

  n = datas [0];
  nn = datas [1];

  v = vec_size( sph->v );

  vec_add( n, n, ch->karkulka );
  cvec( tmp1, ch->kremilek );
  vec_norm( tmp1, tmp1 );
  vec_mul_c( tmp1, tmp1, 2*(cos( ch->coef )) );
  vec_add( nn, nn, tmp1 );
  //  vec_add( sph->v, sph->v, tmp1 );
  
  cvec( tmp, ch->vochomurka );
  vec_norm( tmp, tmp );
  vec_mul_c( tmp, tmp, (1-sph->ball->ball_type->f5)*sin( ch->coef )*v );
  vec_add( sph->omega, sph->omega, tmp );
}
*/

void sphere2plane_end_preproceed( Sphere *sph, PernikovaChaloupka *ch, dbl **datas, int ids, int idp )
{
  dbl	tmp[3], tmp1[3];
  dbl	v;
  dbl	*n, *nn, *rr, *e;

  n = datas [0];
  nn = datas [1];
  rr = datas [2];
  e = datas [3];

  cvec( tmp, null );

  switch( ch->type )
    {
    case 1:
      *e = sph->ball->ball_type->f5;
      vec_add( n, n, ch->n );
      break;

    case 2:
      prec_back_shock [ids] = PREC_BACK_SHOCK_EDGE;
      vec_sub( tmp1, rr, ch->r );
      if ( vec_size2( tmp1 ) > EPSILON )
	{
	  if ( vec_size2( nn ) > EPSILON )
	    {
#ifdef DEBUG
	      printf("type 2\n");
#endif
	      rigid_collision( tmp, sph->v, tmp, sph->omega, nn, rr, 0.0, 
			       1.0 / sph->ball->ball_type->weight, 0.0, 
			       1.0 / sim( sph->ball->ball_type->weight, sph->ball->ball_type->radius ),
			       sph->ball->ball_type->radius, sph->ball->ball_type->radius,
			       sph->ball->ball_type->f9, sph->ball->ball_type->f8, 1 );
	    }
	  cvec( rr, ch->r );
	  cvec( nn, ch->r );
	  /*	  vec_neg( nn ); */
	}
      break;

    case 3:
	    /*      cvec( sph->p, ch->r ); */
      sph->v[Z] = 0.0;
      break;

    case 4:
      *e = sph->ball->ball_type->f9;
      vec_add( n, n, ch->n );
      break;
    default:
#ifdef DEBUG
      printf ("KIIIIIIIIIIIIIKS BUG - this is wrong place!\n");
#endif
    }
}

/***********/

/*
void sphere2plane_end_proceed( Sphere *sph, dbl **datas, int table_col, dbl t )
{
  dbl	tmp[3], tmp1[3];
  dbl	v;
  dbl	*n, *nn;

  n = datas [0];
  nn = datas [1];

  v = vec_size( sph->v );

  vec_norm( n, n );
  // vec_add( tmp, n, nn );
  // vec_norm( nn, nn );
  vec_norm( sph->v, sph->v );
  vec_add( sph->v, sph->v, nn );
  vec_norm( sph->v, sph->v );
  cvec( tmp, n );
  vec_norm( tmp, tmp );
  if ( vec_size( tmp ) > EPSILON )
    rebound( tmp, sph->v );
  if ( table_col )
    vec_mul_c( sph->v, sph->v, (1-sph->ball->ball_type->f9) );
  else
    vec_mul_c( sph->v, sph->v, (1-sph->ball->ball_type->f5) );

    vec_mul_c( sph->omega, sph->omega, (1-sph->ball->ball_type->f8) );

  vec_mul_c( sph->v, sph->v, v );

  if ( (vec_size( sph->omega ) > EPSILON) && (v > EPSILON) )
    {
      cvec( tmp, sph->omega );
      vec_cross( tmp, tmp, sph->v );
      vec_mul_c( tmp, tmp, 1.0 / v  );
      tmp[Z] = 0.0;
      vec_mul_c( tmp, tmp, t*sph->ball->ball_type->f6*sph->ball->ball_type->f8 );
      vec_add( sph->v, sph->v, tmp );
    }
}
*/

void sphere2plane_end_proceed( Sphere *sph, dbl **datas, int table_col, dbl t )
{
  dbl	tmp[3], tmp1[3];
  dbl	v;
  dbl	*n, *nn, *rr, e;
  
  n = datas [0];
  nn = datas [1];
  rr = datas [2];
  e = *datas [3];

  cvec( tmp, null );

  if ( vec_size2( n ) > EPSILON )
    {
      if ( table_col )
	rigid_collision( tmp, sph->v, tmp, sph->omega, n, n, 0.0, 
			 1.0 / sph->ball->ball_type->weight, 0.0, 
			 1.0 / sim( sph->ball->ball_type->weight, sph->ball->ball_type->radius ),
			 sph->ball->ball_type->radius, sph->ball->ball_type->radius,
			 sph->ball->ball_type->f9, sph->ball->ball_type->f8, 1 );
      else
	rigid_collision( tmp, sph->v, tmp, sph->omega, n, n, 0.0, 
			 1.0 / sph->ball->ball_type->weight, 0.0, 
			 1.0 / sim( sph->ball->ball_type->weight, sph->ball->ball_type->radius ),
			 sph->ball->ball_type->radius, sph->ball->ball_type->radius,
			 e, sph->ball->ball_type->f8, 1 );
    }
  else
    {
      cvec( tmp, null );

      if ( vec_size2( nn ) > EPSILON )
	{
#ifdef DEBUG
	  printf("type 2\n");
#endif
	  rigid_collision( tmp, sph->v, tmp, sph->omega, nn, rr, 0.0, 
			   1.0 / sph->ball->ball_type->weight, 0.0, 
			   1.0 / sim( sph->ball->ball_type->weight, sph->ball->ball_type->radius ),
			   sph->ball->ball_type->radius, sph->ball->ball_type->radius,
			   sph->ball->ball_type->f9, sph->ball->ball_type->f8, 1 );
	}
    }
}

/***********/

int ball_out_of_table( Sphere *sph, int id )
{
  int	i, intr;
  dbl	*pta, *ptb, *p;

  if ( id == -1 )
    {
      if ( ball_in_hole (sph) )
	return 0;
    }
  else
    if ( chck_in_holes( sph, id ) )
      return 0;

  p = sph->p;
  pta = pool_table.vertices + 2 * (pool_table.nof_vertices - 1);
  ptb = pool_table.vertices;

  for( intr = i = 0; i < pool_table.nof_vertices; i++, ptb += 2 )
    {
      if ( ((pta[X] < p[X]) && (ptb[X] >= p[X])) || ((pta[X] >= p[X]) && (ptb[X] < p[X])) )
	{
	  if ( pta[Y] > p[Y] && ptb[Y] <= p[Y] )
	    {
	      if ( ((ptb[X]-pta[X])*(p[Y]-pta[Y])/(ptb[Y]-pta[Y])+pta[X]) > p[X] )
		intr++;
	    }
	  else
	    if ( pta[Y] < p[Y] && ptb[Y] >= p[Y] )
	      if ( ((ptb[X]-pta[X])*(p[Y]-pta[Y])/(ptb[Y]-pta[Y])+pta[X]) > p[X] )
		intr++;
	}	
      else
	if ( pta[X] >= p[X] )
	  {
	    if ( pta[Y] > p[Y] && ptb[Y] <= p[Y] )
	      intr++;
	    else
	      if ( pta[Y] < p[Y] && ptb[Y] >= p[Y] )
		intr++;
	  }
      pta = ptb;
    }
  
  if ( (intr&1) == 1 )
    return 0;
  
  if (sph->p[Z] <= FLOOR_DEPTH)
    return 2;
  
  return 1;
}

/***********/

/*
 * ?? TODO: mozna zjistovat trosicku lepeji na podlozce 
 *	    aby se dali delat odrazi i od horejsich vrstev :)
 */

/*
void uno_stepo( Sphere *sph, dbl t )
{
  dbl 	tmp[3], tmp1[3];
  dbl	v;

  cvec( tmp1, null );

  v = vec_size( sph->v );

  if ( chck_hole_bottom( sph, -1, 0 ) || (ball_out_of_table( sph ) == 2) )
    return;

  if ( chck_on_table( sph ) && v > EPSILON && vec_size(sph->omega) > EPSILON )
    {
      cvec( tmp1, sph->omega );
      vec_cross( tmp1, tmp1, sph->v );
      vec_mul_c( tmp1, tmp1, 1.0 / v  );
      v = 1 - (t * tmp1[Z] * sph->ball->ball_type->f2) / v;
      tmp1[Z] = 0;

    }
  else
    v = 1.0;

  vec_mul_c( tmp1, tmp1, t*sph->ball->ball_type->f6 );
  vec_mul_c( tmp, sph->v, t*v );
  vec_add( tmp, tmp, tmp1 );
  vec_add( sph->p, sph->p, tmp );
}
*/

void uno_stepo( Sphere *sph, int ids, dbl t )
{
  dbl 	tmp[3];
  /*  dbl	v; */

  /*  v = vec_size( sph->v ); */

  if ( vs [ids] < EPSILON )
    return;

  if ( chck_hole_bottom( sph, ids, 0 ) || (ball_out_of_table( sph, ids ) == 2) )
    return;

  vec_mul_c( tmp, sph->v, t );
  vec_add( sph->p, sph->p, tmp );
}

/***********/
/*
void fatal_proceed( Sphere *sph, dbl t, dbl t1 )
{
  dbl 	tmp[3], v, vv, n[3], q[4];
  dbl 	omega = vec_size( sph->omega );

  cvec( tmp, sph->omega );
  vv = vec_size( tmp );
  vec_norm( tmp, tmp );

  if ( vv > EPSILON )
    {
      quat_aaxis( q, t * vv, tmp );
      quat_mul( sph->q, q, sph->q );
    }

  n[0] = 0;
  n[1] = 0;
  n[2] = 1;

  v = vec_size( sph->v );

  vec_cross( tmp, n, sph->v );
  vec_norm( tmp, tmp );
  //  vec_mul_c( tmp, tmp, v / sph->ball->ball_type->radius );
  vec_norm( tmp, tmp );

  if ( v > EPSILON )
    {
      quat_aaxis( q, t * v / sph->ball->ball_type->radius, tmp );
      quat_mul( sph->q, q, sph->q );
    }

  quat_norm (sph -> q, sph -> q);

  if ( chck_on_table( sph ) )
    {
      vec_mul_c( sph->v, sph->v, (1-t*sph->ball->ball_type->f7) );
      if ( (vec_size( sph->omega ) > EPSILON) && (v > EPSILON) )
	{
	  cvec( tmp, sph->omega );
	  vec_cross( tmp, tmp, sph->v );
	  vec_mul_c( tmp, tmp, 1.0 / v  );
	  v = 1 - (t * tmp[Z] * sph->ball->ball_type->f2) / v;
	  vec_mul_c( sph->v, sph->v, v );
	  tmp[Z] = 0.0;
	  vec_mul_c( tmp, tmp, t*sph->ball->ball_type->f6 );
	  vec_add( sph->v, sph->v, tmp );
	}
	vec_mul_c( sph->omega, sph->omega, (1-t*sph->ball->ball_type->f2 ) ); // FIXME: mozna f8 :)
    }
  else
    if ( !(chck_hole_bottom( sph, -1, 0 ) || (ball_out_of_table( sph ) == 2)) )
      {
	cvec( tmp, null);
	tmp[Z] = -G;
	vec_mul_c( tmp, tmp, t1 );
	vec_add( sph->v, sph->v, tmp );
      }
}
*/
/**********/

static int is_copl( dbl *v, dbl *om )
{
  dbl	tmp[3];

  vec_cross( tmp, v, om );

  return (vec_size2( tmp ) < EPSILON);
}

/**********/

void fatal_proceed( Sphere *sph, int ids, dbl t, dbl t1 )
{
  dbl 	tmp[3], v, vv, n[3], q[4], mg, a, r, aa, nn;
  int	res, res1;
  
  FyjikSkin	*skin = sph->ball->ball_type;

  r = skin->radius;

  cvec( tmp, sph->omega );
  vv = vec_size2( tmp );
  v = vs [ids];

  res = (res1 = chck_on_table( sph, ids )) || chck_hole_bottom( sph, ids, 0 ) || (ball_out_of_table( sph, ids ) == 2);

  if ( v < EPSILON && vv < EPSILON && res )
    return;

  vv = sqrt (vv);
  vec_norm( tmp, tmp );

  if ( vv > EPSILON )
    {
      quat_aaxis( q, t * vv, tmp );
      quat_mul( sph->q, q, sph->q );
      quat_norm (sph -> q, sph -> q);
    }	

  if ( res1 )
    {
      n [Y] = - sph->omega[X] * r;
      n [X] = sph->omega[Y] * r;
      n [Z] = 0.0;

      nn = vec_size( n );

      vec_cross( tmp, n, sph->v );

      a = sqr( v ) + ((aa=vec_dot( sph->v, n )) >= 0.0 ? (-1):(1))*sqr( nn );

      if ( aa > EPSILON )
	aa /= v * nn;

      mg = vec_size2( tmp );

      if ( fabs( a ) < EPSILON && mg < EPSILON )
	res = 0;	/* equilibrium */
      else
	if ( a > 0.0 || ( aa > 0.0 && ( aa * nn < v ) ) )
	  res = 1;	/* speed > angular speed */
        else
	  res = -1;	/* speed < angular speed */

      a = mg;

      if ( v > EPSILON )
	if ( !res )
	  mg = t * skin->g * skin->f7 / v;
        else
	  mg = t * skin->g * skin->f2 / v;
      else
	mg = 0;

      if ( mg > 1.0 )
	mg = 1.0;

      vec_mul_c( sph->v, sph->v, (1 - mg) );

      cvec( tmp, sph->v );
      tmp [Z] = 0.0;

      vec_norm( tmp, tmp );
      
      if ( vv > EPSILON && (a > EPSILON || res == -1) )
	{
	  vec_norm( n, n );
	  vec_add( tmp, tmp, n );
	  vec_mul_c (n, n, skin->f2 * skin->g * t );
	  vec_add( sph->v, sph->v, n );
	}
      
      /*
      if ( sph->omega[Z] > EPSILON && (mg=(t * skin->g * skin->f7 * 5.0 / (2 * r * sph->omega [Z]))) <= 1.0 )
	sph->omega [Z] *= 1.0 - mg;
      else
	sph->omega [Z] = 0.0;
      */

      if ( (fabs (a) > EPSILON) || ( !((res == 1) && (a > 1.0 - EPSILON)) ) ) 
	{
	  mg = skin->f7 * skin->g * t * 5.0 / (2 * r * vv);
	  if ( mg > 1.0 )
	    vec_mul_c( sph->omega, sph->omega, 0.0 );
	  else
	    vec_mul_c( sph->omega, sph->omega, 1.0 - mg );
	}

      if ( res == 1 )
	{
	  mg = skin->f2 * skin->g * t;
	  vec_norm( tmp, tmp );
	  vec_mul_c( tmp, tmp, mg );
	  SWAP( tmp[X], tmp[Y], a );
	  tmp[Y] = -tmp[Y];
	  vec_mul_c( tmp, tmp, 5.0 / (2 * r) );
	  vec_sub( sph->omega, sph->omega, tmp );
	}
    }
  else
    if ( !res )
      {
	cvec( tmp, null);
	tmp[Z] = -skin->g;
	vec_mul_c( tmp, tmp, t );
	vec_add( sph->v, sph->v, tmp );
      }
}

/***********/

void save_ball_state( Sphere *sph, int n, dbl *state )
{
  int 	i;

  for( i = 0 ; i < n ; i++, state += 3 )
    {
      cvec( state, sph[i].p );	
    }
}

/***********/

void restore_ball_state( Sphere *sph, int n, dbl *state )
{
  int 	i;

  for( i = 0 ; i < nof_crush ; i++ )
    if ( !pocketed [ crushed [i]] )
      {
	cvec( sph[crushed [i]].p, state + 3 * crushed [i] );
      }
}

/***********/

void check_and_proc_cols( Sphere *sphs, dbl *old_pos, DedecekHribecek *hrib,
			  int n, int i, dbl **polys, int *counts, int nofp, 
			  int nofh, TableHole *holes, dbl *cutouts,
			  dbl dt )
{
  int 	r = 0;
  int	j, k;
  dbl	tabl[3*4], tmp[3], g, tmp1 [3];
  dbl	*tbl[4] = { tabl, tabl + 3, tabl + 6, tabl + 9 };

  if ( pocketed [i] )
    return;

  sphere2sphere( sphs, old_pos, n, i, hrib, n + nofp + nofh + DED_HRIB );
  
  for( k = j = 0; j < nofp; k += counts[j++] )
    if (bb_has_ints(bb_balls + 6*i, bb_edges + 6*j) )
      sphere2plane( sphs + i, old_pos + 3*i, hrib, n + nofp + nofh + DED_HRIB, 
		    n, i, polys + k, counts[j], j );

  col_cue2sphere( my_pool_cue.adjusted_top, my_pool_cue.adjusted_bottom, 
		  my_pool_cue.cue_type->top_width, my_pool_cue.cue_type->bottom_width,
		  my_pool_cue.up, my_pool_cue.right, 
		  sphs, i, n, old_pos, nofp + nofh, hrib );

  sphere2holes( sphs + i, i, old_pos + 3 * i, hrib, 
  		n + nofp + nofh + DED_HRIB, n, nofp, nofh, holes, cutouts );
 

  cvec( tmp1, sphs [i].p );
  cvec( sphs [i].p, old_pos + 3 * i );
  if ( (r = ball_out_of_table( sphs + i, i )) )
    {
      cvec( sphs [i].p, tmp1 );
      if ( r == 2 )
	{
	  cvec( sphs[i].v, null );
	  cvec( sphs[i].omega, null );
	  sphs[i].p[Z] = FLOOR_DEPTH;
	  old_pos [3 * i + Z] = FLOOR_DEPTH;
	  ball_out_handler( i, &bum );
	}
      return;
    }
  
  g = sphs [i].ball->ball_type->g;

  if ( (fabs(sphs[i].v[Z])*sphs[i].ball->ball_type->f9) < g*dt && chck_on_table( sphs + i, i ) )
    {
      cvec( sphs [i].p, tmp1 );
      sphs[i].v[Z] = 0;
      sphs[i].p[Z] = sphs[i].ball->ball_type->radius;
      old_pos [3 * i + Z] = sphs[i].ball->ball_type->radius;
      return;
    }  

  if ( chck_near_table( sphs + i, i ) )
    {
      vec_sub( tmp, sphs[i].p, old_pos + 3*i );
      if ( vec_size( tmp ) < DONT_JUMP )
	{
	  cvec( sphs [i].p, tmp1 );
	  sphs[i].v[Z] = 0.0;
	  sphs[i].p[Z] = sphs[i].ball->ball_type->radius;
	  old_pos [3 * i + Z] = sphs[i].ball->ball_type->radius;
	  return;
	}
    }

  if ( chck_hole_bottom( sphs + i, i, 1 ) )
    {
	    /* cvec( sphs [i].p, tmp1 ); */
      cvec( sphs[i].v, null );
      cvec( sphs[i].omega, null );
      return;
    }

  if ( chck_in_holes( sphs + i, i ) )
    {
      cvec( sphs [i].p, tmp1 );
      return;
    }
  
  cvec( tmp, old_pos + 3*i );
  
  tmp[Z] = 0.0;
  tmp[X] -= 10.0;
  tmp[Y] -= 10.0;
  cvec( tabl, tmp );
      
  tmp[X] += 20.0;
  cvec( tabl + 3, tmp );
      
  tmp[Y] += 20.0;
  cvec( tabl + 6, tmp );
      
  tmp[X] -= 20.0;
  cvec( tabl + 9, tmp );
     
  cvec( sphs [i].p, tmp1 ); 
  sphere2plane( sphs + i, old_pos + 3*i, hrib, n + nofp + nofh + DED_HRIB, n, i, tbl, 4, nofp + nofh );
}

/***********/

/*
void babovicka_on_pisecek( DedecekHribecek *hrib, int ns, int np, int nh, int k )
{
  int	i, j;

  if (pisecked [k])
    return;

  pisecked [k] = 1;

  k *= (ns + np + nh + DED_HRIB);
  
  for( i = 0; i < ns; i++ )
    {
      hrib [k + i].b.t = 10.0;
      //	  cvec( hrib [k + i].b.n, null );
    }
  
  for(; i < (ns + np + nh + DED_HRIB); i ++)
    {
      //	  hrib [k + i].ch = chanull;
      hrib [k + i].ch.t = 10.0;
    }
}
*/

/***********/

static inline void pisecek_prepare( Sphere *sphs, int ns )
{
  int	i;

  for (nof_crush = i = 0; i < ns; i ++)
    {
      hrib_lens [i] = 0;
      BA [i].t = 10.0;
      BA [i].idx = -1;
      isected [i] = 0;
      if ( vec_size2 (sphs[i].v) > EPSILON )
	crushed [nof_crush ++] = i;
      else
	vs [i] = 0.0;
    }
}

/***********/

void prepare_bajaja( Bajaja *ba, dbl *tints, int n )
{
  int	i;

  for( i = 0; i < n; i ++ )
    {
      ba [i].idx = -1;
      tints [i] = -1.0;
    }
}

/***********/

int find_min( DedecekHribecek *hrib, int ns, int np, int nh, Bajaja *ba )
{
  int	i, idx = -2, r = 0;
  dbl	min = 2.0;

  for( i = 0; i < ns; i ++ )
    if ( hrib [i].b.t < min )
      {
	min = hrib [i].b.t;
	idx = i;
      }
  
  for( ; i < (ns + np + nh + DED_HRIB); i ++ )
    if ( hrib [i].ch.t < min )
      {
	min = hrib [i].ch.t;
	idx = i;
      }

  ba -> t = min;
  ba -> idx = idx;

  return i;
}

/***********/

void clear_other_cols( DedecekHribecek *hrib, int idb, int ns, int np, int nh, dbl t, Bajaja *ba )
{
  int	i, j, k;

  for( i = 0, k = idb * (ns + np + nh + DED_HRIB);
       i < ns;
       i ++, k ++ )
    {
      if ( i != idb && hrib [k].b.t > (t + ACCUR) )
	{
	  hrib [k].b.t = 10.0;
	  hrib [i * (ns + np + nh + DED_HRIB) + idb].b.t = 10.0;
	  ba [i].idx = -1;
	}
    }

  for( ; i < (ns + np + nh +DED_HRIB); i ++, k ++ )
    if ( hrib [k].ch.t > (t + ACCUR) )
      hrib [k].ch.t = 10.0;
}

/***********/

void compute_lin_interp( dbl *res, dbl *old, dbl *new, dbl t_ints )
{
  dbl 	tmp [3];

  vec_sub( tmp, new, old );
  res [0] = old [0] + tmp [0] * t_ints;
  res [1] = old [1] + tmp [1] * t_ints;
  res [2] = old [2] + tmp [2] * t_ints;
}

/***********/

/*
void compute_proceed_time( Sphere *sphs, dbl *old_pos, int ns, dbl *ts, dbl *t_ints )
{
  int	i;
  dbl	t;
  dbl	tmp [3];

  for ( i = 0; i < ns; i ++ )
    {
      if ( t_ints [i] < 0.0 )
	continue;
      // t = (t_ints [i] > prec_back_shock [i]) ? (t_ints [i] - prec_back_shock [i]) : (t_ints [i]/ 2.0);
      t = t_ints [i];
      if ( t_ints [i] > BALL_MINICUK )
      	t = BALL_MINICUK;
      else
      	continue;
      //      vec_sub( tmp, sphs [i].p, old_pos + 3*i );
      //      old_pos [3 * i + 0] += t_ints [i] * tmp [0];
      //      old_pos [3 * i + 1] += t_ints [i] * tmp [1];
      //      old_pos [3 * i + 2] += t_ints [i] * tmp [2];
      compute_lin_interp( old_pos + 3 * i, old_pos + 3 * i, sphs [i].p, t);

      ts [i] -= ts [i] * t;
      //      if ( ts [i] < EPSILON )
      //      	ts [i] = 0.0;;
    }
}
*/

void compute_proceed_time( Sphere *sphs, dbl *old_pos, int ns )
{
  int	i;
  dbl	t;

  for ( i = 0; i < ns; i ++ )
    if ( vs [i] > EPSILON )
      compute_lin_interp( old_pos + 3 * i, old_pos + 3 * i, 
			  sphs [i].p, NT(T_INTS) );

  cum_scl = (1.0 - T_INTS);
  cum_tr = T_INTS;
  ts -= ts * T_INTS;
}

/***********/

static dbl	N [3], NN [3], RR [3], E;
static dbl*	TMP [4];	

static void *s2p_alc(Sphere *s)
{
  TMP [0] = N;
  TMP [1] = NN;
  TMP [2] = RR;
  TMP [3] = &E;

  cvec( N, null );
  cvec( NN, null );
  cvec( RR, null );

  return TMP;
}

/***********/

static inline int has_same_state( Sphere *sphs, int i, int j, int ns, dbl *p1, dbl *p2 )
{
  dbl	tmp [3];
  int	k = j + i*ns;

  vec_sub( tmp, problemal_cols [k].v1, sphs[i].v );
  if ( vec_size2( tmp ) > EPSILON )
    return 0;

  vec_sub( tmp, problemal_cols [k].v2, sphs[j].v );
  if ( vec_size2( tmp ) > EPSILON )
    return 0;
  /*
  vec_sub( tmp, problemal_cols [k].p1, p1 );
  if ( vec_size2( tmp ) > EPSILON )
    return 0;

  vec_sub( tmp, problemal_cols [k].p2, p2 );
  if ( vec_size2( tmp ) > EPSILON )
    return 0;
  */
#ifdef DEBUG
  printf("HUAHUAHUA %d %d - %f\n", i, j, bum.time);
#endif

  return 1;
}

/***********/

static inline void set_cols_state( Sphere *sphs, int i, int j, int ns, dbl *p1, dbl *p2 )
{
  int	k = j + i*ns, l = i + j*ns;

  problemal_cols [k].t = bum.time;
  /*  cvec( problemal_cols [k].p1, p1 ); */
  /*  cvec( problemal_cols [k].p2, p2 ); */
  cvec( problemal_cols [k].v1, sphs[i].v );
  cvec( problemal_cols [k].v2, sphs[j].v );
  problemal_cols [l].t = bum.time;
  /*  cvec( problemal_cols [l].p1, p2 ); */
  /*  cvec( problemal_cols [l].p2, p1 ); */
  cvec( problemal_cols [l].v1, sphs[j].v );
  cvec( problemal_cols [l].v2, sphs[i].v );
} 

/**********/

static inline int has_same_state1( Sphere *sph, int i, int j, int ns, dbl *p1 )
{
  dbl	tmp [3];
  int	k = j + i*ns;

  vec_sub( tmp, problemal_cols [k].v1, sph->v );
  if ( vec_size2( tmp ) > EPSILON )
    return 0;

  /*
  vec_sub( tmp, problemal_cols [k].p1, p1 );
  if ( vec_size2( tmp ) > EPSILON )
    return 0;

  vec_sub( tmp, problemal_cols [k].p2, p2 );
  if ( vec_size2( tmp ) > EPSILON )
    return 0;
  */
#ifdef DEBUG
  printf("HUAHUAHUA s2p %d %d - %f\n", i, j, bum.time);
#endif

  return 1;
}

/***********/

static inline void set_cols_state1( Sphere *sph, int i, int j, int ns, dbl *p1 )
{
  int	k = j + i*ns;

  problemal_cols [k].t = bum.time;
  /*  cvec( problemal_cols [k].p1, p1 ); */
  /*  cvec( problemal_cols [k].p2, p2 ); */
  cvec( problemal_cols [k].v1, sph->v );
  if ( (fabs(sph->v[Z])*sph->ball->ball_type->f9) < 
       sph->ball->ball_type->g*DT && chck_on_table( sph, i ) )
    problemal_cols [k].v1[Z] = 0.0;
} 

/***********/

static inline void clear_ball (int i, int j, int len)
{
  int	k;

  for (k = 0; k < hrib_lens [i]; k ++)
    if ( hrib_qu [i * len + k] == j )
      {
	hrib_qu [i * len + k] = hrib_qu [i * len + hrib_lens [i] - 1];
	hrib [i * len + k] = hrib [i * len + hrib_lens [i] - 1];
	hrib_lens [i] --;
	return;
      }
}

/***********/

static int	int_ids [1024];
static int	nofid;
static int	break_ids [1024];
static int	nofbr;
static int	int_balls [1024];
static int	nofba;

static inline void ugly_normal_hack (DedecekHribecek *hrib, int j, int len, dbl t, int ns)
{
  int	i, k;
  dbl	tmp [3];

  for (nofba = nofid = i = 0; i < hrib_lens [j]; i++)
    if ( hrib_qu [len * j + i] < ns )
      {
	if ( fabs( hrib[i].b.t - t ) < ACCUR )
	  int_balls [nofba++] = i;
	clear_ball (hrib_qu [len * j + i], j, len);
      }
    else
      if ( fabs( hrib[i].ch.t - t ) < ACCUR )
	int_ids [nofid++] = i;

  for (nofbr = i = 0; i < nofid; i++)
    for (k = i + 1; k < nofid; k++)
      {
	vec_sub (tmp, hrib [int_ids [i]].ch.n, hrib [int_ids [k]].ch.n);
	if ( vec_size2 (tmp) < EPSILON )
	  {
#ifdef DEBUG
	    printf ("BUBAK comes from ??? %d - %d\n", i, k);
#endif
	    break_ids [nofbr++] = k;
	    hrib [int_ids [k]] = hrib [int_ids [--nofid]];
	    hrib [int_ids [nofid]].ch.t  = 10.0;
	    int_ids [k--] = int_ids [nofid];
	  }
      }
}

/***********/

static inline dbl find_nearest_collision (int ns)
{
  int	i;
  dbl	t = 10.0;

  for (i = 0; i < ns; i ++)
    if ( t > BA [i].t )
      t = BA [i].t;

  return t;
}

/***********/

/*
  for all balls do
   if has done (bajaja[i].idx == -2)
     don't proceed it
   
   if ball hasn't valid minimum (bajaja[i].idx == -1)  
     compute it or if it hasn't any intersection finish his proceed

   for all potencial intersectant ( something(ball,edge,table) which may be intersect actual ball )
    if it is ball get his actual minimum and compare it which minimum actual ball
     if both minima is equals 
      compute rebound and other intersection behind this intersection remove from 
      intersection list *.t > min -> *.t = 10.0 

    if it is edge compare intersection t with actual minima and
     if it is equal proceed rebound and other inters behind this intersection ( *.t > min )
      remove from inters list

   and at the end do proceeding time spending before intersection 
      
*/

/* !!!!!!!!!! FIXME: funkce pro handlovani fyziky  !!!!!!!!!! */

/*
int find_true_cols(Sphere *sphs, dbl *old_pos, DedecekHribecek *hrib, int ns, int np, int nh, dbl *ts)
{
int		i, j, k, ints = 1, cflags = 0, cfinal = 0, len, l;
  dbl		t;
  dbl		p1[3], p2[3];
  void*		datas;
  
  prepare_bajaja( BA, T_INTS, ns );
  
  len = ns + np + nh + DED_HRIB;

  while( ints )
  {
  ints = 0;
      
  for( i = 0, k = 0; i < ns; i ++, k += len )
	{
	//	  if ( vec_size( sphs[i].v ) < EPSILON )
	//	    continue;

	  if ( !isected [i] )
	    continue;

	    //	  if ( ba [i].idx == -2 )
	    //	    continue;

	  if ( BA [i].idx == -1 )
	    if ( find_min( hrib + k, ns, np, nh, BA + i ) == -2 )
	      continue;
	  
	  for( nofba = j = 0; j < ns; j ++ )
	    {
	      if ( j == i )
		continue;
	      if ( BA [j].idx == -1 )
		find_min( hrib + j * len, ns, np, nh, BA + j );

	      t = hrib[k + j].b.t;
	      if ( fabs(t - BA [i].t) < ACCUR  && fabs(t - BA [j].t) < ACCUR ) 
		{
		  // FIXME: nepocitat dvakrat , ale jen jednou !!! sebrat z comp_proc_time
		  compute_lin_interp( p1, old_pos + 3*i, sphs[i].p, BA [i].t );
		  compute_lin_interp( p2, old_pos + 3*j, sphs[j].p, BA [j].t );
		  
		  if ( problemal_cols [j + i*len].t == bum.time &&
		       has_same_state( sphs, i, j, len, p1, p2 ) )
		    continue;

		  fyjik_s2s_col_dyn ( sphs, i, j, &hrib[k + j].b );
		  sphere_sphere_handler( i, j, p1, p2, &bum );
		  
		  int_balls [nofba++] = j;

		  if ( !cflags )
		    clear_other_cols( hrib, i, ns, np, nh, t, BA );
		  clear_other_cols( hrib, j, ns, np, nh, t, BA );

		  cflags = 1;
		  hrib [k + j].b.t = hrib [j * len + i].b.t = 10.0;
		  T_INTS [i] = T_INTS [j] = t;
		}
	    }
	  
	  ugly_normal_hack (hrib + k, j, len, BA [i].t); 

	  datas = fyjik_s2p_alc (sphs + i);

	  if ( nofid )
	    for (l = 0; l < nofid; l ++)
	      {
	      j = int_ids [l];
	      compute_lin_interp( p1, old_pos + 3*i, sphs[i].p, BA [i].t );

		if ( problemal_cols [j + i*len].t == bum.time && 
		     (j < (ns + np) || j > (ns + np + nh - 1)) &&
		     has_same_state1( sphs + i, i, j, len, p1 ) )
		  continue;
		
		  
		  sphere_plane_handler( i, j - ns, p1, &bum );
		  fyjik_s2p_col_dyn_preproc ( sphs + i, &hrib [k + j].ch, datas, i, j - ns );
		
		if ( !cflags )
		  clear_other_cols( hrib, i, ns, np, nh, BA [i].t, BA );
		cflags |= 2;
		hrib [k + j].ch.t = 10.0;
		T_INTS [i] = BA [i].t;
		if ( j == (ns + np + nh) )
		  cflags |= 4;
	      }
	
	  if ( cflags & 2 )
	    fyjik_s2p_col_dyn_proc ( sphs + i, datas, cflags & 4, T_INTS [i] * ts [i] );

	  if ( cflags )
	    {
	      for (l = 0; l < nofba; l ++)
		set_cols_state( sphs, i, int_balls [l], len, p1, p2 );
	      for (l = 0; l < nofid; l ++)
		set_cols_state1( sphs + i, i, int_ids [l], len, p1 );
	      for (l = 0; l < nofbr; l ++)
		set_cols_state1( sphs + i, i, break_ids [l], len, p1 );

	      ints = 1;
	      BA [i].idx = -2;
	      cflags = 0;
	    }
	  
	}

      cfinal |= ints;
    }

  compute_proceed_time( sphs, old_pos, ns, ts, T_INTS );

  return cfinal;
}
*/

int find_true_cols(Sphere *sphs, dbl *old_pos, DedecekHribecek *hrib, int ns, int np, int nh)
{
  int		i, j, k, ints = 0, cflags = 0, len, l, m;
  dbl		t;
  dbl		p1[3], p2[3];
  void*		datas;

  len = ns + np + nh + DED_HRIB;

  ints = 0;

  T_INTS = find_nearest_collision (ns);
      
  if ( T_INTS > 1.0 )
    return 0;

  nof_crush = 0;

  for( i = 0, k = 0; i < ns; i ++, k += len )
    {
      if ( !isected [i] )
	continue;
      
      if ( BA [i].t > T_INTS )
	{
	  ints = 1;
	  continue;
	}

      ugly_normal_hack (hrib + k, i, len, T_INTS, ns);

      hrib_lens [i] = 0;
      isected [i] = 0;
      crushed [nof_crush ++] = i;
      BA [i].t = 10.0;

      for( l = 0; l < nofba; l ++ )
	{
	  m = int_balls [l];
	  j = hrib_qu [k + m];

	  /* FIXME: nepocitat dvakrat , ale jen jednou !!! sebrat z comp_proc_time */
	  compute_lin_interp( p1, old_pos + 3*i, sphs[i].p, NT(T_INTS) );
	  compute_lin_interp( p2, old_pos + 3*j, sphs[j].p, NT(T_INTS) );
	      
	  if ( problemal_cols [j + k].t == bum.time &&
	       has_same_state( sphs, i, j, len, p1, p2 ) )
	    continue;
	      
	  fyjik_s2s_col_dyn ( sphs, i, j, &hrib[k + m].b );
	  sphere_sphere_handler( i, j, p1, p2, &bum );
	      
	  cflags = 1;
	}

      if ( nofid )
	{
	  datas = fyjik_s2p_alc (sphs + i);
      
	  for (l = 0; l < nofid; l ++)
	    {
	      m = int_ids [l];
	      j = hrib_qu [k + m];

	      compute_lin_interp( p1, old_pos + 3*i, sphs[i].p, NT(T_INTS) );
	  
	      if ( problemal_cols [j + k].t == bum.time && 
		   (j < (ns + np) || j > (ns + np + nh - 1)) &&
		   has_same_state1( sphs + i, i, j, len, p1 ) )
		continue;
	  
	      if ( (j - ns) < np )
		sphere_plane_handler( i, map_edges [j - ns], p1, &bum );
	      fyjik_s2p_col_dyn_preproc ( sphs + i, &hrib [k + m].ch, 
					  datas, i, j - ns );
	  
	      cflags |= 2;
	      if ( j == (ns + np + nh) )
		cflags |= 4;
	    }
	}
      
      if ( cflags & 2 )
	fyjik_s2p_col_dyn_proc ( sphs + i, datas, 
				 cflags & 4, NT(T_INTS) * ts );
      
      if ( cflags )
	{
	  for (l = 0; l < nofba; l ++)
	    set_cols_state( sphs, i, hrib_qu [k + int_balls [l]], len, p1, p2 );
	  for (l = 0; l < nofid; l ++)
	    set_cols_state1( sphs + i, i, hrib_qu [k + int_ids [l]], len, p1 );
	  for (l = 0; l < nofbr; l ++)
	    set_cols_state1( sphs + i, i, hrib_qu [k + break_ids [l]], len, p1 );
	  
	  ints = 1;
	  cflags = 0;
	}
      
    }
  
  compute_proceed_time( sphs, old_pos, ns );

  return ints;
}

/***********/

int is_time( dbl *ts, int n )
{
  int	i, j = 0;

  for( i = 0; i < n; i++ )
    if ( ts[i] > EPSILON )
      j = 1;

  return j;
}

/***********/

void correct_position( Sphere *sphs, int n )
{
  int	i;

  for( i = 0; i < n; i++ )
    if ( sphs[i].p[Z] > 0.0 )
      sphs[i].p[Z] = sphs[i].ball->ball_type->radius;
}

/***********/

static void prec_init( Sphere *sphs, int n )
{
  int	i;

  for (i = 0; i < n; i++)
    prec_back_shock [i] = 0.0;
    
  for (i = 0; i < nof_crush; i ++)
    vs [crushed [i]] = vec_size( sphs [crushed [i]].v );
}

/***********/

void make_balls_bbs (Sphere *sphs, int n, dbl *old_pos)
{
  int	i, j;

  if ( !bum.round )
    for (i = 0; i < n; i ++)
      make_bb_from_ball_path (bb_balls + 6*i, sphs [i].p, old_pos + 3*i, sphs [i].ball->ball_type->radius);
  else
    for (j = 0; j < nof_crush; j ++)
      {
	i = crushed [j];
	make_bb_from_ball_path (bb_balls + 6*i, sphs [i].p, old_pos + 3*i, sphs [i].ball->ball_type->radius);
      }
}

/***********/


/********************************************************************************
 *		DedecekHribecek structure					*
 *  row \ column: num of balls     |  num of table edge + nof holes + DED_HRIB 	*
 *            --------------------------------------------------		*
 *  number    | 	           |                           |		*
 *   of       |      ball x ball   |	  ball x edge	       |		*
 * balls      |	     intersection  |	 intersection	       |		*
 *            |	      info         |	     info	       |		*
 *            |			   |			       |		*
 *            |			   |			       |		*
 *            |                    |                           |		*
 *            --------------------------------------------------		*
 *                                                                      	*
 ********************************************************************************/

void big_crash( Sphere *sphs, int n, dbl dt, dbl **polys, int *counts, int nofp,
		int nofh, TableHole *holes, dbl *cutout )
{
  int	i, cols;
  dbl	tmp [3];

  bum.round = 0;

  ts = 1.0 / (dbl)DEFAULT_STEPS;
  cum_tr = 0.0;
  cum_scl = 1.0;
  pisecek_prepare (sphs, n);
  save_ball_state( sphs, n, state );

  do
    {
      cols = 0;
      prec_init (sphs, n);
      for( i = 0; i < nof_crush; i++ )
	fyjik_one_step (sphs + crushed [i], crushed [i], ts);
      make_balls_bbs (sphs, n, state);
      for( i = 0; i < nof_crush; i++ )
      	check_and_proc_cols(sphs, state, hrib, n, 
			    crushed [i], polys, counts, 
			    nofp, nofh, holes, cutout, ts);
      cols = find_true_cols(sphs, state, hrib, n, nofp, nofh);
      if ( cols )
      	restore_ball_state(sphs, n, state);
      bum.round ++;
    }
  while( cols /*&& is_time( ts, n ) && ((iter++)<20) */ && bum.round < 100 );

  for( i = 0; i < n; i++ )
    {
      cvec( tmp, sphs[i].p );
      cvec( sphs[i].p, state + 3 * i );
      fyjik_end_proc (sphs + i, i, dt, ts);
      cvec( sphs[i].p, tmp ); 
    }
}

/***********/
/*
int cue2sphere( dbl *p, dbl *v, dbl power, dbl f10, Sphere *sph, int ids )
{
  dbl	ints[2], tmp[3], tmp1[3], t, a;
  int	i;
  
  if ( !ray2sphere( p, v, sph->p, sph->ball->ball_type->radius, ints ) )
    return 0;

  t = MIN( ints[0], ints[1] );
  
  if ( t < 0 || t > 1.0 )
    return 0;

  for( i = 0; i < 3; i++ )
    {
      tmp[i] = v[i] * t + p [i];
      
      pool_cue.adjusted_top [i] = v [i] * (t - CUE_SHOCK / vec_size( v )) + p [i];
      pool_cue.adjusted_bottom [i] = v [i] * (t - CUE_SHOCK / vec_size( v )) + p [i];
    }
  
  vec_sub( tmp, tmp, sph->p ); // normala odrazu na kouly 
  cvec( tmp1, v );
  vec_norm( tmp1, tmp1 );

  vec_norm( tmp, tmp );  
  vec_mul_c( tmp1, tmp1, fabs(vec_dot( tmp1, tmp )) );
  vec_mul_c( tmp1, tmp1, power );
  cvec( sph->v, tmp1 );
  vec_cross( sph->omega, tmp, tmp1 );
  vec_mul_c( sph->omega, sph->omega, f10 );

  if ( (vec_size( sph->omega ) > EPSILON) && (vec_size( sph->v ) > EPSILON) )
    {
      a = vec_size( sph->v );
      cvec( tmp, sph->omega );
      vec_cross( tmp, tmp, sph->v );
      vec_mul_c( tmp, tmp, 1.0 / a  );
      tmp[Z] = 0.0;
      vec_mul_c( tmp, tmp, t*sph->ball->ball_type->f6 );
      vec_add( sph->v, sph->v, tmp );
    }
  //  sph->v[Z] *= sph->ball->ball_type->f9;

  return 1;
}
*/

int cue2sphere( dbl *p, dbl *v, dbl power, dbl f10, Sphere *sph, int ids )
{
  dbl	ints[2], tmp[3], tmp1[3], tmp2[3], t, a;
  int	i;
  dbl	r = sph->ball->ball_type->radius;
  
  if ( !ray2sphere( p, v, sph->p, r, ints ) )
    return 0;

  t = MIN( ints[0], ints[1] );
  
  if ( t < 0 || t > 1.0 )
    return 0;

  for( i = 0; i < 3; i++ )
    {
      tmp[i] = v[i] * t + p [i];
      /* TIDIT */
      pool_cue.adjusted_top [i] = v [i] * (t - CUE_SHOCK / vec_size( v )) + p [i];
      pool_cue.adjusted_bottom [i] = v [i] * (t - CUE_SHOCK / vec_size( v )) + p [i];
    }
  
  cvec( tmp2, v );
  vec_norm( tmp2, tmp2 );
  vec_mul_c( tmp2, tmp2, power );

  vec_sub( tmp, tmp, sph->p ); /* normala odrazu na kouly */
  rigid_collision( sph->v, tmp2, sph->omega, tmp2, v, tmp, 1.0 / sph->ball->ball_type->weight, 0.0, 
		   1.0 / sim( sph->ball->ball_type->weight, sph->ball->ball_type->radius ), 
		   sph->ball->ball_type->radius, sph->ball->ball_type->radius,
		   0.0, 0.0, f10, 0 );

  my_pool_cue = pool_cue;

  return 1;
}

/***********/

int is_one_static( Sphere *sph, dbl dt )
{
  int	j = 0;

  /*
  if ( !(chck_near_table( sph ) || chck_hole_bottom( sph, -1, 0 )) )
    return 0;
  */
  /* TIDIT */
  if ( dt*vec_size( sph->v) < ENDER )
    cvec( sph->v, null );
  else
    j |= 1;
  
  if ( dt*vec_size( sph->omega ) < ENDER )
    {
      if ( !j ) 
	cvec( sph->omega, null );
    }
  else
    j |= 1;

  return !j;
}

/***********/

int is_static( Sphere *sphs, int n, dbl dt )
{
  int	 i, j = 0;

  /* FIXME: jeste rozmyslet kdy uz stojim */

  for( i = 0; i < n; i++ )
    j |= !is_one_static( sphs + i, dt );

  return !j;
}

/***********/

void init_balls( Sphere *sphs, int nos )
{
  int	i;

  for( i = 0; i < nos; i++ )
    {
      cvec( sphs[i].v, null );
      cvec( sphs[i].omega, null );
      /* sphs[i].q[0] = 1.0;
      sphs[i].q[1] = 0.0;
      sphs[i].q[2] = 0.0;
      sphs[i].q[3] = 0.0;*/
    }
}

/***********/

inline dbl ms2sec( dbl ms )
{
  return ms / 1000.0;
}

/************************************
 * TABLE GENERATION FCION FOR FYJIK *
 ************************************/
/*
dbl	table_d[] = { -100,-100,0, 100,-100,0, 100,100,0, -100,100,0 };
dbl	table_h[] = { -100,-100,TABLE_H, 100,-100,TABLE_H, 100,100,TABLE_H, -100,100,TABLE_H };

dbl	*table_sides[] = { table_d, table_d + 3, table_h + 3, table_h,
			   table_d + 3, table_d + 2*3, table_h + 2*3, table_h + 3,
			   table_d + 2*3, table_d + 3*3, table_h + 3*3, table_h + 2*3,
			   table_d + 3*3, table_d, table_h, table_h + 3*3 };

int	counts[] = { 4, 4, 4, 4 };
*/

static	int	*sids;
static	dbl	*tes;

/***********/

static int icomp( dbl *p1, dbl *p2 )
{
  if ( *p1 > *p2 )
    return 1;
  else
    if ( *p1 == *p2 )
      return 0;
    else
      return -1;
}

/***********/

static void bublik( int n )
{
  int 	i, j = 1, k;
  dbl	a;

  while (j)
    {
      j = 0;
      for (i = 0; i < n - 1; i ++)
	if ( tes [2 * i] > tes [2 * (i + 1)] )
	  {
	    SWAP (sids [i], sids [i + 1], k);
	    SWAP (tes [2 * i], tes [2 * (i + 1)], a);
	    SWAP (tes [2 * i + 1], tes [2 * (i + 1) + 1], a);
	    j = 1;
	  }
    }
}

/***********/

void arce( dbl *p, dbl *hcut, dbl *center, int i )
{
  dbl 	norm [2], a;

  norm [0] = p [0] - center [0];
  norm [1] = p [1] - center [1];

  if ( fabs(norm [X]) < EPSILON )
    if ( norm [Y] > 0.0 )
      a = PI / 2;
    else
      a = -PI / 2;
  else
    a = atan( norm [Y] / norm [X] );
  
  if ( norm [X] < 0.0 )
    a += PI;
      
  if ( a < 0.0 )
    a += 2 * PI;

  hcut [i] = a;
}

/***********/

void holebrub( dbl *pc, dbl *center, dbl r, dbl *edges, int *nofe, int sgn )
{
  dbl	n [2], v [2];
  dbl	t = (dbl)sgn * 10.0;

  return;

  n [0] = (center [0] - pc [0]) / r;
  n [1] = (center [1] - pc [1]) / r;

  v [0] = -n [1];
  v [1] = n [0];

  edges [*nofe * 4 + 0] = pc [0];
  edges [*nofe * 4 + 1] = pc [1];
  edges [*nofe * 4 + 2] = v [0] * t + pc [0];
  edges [*nofe * 4 + 3] = v [1] * t + pc [1];

  (*nofe) ++;
}

/***********/

void cute( dbl *e1, dbl *e2, dbl *edges, int *nofe, dbl *hcut, int id )
{
  int	i, ints = 0;
  dbl	ts [1024], a; /* FIXME: may be alloc some memory */
  int	sid [1024]; 
  int	nt = 0;
  dbl	dir [2];

  dir [0] = e2 [0] - e1 [0];
  dir [1] = e2 [1] - e1 [1];

  for( i = 0; i < pool_table.nof_holes; i ++ )
    if ( ray2circle( e1, dir, pool_table.holes [i].p, pool_table.holes [i].radius, ts + 2 * nt ) )
      {
	sid [nt] = i;
	if ( ts [2 * nt] > ts [2 * nt + 1] )
	  SWAP( ts [2 * nt] , ts [2 * nt + 1], a );
	nt ++;
      }
  
  if ( nt )
    {
      tes = ts;
      sids = sid;
      bublik (nt);
    }

  for( i = 0; i < nt; i++ )
    if ( (ts [2 * i] >= 0.0 || ts [2 * i + 1] >= 0.0) && (ts [2*i] <= 1.0) )
      {
	ints = 1;
	break;
      }

  if ( !ints )
    {
      edges [*nofe * 4] = e1 [0];
      edges [*nofe * 4 + 1] = e1 [1];
      edges [*nofe * 4 + 2] = e2 [0];
      edges [*nofe * 4 + 3] = e2 [1];
      map_edges [*nofe] = id;
      (*nofe) ++;
      return;
    }

  if ( ts [2 * i] >= 0.0 )
    {
      edges [*nofe * 4 + 0] = e1 [0];
      edges [*nofe * 4 + 1] = e1 [1];
      edges [*nofe * 4 + 2] = e1 [0] + ts [2 * i] * dir [0];
      edges [*nofe * 4 + 3] = e1 [1] + ts [2 * i] * dir [1];
      map_edges [*nofe] = id;
      arce( edges + *nofe * 4 + 2, hcut + 2 * sid [i], pool_table.holes [sid [i]].p, 0 );
      (*nofe) ++;
      /*      holebrub( edges + ((*nofe)-1) * 4 + 2, pool_table.holes [sid [i]].p, pool_table.holes [sid [i]].radius, edges, nofe, -1 ); */
    }

  for( i ++; i < nt; i ++)
    if ( ts [2 * i] <= 1.0 || ts [2 * i + 1] <= 1.0 ) 
      {
	edges [*nofe * 4 + 0] = e1 [0] + ts [2 * (i - 1) + 1] * dir [0];
	edges [*nofe * 4 + 1] = e1 [1] + ts [2 * (i - 1) + 1] * dir [1];
	edges [*nofe * 4 + 2] = e1 [0] + ts [2 * i] * dir [0];
	edges [*nofe * 4 + 3] = e1 [1] + ts [2 * i] * dir [1];
	map_edges [*nofe] = id;
	arce( edges + *nofe * 4 + 0, hcut + 2 * sid [i - 1], pool_table.holes [sid [i - 1]].p, 1 );
	arce( edges + *nofe * 4 + 2, hcut + 2 * sid [i], pool_table.holes [sid [i]].p, 0 );
	(*nofe) ++;
	/*	holebrub( edges + ((*nofe)-1) * 4 + 0, pool_table.holes [sid [i - 1]].p, pool_table.holes [sid [i - 1]].radius, edges, nofe, 1 ); */
	/*	holebrub( edges + ((*nofe)-2) * 4 + 2, pool_table.holes [sid [i]].p, pool_table.holes [sid [i]].radius, edges, nofe, -1 ); */
      }
    else
      break;

  i --;
  if ( ts [2 * i + 1] < 1.0 )
    {
      edges [*nofe * 4 + 2] = e2 [0];
      edges [*nofe * 4 + 3] = e2 [1];
      edges [*nofe * 4 + 0] = e1 [0] + ts [2 * i + 1] * dir [0];
      edges [*nofe * 4 + 1] = e1 [1] + ts [2 * i + 1] * dir [1];
      arce( edges + *nofe * 4 + 0, hcut + 2 * sid [i], pool_table.holes [sid [i]].p, 1 );
      map_edges [*nofe] = id;
      (*nofe) ++;
      /*      holebrub( edges + ((*nofe)-1) * 4 + 0, pool_table.holes [sid [i]].p, pool_table.holes [sid [i]].radius, edges, nofe, 1 ); */
    }
}

/***********/

void cut_edge( dbl *edges, int *nofe, dbl *hcut )
{
  int	i, j;

  for( i = 0; i < pool_table.nof_holes; i ++ )
    {
      hcut [2 * i] = 2 * PI; 
      hcut [2 * i + 1] = 0;
    }

  *nofe = 0;

  j = pool_table.nof_vertices - 1;
  for( i = 0; i < pool_table.nof_vertices; j = i, i++ )
    cute( pool_table.vertices + 2 * j, pool_table.vertices + 2 * i, edges, nofe, hcut, i );
}

/***********/

void generate_table()
{
  int	 i, j;
  dbl	*v;

  if ( !pool_table.update_fyjik_info )
    return;

  if ( table_d )
    {
      g_free( table_d );
      g_free( table_h );
      g_free( table_sides );
      g_free( counts );
      if ( hole_cuts )
	{
	  g_free( hole_cuts );
	  hole_cuts = NULL;
	}
    }

  if ( pool_table.nof_holes )
    {
      hole_cuts = g_new( dbl, 2 * pool_table.nof_holes );
      assert( hole_cuts );
    }

  v = g_new( dbl, 2 * 4 * (pool_table.nof_vertices + pool_table.nof_holes) );
  assert( v );
  map_edges = g_new( int, 2 * 4 * (pool_table.nof_vertices + pool_table.nof_holes) );
  assert (map_edges);

  cut_edge( v, & nof_edges, hole_cuts );

  table_d = g_new( dbl, 3 * 2 * nof_edges );
  assert( table_d );
  table_h = g_new( dbl, 3 * 2 * nof_edges );
  assert( table_h );
  table_sides = g_new( dbl*, 4 * nof_edges );
  assert( table_sides );
  counts = g_new( int, nof_edges );
  assert( counts );

  /*  j = nof_edges - 1; */
  for( i = 0; i < nof_edges; i ++ )
    {
      /*
      table_h [3 * i + 0] = table_d [3 * i + 0] = v [2*i + 0];
      table_h [3 * i + 1] = table_d [3 * i + 1] = v [2*i + 1];
      table_d [3 * i + 2] = 0.0;
      table_h [3 * i + 2] = TABLE_H;
      
      table_sides [4 * i + 0] = table_d + 3 * j;
      table_sides [4 * i + 1] = table_d + 3 * i;
      table_sides [4 * i + 2] = table_h + 3 * i;
      table_sides [4 * i + 3] = table_h + 3 * j;
      counts [i] = 4;
      j = i;
      */

      table_h [3 * 2 * i + 0] = table_d [3 * 2 * i + 0] = v [4 * i + 0];
      table_h [3 * 2 * i + 1] = table_d [3 * 2 * i + 1] = v [4 * i + 1];
      table_d [3 * 2 * i + 2] = 0.0;
      table_h [3 * 2 * i + 2] = TABLE_H;

      table_h [3 * (2 * i + 1) + 0] = table_d [3 * (2 * i + 1) + 0] = v [4 * i + 2];
      table_h [3 * (2 * i + 1) + 1] = table_d [3 * (2 * i + 1) + 1] = v [4 * i + 3];
      table_d [3 * (2 * i + 1) + 2] = 0.0;
      table_h [3 * (2 * i + 1) + 2] = TABLE_H;
      
      table_sides [4 * i + 0] = table_d + 3 * 2 * i;
      table_sides [4 * i + 1] = table_d + 3 * (2 * i + 1);
      table_sides [4 * i + 2] = table_h + 3 * (2 * i + 1);
      table_sides [4 * i + 3] = table_h + 3 * 2 * i;
      counts [i] = 4;
    }

  g_free( v );

  pool_table.update_fyjik_info = 0;
}

/***********/

void init_pythie( int nob, Sphere *sphs )
{
  int 	i, j;

  my_pool_cue = pool_cue;

  for( j = i = 0; i < MAX_BALLS; i++, j += 3 * MAX_PYTHIE_COLS )
    {
      ppythie [i] = pythie + j; 
      if ( i < nob )
	{
	  pythie_count [i] = 1;
	  cvec( ppythie [i], sphs[i].p );
	  ppythie [i] [Z] = 2 * SLIGHTLY_ABOVE_TABLE;
	}
    }
}

/***********/

static void set_fyjik_fcions() 
{ 
  fyjik_one_step = uno_stepo;
  fyjik_end_proc = fatal_proceed; 
  fyjik_cue_dyn = (void	(*) (dbl *, dbl *, dbl , dbl , Sphere *, int ))cue2sphere;
  fyjik_s2s_col_dyn = (void (*) (Sphere *, int , int , Babovicka *))sphere2sphere_endproc; 
  fyjik_s2p_alc = s2p_alc;
  fyjik_s2p_col_dyn_preproc = (void (*) (Sphere *, PernikovaChaloupka *, void *, int , int ))sphere2plane_end_preproceed;
  fyjik_s2p_col_dyn_proc = (void (*) (Sphere *, void *, int , dbl ))sphere2plane_end_proceed;
}

/***********/

static void init_fyjik( int n, int nofp, int nofh )
{
  int	i, j;

  set_fyjik_fcions ();

  state = g_new (dbl, 3*n);
  assert( state );

  hrib = g_new (DedecekHribecek, (n + nofp + nofh + DED_HRIB) * n);
  assert( hrib );

  hrib_qu = g_new (int, (n + nofp + nofh + DED_HRIB) * n);
  assert( hrib_qu );

  hrib_lens = g_new (int, n);
  assert( hrib_lens );

  crushed = g_new (int, n);

  pocketed = g_new (int, n);
  assert( pocketed );

  bb_balls = g_new (dbl, 6*n);
  assert( bb_balls );

  if ( nofh )
    {
      bb_holes = g_new (dbl, 6*nofh);
      assert( bb_holes );
    }

  bb_edges = g_new (dbl, 6*nofp);
  assert( bb_edges );

  problemal_cols = g_new (Snehurka, n*(n + nofp + nofh + DED_HRIB));
  assert( problemal_cols );

  isected = g_new (int, n);
  assert( isected );

  vs = g_new (dbl, n);
  assert( vs );

  BA = g_new( Bajaja, n );
  assert( BA );

  for (i = 0; i < n; i ++)
    pocketed [i] = 0;

  for (i = 0; i < nofh; i ++)
    make_bb_from_hole( bb_holes + 6*i, pool_table.holes[i].p, pool_table.holes[i].radius, 
		       HOLE_DEPTH, TABLE_H );

  for (j = i = 0; i < nofp; j += counts [i], i ++)
    make_bb_from_poly( bb_edges + 6*i, table_sides + j, counts [i] );

  for (i = 0; i < n*(n + nofp + nofh + DED_HRIB); i++)
    {
      problemal_cols [i].t = -1.0;
      cvec( problemal_cols [i].v1, null );
      cvec( problemal_cols [i].v2, null );
      /*      cvec( problemal_cols [i].p1, null ); */
      /*      cvec( problemal_cols [i].p2, null ); */
    }

  /*
  for (i = 0; i < n; i ++)
    babovicka_on_pisecek (hrib, n, nofp, nofh, i);
  */

#ifdef DEBUG
  fyjik_inited = 1;
#endif
}

void clear_fyjik()
{
  if ( state )
    {
      g_free(state);
      state = NULL;
    }
  if ( hrib )
    {
      g_free(hrib);
      hrib = NULL;
    }
  if ( pocketed )
    {
      g_free(pocketed);
      pocketed = NULL;
    }
  if ( bb_balls )
    {
      g_free(bb_balls);
      bb_balls = NULL;
    }
  if ( bb_holes )
    {
      g_free(bb_holes);
      bb_holes = NULL;
    }
  if ( bb_edges )
    {
      g_free(bb_edges);
      bb_edges = NULL;
    }
  if ( isected )
    {
      g_free(isected);
      isected = NULL;
    }
  if ( problemal_cols )
    {
      g_free(problemal_cols);
      problemal_cols = NULL;
    }
  if ( vs )
    {
      g_free(vs);
      vs = NULL;
    }
  if ( BA )
    {
      g_free( BA );
      BA = NULL;
    }
  if ( hrib_qu )
    {
      g_free( hrib_qu );
      hrib_qu = NULL;
    }
  if ( hrib_lens )
    {
      g_free( hrib_lens );
      hrib_lens = NULL;
    }
  if ( crushed )
    {
      g_free( crushed );
      crushed = NULL;
    }
  
#ifdef DEBUG
  fyjik_inited = 0;
#endif
}

/***********/

static void debuz(Sphere *sphs)
{
  sphs[0].p[0] = 40;
  sphs[0].p[1] = 46;
  sphs[0].p[2] = 3.0;
  sphs[0].ball->ball_type->radius = 3.0;
  
  sphs[1].p[0] = 40;
  sphs[1].p[1] = 40;
  sphs[1].p[2] = 3.0;
  sphs[1].ball->ball_type->radius = 3.0;
}

/************** 
 * TULPAS API *
 **************/

gint pool_cue_ball (PoolBallSpaceInfo *inf, int nof_balls,   
		    dbl *startpos, dbl *endpos)
{
  dbl	ints[2], tmp[3], t;
  int	i;

  vec_sub( tmp, endpos, startpos );

  for( i = 0; i < nof_balls; i++ )
    if ( ray2sphere( startpos, tmp, inf[i].p, inf[i].ball->ball_type->radius, ints ) )
      if ( (t = MIN(ints[0],ints[1])) >= 0.0 )
	if ( t <= 1.0 )
	  {
#ifdef DEBUG
	    printf( "koule tuk - s: %f %f %f e:%f %f %f sph:%f %f %f t:%f\n",
		     startpos[0], startpos[1], startpos[2],
		     endpos[0], endpos[1], endpos[2],
		     inf[i].p[0], inf[i].p[1], inf[i].p[2], t );
#endif
	    return 1;
	  }
  
  return 0;
}

/***********/

/* pool_table */

void pool_fyjik( Animation *anim, int nop, dbl cue_burr )
{
  int			actual_size, nof_balls, nof_steps, act_step, noe;
  PoolBallSpaceInfo 	*binf;

#ifdef DEBUG
  int			initing = 0;
#endif

  binf = anim->balls;
  nof_balls = anim->nof_balls;
  nof_steps = anim->length - 1;
  actual_size = anim->ball_array_size;

  bum.time = (dbl)nof_steps / (dbl)DEFAULT_STEPS;
  bum.ani = anim;

#ifdef DEBUG
  if ( !fyjik_inited )
    {
      init_fyjik( nof_balls, nof_edges, pool_table.nof_holes );
      initing = 1;
    }
#endif
  
  for( act_step = 0; 
       !is_static( binf + nof_balls*nof_steps, nof_balls, 1.0 / (dbl)DEFAULT_STEPS ) 
	 && ( act_step < nop );
       nof_steps++, act_step++ )
    {
      if ( actual_size == (nof_steps+1) )
	{
	  actual_size += ADD_STEPS;
	  binf = g_renew (Sphere, binf, actual_size * nof_balls);
	  assert( binf );
	}

      bum.act_sphs = binf + nof_balls*(nof_steps+1);
      memcpy( binf + nof_balls*(nof_steps+1), binf + nof_balls*nof_steps, sizeof(Sphere)*nof_balls );
      big_crash( binf + nof_balls*(nof_steps+1), 
		 nof_balls, 1.0 / (dbl)DEFAULT_STEPS, table_sides, counts, nof_edges, 
		 pool_table.nof_holes, 
		 pool_table.holes, hole_cuts );

      bum.time += 1.0 / (dbl)DEFAULT_STEPS;
    }

  /*  correct_position( binf + nof_balls*(nof_steps), nof_balls ); */

  anim->length = nof_steps + 1;
  anim->balls = binf;
  anim->ball_array_size = actual_size;

  if ( is_static( binf + nof_balls*nof_steps, nof_balls, 1.0 / (dbl)DEFAULT_STEPS ) )
    {
      clear_fyjik ();
      animation_add_event( anim, bum.time * 1000.0, S_End_Of_Shoot, 0, 0 );
      anim -> finished = TRUE;
    }

#ifdef DEBUG  
  if ( fyjik_inited && initing )
    clear_fyjik ();
#endif

}

/***********/

/* FIXME: smazat */
#define SILA	  4000
/* #define SILA	 142.464095 */

void pool_big_mover (PoolBallSpaceInfo *inf, int nof_balls, int nop,
		     dbl *startpos, dbl *endpos, dbl speed, dbl cue_burr,
		     Animation **newanim )
{
  PoolBallSpaceInfo *binf;

  /* FIXME: dodelat udelatko, ktere prevede stul na mnou pozadovany nabodenicek :) */
  int		    nof_steps = 0;
  int		    actual_size, i, j;

  dbl		    tmp[3], ints[2], t, mint, power;

  sphere_sphere_handler = spherespherehandler;
  sphere_plane_handler = sphereplanehandler;
  sphere_hole_handler = sphereholehandler;
  ball_out_handler = holeouthandler;
  fyjik_state = STATE_FYJIK;

  vec_sub( tmp, endpos, startpos );

  power = speed;
#if 1
  if ( power > CUE_MAX_POWER )
    power = CUE_MAX_POWER;
#endif

  actual_size = power * (dbl)DEFAULT_STEPS / ( inf->ball->ball_type->f7 * inf->ball->ball_type->g );
  actual_size = ( actual_size < nop ) ? ( nop ) : (actual_size );

  init_balls( inf, nof_balls );
  generate_table();
  init_fyjik ( nof_balls, nof_edges, pool_table.nof_holes );
  
#ifdef DEBUG
  printf("!!!********sila --- %f ---- \n", power );
#endif

  *newanim = animation_new (nof_balls, DEFAULT_STEPS, actual_size);
  (*newanim)->finished = FALSE;

  binf = (*newanim) -> balls;
  memcpy( binf , inf, sizeof(Sphere) * nof_balls );

  for( mint = 10.0, i = 0; i < nof_balls; i ++ )
    if ( ray2sphere( startpos, tmp, binf[i].p, binf[i].ball->ball_type->radius, ints ) )
      if ( (t = MIN( ints [0], ints [1] )) >= 0.0 )
	if ( t <= 1.0 )
	  {
	    mint = t;
	    j = i;
	  }

#if 0
  debuz( binf );
#endif

#if 1
  fyjik_cue_dyn ( startpos, tmp, power, cue_burr, binf + j, j );
#else
  fyjik_cue_dyn ( startpos, tmp, SILA, cue_burr, binf + j, j );
#endif
  animation_add_event (*newanim, 0, S_Cue_Ball_Collision, j, 0);

  pool_fyjik( *newanim, nop, cue_burr );
}

/***********/

void pool_pythie (PoolBallSpaceInfo *inf, int nof_balls, int active_ball, dbl *cue_orient)
{
  Sphere	*sphs;
  int		 i, j;

#ifdef DEBUG
  printf("----------\n");
#endif

  bum.time = 0;
  bum.act_sphs = inf;

  sphs = g_new( Sphere, nof_balls );

  assert( sphs );

  init_balls( inf, nof_balls );

  generate_table();
  set_fyjik_fcions ();
  init_fyjik ( nof_balls, nof_edges, pool_table.nof_holes );
  
  init_pythie( nof_balls, inf );
  
  memcpy( sphs, inf, sizeof( Sphere ) * nof_balls );

  sphs [active_ball].v [0] = cue_orient[0];
  sphs [active_ball].v [1] = cue_orient[1];

  vec_norm( sphs [active_ball].v, sphs [active_ball].v );
  vec_mul_c( sphs [active_ball].v, sphs [active_ball].v, 100 );

  sphere_sphere_handler = sphere_sphere_pythie;
  sphere_plane_handler = sphere_plane_pythie;
  sphere_hole_handler = sphere_hole_pythie;
  ball_out_handler = hole_out_pythie;

  fyjik_state = STATE_PYTHIE;

  bum.act_sphs = sphs;

  for ( i = 0; i < 50; i++ )
    {
      big_crash( sphs, nof_balls, 1.0 / 20.0, table_sides, counts, nof_edges,
		 pool_table.nof_holes, 
		 pool_table.holes, hole_cuts );
      for( j = 0; j < nof_balls; j++ )
	if ( vec_size (sphs [j].v) > 1 )
	  {
	    vec_norm( sphs [j].v, sphs [j].v );
	    vec_mul_c( sphs [j].v, sphs [j].v, 100 );
	  }
      bum.time ++;
    }

  sphere_sphere_handler = spherespherehandler;
  sphere_plane_handler = sphereplanehandler;
  sphere_hole_handler = sphereholehandler;
  ball_out_handler = holeouthandler;
  fyjik_state = STATE_FYJIK;

  clear_fyjik ();
  g_free( sphs );
}

/*************
 * TEMPLATES *
 *************/

/*
  Basic kinematics template - 

  do
{
  save actual state of balls
  proceed one step with actual physics
  if ( collision )
  {
    restore state
    proceed collisions; 
    cols = true;
  }
} while( cols );

proceed external forces
*/

/*
  // v_uhl = omega*r

  pos = v*t + pos + (na podlozce )?(omega*r*t*f6):(0)

  v = v + (nad podlozkou)?(g*t):(- treni)
  omega -= ( na podlozce)?(treni):(0)
*/
