/* game engine7 1999/11/14 20:27:205 2000/01/09 16:43:42
   $Id: game.c,v 1.47 2000/02/29 16:23:10 0rfelyus Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <guile/gh.h>
#include <glib.h>


#include "debug.h"
#include "game.h"
#include "gtulpas.h"
#include "query.h"
#include "variant.h"
#include "player.h"
#include "pool.h"

#include "cue.h"
#include "table.h"

#include "animation.h"
#include "guileglue.h"
#include "poolfeedgl.h"

#include <libgnome/gnome-config.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gnome-dialog.h>
#include <libgnomeui/gnome-dialog-util.h>

/* static part */

GList * game_list = NULL;      /* list of all games, created during gtulpas_guile_init */
gint    game_max_id = 0;       /* maximal game id */

/* game */

Game *
game_new (gint minp, gint maxp, gint nof_balls)
{
	Game * g = (Game *) g_new (Game, 1);
	gint   i;

	g->ball_out          = NULL;
	g->animation         = NULL;
	g->events            = NULL;
	g->events_processed  = NULL;
	g->view_answer       = 0;
	g->anim_event_mask  = S_Cue_Ball_Collision | S_Ball_Ball_Collision | S_Ball_Jump_Off;
	g->cur_player       = 0;
	g->help_file	    = NULL;

	g->min_players_nof   = minp;
	g->max_players_nof   = maxp;
	g->player            = (Player **) g_new (Player *, maxp);
	g->score_type        = NULL;
	g->score_name        = NULL;
	g->nscore            = 0;

	g->query_list = NULL;
	g->query_cur = NULL;
	g->variant_list = NULL;

	g -> default_view_position = -1;
	for (i = 0; i < GAME_MAX_NOF_VIEW_POS; i++)
	  g -> view_positions [i] . eye_focus_distance = -1; /* invalid */

	g -> setup_table_func = SCM_BOOL_F;
	g -> new_game_func = SCM_BOOL_F;     
	g -> game_rules_func = SCM_BOOL_F;   

	g -> end_of_shot_func = SCM_BOOL_F;
	g -> next_stroke_func = SCM_BOOL_F;

	g -> ball_move_verify_func = SCM_BOOL_F;
	g -> ball_move_id = -1;

	g -> safe_to_undo = FALSE;
	g -> game_state_deep_copy_func = SCM_BOOL_F;

	g->balls = NULL;
	g->balls_in_hand_list = NULL;
	game_set_nof_balls (g, nof_balls);

	return g;
}

void
game_set_nof_balls (Game *g, gint nof_balls)
{
  int	i;

  if (g->balls)
    g_free (g->balls);

  g -> balls = g_new (PoolBall, nof_balls);
  g->balls_nof = nof_balls;

  for (i = 0; i < nof_balls; i++)
  {
    g -> balls [i] . name = g_strdup_printf (_("%d. ball"), i + 1);

    g -> balls [i] . ball_skin = default_ball_skin;
    g -> balls [i] . ball_type = default_fyjik_skin;

    g -> balls [i] . GL_ball_data = NULL;
    g -> balls [i] . GL_mark_data = NULL;
    g -> balls [i] . GL_strip_data = NULL;
    g -> balls [i] . GL_cap_data = NULL;
    
    g -> balls [i] . is_invalid = FALSE;
  }
}

static void
game_clear_balls (Game *game)
     /* clear all flags, balls out,... */
{
  int i;

  g_list_free (game -> ball_out);
  g_list_foreach (game -> balls_in_hand_list, (GFunc) g_free, NULL);
  g_list_free (game -> balls_in_hand_list);
  game -> balls_in_hand_list = NULL;
  game -> ball_out = NULL;

  for (i = 0; i < game->balls_nof; i++)
    game->balls [i] . is_invalid = FALSE;
}

void
game_save (Game *game, gchar *prefix)
{
	gchar  num [16];
	gchar *pr;
	gint   i;

	debug ("game_save %d\n", game->id);

	g_snprintf (num, 15, "%d", game->id);

	pr = g_strconcat (prefix, "game", num, "/", NULL);
	gnome_config_push_prefix (pr);
	g_free (pr);

	gnome_config_set_int ("players", game->nof_players);

	for (i=0; i<game->nof_players; i++) {
		g_snprintf (num, 15, "%d", i);
		pr = g_strconcat ("player", num, "_name", NULL);
		gnome_config_set_string (pr, game->player [i]->name);
		g_free (pr);
	}

	gnome_config_pop_prefix ();
}

void
game_load (Game *game, gchar *prefix)
{
	gchar  num [16];
	gchar *pr;
	gint   i;

	debug ("game_load %d\n", game->id);

	g_snprintf (num, 15, "%d", game->id);

	pr = g_strconcat (prefix, "game", num, "/", NULL);
	gnome_config_push_prefix (pr);
	g_free (pr);

	game->nof_players = gnome_config_get_int ("players=1");
	if (game->nof_players < game->min_players_nof)
		game->nof_players = game->min_players_nof;

	for (i=0; i<game->nof_players; i++) {
		g_snprintf (num, 15, "%d", i);
		pr = g_strconcat ("player", num, "_name=Unknown", NULL);
		game->player [i] = player_new (gnome_config_get_string (pr), game);
		g_free (pr);
	}

	gnome_config_pop_prefix ();
}

void
game_add_score (Game *g, gint t, gchar *name)
{
	g_assert (t < ScoreTypeMax);

	g->score_type = g_list_append (g->score_type, (gpointer) t);
	g->score_name = g_list_append (g->score_name, g_strdup (name));
	g->nscore++;
}

void
game_start (Game * game)
{
	gint i, j;
	PoolBallSpaceInfo	*binfo;

	g_return_if_fail (game != NULL);

	gtulpas_status_clear ();
	gtulpas_status_message ("Welcome to Gnome Tulpas :*)");

	g_slist_free (game -> events_processed);
	game -> events_processed = NULL;

	game->stroke_number = 0;
	game->state         = GAME_RUNNING;

	animation_kill ();
	if (game->animation)
		animation_destroy (game->animation);

	game->animation = animation_new (game->balls_nof, 1, 1);

	game_clear_balls (game);

	/* set anim ball */
	for (i = 0, binfo = game -> animation -> balls;
	     i < game->balls_nof; 
	     i++, binfo++)
	  {
	    binfo -> ball = game -> balls + i;
	    binfo -> p [X] = 0;
	    binfo -> p [Y] = ((dbl) i)*(binfo -> ball -> ball_type -> radius * 3.);
	    binfo -> p [Z] = binfo -> ball -> ball_type -> radius;
	  }
	/* clear score */
	for (i=0; i<game->nof_players; i++)
		for (j=0; j<game->nscore; j++)
			score_reset (game->player [i]->score [j]);

	/* fixme */ pool_play_current_animation ();

	glue_safe_apply (game->new_game_func,
			 variant_vlistval2scm (game -> variant_list));

	game_add_event_nop_type (game, PLACE_BALL_NEXT);

	gnome_tulpas_score_refresh    (gtulpas->score);
	game_set_player (game, 0);
}

void
game_unselect (Game *game)
{
  int i;

  if (! game)
    return;

  if (pool_table.vertices != NULL)
    { g_free (pool_table.vertices); pool_table.vertices = NULL; }
  
  if (pool_table.holes != NULL)
    { g_free (pool_table.holes); pool_table.holes = NULL; }
  
  if (pool_table.mark_list != NULL)
    { 
      g_list_foreach (pool_table.mark_list, (GFunc) g_free, NULL);
      g_list_free (pool_table.mark_list);
      pool_table.mark_list = NULL;
    }

  for (i = 0; i < game -> balls_nof; i++)
    LOD_finish_ball (game -> balls +i);
}

void
game_select (Game * game)
{
#ifdef DEBUG
	g_message ("Game select: %s", game->name);
#endif
	inside_set_rules = game;
	glue_safe_apply (game->setup_table_func,
			 variant_vlistval2scm (game->variant_list));
	inside_set_rules = NULL;

	exit_if_fail (pool_table.vertices != NULL, "no (define-table) inside set-rules function");
}

gint
game_add_player (Game *game, gchar *name)
{
	if (game->nof_players >= game->max_players_nof)
		return FALSE;

	game->player [game->nof_players] = player_new (name, game);
	game->nof_players++;

	return TRUE;
}

void
game_set_player (Game *g, gint p)
{
	g_assert (p>=0 && p<g->nof_players);
	g->cur_player = p;
	gnome_tulpas_score_set_player (gtulpas->score, p);
}

void
game_do_queries (Game *game)
{
  GList		 *ptr;

  debug ("game_do_queries\n");
  
  game->view_answer = TRUE;

  game->query_cur = game->query_list;
  game_add_event_nop_type (game, NEXT_QUERY);
}

int
game_cue_active (Game *game)
{
  return animation_finished_p () &&
    (game->state == GAME_RUNNING) &&
    (game->balls_in_hand_list == NULL);

}

static int
game_place_ball_next (Game *game)
{
  if (game -> balls_in_hand_list != NULL)
    {
      BallInHand     * bih;

      bih = (BallInHand *) (game -> balls_in_hand_list -> data);
      gtulpas_status_start (bih->prompt);

      return 1;
    }
  else
    return 0;
}

BallInHand*
ball_in_hand_new (int ball_id)
{
  BallInHand	*bih;

  bih = g_new (BallInHand, 1);

  bih -> ball_id = ball_id;
  bih -> validate = SCM_BOOL_T;
  bih -> after = SCM_BOOL_T;
  bih -> prompt = 
    g_strdup_printf (_("You have %s in your hand"), 
		     gtulpas -> game -> balls [ball_id] . name);

  return bih;
}

void
ball_in_hand_free (BallInHand *bih)
{
  if (SCM_NIMP (bih -> validate))
    scm_unprotect_object (bih -> validate);

  if (SCM_NIMP (bih -> after))
    scm_unprotect_object (bih -> after);

  g_free (bih -> prompt);
  g_free (bih);
}

void 
game_ball_in_hand_answer (Game *game)
{
  GList		*ptr;
  BallInHand	*bih;

  ptr = game -> balls_in_hand_list;
  bih = (BallInHand *) (ptr -> data);

  if (bih -> after != SCM_BOOL_T)
    glue_safe_call0 (bih -> after);

  game -> balls_in_hand_list = ptr -> next;
  ptr -> next = NULL;

  g_free (bih);
  g_list_free (ptr);

  gtulpas_status_end ();
  game_add_event_type (game, PLACE_BALL_NEXT);
}

void
game_next_query (Game *game)
{
	Query        *  query;
	
	/* find next query to ask */
	while (game->query_cur) {

		query  = (Query *)        game->query_cur->data;

		if (query_needs_answer (query, game->stroke_number, game->cur_player)) {
			gtulpas_answers_undo_sensitivity (TRUE);
			gtulpas_status_start (query->prompt);
			return;
		}
		game->query_cur = game->query_cur->next;
	}

	debug ("all queries done\n");
	game->view_answer = FALSE;

	/* do stroke as all queries are answered */
	game_add_event_nop_type (game, DO_STROKE);
}

void
game_answer_query (Game *game, PoolViewGL *view, int x, int y)
{
	Query * query = game->query_cur->data;
	gint answer;

	g_assert (query != NULL);

	answer = pool_view_do_select (view, x, y, query -> type);

	if (answer < 0)
		return;

	debug ("query answered\n");

	/* query answered */
	gtulpas_status_end ();

	if (query_is_valid (query, game->stroke_number, game->cur_player, answer)) {
	        view_blink_start (view, Blink_Good);
		game->query_cur = game->query_cur->next;
		game_add_event_type (game, NEXT_QUERY);
	} else {
	        view_blink_start (view, Blink_Bad);
		gtulpas_status_start (query->error);
	}
}

void
game_add_event (Game * game, GameEvent *ev)
{
	game_add_event_nop  (game, ev);
	game_process_events (game);
}

void
game_add_event_nop (Game * game, GameEvent *ev)
{
	game->events = g_slist_append (game->events, ev);
}

void
game_add_event_type (Game * game, gint t)
{
	game_add_event (game, game_event_new (t, NULL));
}

void
game_add_event_nop_type (Game * game, gint t)
{
	game_add_event_nop (game, game_event_new (t, NULL));
}

void
game_add_event_type_data (Game * game, gint t, gpointer * d)
{
	game_add_event (game, game_event_new (t, d));
}

void
game_add_event_nop_type_data (Game * game, gint t, gpointer * d)
{
	game_add_event_nop (game, game_event_new (t, d));
}

static gchar * game_event_name [EVENTS] = {
	"new game",
	"que x ball",
	"my move",
	"other move",
	"end, you win",
	"end, you lose",
	"end, you equal with other",
	"faul",
	"do queries",
	"next query",
	"do query",
	"shot end",
	"next player",
	"set player",
	"next stroke",
	"do stroke",

	"point",
	"penalty",

	"pocketed",
	"ball out",
	"ball in hand",
	"place ball next",

	"undo",
	"undo answers"
};

void
game_end (Game *game, gint type)
{
	int  p, s, i;
	GtkWidget *dialog;
	gchar *msg;

	game->state = GAME_ENDED;

	switch (type) {
	case END_WIN:

		msg = _("You won!");
		break;

	case END_LOSE:

		msg = _("You lost!");
		break;

	case END_EQ:

		msg = _("Remise!");
		break;

	case END_HIGHER_SCORE:
		/* p = 0;
	  s = game -> player [0] -> score -> n;
	  for (i = 1; i < game -> nof_players; i++)
	    if (s < game -> player [i] -> score -> n)
	      { p = i; s = game -> player [i] -> score -> n; }
	  
	      msg = g_strdup_printf (_("%s wins the game!"), game -> player [p] -> name); */
#ifdef DEBUG
		g_message ("FIXME!!! END_HIGHER_SCORE");
#endif
	  break;

	default:
		g_assert_not_reached ();
	};

	dialog = gnome_ok_dialog_parented (msg, GTK_WINDOW (gtulpas->app));
	gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
}

static void
game_add_point (Game *game, gint counter, gint points)
{
	g_assert (counter < game->nscore);

	score_inc (game->player [game->cur_player]->score [counter], points);
	gtulpas_score_refresh (gtulpas);
}

static void
game_ball_pocketed (Game *game, gint counter, gint bid)
{
	g_assert (counter < game->nscore);

	score_sunk (game->player [game->cur_player]->score [counter], bid);
	gtulpas_score_refresh (gtulpas);
}

static void
game_add_penalty (Game *game, gint counter, gint points)
{
  int	i;

  g_assert (counter < game->nscore);

  for (i = 0 ; i < game -> nof_players; i++)
    if (i != game->cur_player)
      score_inc (game->player [i] -> score [counter], points);
			  
  gtulpas_score_refresh (gtulpas);
}

int
game_get_player_points (Game *g, gint  pid, gint counter)
{
  GSList	*gl;
  GameEvent	*ge;
  EventScore	*es;
  EventResetScore *rs;

  int		points;
  int		curpid;

  points = g->player [pid] -> score [counter] -> hits;

  for (gl = g -> events, curpid = g->cur_player;
       gl;
       gl = gl -> next)
    {
      ge = (GameEvent *) gl -> data;
      switch (ge -> type)
	{
	case POINT:
	  es = (EventScore *) ge -> data;
	  if ((pid == curpid) && (es->counter == counter))
	    points += es -> data;
	  break;
	case PENALTY:
	  es = (EventScore *) ge -> data;
	  if ((pid != curpid) && (es->counter == counter))
	    points += es -> data;
	  break;
	case RESET_SCORE:
	  rs = (EventResetScore *) ge -> data;
	  if (((curpid == rs->pid) || (rs->pid == -1)) && (counter == rs->counter))
	    points = 0;
	  break;
	case NEXT_PLAYER:
	  curpid = (curpid + 1) % g->nof_players;
	  break;
	case SET_PLAYER:
	  curpid = (int) (ge->data);
	  break;
	}
    }
  return points;
}

static void
game_reset_score (Game *g, EventResetScore *ev)
{
	gint i;

	g_assert (ev->counter < g->nscore);

	if (ev->data)
		g_free (ev->data);
	ev->data = g_new (gint, g->nof_players);

	for (i=0; i<g->nof_players; i++)
	  {
	    ev->data [i] = g->player [i]->score [ev->counter]->hits;
	    if (ev -> pid == -1 || ev -> pid == i)
	      g -> player [i] -> score [ev->counter] -> hits = 0;
	  }

	gtulpas_score_refresh (gtulpas);
}

static void
game_reset_score_undo (Game *g, EventResetScore *ev)
{
	gint i;

	g_assert (ev->counter < g->nscore);

	if (ev->data)
		for (i=0; i<g->nof_players; i++)
			g->player [i]->score [ev->counter]->hits = ev->data [i];

	gtulpas_score_refresh (gtulpas);
}

void
game_undo_events_till_next_game_event (Game *game, int ignore)
{
  GSList	*start, *previous;
  GameEvent	*cur;
  EventScore    *sc;


  start = game -> events_processed;
  previous = NULL;

#ifdef DEBUG
  g_message ("-*- UNDO START (%d) -*-", ignore);
#endif

  while ((ignore != -1) && (game -> events_processed))
    {
      cur = (GameEvent *) game -> events_processed -> data;
#ifdef DEBUG
      g_message ("  undo: %s", game_event_name [cur->type]);
#endif
      switch (cur -> type)
	{
	case NEXT_STROKE:
#ifdef DEBUG
	  gh_write (GAME_STATE (game)); gh_newline ();
#endif
	  glue_game_state_restore (game);
#ifdef DEBUG
	  gh_write (GAME_STATE (game)); gh_newline ();
#endif
	  ignore --;
	  break;
	 
	case END_WIN:	case END_LOSE:
	case END_EQ:	case END_HIGHER_SCORE:
	  game -> state = GAME_RUNNING;
	  break;
 
	case SHOT_END:
	  game -> stroke_number--;
	  break;

	case POINT:
		sc = (EventScore *) cur->data;
		game_add_point (game, sc->counter, -sc->data);
		break;

	case PENALTY:
		sc = (EventScore *) cur->data;
		game_add_penalty (game,  sc->counter, -sc->data);
		break;

	case RESET_SCORE:

		game_reset_score_undo (game, (EventResetScore *) cur->data);
		break;

	case NEXT_PLAYER:
	case SET_PLAYER:
	  game_set_player (game, (int) (cur -> data));
	  break;
	}
      
      previous = game -> events_processed;
      game->events_processed = previous -> next;
    }
  
  if (previous)
    previous -> next = NULL;

  game_clear_balls (game);

  g_slist_free (start);
#ifdef DEBUG
  g_message ("-*- UNDO STOP -*-");
#endif
}

static int game_process_inside = 0;

int
game_process_events_inside (void)
{
  return game_process_inside ? 1 : 0;
}

void
game_process_events (Game *game)
{
	static int ignore_steps = 0;
	GSList     * del;
	GameEvent  * cur;
	EventScore * sc;

	game_process_inside++;

	while (game->events) {

		cur = (GameEvent *) game->events->data;

		debug ("game event: %s\n", game_event_name [cur->type]);

		switch (cur->type) {
		case NEW_GAME:
			glue_gc_enable ();
			game_start (game);
			break;

		case CUE_X_BALL: {
			EventCueXBall * ev = (EventCueXBall *) cur->data;
			ignore_steps = 0;
			glue_gc_disable ();		  
			gtulpas_answers_undo_sensitivity (FALSE);

			if (game->state != GAME_ENDED)
				cue_boom (ev->top, ev->newpos, ev->speed);

			break;
		}
		case SHOT_END:
			glue_gc_enable ();
			glue_call_end_of_shot (game);
			
			g_list_free (game->ball_out);
			game->ball_out = NULL;

			game->stroke_number++;
				/* place all balls in hand */
			game_add_event_nop_type (game, PLACE_BALL_NEXT);

			break;

		case NEXT_STROKE:
			glue_game_state_backup (game);
			if (game->next_stroke_func != SCM_BOOL_F)
			  glue_safe_call2 (game->next_stroke_func, 
					   SCM_MAKINUM (game->cur_player),
					   GAME_STATE (game));

			cue_reset_position (&pool_cue);
			game_add_event_nop_type (game, DO_QUERIES);
			break;

		case DO_STROKE:

			/* clean query flag in views */
			break;

		case DO_QUERIES:
			game_do_queries (game);
			break;

		case NEXT_QUERY:
			game_next_query (game);
			break;

		case END_WIN:
		case END_LOSE:
		case END_EQ:
		case END_HIGHER_SCORE:
			game_end (game, cur->type);
			/* thrown from Rules system */
			break;

		case POINT:
			sc = (EventScore *) cur -> data;
			game_add_point (game, sc->counter, sc->data);
			break;

		case PENALTY:
			sc = (EventScore *) cur -> data;
			game_add_penalty (game, sc->counter, sc->data);
			break;

		case RESET_SCORE:
			game_reset_score (game, (EventResetScore *) cur->data);
			break;

		case BALL_ON:
			gnome_tulpas_score_ball_on (gtulpas->score, (gint) cur->data);
			break;

		case POCKETED:
			sc = (EventScore *) cur -> data;
			game_ball_pocketed (game, sc->counter, sc->data);
			gtulpas_score_refresh (gtulpas);
			break;

		case NEXT_PLAYER:
			cur -> data = (gpointer) (game -> cur_player);
			game_set_player (game, (game->cur_player + 1) % game->nof_players);
			break;

		case SET_PLAYER:
			cur -> data = (gpointer) (game -> cur_player);
			game_set_player (game, (int) cur->data);
			break;

		case FAUL:
		        g_message ("Faul! :)");
			break;

		case PLACE_BALL_NEXT:
		  if (! game_place_ball_next (game))
		    {
		      ignore_steps = 1;
		      game_add_event_nop_type (game, NEXT_STROKE);
		    }
		  break;

		case BALL_OUT:
		  game->ball_out = g_list_append (game->ball_out, cur->data);
		  break;

		case BALL_IN_HAND:
		  if (g_list_find (game -> balls_in_hand_list, cur->data) == NULL)
		    game->balls_in_hand_list = g_list_append (game->balls_in_hand_list, cur->data);
		  
		  break;

		case UNDO:
		  game_undo_events_till_next_game_event (game, ignore_steps);
		  ignore_steps = 0;
		  animation_undo (game->animation); 
		  game_add_event_nop_type (game, NEXT_STROKE);

		  gtulpas_status_clear ();
		  gtulpas_status_message (_("Undo!"));

		  cue_reset_position (&pool_cue);
		  break;

		case UNDO_ANSWERS:
		  game_undo_events_till_next_game_event (game, 0);
		  /* no animation_undo because no animation was undone */
		  game_add_event_nop_type (game, NEXT_STROKE);
		  cue_reset_position (&pool_cue);


		  gtulpas_status_clear ();
		  gtulpas_status_message (_("Answers undone!"));

		  break;


		default:
			g_assert_not_reached ();
		}

		/* remove processed event and store it in game_events_processed */
		del = game->events;
		game->events = game->events->next;

		game->events_processed = g_slist_prepend (game->events_processed, del->data);
		g_slist_free_1 (del);
	}
	game_process_inside--;
}

