/* game engine
   $Id: game.h,v 1.25 2000/02/29 16:23:10 0rfelyus Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef __GAME_H__
#define __GAME_H__

typedef struct _Game             Game;
typedef enum   _GameState        GameState;

#include "mathutils.h"
#include "query.h"
#include "player.h"
#include "pool.h"
#include "score.h"
#include "gameevent.h"
#include "poolview.h"
#include "animation.h"
#include <libguile/tags.h>
#include <glibconfig.h>

#define  GAME_MAX_NOF_VIEW_POS		9
#define  GAME_STATE(GAME)		(SCM_VELTS ((GAME)->game_state) [0])
#define  GAME_STATE_BUP_1(GAME)		(SCM_VELTS ((GAME)->game_state) [1])
#define  GAME_STATE_BUP_2(GAME)		(SCM_VELTS ((GAME)->game_state) [2])

typedef struct {
	int		ball_id;
	char	       *prompt;
	SCM		validate;
	SCM		after;
} BallInHand;

enum _GameState {
	GAME_RUNNING,
	GAME_ENDED,
};

struct _Game {

	/* general info */
	gchar           * name;
	gint	            id;
	gchar		* help_file;
	
	gint stroke_number;
	GameState state;

	GList * ball_out;

	/* anim events */
	AnimEventType     anim_event_mask;
	gint		  anim_event_ball_mantinel_num;
	gint		  anim_event_ball_ball_num;

	/* queries */
	GList * query_cur;
	GList * query_list;

	/* variants */
	GList * variant_list;

	/* list of game events to process and processed (history) */
	GSList * events;
	GSList * events_processed;

	/* balls */
	gint       balls_nof;
	PoolBall * balls;
	GList	 * balls_in_hand_list;

	/* animation */
	Animation	* animation;

	/* players */
	gint     min_players_nof;
	gint     max_players_nof;
	gint     nof_players;
	gint     cur_player;
	Player **player;
	gint     nscore;
	GList   *score_type;
	GList   *score_name;

        /* views */

	gint		default_view_position;
	ViewPosition	view_positions [GAME_MAX_NOF_VIEW_POS];

	/* flags */
	gint view_answer;

	/* callbacks */

	SCM	setup_table_func;
	SCM	new_game_func;
	SCM	game_rules_func;

	SCM	end_of_shot_func;
	SCM     next_stroke_func;

	SCM	ball_move_verify_func;
	int	ball_move_id;

	/* game state */
	int	safe_to_undo;
	SCM	game_state_deep_copy_func;
	SCM	game_state;
};

Game * game_new        (gint, gint, gint);
void   game_set_nof_balls (Game *g, gint nof_balls);

void   game_start      (Game *game);
void   game_end        (Game *game, gint type);
void   game_select     (Game *game);
void   game_unselect   (Game *game);
gint   game_add_player (Game *game, gchar *name);
void   game_set_player (Game *g, gint p);

void   game_save (Game *game, gchar *prefix);
void   game_load (Game *game, gchar *prefix);

void   game_add_score (Game *g, gint t, gchar *name);

void   game_ball_out_recover (Game *game);
void   game_next_query   (Game *game);
void   game_answer_query (Game *game, PoolViewGL *view, int x, int y);
void   game_ball_in_hand_answer (Game *game);

int    game_get_player_points (Game *, gint pid, gint counter);
int    game_cue_active (Game *game);

BallInHand* ball_in_hand_new (int ball_id);
int    pool_place_ball_is_correct (int, BallInHand *, PoolBallSpaceInfo *, PoolBallSpaceInfo *);

void   game_add_event                 (Game *, GameEvent *);
void   game_add_event_type            (Game *, gint t);
void   game_add_event_type_data       (Game *, gint t, gpointer * d);

/* these do not run process event - use it if you want add more events */
void   game_add_event_nop             (Game *, GameEvent *);
void   game_add_event_nop_type        (Game *, gint t);
void   game_add_event_nop_type_data   (Game *, gint t, gpointer * d);

void   game_process_events            (Game *game);
int    game_process_events_inside     (void);

/* static part */

extern GList * game_list;       /* list of all games, created during gtulpas_guile_init */
extern gint    game_max_id;     /* maximal game id */

#endif
