/* game events
   $Id: gameevent.h,v 1.12 2000/02/29 12:19:36 rodo Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef __GAME_EVENT_H__
#define __GAME_EVENT_H__

#include "mathutils.h"
#include <glib.h>

typedef enum   _GameEventType    GameEventType;
typedef struct _GameEvent        GameEvent;
typedef struct _EventCueXBall    EventCueXBall;
typedef struct _EventScore       EventScore;
typedef struct _EventResetScore  EventResetScore;

/* game events */

enum _GameEventType
{
	NEW_GAME,              /* start new game */
	CUE_X_BALL,            /* shot begin */
	MOVE_MY,
	MOVE_OTHER,
	END_WIN,
	END_LOSE,
	END_EQ,
	END_HIGHER_SCORE,
	FAUL,
	DO_QUERIES,
	NEXT_QUERY,
	DO_QUERY,
	SHOT_END,
	NEXT_PLAYER,
	SET_PLAYER,
	NEXT_STROKE,
	DO_STROKE,

	POINT,                  /* ball's hit - adds one point to score */
	PENALTY,		/* penalty - add points to opponent */
	RESET_SCORE,

	POCKETED,
	BALL_OUT,
	BALL_IN_HAND,
	PLACE_BALL_NEXT,
	BALL_ON,

	UNDO,
	UNDO_ANSWERS,

	EVENTS
};

struct _GameEvent
{
	gint       type;
	gpointer * data;
};

GameEvent * game_event_new                 (gint t, gpointer * d);

struct _EventCueXBall
{
	dbl top    [3];
	dbl newpos [3];
	dbl speed;
};

struct _EventScore
{
	gint counter;
	gint data;
};

struct _EventResetScore
{
	gint   counter;
	gint   pid;
	gint * data;
};

EventCueXBall   * event_cue_x_ball_new  (dbl *t, dbl *n, dbl s);
EventScore      * event_score_new       (gint c, gint d);
EventResetScore * event_reset_score_new (gint c, gint pid);

#endif
