/* widget/dialog for cue x ball positioning
   $Id: gnome-ball.c,v 1.11 2000/02/28 16:03:35 rodo Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <gnome.h>
#include "config.h"
#include "gnome-ball.h"
#include "point_mark_c.xpm"
#include "debug.h"
#include "gtulpas.h"
#include "cue.h"
#include <math.h>

static GdkPixmap * point = NULL;
static GdkBitmap * point_mask;
static GdkGC     * point_gc;

#define BALL_CENTER_X	52
#define BALL_CENTER_Y	52
#define BALL_RADIUS	45
#define BALL_RADIUS_SQUARED	(BALL_RADIUS * BALL_RADIUS)

#define BALL_WIDTH  100
#define BALL_HEIGHT 100

#define CROSS_CENTER_X  11
#define CROSS_CENTER_Y  11

#define CROSS_WIDTH  24
#define CROSS_HEIGHT 24

static void
init_point (GdkWindow *w)
{
	point = gdk_pixmap_create_from_xpm_d (w,
					      &point_mask,
					      NULL,
					      point_mark_xpm);
	point_gc = gdk_gc_new (w);
	gdk_gc_set_clip_mask (point_gc, point_mask);
}

#define PX(x) ((x < 0) ? 0 : (x>=BALL_WIDTH)  ? BALL_WIDTH-1 : x)
#define PY(y) ((y < 0) ? 0 : (y>=BALL_HEIGHT) ? BALL_HEIGHT-1 : y)

static void
place_point_mark (GnomeBall *ball, gint x, gint y)
{
	gint xi = (ball->pixmap->allocation.width  - BALL_WIDTH)>>1;
	gint yi = (ball->pixmap->allocation.height - BALL_HEIGHT)>>1;

	if (ball->point_bg) {
		/* restore original */
		gdk_gc_set_clip_origin (point_gc, ball->x, ball->y);
		gdk_window_copy_area (GNOME_PIXMAP (ball->pixmap)->pixmap,
				      point_gc, ball->x, ball->y, ball->point_bg, 0, 0, 24, 24);
		gdk_gc_set_clip_origin (point_gc, ball->x+xi, ball->y+yi);
		gdk_window_copy_area (ball->pixmap->window,
				      point_gc, ball->x+xi, ball->y+yi, ball->point_bg, 0, 0, 24, 24);
	} else
		ball->point_bg = gdk_pixmap_new (ball->pixmap->window, 24, 24, -1);

	/* backup original */
	gdk_gc_set_clip_origin (point_gc, 0, 0);
	gdk_window_copy_area (ball->point_bg,
			      point_gc, 0, 0, GNOME_PIXMAP (ball->pixmap)->pixmap, x, y, 24, 24);

	/* draw new */
	gdk_gc_set_clip_origin (point_gc, x, y);
	gdk_window_copy_area (GNOME_PIXMAP (ball->pixmap)->pixmap, point_gc, x, y, point, 0, 0, 24, 24);
	gdk_gc_set_clip_origin (point_gc, x+xi, y+yi);
	gdk_window_copy_area (ball->pixmap->window, point_gc, x+xi, y+yi, point, 0, 0, 24, 24);

	ball->x = x;
	ball->y = y;

	cue_adjust (&pool_cue, 
		    ((dbl) (x - BALL_CENTER_X + CROSS_CENTER_X)) / (dbl) BALL_RADIUS, 
		    ((dbl) (y - BALL_CENTER_Y + CROSS_CENTER_Y)) / (dbl) BALL_RADIUS);
}

static void
realize (GtkWidget *w, GnomeBall *ball)
{
	debug ("realize event\n");

	if (!point)
		init_point (w->window);

	place_point_mark (ball, ball->x, ball->y);
}

#undef X
#undef Y

#define X (*x)
#define Y (*y)

static void
correct_xy (gint *x, gint *y)
{
	gdouble p;

	/* place it in the origin */

	X -= BALL_CENTER_X;
	Y -= BALL_CENTER_Y;

	if (X*X + Y*Y > BALL_RADIUS_SQUARED) {

		p = sqrt ((gdouble) BALL_RADIUS_SQUARED / (X*X + Y*Y));

		X = p*X;
		Y = p*Y;

		/* now we have corrected coordinates */
	}

	/* move it to center */

	X += BALL_CENTER_X - CROSS_CENTER_X;
	Y += BALL_CENTER_Y - CROSS_CENTER_Y;
}

#define GRAB_MASK   GDK_ENTER_NOTIFY_MASK	\
	          | GDK_LEAVE_NOTIFY_MASK  	\
		  | GDK_BUTTON_PRESS_MASK 	\
		  | GDK_BUTTON_RELEASE_MASK 	\
		  | GDK_POINTER_MOTION_MASK	\
		  | GDK_POINTER_MOTION_HINT_MASK 

static void
button_press_event (GtkWidget *w, GdkEventButton *e, GnomeBall *ball)
{
	gint    x, y;

	x = (int) e->x;
	y = (int) e->y;

	x -= (ball->pixmap->allocation.width  - BALL_WIDTH)>>1;
	y -= (ball->pixmap->allocation.height - BALL_HEIGHT)>>1;

	if (e->type == GDK_BUTTON_PRESS) {

		correct_xy (&x, &y);
		place_point_mark (ball, x, y);

		debug ("grab\n");
  
		ball->pressed = TRUE;
		gdk_pointer_grab (w->window, TRUE, GRAB_MASK, 
				  NULL, NULL, e->time);
		gtk_grab_add (w);
	} else {

		debug ("ungrab\n");

		ball->pressed = FALSE;
		gdk_pointer_ungrab (e->time);
		gtk_grab_remove (w);
	}
  
}

static void
motion_notify_event (GtkWidget *w, GdkEventMotion *e, GnomeBall *ball)
{
	GdkModifierType state;
	gint    x, y;
	gdouble p;

	if (!ball->pressed)
		return;

	if (e->is_hint)
		gdk_window_get_pointer (w->window, &x, &y, &state);
	else {
		x = (int) e->x;
		y = (int) e->y;
	}

	x -= (ball->pixmap->allocation.width  - BALL_WIDTH)>>1;
	y -= (ball->pixmap->allocation.height - BALL_HEIGHT)>>1;

	correct_xy (&x, &y);

	place_point_mark (ball, x, y);
}

GnomeBall *
gnome_ball_new (GtkWidget *w)
{
	GnomeBall * new_ball = g_new (GnomeBall, 1);
	GtkWidget * ebox = gtk_event_box_new ();
	gchar *fname;

	/* initial place it on the center 50 - 12 */
	new_ball->x =  BALL_CENTER_X - CROSS_CENTER_X;
	new_ball->y =  BALL_CENTER_Y - CROSS_CENTER_Y;
	new_ball->point_bg        = NULL;
	new_ball->pressed         = FALSE;

	new_ball->frame  = gtk_frame_new (NULL /* _("Cue position") */);
	gtk_container_set_border_width (GTK_CONTAINER (new_ball->frame), 2);
	gtk_frame_set_shadow_type (GTK_FRAME (new_ball->frame), GTK_SHADOW_OUT); 

	fname = gnome_unconditional_datadir_file ("gtulpas/pics/ball.png");
	new_ball->pixmap = gnome_pixmap_new_from_file (fname);
	g_free (fname);

	/* GTK_WIDGET_SET_FLAGS (new_ball->pixmap, GTK_CAN_FOCUS | GTK_CAN_DEFAULT); */
	gtk_widget_add_events (new_ball->pixmap,
			       GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK);

	gtk_signal_connect (GTK_OBJECT (new_ball->pixmap), "realize",
			    GTK_SIGNAL_FUNC (realize), new_ball);
	gtk_signal_connect (GTK_OBJECT (ebox), "button_press_event",
			    GTK_SIGNAL_FUNC (button_press_event), new_ball);
	gtk_signal_connect (GTK_OBJECT (ebox), "motion_notify_event",
			    GTK_SIGNAL_FUNC (motion_notify_event), new_ball);
	gtk_signal_connect (GTK_OBJECT (ebox), "button_release_event",
			    GTK_SIGNAL_FUNC (button_press_event), new_ball);

	gtk_container_add (GTK_CONTAINER (ebox), new_ball->pixmap);
	gtk_widget_show_all (GTK_WIDGET (new_ball -> frame));

	gtk_container_add (GTK_CONTAINER (new_ball->frame), ebox);
	gtk_widget_show_all (GTK_WIDGET (new_ball->frame));

	new_ball->ebox = ebox;

	return new_ball;
}

