/* gnome wrapper around poolview
   $Id: gnome-poolview.c,v 1.4 1999/12/13 18:51:09 0rfelyus Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "debug.h"
#include "gnome-poolview.h"
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-config.h>
#include <libgnomeui/gnome-geometry.h>

GnomePoolView *
gnome_pool_view_new (GnomePoolView *to_clone)
{
	GnomePoolView * view = g_new (GnomePoolView, 1);

	view->gl_view = pool_view_new ((to_clone) ? GTK_WIDGET (to_clone->gl_view->area) : NULL);
	view->widget  = gtk_frame_new (NULL);
	view->topwindow = NULL;

	gtk_frame_set_shadow_type (GTK_FRAME (view->widget), GTK_SHADOW_NONE);
	gtk_container_set_border_width (GTK_CONTAINER (view->widget), 1);
	gtk_container_add (GTK_CONTAINER (view->widget), GTK_WIDGET (view->gl_view->area));

	return view;
}

void
gnome_pool_view_save (GnomePoolView *view, GtkWidget *window, gchar *prefix)
{
	gnome_config_push_prefix (prefix);

	/* debug ("save prefix %s %f %f %f\n", prefix,
	       view->gl_view->eye [0],
	       view->gl_view->eye [1],
	       view->gl_view->eye [2]); */

	gnome_config_set_float ("center_of_rotation_x", view->gl_view->vp.center_of_rotation [0]);
	gnome_config_set_float ("center_of_rotation_y", view->gl_view->vp.center_of_rotation [1]);
	gnome_config_set_float ("center_of_rotation_z", view->gl_view->vp.center_of_rotation [2]);

	gnome_config_set_float ("eye_focus_distance", view->gl_view->vp.eye_focus_distance);
	gnome_config_set_float ("eye_center_distance", view->gl_view->vp.eye_center_distance);
	gnome_config_set_float ("alfa", view->gl_view->vp.alfa);
	gnome_config_set_float ("beta", view->gl_view->vp.beta);

	gnome_config_set_int ("lod", view->gl_view->lod);
	gnome_config_set_int ("lights", view->gl_view->with_lights);
	gnome_config_set_int ("shading", view->gl_view->shading);
	gnome_config_set_int ("smooth", view->gl_view->smooth_shading);

	if (window)
	  {
	    char *geom_string;

	    geom_string = gnome_geometry_string (window->window);
	    gnome_config_set_string ("geometry", geom_string);

	    g_free (geom_string);
	  }

	gnome_config_pop_prefix ();
}

void
gnome_pool_view_load (GnomePoolView *view, GtkWidget *window, gchar *prefix)
{
	gnome_config_push_prefix (prefix);

	view->gl_view->vp.center_of_rotation [0] = gnome_config_get_float ("center_of_rotation_x=0");
	view->gl_view->vp.center_of_rotation [1] = gnome_config_get_float ("center_of_rotation_y=0");
	view->gl_view->vp.center_of_rotation [2] = gnome_config_get_float ("center_of_rotation_z=0");

	view->gl_view->vp.eye_focus_distance  = gnome_config_get_float ("eye_focus_distance=100");
	view->gl_view->vp.eye_center_distance = gnome_config_get_float ("eye_center_distance=100");
	view->gl_view->vp.alfa = gnome_config_get_float ("alfa=.7");
	view->gl_view->vp.beta = gnome_config_get_float ("beta=.7");

	view->gl_view->lod            = gnome_config_get_int ("lod=1");
	view->gl_view->with_lights    = gnome_config_get_int ("lights=1");
	view->gl_view->shading        = gnome_config_get_int ("shading=1");
	view->gl_view->smooth_shading = gnome_config_get_int ("smooth=1");

	view_update_orientation (view->gl_view);
	view_update_position    (view->gl_view);

	if (window)
	  {
	    char	*geomstring;
	    gint	x,y,width, height;

	    geomstring = gnome_config_get_string ("geometry=200x200+10+10");

	    gnome_parse_geometry (geomstring, &x, &y, &width, &height);
	    gtk_widget_set_uposition (window, x,y);
	    gtk_widget_set_usize (window, width, height);

	    g_free (geomstring);
	  }

	gnome_config_pop_prefix ();
}

