/* widget for scores displayed using color balls
   $Id: gnome-score-balls.c,v 1.24 2000/03/02 12:08:48 0rfelyus Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "debug.h"
#include "gnome-score-balls.h"
#include "digital.h"
#include "score.h"
#include "game.h"
#include "poolfeedgl.h"
#include "c_disable.xpm"
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-util.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gnome-pixmap.h>
#include <string.h>
#include <assert.h>

#define COUNTS          4
#define COLORED         6
#define PIXMAPS         COUNTS+COLORED
#define SCORE_WIDTH     100
#define SCORE_BALLS     5

#define COUNTS_OFF      COLORED

static GdkPixmap   * cell_pixmap [CELL_PIXMAPS];
static GdkPixmap   * cell_ball = NULL;
static GdkPixmap   * cell_faul;


static GdkColor   bg_color = {0, 0x0000, 0x0000, 0x0000};
static GdkColor   gray_color = {0, 0x0000, 0x0000, 0x0000};
static GdkGC     *bg_gc;
static GdkPixmap *disable_pixmap = NULL;
static GdkPixmap *disable_pixmap_svballs;
static gint       count [COUNTS] = {10, 5, 2, 1};

static void
create_cells (GtkWidget *w)
{
	gint i;
	gchar * fname, *fn;

	gdk_color_alloc (w->style->colormap, &bg_color);
	gdk_color_alloc (w->style->colormap, &gray_color);
	bg_gc = gdk_gc_new (w->window);
	gdk_gc_set_foreground (bg_gc, &bg_color);
	gdk_gc_set_background (bg_gc, &bg_color);

	for (i=0; i<CELL_PIXMAPS; i++)
		cell_pixmap [i] = gdk_pixmap_create_from_xpm_d (w->window, NULL, &bg_color, cell_digit [i]);

	cell_ball = gdk_pixmap_create_from_xpm_d (w->window, NULL, &bg_color, ball_xpm);
	cell_faul = gdk_pixmap_create_from_xpm_d (w->window, NULL, &bg_color, faul_xpm);
}

static void
realize (GtkWidget *w, GnomeScoreBalls *sc)
{
	debug ("realize area\n"); 

	if (!cell_ball)
		create_cells (w);

	debug ("realize w:%d h:%d\n", w->allocation.width, w->allocation.height);
	sc->pixmap = gdk_pixmap_new (w->window, w->allocation.width, w->allocation.height, -1);

	gnome_score_balls_refresh (sc);

	if (!disable_pixmap) {
		disable_pixmap = gdk_bitmap_create_from_data (w->window, "\xff\x00\xff", 8, 3);
		disable_pixmap_svballs = gdk_bitmap_create_from_data (w->window, "\x55", 2, 2);
	}
}

static void
expose (GtkWidget *w, GdkEventExpose *e, GnomeScoreBalls *sc)
{
	debug ("expose\n");

	if (sc->refresh) {
		gnome_score_balls_refresh (sc);
		sc->refresh = 1;
	} else
		gdk_window_copy_area (w->window, bg_gc, 0, 0, sc->pixmap, 0, 0, w->allocation.width, w->allocation.height);
}

static void
size_request (GtkWidget       * widget,
	      GtkRequisition  * requisition,
	      GnomeScoreBalls * sc)
{
	requisition->width  = (sc->score->type != SVBalls) ? CELL_WIDTH*CELLS : BALLS_PER_LINE*BALL_OFF;
	requisition->height = CELL_HEIGHT*sc->lines;

	debug ("size request %d %d\n", requisition->width, requisition->height);
}

static void
size_allocate (GtkWidget       * widget,
	       GtkAllocation   * allocation,
	       GnomeScoreBalls * sc)
{
	allocation->width  = (sc->score->type != SVBalls) ? CELL_WIDTH*CELLS : BALLS_PER_LINE*BALL_OFF;
	allocation->height = CELL_HEIGHT*sc->lines;
	widget->allocation = *allocation;

	debug ("size allocate %d %d\n", allocation->width, allocation->height);

	if (GTK_WIDGET_REALIZED (widget))
		realize (widget, sc);
}

GnomeScoreBalls *
gnome_score_balls_new (GnomeTulpasScore *gs, Score *s)
{
	GnomeScoreBalls * new_score;

	new_score = g_new (GnomeScoreBalls, 1);

	new_score->gs      = gs;
	new_score->score   = s;
	new_score->lines   = 1;
	new_score->active  = 1;
	new_score->refresh = 1;

	gtk_widget_push_colormap (gdk_imlib_get_colormap ());
	gtk_widget_push_visual (gdk_imlib_get_visual ());

	new_score->area   = gtk_drawing_area_new ();
	new_score->pixmap = NULL;

	gtk_widget_pop_visual ();
	gtk_widget_pop_colormap ();

	// gtk_widget_set_usize (new_score->area, CELLS*CELL_WIDTH, CELL_HEIGHT);

	gtk_signal_connect (GTK_OBJECT (new_score->area), "realize",
			    GTK_SIGNAL_FUNC (realize), new_score);
	gtk_signal_connect (GTK_OBJECT (new_score->area), "expose_event",
			    GTK_SIGNAL_FUNC (expose), new_score);
	gtk_signal_connect (GTK_OBJECT (new_score->area), "size_request",
			    GTK_SIGNAL_FUNC (size_request), new_score);
	gtk_signal_connect (GTK_OBJECT (new_score->area), "size_allocate",
			    GTK_SIGNAL_FUNC (size_allocate), new_score);

	return new_score;
}

static gint
count_lines (gint n) {
	gint i, balls=0;

	for (i=0; i<COUNTS; i++) {
		balls += n/count[i];
		n %= count [i];
	}

	return balls / SCORE_BALLS + ((balls % SCORE_BALLS) ? 1 : 0);
}

void
gnome_score_balls_refresh (GnomeScoreBalls *sc)
{
	gint new_lines, i;

	if (!sc->pixmap || !sc->area->window)
		return;

	switch (sc->score->type) {
	case SNumbers: {
		gint n = sc->score->hits;
		for (i=0; i<CELLS; i++, n=n/10)
			gdk_window_copy_area (sc->pixmap, bg_gc,
					      (CELLS - i - 1)*CELL_WIDTH, 0,
					      (!i || n>0) ? cell_pixmap [n % 10] : cell_pixmap [CELL_BLANK],
					      0, 0, CELL_WIDTH, CELL_HEIGHT);
		break;
	}
	case SBalls:
	case SFauls: {
		gint n = sc->score->hits;
		if (n<5) {
			for (i=0; i<CELLS; i++)
				gdk_window_copy_area (sc->pixmap, bg_gc,
						      (CELLS - i - 1)*CELL_WIDTH, 0,
						      (i<n) ? (sc->score->type == SBalls) ? cell_ball : cell_faul : cell_pixmap [CELL_BLANK],
						      0, 0, CELL_WIDTH, CELL_HEIGHT);
				
		} else {
			gdk_window_copy_area (sc->pixmap, bg_gc,
					      0, 0,
					      (sc->score->type == SBalls) ? cell_ball : cell_faul,
					      0, 0, CELL_WIDTH, CELL_HEIGHT);
			for (i=0; i<CELLS-1; i++, n=n/10)
				gdk_window_copy_area (sc->pixmap, bg_gc,
						      (CELLS - i - 1)*CELL_WIDTH, 0,
						      (!i || n>0) ? cell_pixmap [n % 10] : cell_pixmap [CELL_BLANK],
						      0, 0, CELL_WIDTH, CELL_HEIGHT);
		}
		break;
	}
	case SVBalls: {
		
		gint i;

		debug ("lines: %d n: %d\n", sc->lines, sc->score->sunken_count);

		if (sc->score->sunken_count / BALLS_PER_LINE + ((sc->score->sunken_count % BALLS_PER_LINE) ? 1 : 0) > sc->lines) {
			if (sc->pixmap) {
				gdk_pixmap_unref (sc->pixmap);
			}
			sc->lines = sc->score->sunken_count / BALLS_PER_LINE + ((sc->score->sunken_count % BALLS_PER_LINE) ? 1 : 0);

			debug ("resize lines %d w:%d h:%d\n", sc->lines, BALLS_PER_LINE*BALL_OFF, CELL_HEIGHT*sc->lines);
			sc->pixmap = gdk_pixmap_new (sc->area->window, BALLS_PER_LINE*BALL_OFF, CELL_HEIGHT*sc->lines, -1);
			gtk_widget_queue_resize (sc->area);
		}

		gdk_draw_rectangle (sc->pixmap, bg_gc, TRUE, 0, 0, BALLS_PER_LINE*BALL_OFF, CELL_HEIGHT*sc->lines);

		for (i=0; i<sc->score->sunken_count; i++) {
			if (!sc->gs->ball_pixmap [sc->score->sunken [i]])
				render_score_ball (sc->gs->game, sc->score->sunken [i], BALL_PIXMAP_WIDTH, BALL_PIXMAP_WIDTH,
						   &sc->gs->ball_pixmap [sc->score->sunken [i]],
						   &sc->gs->ball_bitmap [sc->score->sunken [i]]);
			
			debug ("ball: %d\n", sc->score->sunken [i]);

			gdk_window_copy_area (sc->pixmap, bg_gc,
					      (i%BALLS_PER_LINE)*BALL_OFF+1, (i / BALLS_PER_LINE)*BALL_OFF+1,
					      sc->gs->ball_pixmap [sc->score->sunken [i]],
					      0, 0, BALL_PIXMAP_WIDTH, BALL_PIXMAP_WIDTH);
		}
		
		break;
	}
	default:
		g_assert_not_reached ();
	}

	if (!sc->active) {
		gdk_gc_set_foreground (bg_gc, &gray_color);
		gdk_gc_set_background (bg_gc, &gray_color);
		gdk_gc_set_stipple (bg_gc, (sc->score->type == SVBalls) ? disable_pixmap_svballs : disable_pixmap);
		gdk_gc_set_ts_origin (bg_gc, 1, 1);
		gdk_gc_set_fill    (bg_gc, GDK_STIPPLED);
		gdk_draw_rectangle (sc->pixmap, bg_gc, TRUE, 0, 0, sc->area->allocation.width, sc->area->allocation.height);
		gdk_gc_set_fill    (bg_gc, GDK_SOLID);
	}

	gdk_window_copy_area (sc->area->window, bg_gc,
			      0, 0, sc->pixmap,
			      0, 0, sc->area->allocation.width, sc->area->allocation.height);
}

/* tulpas score begin */

GnomeTulpasScore *
gnome_tulpas_score_new (Game *g)
{
	GnomeTulpasScore * new_score;
	GtkWidget *frame, *vbox, *label, *hbox;
	gint i,j,k, vb, vbb;

	new_score = g_new (GnomeTulpasScore, 1);

	new_score->players = g->nof_players;
	new_score->game    = g;

	new_score->active_player = -1;

	new_score->ball_pixmap = g_new (GdkPixmap *, g->balls_nof);
	new_score->ball_bitmap = g_new (GdkBitmap *, g->balls_nof);
	for (i=0; i<g->balls_nof; i++) {
		new_score->ball_pixmap [i] = NULL;
		new_score->ball_bitmap [i] = NULL;
	}

	new_score->score = (GnomeScoreBalls **) g_new (GnomeScoreBalls *, new_score->players*g->nscore);
	/* player name */
	new_score->label   = (GtkWidget **)       g_new (GtkWidget *, new_score->players);
	new_score->tab     = (GtkWidget **)       g_new (GtkWidget *, new_score->players);

	new_score->ball_on    = (GtkWidget **)       g_new (GtkWidget *, new_score->players);
	new_score->ball_frame = (GtkWidget **)       g_new (GtkWidget *, new_score->players);

	new_score->frame = gtk_frame_new (NULL /* _("Score") */ );
	gtk_container_set_border_width (GTK_CONTAINER (new_score->frame), 0);
	gtk_frame_set_shadow_type (GTK_FRAME (new_score->frame), GTK_SHADOW_NONE); 

	new_score->vbox  = gtk_vbox_new (FALSE, 0);

	for (i=0, k=0; i<new_score->players; i++) {

		vb  = 0;
		vbb = 1;

		new_score->ball_on [i] = NULL;

		hbox = gtk_hbox_new (FALSE, 0);

		frame = gtk_frame_new (NULL);
		gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT); 
		new_score->label [i]         = gtk_label_new (g->player [i]->name);
		gtk_container_add (GTK_CONTAINER (frame), new_score->label [i]);
		gtk_box_pack_start_defaults (GTK_BOX (hbox), frame);

		new_score->ball_frame [i] = gtk_frame_new (NULL);
		gtk_widget_set_usize (new_score->ball_frame [i], BALL_PIXMAP_WIDTH, BALL_PIXMAP_WIDTH);
		gtk_frame_set_shadow_type (GTK_FRAME (new_score->ball_frame [i]), GTK_SHADOW_OUT); 
		gtk_box_pack_start (GTK_BOX (hbox), new_score->ball_frame [i], FALSE, FALSE, 0);

		/* add active ball pixmap here */

		gtk_box_pack_start (GTK_BOX (new_score->vbox), hbox, FALSE, FALSE, 0);

		frame = gtk_frame_new (NULL);
		gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT); 
		new_score->tab   [i]         = gtk_table_new (g->nscore+1, 3, FALSE);
		gtk_table_set_col_spacing (GTK_TABLE (new_score->tab [i]), 0, 2);
		gtk_container_set_border_width (GTK_CONTAINER (new_score->tab [i]), 3);
		gtk_container_add (GTK_CONTAINER (frame), new_score->tab [i]);
		gtk_box_pack_start (GTK_BOX (new_score->vbox), frame, FALSE, FALSE, 0);
		vbox = gtk_vbox_new (FALSE, 0);

		for (j=0; j<g->nscore; j++, k++) {
			new_score->score [k]         = gnome_score_balls_new (new_score, g->player [i]->score [j]);
			new_score->score [k]->active = (i==g->cur_player) ? TRUE : FALSE;
			if (new_score->score [k]->score->type != SVBalls) {
				label = gtk_label_new (new_score->score [k]->score->name);
				gtk_misc_set_alignment (GTK_MISC (label), 1.0, 1.0);
			
				gtk_table_attach (GTK_TABLE (new_score->tab [i]), label,
						  0,1,j+1,j+2, GTK_FILL|GTK_EXPAND|GTK_SHRINK, GTK_FILL|GTK_EXPAND|GTK_SHRINK, 0, 0);
				gtk_box_pack_start (GTK_BOX (vbox), new_score->score [k]->area,
						    FALSE, FALSE, 0);
				vb++;
			} else {
				
				if (vb) {
					frame = gtk_frame_new (NULL);
					gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN); 
					gtk_container_set_border_width (GTK_CONTAINER (frame), 0);
					gtk_container_add (GTK_CONTAINER (frame), vbox);
					gtk_table_attach (GTK_TABLE (new_score->tab [i]), frame,
							  1, 3, vbb, vbb+vb, 0, 0, 0, 0);
					vbox = gtk_vbox_new (FALSE, 0);
				}

				frame = gtk_frame_new (NULL);
				hbox  = gtk_hbox_new (FALSE, 0);
				gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN); 
				gtk_container_set_border_width (GTK_CONTAINER (frame), 0);
				gtk_container_add (GTK_CONTAINER (frame), new_score->score [k]->area);
				gtk_box_pack_end (GTK_BOX (hbox), frame,
						  FALSE, FALSE, 0);
				gtk_table_attach (GTK_TABLE (new_score->tab [i]), hbox,
						  0,3,j+1,j+2, GTK_FILL|GTK_EXPAND|GTK_SHRINK, GTK_FILL|GTK_EXPAND|GTK_SHRINK, 0, 0);
				vbb = j+2;
				vb  = 0;
			}
		}
		if (vb) {
			frame = gtk_frame_new (NULL);
			gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN); 
			gtk_container_set_border_width (GTK_CONTAINER (frame), 0);
			gtk_container_add (GTK_CONTAINER (frame), vbox);
			gtk_table_attach (GTK_TABLE (new_score->tab [i]), frame,
					  1, 3, vbb, vbb+vb, 0, 0, 0, 0);
		}
	}

	gtk_container_add (GTK_CONTAINER (new_score->frame), new_score->vbox);

	return new_score;
}


void
gnome_tulpas_score_refresh (GnomeTulpasScore *s)
{
	gint i,j,k;

	for (i=0,k=0; i<s->players; i++) {
		gtk_label_set_text (GTK_LABEL (s->label [i]), s->game->player [i]->name);
		for (j=0; j<s->game->nscore; j++,k++) {
			gnome_score_balls_refresh (s->score [k]);
		}
	}
}

void
gnome_tulpas_score_set_player (GnomeTulpasScore *s, gint p)
{
	gint i,j,k;

	s->active_player = p;

	if (s->game->nscore) {
		for (i=0,k=0; i<s->players; i++) {
			gtk_widget_set_sensitive (s->label [i], (i==p) ? TRUE : FALSE);
			for (j=0; j<s->game->nscore; j++,k++)
				s->score [k]->active = (p == i) ? TRUE : FALSE;
		}

		gnome_tulpas_score_refresh (s);
	}
}

void
gnome_tulpas_score_ball_on (GnomeTulpasScore *s, gint bid)
{
	gint i;

	for (i=0; i<s->players; i++) {
		if (i == s->active_player) {
			if (s->ball_on [i])
				gtk_container_remove (GTK_CONTAINER (s->ball_frame [i]), s->ball_on [i]);
			if (!s->ball_pixmap [bid])
				render_score_ball (s->game, bid, BALL_PIXMAP_WIDTH, BALL_PIXMAP_WIDTH,
						   &s->ball_pixmap [bid], &s->ball_bitmap [bid]);
			s->ball_on [i] = gtk_pixmap_new (s->ball_pixmap [bid], s->ball_bitmap [bid]);
			gtk_container_add (GTK_CONTAINER (s->ball_frame [i]), s->ball_on [i]);
			gtk_widget_show_all (s->ball_frame [i]);
		} else {
			if (s->ball_on [i]) {
				gtk_container_remove (GTK_CONTAINER (s->ball_frame [i]), s->ball_on [i]);
				s->ball_on [i] = NULL;
			}
		}
	}
}
