#include <gnome.h>
#include "gnome-ball.h"
#include "gtulpas.h"

#include <gdk/gdkprivate.h>
#include <string.h>

void
gtulpas_status_start (char *message)
{
  gnome_appbar_push (gtulpas->bar, message);
}

void
gtulpas_status_end (void)
{
  gnome_appbar_pop (gtulpas->bar);
}

void
gtulpas_status_message (char *message)
{
  gnome_appbar_set_status (gtulpas->bar, message);
}

void
gtulpas_status_clear (void)
{
  gnome_appbar_clear_stack(gtulpas->bar);
}

static GtkAdjustment * panel_position_adjustment = NULL;
static GtkAdjustment * panel_speed_adjustment = NULL;

static dbl panel_max = 1;
static int lock = 0;
static int ignore = 0;

static void
panel_posiotion_cb (GtkAdjustment *adj)
{
  if (ignore > 0)
    { ignore --; return; }

  lock = 1;
  animation_set_position (adj -> value);
  lock = 0;
}

void
anim_panel_position_max (dbl maxpos)
{
  panel_max = maxpos;

  if (! panel_position_adjustment)
    return;

  panel_position_adjustment -> upper = maxpos;
  gtk_adjustment_changed (panel_position_adjustment);
}

void 
anim_panel_set_position (dbl pos)
{
  if (lock)
    return;

  if (! panel_position_adjustment)
    return;

  if (panel_position_adjustment -> value == pos)
    return;

  ignore ++;
  gtk_adjustment_set_value (panel_position_adjustment, pos);
}

static void
panel_speed_cb (GtkAdjustment *adj)
{
  animation_set_speed (adj -> value);
}

static void
panel_start_cb (void)
{
  gtk_adjustment_set_value (panel_position_adjustment, 0);
}

static void
panel_last_cb (void)
{
  gtk_adjustment_set_value (panel_position_adjustment, panel_position_adjustment -> upper);
}

static void
panel_very_last_cb (void)
{
  animation_force_complete ();
  panel_last_cb ();
}

static void
panel_fyjik_cb (void)
{
  animation_reset_fyjik (-1);
}


static char *file_name = NULL;
static GtkFileSelection *fs_dialog = NULL;

static void
panel_fs_ok_callback (void)
{
  file_name = gtk_file_selection_get_filename (fs_dialog);
  gtk_widget_hide (GTK_WIDGET (fs_dialog));
  gtk_main_quit ();
}

static void
panel_fs_cancel_callback (void)
{
  file_name = NULL;
  gtk_widget_hide (GTK_WIDGET (fs_dialog));
  gtk_main_quit ();
}

static char*
panel_get_file_name (void)
{
  if (fs_dialog == NULL)
    {
      fs_dialog = (GtkFileSelection *) gtk_file_selection_new ("Select saved animation");
      gtk_file_selection_set_filename (fs_dialog, "tulpas.anim");

      gtk_signal_connect (GTK_OBJECT (fs_dialog ->ok_button), "clicked",
			  (GtkSignalFunc) panel_fs_ok_callback, NULL);
      gtk_signal_connect (GTK_OBJECT (fs_dialog ->cancel_button), "clicked",
			  (GtkSignalFunc) panel_fs_cancel_callback, NULL);
      gtk_signal_connect (GTK_OBJECT (fs_dialog), "delete",
			  (GtkSignalFunc) panel_fs_cancel_callback, NULL);
    }

  gtk_window_set_modal (GTK_WINDOW (fs_dialog), TRUE);
  gtk_widget_show_all (GTK_WIDGET (fs_dialog));

  gtk_main ();
  return file_name;
}

static void
panel_save_cb (void)
{
  if (panel_get_file_name ())
    animation_save (file_name);
}

static void
panel_load_cb (void)
{
  if (panel_get_file_name ())
    animation_load (file_name);
}

static GtkWidget*
panel_button (GtkWidget *window, const char *icon, GtkSignalFunc func)
{
  GtkWidget	* button;
  GtkWidget	* pixmap;
  GtkWidget	* hbox;
  
  button = gtk_button_new ();

  if (icon)
    {
      pixmap = gnome_stock_pixmap_widget (window, icon);
      gtk_container_add (GTK_CONTAINER (button), pixmap);
    }
  else
    {
      hbox = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start_defaults (GTK_BOX (hbox),
	gnome_stock_pixmap_widget (window, GNOME_STOCK_PIXMAP_FORWARD));
      gtk_box_pack_start_defaults (GTK_BOX (hbox),
	gnome_stock_pixmap_widget (window, GNOME_STOCK_PIXMAP_LAST));
      gtk_container_add (GTK_CONTAINER (button), hbox);
    }
  gtk_signal_connect (GTK_OBJECT (button), "clicked", func, NULL);

  return button;
}

void
anim_panel_show (void)
{
  static GtkWidget *window = NULL;
  
  if (window)
    if (GTK_WIDGET_VISIBLE (window))
      gtk_widget_hide (window);
    else
      gtk_widget_show (window);
  else
    {
      GtkWidget		*vbox, *hbox;
      GtkWidget		*table;
      GtkWidget		*label;
      GtkWidget		*scale;
      GtkAdjustment	*adj;

      window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_window_set_title (GTK_WINDOW (window), "Tulpas - Animation Debug Window");
      gtk_container_set_border_width (GTK_CONTAINER (window), 10);
      gtk_signal_connect (GTK_OBJECT (window), "delete_event",
			  (GtkSignalFunc) gtk_widget_hide, NULL);

      vbox = gtk_vbox_new (FALSE, 0);
      gtk_container_add (GTK_CONTAINER (window), vbox);
      
      table = gtk_table_new (2,2,FALSE);
      gtk_box_pack_start (GTK_BOX (vbox), table, FALSE, FALSE, 0);
      
      gtk_table_set_col_spacing (GTK_TABLE (table), 0, 10);

      label = gtk_label_new ("Position:");
      gtk_misc_set_alignment (GTK_MISC (label), 0, 1);
      gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

      label = gtk_label_new ("Animation speed:");
      gtk_misc_set_alignment (GTK_MISC (label), 0, 1);
      gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);

      panel_position_adjustment = adj = 
	(GtkAdjustment *) gtk_adjustment_new (0, 0, panel_max, 1, 1,1);
      scale = gtk_hscale_new (adj);
      gtk_scale_set_digits (GTK_SCALE (scale), 0);
      gtk_signal_connect (GTK_OBJECT (adj), "value_changed", 
			  (GtkSignalFunc) panel_posiotion_cb, NULL);
      gtk_table_attach (GTK_TABLE (table), scale, 1,2, 0,1,GTK_EXPAND|GTK_FILL,GTK_FILL, 0,0);


      panel_speed_adjustment = adj = 
	(GtkAdjustment *) gtk_adjustment_new (1, 0.01, 5, 0.1, 1,1);
      gtk_signal_connect (GTK_OBJECT (adj), "value_changed", 
			  (GtkSignalFunc) panel_speed_cb, NULL);
      scale = gtk_hscale_new (adj);
      gtk_table_attach (GTK_TABLE (table), scale, 1,2, 1,2,GTK_EXPAND|GTK_FILL,GTK_FILL, 0,0);

      hbox = gtk_hbox_new (FALSE, 0);
      gtk_box_pack_end (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

      gtk_box_pack_start (GTK_BOX (hbox), 
			  panel_button (window, GNOME_STOCK_PIXMAP_FIRST, panel_start_cb),
			  FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), 
			  panel_button (window, GNOME_STOCK_PIXMAP_FORWARD, animation_play),
			  FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), 
			  panel_button (window, GNOME_STOCK_PIXMAP_STOP, animation_stop),
			  FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), 
			  panel_button (window, GNOME_STOCK_PIXMAP_LAST, panel_last_cb),
			  FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), 
			  panel_button (window, NULL, panel_very_last_cb),
			  FALSE, FALSE, 0);

      gtk_box_pack_end  (GTK_BOX (hbox),
			 panel_button (window, GNOME_STOCK_PIXMAP_EXEC, panel_fyjik_cb),
			 FALSE, FALSE, 0);
      gtk_box_pack_end  (GTK_BOX (hbox),
			 panel_button (window, GNOME_STOCK_PIXMAP_SAVE_AS, panel_save_cb),
			 FALSE, FALSE, 0);
      gtk_box_pack_end  (GTK_BOX (hbox),
			 panel_button (window, GNOME_STOCK_PIXMAP_OPEN, panel_load_cb),
			 FALSE, FALSE, 0);

      gtk_widget_set_usize (window, 330, 140);
      gtk_widget_show_all (window);
    }
}

static char * ag_form_save = NULL;
static char * ag_str_save = NULL;
static int    ag_str_save_length;

static int    ag_str_main_win;

static GtkWidget  * ag_combo_box;
static GtkWidget  * ag_interleave_spin;
static GtkWidget  * ag_length_spin;
static GtkWidget  * ag_combo_box;

static void
ag_repleace_format_strings (int frame)
{
  Window		  pwin;
  GtkWidget		* widget;

  widget = ag_str_main_win 
    ?  gtulpas -> app 
    : (GtkWidget *) (gtulpas -> main_view -> gl_view -> area);
    
  pwin = ((GdkWindowPrivate *) (widget -> window)) -> xwindow;
  
  snprintf (ag_str_save, ag_str_save_length, ag_form_save, frame, pwin);
}

static void
ag_cancel_cb (GtkWidget *widget, GtkWidget *window)
{
  gtk_widget_hide (window);
}

static void
ag_ok_cb (GtkWidget *widget, GtkWidget *window)
{
  if (ag_form_save)
    g_free (ag_form_save);

  if (ag_str_save)
    g_free (ag_str_save);

  ag_form_save = g_strdup (gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (ag_combo_box)->entry)));
  ag_str_save_length =  strlen (ag_form_save) + 64;
  ag_str_save = g_new (char,ag_str_save_length + 1);

  gtk_widget_hide (window);
  anim_grab_start_grab (
      gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (ag_length_spin)),
      gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (ag_interleave_spin)));
}

static void
ag_toggle_cb (GtkWidget *widget, int value)
{
  if (GTK_TOGGLE_BUTTON (widget) -> active)
    ag_str_main_win = value;
}

void
anim_grab_panel_show (void)
{
  static char	 * predefined [] = {
    "xwd -root -out tulpas-%03d.xwd",
    "xwd -out tulpas-%03d.xwd -id %#d",
    "import -window root tulpas-%03d.png",
    "import tulpas-%03d.png -window %#d",
  };

  static GtkWidget	*window;

  animation_stop ();
  if (window == NULL)
    {
      GtkWidget *vbox;
      GtkWidget *hbox;
      GtkWidget *button;
      GtkWidget *frame;
      GtkWidget *label;

      window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_window_set_title (GTK_WINDOW (window), "Tulpas - Screenshot maker");
      gtk_container_set_border_width (GTK_CONTAINER (window), 10);
      gtk_signal_connect (GTK_OBJECT (window), "delete_event",
			  (GtkSignalFunc) gtk_widget_hide, NULL);
      
      vbox = gtk_vbox_new (FALSE, 5);
      gtk_container_add (GTK_CONTAINER (window), vbox);

      label = gtk_label_new ("Grab Animation\n\n\
This Tulpas feature allows you to save generated animation \
to images. For this task Tulpas calls external program - \
you can specify it and its parameters in combo box in this \
window. String is formated using sprintf with two number \
arguments - frame number (first) and Window id (second). You \
can use %d, %x format strings to print these parameters to \
executed string");
      
      gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_CENTER);
      gtk_box_pack_start (GTK_BOX (vbox), label,
			  FALSE, FALSE, 5);
      
      frame = gtk_frame_new ("Window ID");
      gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);
      gtk_container_set_border_width (GTK_CONTAINER (frame), 5);

      hbox = gtk_vbox_new (FALSE, 5);
      gtk_container_add (GTK_CONTAINER (frame), hbox);

      button = gtk_radio_button_new_with_label (NULL, "Main window");
      gtk_box_pack_start (GTK_BOX (hbox), button,
			  FALSE, FALSE, 0);
      gtk_signal_connect (GTK_OBJECT (button), "toggled",
			  (GtkSignalFunc) ag_toggle_cb, (gpointer) TRUE);

      button = gtk_radio_button_new_with_label_from_widget 
	(GTK_RADIO_BUTTON (button), "Main GL view");
      gtk_box_pack_start (GTK_BOX (hbox), button,
			  FALSE, FALSE, 0);
      gtk_signal_connect (GTK_OBJECT (button), "toggled",
			  (GtkSignalFunc) ag_toggle_cb, (gpointer) FALSE);

      hbox = gtk_hbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

      label = gtk_label_new ("Grab interleave:");
      gtk_misc_set_alignment (GTK_MISC (label), 0, 1);
      gtk_box_pack_start (GTK_BOX (hbox), label, 
			  FALSE, FALSE, 0);

      ag_interleave_spin = 
	gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (1, 1, 100, 1, 10, 10)),
			     1,0);
      gtk_box_pack_end (GTK_BOX (hbox), ag_interleave_spin, 
			FALSE, FALSE, 0);

      hbox = gtk_hbox_new (FALSE, 0);
      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

      label = gtk_label_new ("Grab length:");
      gtk_misc_set_alignment (GTK_MISC (label), 0, 1);
      gtk_box_pack_start (GTK_BOX (hbox), label, 
			  FALSE, FALSE, 0);

      ag_length_spin = 
	gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (1, 1, 1000, 1, 10, 10)),
			     1,0);
      gtk_box_pack_end (GTK_BOX (hbox), ag_length_spin, 
			FALSE, FALSE, 0);

      ag_combo_box = gtk_combo_new ();
      gtk_box_pack_start (GTK_BOX (vbox), ag_combo_box,
			  FALSE, FALSE, 5);

      {
	int    i;
	GList *tmp = NULL;
      
	for (i = 0; i < (int) (sizeof (predefined) / sizeof (char *)); i++)
	  tmp = g_list_prepend (tmp, predefined [i]);

	gtk_combo_set_popdown_strings (GTK_COMBO (ag_combo_box), 
				       tmp);
      }

      gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (ag_combo_box) -> entry),
			  predefined [3]);

      hbox = gtk_hbox_new (TRUE, 5);
      gtk_box_pack_end (GTK_BOX (vbox), hbox,
			FALSE, FALSE, 0);

      button = gnome_stock_button (GNOME_STOCK_BUTTON_OK);
      gtk_box_pack_start (GTK_BOX (hbox), button,
			  FALSE, FALSE, 0);
      gtk_signal_connect (GTK_OBJECT (button), "clicked",
			  (GtkSignalFunc) ag_ok_cb, window);

      button = gnome_stock_button (GNOME_STOCK_BUTTON_CANCEL);
      gtk_box_pack_start (GTK_BOX (hbox), button,
			  FALSE, FALSE, 0);
      gtk_signal_connect (GTK_OBJECT (button), "clicked",
			  (GtkSignalFunc) ag_cancel_cb, window);
    }
  gtk_widget_show_all (window);
}

void 
anim_grab_do (int frame)
{

#ifdef DEBUG
  g_message ("Grab!");
#endif
  ag_repleace_format_strings (frame);

  system (ag_str_save);
}



static GtkWidget *ge_text;
static GtkAdjustment *ge_adjustment;;

static void
ge_clicked_cb (GtkWidget *w, int bid)
{
  GtkText	*text;

  switch (bid)
    {
    case 0:
      gtk_widget_hide (w);
      /* no break */
    case 1:
      text = (GtkText *) ge_text;
      gtk_text_set_point (text, 0);
      gtk_text_forward_delete (text, gtk_text_get_length (text));
      
      break;

    case 2:
      gtk_widget_hide (w);
      break;

    default:
      g_assert_not_reached ();
    }
}

void
gtulpas_glue_error (char *errmsg, int length)
{
  static GtkWidget *window = NULL;

  if (window == NULL)
    {
      GtkWidget		*hbox;
      GtkWidget		*scroll;
      GtkWidget		*button;

      window = gnome_dialog_new ("Tulpas - Glue error output",
				 GNOME_STOCK_BUTTON_OK,
				 _("Clear"),
				 GNOME_STOCK_BUTTON_CLOSE,
				 NULL);
      gnome_dialog_close_hides (GNOME_DIALOG (window), TRUE);
      gtk_window_set_policy (GTK_WINDOW (window), FALSE, TRUE, FALSE);

      hbox = gtk_hbox_new (FALSE, 2);
      gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (window)->vbox), hbox,
			  TRUE, TRUE, 0);
      
      ge_adjustment = (GtkAdjustment *) gtk_adjustment_new (0,0,1,1,1,1);
      ge_text = gtk_text_new  (NULL, ge_adjustment);

      gtk_box_pack_start (GTK_BOX (hbox), ge_text,
			  TRUE, TRUE, 0);

      scroll = gtk_vscrollbar_new (ge_adjustment);
      gtk_box_pack_end (GTK_BOX (hbox), scroll,
			FALSE, FALSE, 0);

      gtk_signal_connect (GTK_OBJECT (window), "clicked",
			  (GtkSignalFunc)ge_clicked_cb, NULL);
    }

  gtk_widget_show_all (window);

  if (errmsg)
    {
      gtk_text_freeze (GTK_TEXT (ge_text));

      if (gtk_text_get_length (GTK_TEXT (ge_text)) > 0)
	gtk_text_insert (GTK_TEXT (ge_text), NULL, NULL, NULL,
			"\n-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n", -1);
      gtk_text_insert (GTK_TEXT (ge_text), NULL, NULL, NULL, errmsg, length);
      gtk_text_thaw (GTK_TEXT (ge_text));

      gtk_adjustment_set_value (ge_adjustment, ge_adjustment->upper - ge_adjustment->page_size);
    }
}

void gtulpas_glue_error_show_console (void)
{
  gtulpas_glue_error (NULL, -1);
}
