/* gtulpas.c -- app5 1999/11/27 16:10:11 0rfelyus4 1999/12/12 15:40:43
   $Id: gtulpas.c,v 1.33 2000/02/28 16:03:35 rodo Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <config.h>
#include <gnome.h>

#include "debug.h"
#include "dialog_players.h"

#include <guile/gh.h>

#include "gtulpas.h"
#include "game.h"
#include "variant.h"
#include "gnome-ball.h"
#include "properties.h"

#include "pool.h"
#include "poolkeys.h"
#include "guileglue.h"
#include "poolview.h"
#include "poolfeedgl.h"

GTulpas *gtulpas;

static void exit_cb     (GtkWidget* widget, gpointer data);
static void about_cb    (GtkWidget *widget, gpointer data);
static void new_game_cb (void);
static void answers_undo_cb (void);
static void shot_undo_cb (void);
static void help_game_cb (void);
static void game_options_cb (void);
static void players_cb (void);
static void new_view_cb (void);
static void miniviews_cb (GtkWidget *);

#define GNOMEUIINFO_ITEM_STOCK_ACCEL(label,tooltip,callback, stock_id,modifier,key) \
  { GNOME_APP_UI_ITEM, label, tooltip, (gpointer) callback, NULL, NULL, \
    GNOME_APP_PIXMAP_STOCK, stock_id, key, (GdkModifierType)modifier, NULL } 

#define GNOMEUIINFO_ITEM_ACCEL(label,tooltip,callback, modifier,key) \
  { GNOME_APP_UI_ITEM, label, tooltip, (gpointer) callback, NULL, NULL, \
    GNOME_APP_PIXMAP_NONE, NULL, key, (GdkModifierType)modifier, NULL } 

static GnomeUIInfo file_menu [] = {
	GNOMEUIINFO_MENU_NEW_ITEM(N_("_New game"),
				  N_("Start a new game"),
				  new_game_cb, NULL),
#define MENU_POS_SELECT_GAME 1
	GNOMEUIINFO_SUBTREE (N_("_Select game"), NULL),
#define MENU_POS_GAME_OPTIONS 2
	GNOMEUIINFO_ITEM_NONE (N_("Game _options..."), N_("Change game options"), game_options_cb),
	GNOMEUIINFO_ITEM_NONE (N_("Players..."), N_("Add/remove players"), players_cb),
	/* GNOMEUIINFO_MENU_PREFERENCES_ITEM (properties_show, NULL), */
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_EXIT_ITEM (exit_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo play_menu [] = {
  GNOMEUIINFO_ITEM_STOCK (N_("Undo _answers"), N_("Undo ansers"),
			  answers_undo_cb, GNOME_STOCK_PIXMAP_UNDO),
#define MENU_POS_UNDO	1
  GNOMEUIINFO_ITEM_STOCK_ACCEL (N_("_Undo last shot"), N_("Undo last shot"),
				shot_undo_cb, GNOME_STOCK_PIXMAP_UNDO, 0, 'U'),
  GNOMEUIINFO_SEPARATOR,
  GNOMEUIINFO_ITEM_ACCEL (N_("_Replay"), N_("Replay last animation"),
			  pool_replay_current_animation, GDK_CONTROL_MASK, 'R'),
  GNOMEUIINFO_ITEM_ACCEL (N_("Skip till _end"), N_("Skip current animation till end"),
			  animation_skip_till_end, 0, 'E'),
  GNOMEUIINFO_END
};

static GnomeUIInfo view_menu [] = {
  GNOMEUIINFO_TOGGLEITEM (N_("Show _mini views"), N_("Show/hide mini views"), miniviews_cb, NULL),
  GNOMEUIINFO_ITEM_NONE (N_("_New view"), N_("Create view in new window"), new_view_cb),
  GNOMEUIINFO_END
};


#ifdef DEBUG
static GnomeUIInfo debug_menu [] = {
	GNOMEUIINFO_ITEM_STOCK ("Reload init files", "Reload scheme init files",
			       glue_load_scheme_files, GNOME_STOCK_PIXMAP_REFRESH),
	GNOMEUIINFO_ITEM_STOCK ("Scheme error console",
				"Show scheme error console",
				gtulpas_glue_error_show_console, GNOME_STOCK_PIXMAP_TRASH_FULL),
	GNOMEUIINFO_ITEM_NONE ("Scheme debug console", N_("Show scheme debug console"),
			       glue_debug_console),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK ("Animation debug",
				"Show animation debug window",
				anim_panel_show, GNOME_STOCK_PIXMAP_EXEC),
	GNOMEUIINFO_ITEM_STOCK ("Screenshot Maker",
				"Grab animation to graphics files",
				anim_grab_panel_show, GNOME_STOCK_PIXMAP_SAVE),
	GNOMEUIINFO_END
};
#endif

static GnomeUIInfo help_menu [] = {
	GNOMEUIINFO_ITEM_NONE (N_("Game rules"), 
			       N_("Description of current game rules"),
			       help_game_cb),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_HELP ("gtulpas"),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_ABOUT_ITEM (about_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo menu [] = {
	GNOMEUIINFO_MENU_FILE_TREE(file_menu),
	GNOMEUIINFO_SUBTREE (N_("_Play"), play_menu),
	GNOMEUIINFO_SUBTREE (N_("_Views"), view_menu),
#ifdef DEBUG
	GNOMEUIINFO_SUBTREE (N_("_Debug"), debug_menu),
#endif
	GNOMEUIINFO_MENU_HELP_TREE(help_menu),
	GNOMEUIINFO_END
};

static GnomeUIInfo toolbar [] = {
	GNOMEUIINFO_ITEM_STOCK (N_("New game"), N_("Start a new game"), 
				new_game_cb, GNOME_STOCK_PIXMAP_NEW),
	{ GNOME_APP_UI_ITEM, N_("Properties"), N_("Set gtulpas properties"),
	  properties_show, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
	  GNOME_STOCK_PIXMAP_PROPERTIES, 0, 0, NULL },
	GNOMEUIINFO_ITEM_STOCK (N_("Replay"), N_("Replay last animation"),
				pool_replay_current_animation,
				GNOME_STOCK_PIXMAP_FORWARD),
#define TBAR_POS_UNDO 3
	/* code not yet matured */
	GNOMEUIINFO_ITEM_STOCK (N_("Undo"), N_("Undo last shot"),
				shot_undo_cb, GNOME_STOCK_PIXMAP_UNDO),
	GNOMEUIINFO_ITEM_STOCK (N_("Rules"), N_("Description of current game rules"),
				help_game_cb, GNOME_STOCK_PIXMAP_HELP),
	GNOMEUIINFO_END
};

static DialogPlayers *dialog_players = NULL;

static void
extra_view_destroy_cb (GtkWidget *w, GnomePoolView *pv)
{
  gtulpas->extra_views--;
  gtulpas->extra_view_list = g_list_remove (gtulpas->extra_view_list, pv);
}


static GnomePoolView*
gtulpas_new_extra_view ()
{
  GnomePoolView	*pv;
  GtkWidget	*win;

  win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (win), _("Gnome Tulpas - Extra View"));
  gtk_window_set_policy (GTK_WINDOW (win), TRUE, TRUE, FALSE);

  pv = gnome_pool_view_new (NULL);
  gtk_container_add (GTK_CONTAINER (win), pv -> widget);
  pv -> topwindow = win;

  gtulpas->extra_views++;
  gtulpas->extra_view_list = g_list_prepend (gtulpas->extra_view_list, pv);

  gtk_signal_connect (GTK_OBJECT (win), "destroy",
		      (GtkSignalFunc) extra_view_destroy_cb, pv);


  gtk_widget_set_usize (win, 200, 200);
  return pv;
}

static void
new_view_cb (void)
{
  GnomePoolView		*pv;

  pv = gtulpas_new_extra_view ();
  gtk_widget_show_all (pv -> topwindow);

  gtk_idle_add ((GtkFunction) views_force_size, NULL);
}

static int  miniview_cb_lock = 0;

static void
miniview_check_update (void)
{
  miniview_cb_lock = TRUE;
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view_menu[0].widget),
				  properties.show_mini_views);
  miniview_cb_lock = FALSE;
}

void gtulpas_update_ugly_box (GTulpas *);
void gtulpas_clear_ugly_box  (GTulpas *);

static void 
miniviews_cb (GtkWidget *toggle)
{
  int newval;

  if (miniview_cb_lock)
    return;

  newval = GTK_CHECK_MENU_ITEM (toggle) -> active;
  if (newval != properties.show_mini_views)
    {
      properties.show_mini_views = newval;

      gtulpas_clear_ugly_box (gtulpas);
      gtulpas_update_ugly_box (gtulpas);
    }

}

static void
game_options_cb (void)
{
  variant_dialog_new (gtulpas -> game, 0);
}

static void
players_cb ()
{
	if (!dialog_players) {
		dialog_players = dialog_players_new (gtulpas, &dialog_players);
	} else {
		gdk_window_raise (dialog_players->win->window);
	}
}

static void 
exit_cb (GtkWidget* widget, gpointer data)
{
	gtk_main_quit();
}

static gint 
delete_event_cb(GtkWidget* window, GdkEventAny* e, gpointer data)
{
	gtk_main_quit ();

	return TRUE;
}

static void
about_cb (GtkWidget *widget, gpointer data)
{
	GtkWidget *about;
	const gchar *authors[] = {
		"Tomas Bily   (tomby@suse.cz)",
		"Jan Bily   (bilyj@seznam.cz)",
		"Radek Doulik   (rodo@suse.cz)",
		"Daniel Skarda   (0rfelyus@suse.cz)",
		NULL
	};
	
	about = gnome_about_new ( _("Tulpas"), VERSION,
				  /* copyright notice */
				  _("(C) 1999 Tuhyk Labs"),
				  authors,
				  /* another comments */
				  _("Tulpas is a pool game"),
				  NULL);
	gtk_widget_show (about);
	
	return;
}

static void
shot_undo_cb  (void)
{
  game_add_event_type (gtulpas->game, UNDO);
}

static void
answers_undo_cb (void)
{
  game_add_event_type (gtulpas->game, UNDO_ANSWERS);
}

void
gtulpas_answers_undo_sensitivity (int sensitivity)
{
  gtk_widget_set_sensitive (play_menu[0].widget, sensitivity);
}


static void
new_game_cb ()
{
	game_start (gtulpas->game);
	game_process_events (gtulpas->game);
}

static void
help_game_cb ()
{
  gchar	*hpath, *ghelp;

  hpath = gnome_help_file_path ("gtulpas", 
     (gtulpas -> game -> help_file == NULL) 
   ? "default-page.html" : gtulpas -> game -> help_file);

  if (! hpath)
    hpath = gnome_help_file_path ("gtulpas", "default-page.html");

  ghelp = g_strconcat ("ghelp:", hpath, NULL);
  gnome_help_goto (NULL, ghelp);

  g_free (hpath); g_free (ghelp);
}

static void
game_menu_select_game_cb (GtkWidget *w, gint id)
{
	debug ("select game id: %d\n", id);

	gtulpas_select_game_by_id (gtulpas, id);
	game_add_event_type (gtulpas->game, NEW_GAME);
}

static void
add_game_menu ()
{
	GnomeUIInfo * uiinfo;	
	Game  *g;
	GList *cur;
	gint n;

	n = g_list_length (game_list);
	uiinfo = g_new (GnomeUIInfo, n+1);
	bzero (uiinfo, (n+1)*sizeof (GnomeUIInfo));
	uiinfo [n].type = GNOME_APP_UI_ENDOFINFO;

	cur = game_list;
	n   = 0;

	while (cur) {

		g = (Game *) cur->data;
		uiinfo [n].label       = g->name;
		uiinfo [n].type        = GNOME_APP_UI_ITEM;
		uiinfo [n].pixmap_type = GNOME_APP_PIXMAP_NONE;
		uiinfo [n].moreinfo    = game_menu_select_game_cb;
		uiinfo [n].user_data   = (gpointer) g->id;
		n++;
		cur = cur->next;
	}

	file_menu [MENU_POS_SELECT_GAME].moreinfo = uiinfo;
}

/*
   GTulpas implementation
*/

GTulpas *
gtulpas_new ()
{
	GTulpas * gt;

	gt = g_new (GTulpas, 1);

	gt->views = NULL;
	gt->game  = NULL;
	gt->score = NULL;
	gt->app   = gnome_app_new (PACKAGE, _("Gnome Tulpas"));

	gnome_app_create_menus   (GNOME_APP (gt->app), menu);
	gnome_app_create_toolbar (GNOME_APP (gt->app), toolbar);

	gtk_signal_connect (GTK_OBJECT(gt->app),
			    "delete_event",
			    GTK_SIGNAL_FUNC(delete_event_cb),
			    NULL);

	return gt;
}

void
gtulpas_select_game_by_id (GTulpas *gt, gint id)
{
	GList *c = game_list;
	Game  *g;

	while (c) {

		g = (Game *) c->data;
		if (g->id == id) {
			gtulpas_select_game (gt, g);
			break;
		}
		c = c->next;
	}
	if (!c)
	  gtulpas_select_game (gt, (Game *)game_list->data);
}

void
gtulpas_select_game (GTulpas *gt, Game *g)
{
	gchar	*title;

	gint mo = ((properties.show_mini_views) ? 1 : 0);

	debug ("gtulpas_select_game\n");

	gt->game_id = g->id;
	if (gt->score) {
		/* FIX ME!!!: destroy old score */
		gtk_container_remove (GTK_CONTAINER (gt->sc_vbox), gt->score->frame);
	}

	game_unselect (gtulpas->game);

	gtulpas->game = g;
	gt->score = gnome_tulpas_score_new (g);
	gnome_tulpas_score_refresh (gt->score);

	gtk_box_pack_start_defaults (GTK_BOX (gt->sc_vbox), gt->score->frame);

	gtk_widget_show_all (gt->app);

	title = g_strdup_printf (_("Gnome Tulpas - %s"), g-> name);
	gtk_window_set_title (GTK_WINDOW (gtulpas->app), title);
	g_free (title);

	/* game variants */

	variant_dialog_new (g, TRUE);
	gtk_widget_set_sensitive (file_menu [MENU_POS_GAME_OPTIONS] . widget,
				  g -> variant_list == NULL ? FALSE : TRUE);
	gtulpas_undo_sensitivity (0);
	game_select (g);
}

void
gtulpas_undo_sensitivity (int sensitivity)
{
  gtk_widget_set_sensitive (play_menu [MENU_POS_UNDO] . widget,
			    sensitivity & (gtulpas->game->safe_to_undo));
  gtk_widget_set_sensitive (toolbar [TBAR_POS_UNDO] . widget,
			    sensitivity & (gtulpas->game->safe_to_undo));
}

void
gtulpas_score_refresh (GTulpas *gt)
{
	gint i;

	for (i=0; i<gt->game->nof_players; i++)
		gnome_score_balls_refresh (gt->score->score [i]);
}

void
gtulpas_create_and_load (GTulpas *gt, gchar *prefix)
{
	GnomePoolView  *pv;
	
	GList *cur;
	gint   i, nof_ev;
	gchar *new_prefix;
	gchar  num [2];

	debug ("gtulpas_create_and_load\n");

	new_prefix = g_strconcat (prefix, "properties/", NULL);
	properties_load (&properties, new_prefix);
	g_free (new_prefix);

	new_prefix = g_strconcat (prefix, "gtulpas/", NULL);
	gnome_config_push_prefix (new_prefix);
	g_free (new_prefix);

	/* create main view */
	gt->main_view  = gnome_pool_view_new (NULL);
	new_prefix = g_strconcat (prefix, "main_view/", NULL);
	gnome_pool_view_load (gt->main_view, NULL, new_prefix);
	g_free (new_prefix);

	gtk_widget_set_usize (GTK_WIDGET (gt->main_view->gl_view->area), 400, 300);
	gt->mini_views = gnome_config_get_int ("mini_views=3");
	nof_ev = gnome_config_get_int ("extra_views=0");

	num [1]=0;

	/* create and load miniviews */
	for (i=0; i<gt->mini_views; i++) {
		gt->mini_view [i] = gnome_pool_view_new (gt->main_view);
		num [0] = '0' + i;
		new_prefix = g_strconcat (prefix, "mini_view_", num, "/", NULL);
		gnome_pool_view_load (gt->mini_view [i], NULL, new_prefix);
		g_free (new_prefix);
	}

	/* create and load extra views */

	gtulpas->extra_view_list = NULL;
	gtulpas->extra_views = 0;

	for (i = 0; i < nof_ev; i++)
	  {
	    pv = gtulpas_new_extra_view ();
	    num [0] = '0' + i;
	    new_prefix= g_strconcat (prefix, "extra_view_", num, "/", NULL);
	    gnome_pool_view_load (pv, pv->topwindow, new_prefix);
	    g_free (new_prefix);

	    gtk_widget_show_all (pv -> topwindow);
	  }

	gt->ball  = gnome_ball_new (gt->app);

	gtulpas_create_contents (gt);
	gtulpas_fill_table      (gt);
	gtulpas_update_ugly_box (gt);

	gtk_paned_set_position (GTK_PANED (gt->paned), gnome_config_get_int ("paned_pos=100"));

	/* load games state from config */
	cur = game_list;
	while (cur) {
		game_load ((Game *) cur->data, prefix);
		cur = cur->next;
	}

	gtk_window_set_default_size (GTK_WINDOW (gt->app),
				     gnome_config_get_int ("win_w=600"),
				     gnome_config_get_int ("win_h=440"));
	gtk_window_set_policy (GTK_WINDOW (gt->app), TRUE, TRUE, FALSE);

	gt->game_id = gnome_config_get_int ("game_id=0");
	gtulpas_select_game_by_id (gt, gt->game_id);

	gnome_config_pop_prefix ();
}

void
gtulpas_save (GTulpas *gt, gchar *prefix)
{
	GnomePoolView *pv;
	GList *cur;

	gint   i;
	gchar *new_prefix;
	gchar  num [2];

	debug ("gtulpas_save\n");

	new_prefix = g_strconcat (prefix, "properties/", NULL);
	properties_save (&properties, new_prefix);
	g_free (new_prefix);

	new_prefix = g_strconcat (prefix, "gtulpas/", NULL);
	gnome_config_push_prefix (new_prefix);
	g_free (new_prefix);

	gnome_config_set_int ("mini_views", gt->mini_views);
	gnome_config_set_int ("extra_views", gt->extra_views);
	gnome_config_set_int ("game_id", gt->game->id);
	gnome_config_set_int ("paned_pos", GTK_PANED (gt->paned)->child1_size);
	gnome_config_set_int ("win_w", GTK_WIDGET (gt->app)->allocation.width);
	gnome_config_set_int ("win_h", GTK_WIDGET (gt->app)->allocation.height);

	/* save main view */
	new_prefix = g_strconcat (prefix, "main_view/", NULL);
	gnome_pool_view_save (gt->main_view, NULL, new_prefix);
	g_free (new_prefix);

	num [1] = 0;
	/* save miniviews */
	for (i=0; i<gt->mini_views; i++) {
		num [0] = '0' + i;
		new_prefix = g_strconcat (prefix, "mini_view_", num, "/", NULL);
		gnome_pool_view_save (gt->mini_view [i],NULL, new_prefix);
		g_free (new_prefix);
	}

	/* save extra views */
	for (i = 0, cur = gt->extra_view_list; 
	     i < gt->extra_views;
	     i++, cur = cur -> next)
	  {
	    pv = (GnomePoolView *) (cur->data);
	    num [0] = '0' + i;
	    new_prefix= g_strconcat (prefix, "extra_view_", num, "/", NULL);
	    gnome_pool_view_save (pv, pv->topwindow, new_prefix);
	    g_free (new_prefix);
	  }

	/* save games state to config */
	cur = game_list;
	while (cur) {
		game_save ((Game *) cur->data, prefix);
		cur = cur->next;
	}

	gnome_config_pop_prefix ();
}

void
gtulpas_clear_table (GTulpas *gt)
{
	GList *ch = gtk_container_children (GTK_CONTAINER (gt->mv_vbox));

	while (ch) {
		debug ("container_remove\n");
		gtk_widget_ref  (GTK_WIDGET (ch->data));
		gtk_widget_hide (GTK_WIDGET (ch->data));
		gtk_container_remove (GTK_CONTAINER (gt->mv_vbox), GTK_WIDGET (ch->data));
		ch = ch->next;
	}
}

void
gtulpas_clear_ugly_box (GTulpas *gt)
{
  if (gt->paned->parent == gt->ugly_box)
    {
      gtk_widget_ref (gt->paned);
      gtk_widget_hide (gt->paned);
      gtk_container_remove (GTK_CONTAINER (gt->ugly_box), gt->paned);
    }

  if (gt->main_view->widget->parent == gt->paned)
    {
      gtk_widget_ref (gt->main_view->widget);
      gtk_widget_hide (gt->main_view->widget);
      gtk_container_remove (GTK_CONTAINER (gt->paned), gt->main_view->widget);
    }

  if (gt->main_view->widget->parent == gt->ugly_box)
    {
      gtk_widget_ref (gt->main_view->widget);
      gtk_widget_hide (gt->main_view->widget);
      gtk_container_remove (GTK_CONTAINER (gt->ugly_box), gt->main_view->widget);
    }
}

void
gtulpas_update_ugly_box (GTulpas *gt)
{
  if (properties.show_mini_views)
    {
      gtk_box_pack_start (GTK_BOX (gt->ugly_box), gt->paned, TRUE, TRUE,0);
      gtk_widget_show (gt->paned);
      gtk_paned_add2 (GTK_PANED (gt->paned), GTK_WIDGET (gt->main_view->widget));
      gtk_widget_show (gt->main_view->widget);
    }
  else
    {
      gtk_box_pack_start (GTK_BOX (gt->ugly_box), GTK_WIDGET (gt->main_view->widget),
			  TRUE, TRUE,0);
      gtk_widget_show (gt->main_view->widget);
    }

  gtk_widget_show_all (gt->ugly_box);
}

void
gtulpas_fill_table (GTulpas *gt)
{
	gint i;

	gint tw = 4 
#if 0
		+ ((properties.show_mini_views) ? 1 : 0)
#else
		+ ((TRUE) ? 1 : 0)
#endif
		+ ((properties.show_game_status) ? 1 : 0);

	gint mo = ((properties.show_mini_views) ? 1 : 0);

	g_list_free (gt->views);

	/* gtk_table_resize (GTK_TABLE (gt->table), 3, tw);

	gtk_table_attach_defaults (GTK_TABLE (gt->table), GTK_WIDGET (gt->main_view->widget),
	mo, mo+4, 0, 3); */


	/* gtk_paned_add2 (GTK_PANED (gt->paned), GTK_WIDGET (gt->main_view->widget)); */
	gt->views = g_list_prepend (gt->views, gt->main_view);

#if 0
	if (properties.show_mini_views)
#endif
		for (i=0; i<gt->mini_views; i++) {
			/* gtk_table_attach_defaults (GTK_TABLE (gt->table), GTK_WIDGET (gt->mini_view [i]->widget),
			   0, 1, i, i+1); */
			gtk_box_pack_start_defaults (GTK_BOX (gt->mv_vbox), GTK_WIDGET (gt->mini_view [i]->widget));
			gt->views = g_list_prepend (gt->views, gt->mini_view [i]);
		}

	if (properties.show_game_status) {
		/* gtk_table_attach (GTK_TABLE (gt->table), gt->ball->frame,
				  mo+4, mo+5, 0, 1,
				  FALSE, FALSE, 0, 0); */
		gtk_box_pack_start (GTK_BOX (gt->ball_hbox), gt->ball->frame, FALSE, FALSE, 0);
	}

}

void
gtulpas_create_contents (GTulpas *gt)
{
	GtkWidget *hbox;

	hbox           = gtk_hbox_new (FALSE, 0);
	gt->sc_vbox    = gtk_vbox_new (FALSE, 0);
	gt->ball_hbox  = gtk_hbox_new (FALSE, 0);
	gt->mv_vbox    = gtk_vbox_new (FALSE, 0);

	gt->ugly_box = gtk_vbox_new (FALSE,0);
	gt->paned = gtk_hpaned_new ();

	gtk_box_pack_start (GTK_BOX (gt->sc_vbox), gt->ball_hbox, FALSE, FALSE, 0);
	gtk_paned_add1 (GTK_PANED (gt->paned), gt->mv_vbox);

	/* gtk_box_pack_start (GTK_BOX (gt->ugly_box), gt->paned, TRUE,TRUE,0); */
	gtk_box_pack_start_defaults (GTK_BOX (hbox), gt->ugly_box);
	gtk_box_pack_start (GTK_BOX (hbox), gt->sc_vbox, FALSE, FALSE, 0);

	gt->bar = GNOME_APPBAR (gnome_appbar_new (FALSE, TRUE, FALSE));

	gnome_app_set_contents (GNOME_APP (gt->app), hbox);
	gnome_app_set_statusbar (GNOME_APP (gt->app), GTK_WIDGET (gt->bar));
}

void
gtulpas_layout_change (GTulpas *gt)
{
	gtulpas_clear_table (gt);
	gtulpas_fill_table (gt);
}

/* end of GTulpas implementations */

static gint
session_save (GnomeClient *client, gint phase, GnomeSaveStyle save_style,
              gint is_shutdown, GnomeInteractStyle interact_style,
              gint is_fast, gpointer client_data)
{
	debug ("session_save\n");
	return TRUE;
}

static void
session_die (GnomeClient* client, gpointer client_data)
{
	debug ("session_die\n");
	gtk_main_quit ();
}

static GtkWidget *splash_win;
static GtkWidget *splash_pixmap;

static gint
splash_end ()
{
	gtk_widget_hide (splash_win);
	gtk_widget_show_all (gtulpas->app);
	poolanim_setup ();
	views_force_size ();

	return FALSE;
}

static void
splash_begin ()
{
	gchar * fname;

	fname = gnome_unconditional_datadir_file ("gtulpas/pics/splash.png");
	splash_pixmap = gnome_pixmap_new_from_file (fname);
	g_free (fname);

	splash_win = gtk_window_new (GTK_WINDOW_POPUP);
	gtk_window_set_position (GTK_WINDOW (splash_win), GTK_WIN_POS_CENTER);

	gtk_container_add (GTK_CONTAINER (splash_win), splash_pixmap);

	gtk_widget_show_all (splash_win);

	/* refresh gtk to show our splash screen */
	while (gtk_events_pending ()) { gtk_main_iteration_do (FALSE); gdk_flush (); }

	gtk_timeout_add (2000, splash_end, NULL);
}

void
inner_main (void *closure, int argc, char* argv[])
{
	GnomeClient * client;

	bindtextdomain (PACKAGE, GNOMELOCALEDIR);  
	textdomain (PACKAGE);

	gnome_init (PACKAGE, VERSION, argc, argv);

	client = gnome_master_client ();
	gtk_signal_connect (GTK_OBJECT (client), "save_yourself",
			    GTK_SIGNAL_FUNC (session_save), argv[0]);
	gtk_signal_connect (GTK_OBJECT (client), "die",
			    GTK_SIGNAL_FUNC (session_die), NULL);

	#ifndef DEBUG
	/* splash screen */
	splash_begin ();
	#endif

	/* our guile bootstrap */
	gtulpas_guile_init ();
	add_game_menu ();

	gtulpas = gtulpas_new ();
	gtulpas_create_and_load (gtulpas, "gtulpas/");

	if (splash_win) {
		gdk_window_raise (splash_win->window);
	} 

	miniview_check_update ();

	game_add_event_type (gtulpas->game, NEW_GAME);

	pool_key_init_default ();

#ifdef DEBUG
	gtk_widget_show_all (gtulpas->app);
	poolanim_setup ();
	views_force_size ();
#endif

	gtk_main();

	gtulpas_save (gtulpas, "gtulpas/");
	gnome_config_sync ();
}

int
main (int argc, char *argv[])
{
  scm_boot_guile (argc, argv, (void (*) ()) inner_main, NULL);
  return 0;
}
