#ifndef __POOL_POOL_H__
#define __POOL_POOL_H__

#include "mathutils.h"
#include <glib.h>
#include "config.h"

#define POOL_SPHERE_RADIUS	4.0

typedef struct
{
  char	       * name;
  dbl		radius;
  dbl		weight;

  /* Basic Tomby's physics world constant :) */
  /* Alias TOPHYC - TeX translation : topi\v c - ISO8852: topi */

  dbl		f1; 	/* coef nepruzneho narazu kouli */
  dbl		f2; 	/* treni o podlozku */
  dbl		f3; 	/* coef treni dvou kouli pro predani rychlost -> uhlova rychlost */
  dbl		f4; 	/* coef treni kouli pro predani uhlove rychlosti uhl r -> uhl r. :) */
  dbl		f5; 	/* koef nepruzneho narazu o hranu*/
  dbl		f6; 	/* ubytek tychlosti kolem osy z */
  dbl		f7; 	/* valivy odpor */
  dbl		f8; 	/* ubytek uhlove rychlosti pri odrazu o hranu */
  dbl		f9; 	/* nepruzny odraz koule od podlozky stolu */
  dbl		g;	/* gravity acceleration */

} FyjikSkin;

typedef struct 
{
  char	        * name;

  dbl		color1 [3];
  dbl		color2 [3];

  char		* texture_file;
  dbl		dvalue;
  /* FIXME: data for loaded texture */

     /* flags */
  guint		has_strip : 1;
  guint		has_mark : 1;
} Skin;

typedef struct 
{
  char		*name;
  FyjikSkin	*ball_type;
  Skin		*ball_skin;

  void		*GL_ball_data;
  void		*GL_mark_data;
  void		*GL_strip_data;
  void		*GL_cap_data;

  /* various flags */

  guint		 is_invalid : 1;  /* ball in hand but at bad position */
} PoolBall;

typedef struct
{
  PoolBall     *ball;

  dbl		p[3];		/* position */
  dbl		q[4]; 		/* quaternion of actual transformation */

  dbl		v[3];		/* vector of speed */
  dbl		omega[3];	/* vector of angular speed */

} PoolBallSpaceInfo;

void pool_prepare_for_next_stroke (void);

extern Skin		* default_ball_skin;
extern Skin		* default_cue_skin;
extern Skin		* default_table_skin;
extern FyjikSkin	* default_fyjik_skin;

#endif
