#include <gtk/gtk.h>
#include <gdk/gdkprivate.h>
#include <gtkgl/gtkglarea.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include <math.h>
#include <stdlib.h>

#include "mathutils.h"
#include "pool.h"

#include "gtulpas.h"


#include "poolview.h"
#include "poolfeedgl.h"
#include "render.h"

#include "cue.h"
#include "table.h"

#include "animation.h"
#include "fyjik.h"

#include "properties.h"

static LOD_array pool_GL_cues = {
  -1, -1, -1, -1, -1, -1
};


/*
 *
 *		SPHERES & LOD_info
 */

LOD_info*
LOD_ref (LOD_info *ptr)
{
  if (! ptr)
    {
      int	i;

      ptr = g_new (LOD_info, 1);
      ptr -> refcount = 1;

      LOD_array_init (ptr -> lists);
    }
  else
    (ptr -> refcount)++;

  return ptr;
}

void
LOD_unref (LOD_info *ptr)
{
  if (! ptr)
    return;

  if ( --(ptr->refcount) == 0)
    {
      LOD_array_GL_free (ptr->lists);
      g_free (ptr);
    }
}

void
LOD_finish_ball (PoolBall *ball)
{
#define CLEAR(X)	if (ball -> X) LOD_unref (ball -> X); ball -> X = NULL
  CLEAR (GL_ball_data);
  CLEAR (GL_mark_data);
  CLEAR (GL_strip_data);
  CLEAR (GL_cap_data);
#undef CLEAR
}

static LOD_info*
sphere_find_sphere_LOD (PoolBall *ball)
{
  Game		*g;
  PoolBall	*b;
  int		 i;

  if (ball -> GL_ball_data)
    return ball -> GL_ball_data;

  g = gtulpas -> game;
  b = g -> balls;
  for (i = 0, b = g -> balls;
       i < g -> balls_nof; 
       i++, b++)
    if (ball != b && b->GL_ball_data != NULL &&
	ball->ball_type->radius == b->ball_type->radius)
      return LOD_ref (ball->GL_ball_data = b -> GL_ball_data);
      
  return (ball -> GL_ball_data = LOD_ref (NULL));
}

static LOD_info*
sphere_find_mark_LOD (PoolBall *ball)
{
  Game		*g;
  PoolBall	*b;
  int		 i;

  if (ball -> GL_mark_data)
    return ball -> GL_mark_data;

  g = gtulpas -> game;
  b = g -> balls;
  for (i = 0, b = g -> balls;
       i < g -> balls_nof; 
       i++, b++)
    if (ball != b && b->GL_mark_data != NULL &&
	ball->ball_type->radius == b->ball_type->radius &&
	b->ball_skin->has_mark &&
	ball->ball_skin->dvalue == b->ball_skin->dvalue)
      return LOD_ref (ball->GL_ball_data = b -> GL_mark_data);
      
  return (ball -> GL_mark_data = LOD_ref (NULL));
}

static void
sphere_find_strip_LOD (PoolBall *ball, LOD_info **si, LOD_info **ci)
{
  Game		*g;
  PoolBall	*b;
  int		 i;

  if (ball -> GL_strip_data)
    {
      *si = ball -> GL_strip_data;
      *ci = ball -> GL_cap_data;
    }

  g = gtulpas -> game;
  b = g -> balls;
  for (i = 0, b = g -> balls;
       i < g -> balls_nof; 
       i++, b++)
    if (ball != b && b->GL_strip_data != NULL &&
	ball->ball_type->radius == b->ball_type->radius &&
	b->ball_skin->has_strip &&
	ball->ball_skin->dvalue == b->ball_skin->dvalue)
      {
	*si = b -> GL_strip_data;
	*ci = b -> GL_cap_data;
	return;
      }

  *si = ball -> GL_strip_data = LOD_ref (NULL);
  *ci = ball -> GL_cap_data = LOD_ref (NULL);
}

static int
sphere_arrays_to_list (void)
{
  int	listid;
  GLint *ptr;

  listid = glGenLists (1);

  glEnableClientState (GL_NORMAL_ARRAY);
  glVertexPointer (3, GL_DOUBLE, 0, ts_vertices);
  glNormalPointer (GL_DOUBLE, 0, ts_normals);

  glNewList (listid, GL_COMPILE);
  for (ptr = ts_indices;
       *ptr != 0 && *(ptr + 1) != 0;
       ptr += 2 + *(ptr + 1))
    glDrawElements (*ptr, *(ptr + 1), GL_UNSIGNED_INT, ptr + 2);
  glEndList ();

  glDisableClientState (GL_NORMAL_ARRAY);
  ts_free ();

  return listid;
}

static int
sphere_arrays_to_mark_list (void)
{
  int	listid;
  GLint *ptr;

  listid = glGenLists (1);

  glEnableClientState (GL_NORMAL_ARRAY);
  glVertexPointer (3, GL_DOUBLE, 0, ts_vertices);
  glNormalPointer (GL_DOUBLE, 0, ts_normals);

  glNewList (listid, GL_COMPILE);

  glPolygonOffset (-2,-2);
  glEnable (GL_POLYGON_OFFSET_FILL);

  for (ptr = ts_indices;
       *ptr != 0 && *(ptr + 1) != 0;
       ptr += 2 + *(ptr + 1))
    glDrawElements (*ptr, *(ptr + 1), GL_UNSIGNED_INT, ptr + 2);

  glDisable (GL_POLYGON_OFFSET_FILL);
  glEndList ();

  glDisableClientState (GL_NORMAL_ARRAY);
  ts_free ();

  return listid;
}

static void
feed_GL_with_sphere (PoolViewGL *view, PoolBallSpaceInfo *sinfo, int lod)
{
  PoolBall	*ball;
  Skin		*skin;
  LOD_info	*l1, *l2, *l3;
  GLdouble	m[4][4];
  int		was_invalid = 0;

  ball = sinfo -> ball;
  skin = ball -> ball_skin;

	/* prelude */

  if (skin -> has_strip)
    {
      l1 = ball -> GL_strip_data;
      l2 = ball -> GL_cap_data;

      if (l1 == NULL)
	sphere_find_strip_LOD (ball, &l1, &l2);

      if (l1 -> lists [lod] == -1)
	{
	  int		xy, z;
	  dbl		radius, s_width;
	  

	  z = LOD_TO_SPHERE_PREC (lod);
	  xy = 2 * z;

	  radius = ball->ball_type->radius;
	  s_width = ball->ball_skin->dvalue;

	  if (s_width < 0.0 + EPSILON || s_width >= 1.0 - EPSILON)
	    s_width = 0.5;

	  /* strip first */
	  ts_sphere_strip (radius, s_width, xy,z);
	  l1 -> lists[lod] = sphere_arrays_to_list ();

	  /* then cap */
	  ts_sphere_cap (radius, s_width, xy,z);
	  l2 -> lists[lod] = sphere_arrays_to_list ();
	}
    }
  else /* ball without strip */
    {
      l1 = ball -> GL_ball_data;
      if (l1 == NULL)
	l1 = sphere_find_sphere_LOD (ball);
	
      if (l1 -> lists [lod] == -1)
	{
	  int		xy, z;	  

	  z = LOD_TO_SPHERE_PREC (lod);
	  xy = 2 * z;

	  ts_sphere_create (ball->ball_type->radius, xy,z);
	  l1 -> lists [lod] = sphere_arrays_to_list ();
	}
    }


  if (skin -> has_mark)
    {
      l3 = ball -> GL_mark_data;
      if (l3 == NULL)
	l3 = sphere_find_mark_LOD (ball);

      if (l3 -> lists [lod] == -1)
	{
	  int		xy, z;	  

	  z = LOD_TO_SPHERE_PREC (lod);
	  xy = 2 * z;

	  ts_sphere_mark (ball->ball_type->radius, skin->dvalue,xy,z);
	  l3 -> lists [lod] = sphere_arrays_to_mark_list ();
	}
    }

	/* invalid ball in hand ... */
  if (ball -> is_invalid && view -> shading)
    glPolygonMode (GL_FRONT, GL_LINE);

	/* Transform */

  glPushMatrix ();  
  glTranslated (sinfo->p[X], sinfo->p[Y], sinfo->p[Z]);

  quat_to_rmat (m, sinfo -> q);
  glMultMatrixd ((dbl *) m);

        /* here we go! */
 
  if (skin -> has_strip)
    {
      glColor3dv (skin -> color2);
      glCallList (l1 -> lists [lod]);

      glColor3dv (skin -> color1);
      glCallList (l2 -> lists [lod]);

      glPushMatrix ();
      glRotated (180, 1,0,0);

      glCallList (l2 -> lists [lod]);

      glPopMatrix ();
    }
  else
    {
      glColor3dv (skin -> color1);
      glCallList (l1 -> lists [lod]);
    }

  if (skin -> has_mark)
    {
      glColor3dv (skin -> color2);
      glCallList (l3 -> lists [lod]);
    }

  glPopMatrix ();

	/* restore invalid ball in hand ... */
  if (ball -> is_invalid && view -> shading)
    glPolygonMode (GL_FRONT, GL_FILL);
}


static void
sphere_array_draw (void)
{
  GLint	*ptr;

  glVertexPointer (3, GL_DOUBLE, 0, ts_vertices);
  glNormalPointer (GL_DOUBLE, 0, ts_normals);

  for (ptr = ts_indices;
       *ptr != 0 && *(ptr + 1) != 0;
       ptr += 2 + *(ptr +1))
    glDrawElements (*ptr, *(ptr+1), GL_UNSIGNED_INT, ptr+2);
}

static void
feed_GL_with_raw_sphere (PoolBall *ball,  int lod, int colors)
     /* this is simpler version of feed_GL_with_sphere. it does not use
      *	lists (so NVidia X server will not crash when rendering to Pixmaps...)
      * sphere is rendered at <0,0,0>
      */ 
{
  FyjikSkin *fs;
  Skin	    *skin;
  int	     z,xy;

#define COLOR(X)	if (colors) glColor3dv (X)

  z = LOD_TO_SPHERE_PREC (lod);
  xy = 2 * z;

  fs = ball -> ball_type;
  skin = ball -> ball_skin;

  glEnableClientState (GL_NORMAL_ARRAY);
  glEnableClientState (GL_VERTEX_ARRAY);

  if (skin -> has_strip)
    {
      COLOR (skin -> color2);
      ts_sphere_strip (fs->radius, skin->dvalue, xy, z);
      sphere_array_draw ();

      ts_sphere_cap (fs->radius, skin->dvalue, xy, z);
      COLOR (skin->color1);
      sphere_array_draw ();
      glPushMatrix ();
      glRotated (180,1,0,0);
      sphere_array_draw ();
      glPopMatrix ();
    }
  else
    {
      ts_sphere_create (fs->radius, xy,z);
      COLOR (skin -> color1);
      sphere_array_draw ();
    }

  if (skin -> has_mark)
    {
      ts_sphere_mark (fs->radius, skin -> dvalue, xy, z);
      COLOR (skin -> color1);
      sphere_array_draw ();
    }

  glDisableClientState (GL_NORMAL_ARRAY);
}

/*
 *
 *		TABLE
 */


static void
pool_feed_GL_with_sphere_shadow (PoolViewGL *data, PoolBallSpaceInfo *ball)
{
  dbl		radius;

  if (ball->p [Z] < 0.0)
    return;

  radius = POOL_SHADOW_SIZE_COEFF * ball -> ball -> ball_type -> radius;

  glPushMatrix ();
  glTranslated (ball->p[X], ball->p[Y], SLIGHTLY_ABOVE_TABLE);
  glScaled (radius, radius, radius);

  render_disk (data);

  glPopMatrix ();
}

static void
pool_GL_cue_generate (int lod)
{
  GLUquadricObj	   *qptr;
  Skin		   *skin;
  PoolCueTypeInfo  *ti;
  dbl		    rm, s1;
  int		    a;

  ti = pool_cue.cue_type;
  skin = pool_cue.cue_skin; /* FIXME */

  g_return_if_fail (ti != NULL);

  qptr = render_quadric_init (RENDER_QUADRIC_WITH_NORMALS);

  glColor3dv (skin -> color2);

  glFrontFace (GL_CW);
  gluDisk (qptr,
	   0, pool_cue.cue_type -> top_width / 2.0,
	   (lod + 1) * 2 + 1, 1);
  glFrontFace (GL_CCW);

  if (lod != 0 && skin -> dvalue > EPSILON)
    {
      s1 = skin -> dvalue;
      rm = ti->top_width * s1 +  ti->bottom_width * (1. - s1);
      a = 0;

      gluCylinder (qptr,
		   ti -> top_width / 2.0, rm / 2.0,
		   s1,
		   (lod + 1) * 2 + 1, 1);
    }
  else
    {
      rm = ti -> top_width;
      s1 = 0.0;
      a = 1;
    }


  glColor3dv (skin -> color1);

  if (s1 != 0.0)
    glTranslated (0,0,s1);

  gluCylinder (qptr, 
	       rm / 2.0,
	       ti -> bottom_width / 2.0,
	       ti -> length - s1,
	       (lod + 1) * 2 + 1,
	       lod + a);

  glTranslated (0, 0, pool_cue.cue_type -> length - s1);
  gluDisk (qptr,
	   0, ti -> bottom_width / 2.0,
	   (lod + 1) * 2 + 1,1);

  render_quadric_finish (qptr);
}

static void 
pool_feed_GL_with_cue (PoolViewGL *data)
{
  static dbl cue_default_orientation [3] = {0, 0, 1};

  int		i,j;
  GLint	       *iptr;
  dbl		rvector [3], 
		delta;

  if (pool_cue.rebuild_gl_list)
    {
      cue_update_orientation (&pool_cue);
      cue_update_position (&pool_cue);
      pool_cue.rebuild_gl_list = 0;

      LOD_array_GL_free (pool_GL_cues);
    }

  glPushMatrix ();
  glTranslated (pool_cue.adjusted_top [X], 
		pool_cue.adjusted_top [Y],
		pool_cue.adjusted_top [Z]);

  glRotated (RAD2DEG (pool_cue.alfa) + 180., 0, 0, 1);
  glRotated (RAD2DEG (pool_cue.beta) - 90., 0,1,0);

  iptr = &(pool_GL_cues [data->lod]);
  if (*iptr == -1)
    {
      *iptr = glGenLists (1);

      glNewList (*iptr, GL_COMPILE_AND_EXECUTE);
        pool_GL_cue_generate (data->lod);
      glEndList ();
    }
  else
    glCallList (*iptr);

  glPopMatrix ();
}

static dbl pool_cue_shadow_vertices [4][3];
static dbl pool_cue_shadow_planes [5][4];


static void
pool_plane_equation (dbl *equation, dbl *v1, dbl *v2)
{
  equation [X] = v2[Y] - v1[Y];
  equation [Y] = v1[X] - v2[X];
  equation [Z] = 0;

  vec_norm (equation, equation);
  equation [W] = - equation [X] * v1[X] - equation [Y] * v1[Y];
}

static void
pool_fill_cue_shadow_vertices (void)
{
  dbl		begin [3];
  dbl		end [3];
  dbl	       *normal;
  
  dbl		r1, r2, tmp;

  r1 = pool_cue.cue_type -> top_width / 2.0;
  r2 = pool_cue.cue_type -> bottom_width / 2.0;

  vec_cpy (begin, pool_cue.adjusted_top);
  vec_cpy (end, pool_cue.adjusted_bottom);
  normal = pool_cue.right;

  begin [Z] = end [Z] = SLIGHTLY_ABOVE_TABLE ;

  normal [Z] = 0;			/* should be - but we should be sure */
  vec_norm (normal, normal);

  pool_cue_shadow_vertices [0][X] = begin [X] + r1 * normal [X];
  pool_cue_shadow_vertices [0][Y] = begin [Y] + r1 * normal [Y];
  pool_cue_shadow_vertices [0][Z] = begin [Z];

  pool_cue_shadow_vertices [1][X] = begin [X] - r1 * normal [X];
  pool_cue_shadow_vertices [1][Y] = begin [Y] - r1 * normal [Y];
  pool_cue_shadow_vertices [1][Z] = begin [Z];

  pool_cue_shadow_vertices [2][X] = end [X] - r2 * normal [X];
  pool_cue_shadow_vertices [2][Y] = end [Y] - r2 * normal [Y];
  pool_cue_shadow_vertices [2][Z] = end [Z];

  pool_cue_shadow_vertices [3][X] = end [X] + r2 * normal [X];
  pool_cue_shadow_vertices [3][Y] = end [Y] + r2 * normal [Y];
  pool_cue_shadow_vertices [3][Z] = end [Z];


  pool_plane_equation (pool_cue_shadow_planes [0], 
		       pool_cue_shadow_vertices [0],
		       pool_cue_shadow_vertices [3]);
  pool_plane_equation (pool_cue_shadow_planes [1], 
		       pool_cue_shadow_vertices [3],
		       pool_cue_shadow_vertices [2]);  
  pool_plane_equation (pool_cue_shadow_planes [2], 
		       pool_cue_shadow_vertices [2],
		       pool_cue_shadow_vertices [1]);
  pool_plane_equation (pool_cue_shadow_planes [3], 
		       pool_cue_shadow_vertices [1],
		       pool_cue_shadow_vertices [0]);
  
  vec_mul_c (pool_cue_shadow_planes [4], pool_cue.up, -1.);
  pool_cue_shadow_planes [4][3] = pool_cue.up [X] * pool_cue.adjusted_top [X] + 
				  pool_cue.up [Y] * pool_cue.adjusted_top [Y] + 
				  pool_cue.up [Z] * pool_cue.adjusted_top [Z];
}

static int
pool_plane_to_sphere (dbl *equation, dbl *point, dbl radius)
{
  if (equation[X] * point [X] + equation[Y] * point [Y] + 
      equation[Z] * point [Z] + equation [3] - radius < 0)
    return FALSE;
  else
    return TRUE;
}


static int
pool_cue_shadows_sphere (PoolBallSpaceInfo *ball)
{
  int i;
  
  for (i = 0; i < 3; i++)
    if (pool_plane_to_sphere (pool_cue_shadow_planes [i],
			      ball -> p,
			      ball -> ball -> ball_type -> radius) == FALSE)
      return FALSE;
  
  return TRUE;
}

static void
pool_feed_GL_with_cue_clips (int on)
{
  GLdouble	 equation [4];

  if (! on)
    {
      glDisable (GL_CLIP_PLANE0);
      glDisable (GL_CLIP_PLANE1);
      glDisable (GL_CLIP_PLANE2);
      glDisable (GL_CLIP_PLANE3);
      glDisable (GL_CLIP_PLANE4);

      return;
    }

  glEnable (GL_CLIP_PLANE0);
  glEnable (GL_CLIP_PLANE1);
  glEnable (GL_CLIP_PLANE2);
  glEnable (GL_CLIP_PLANE3);
  glEnable (GL_CLIP_PLANE4);

  glClipPlane (GL_CLIP_PLANE0, pool_cue_shadow_planes [0]);
  glClipPlane (GL_CLIP_PLANE1, pool_cue_shadow_planes [1]);
  glClipPlane (GL_CLIP_PLANE2, pool_cue_shadow_planes [2]);
  glClipPlane (GL_CLIP_PLANE3, pool_cue_shadow_planes [3]);
  glClipPlane (GL_CLIP_PLANE4, pool_cue_shadow_planes [4]);
}

static void
pool_feed_GL_with_simple_cue_shadow (PoolViewGL *data)
{
  dbl *ptr;
  int i;

  glBegin (GL_QUADS);

  ptr = pool_cue_shadow_vertices[0];
  for (i = 0; i < 4; i++, ptr +=3)
    glVertex3d (ptr [X], ptr [Y], ptr [Z]);
  glEnd ();
}

static void
pool_feed_GL_with_marks (TableMark *mark, PoolViewGL *data)
{
  g_return_if_fail (mark != NULL);

  glNormal3dv (render_up_vector);
  glColor3dv (mark -> skin -> color1);

  switch (mark -> mark_type)
    {
    case TableMark_Line:
      glBegin (GL_LINES);
	glVertex3d (mark->p1 [X], mark -> p1 [Y], SLIGHTLY_ABOVE_TABLE);
	glVertex3d (mark->p2 [X], mark -> p2 [Y], SLIGHTLY_ABOVE_TABLE);
      glEnd ();
      break;
      
    case TableMark_Point:
      {
	dbl   *p;
	dbl   size;

	p = mark -> p1;
	size = mark -> dparam / 1.41421356237 /* sqrt (2) */;
     
	glBegin (GL_QUADS);
          glVertex3d (p [X]       , p [Y] + size  , SLIGHTLY_ABOVE_TABLE);
          glVertex3d (p [X] - size, p [Y]		, SLIGHTLY_ABOVE_TABLE);
          glVertex3d (p [X]       , p [Y] - size  , SLIGHTLY_ABOVE_TABLE);     
          glVertex3d (p [X] + size, p [Y]		, SLIGHTLY_ABOVE_TABLE);
	glEnd ();
	break;
      }

    case TableMark_Diamond:
      {
	dbl	*p, *s;

	p = mark -> p1;
	s = mark -> p2;

	glBegin (GL_QUADS);
	 glVertex3d (p [X]        , p [Y] + s [Y], pool_table.table_height);
	 glVertex3d (p [X] - s [X], p [Y]        , pool_table.table_height);
	 glVertex3d (p [X]        , p [Y] - s [Y], pool_table.table_height);
	 glVertex3d (p [X] + s [X], p [Y]        , pool_table.table_height);
        glEnd ();

	break;
      }
    case TableMark_Arc:
      {
	int	 prec;
	dbl	*ptr;

	prec = mark -> iparam + 1;

	if (mark -> data == NULL)
	  {
	    int		 i,j;
	    dbl		*ptr, delta,angle, alfa, radius;

	    mark -> data = ptr = g_new (dbl, 3 * (prec + 2 * prec + 4 * prec));
	    angle = mark->p2 [1] - mark -> p2 [0];
	    radius = mark -> dparam;

	    for (j = prec ; j <= 4 * prec; j *= 2)
	      for (i = 0, alfa = mark->p2[0], delta = angle / (dbl) (j - 1); 
		   i < j; 
		   i++, alfa += delta)
		{
		  *ptr++ = mark -> p1 [X] + radius * cos (alfa);
		  *ptr++ = mark -> p1 [Y] + radius * sin (alfa);
		  *ptr++ = SLIGHTLY_ABOVE_TABLE;
		}
	  }

	ptr = mark -> data;
	if (data -> lod > 0)
	  { ptr += 3 * prec; prec *= 2; }

	if (data -> lod > 1)
	  { ptr += 3 * prec; prec *= 2; }

	glVertexPointer (3, GL_DOUBLE, 0, ptr);
	glDrawArrays (GL_LINE_STRIP, 0, prec); 

	break;
      }
    default:
      g_assert_not_reached ();
    }
}

static void
pool_feed_GL_with_hints (PoolViewGL *data, Animation *anim)
{
#if MAX_PYTHIE_COLS <= 15
  static GLuint indices [] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
#else
# error Please update above lines to higher MAX_PYTHIE_COLS
#endif
  int		i;

  /* FIXME : should follow ball's colour */

#ifndef DEBUG
  if (! animation_finished_p ())
    return;
#endif

  glColor3d (1,1,1);

  for (i = 0; i < anim -> nof_balls; i++)
    if (pool_pythie_counts [i] > 1)
      {
	glVertexPointer (3, GL_DOUBLE, 0, pool_pythie_positions [i]);
	glDrawElements (GL_LINE_STRIP,  pool_pythie_counts [i],
			GL_UNSIGNED_INT, indices);
      }
}

static void
render_material (Skin *tmp)
{
  /* FIXME: use tmp... */
  static GLfloat specular[] = { .75, .75, .75, 1. };
  static GLfloat shininess[] = { 100 };
  static GLfloat diffuse[] = { 0.6, 0.6, 0.6 };

  glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, shininess);
}

static void
pool_feed_GL (PoolViewGL *data)
{
  Animation		*a;
  PoolBallSpaceInfo	*ball;
  
  int	i, frame, nob, draw_cue, draw_shadows;

  /* FIXME - skins,.... :) */
  render_material (NULL);
  
  a = gtulpas->game->animation;

  /* FIXME - should interpolate between frames !!! */
  nob = a->nof_balls;
  frame = floor (a->current_frame);

  draw_cue = game_cue_active (gtulpas->game);
  draw_shadows = (data->draw_adepts & PV_Draw_Umbrae) && properties.draw_shadows;

  ball = a->balls + nob * frame;

  if (data->with_lights)
    glEnable (GL_LIGHTING); 

  if (data->draw_adepts & PV_Draw_Balls)
    for (i = 0; i < nob; i++, ball++)
      feed_GL_with_sphere (data, ball, data->lod);

  if ((data->draw_adepts & PV_Draw_Cue) && draw_cue)
    pool_feed_GL_with_cue (data);

  if (data->draw_adepts & PV_Draw_Table)
    table_render (data, TRUE);

  if (data->draw_adepts & PV_Draw_Marks && pool_table.mark_list)
    {
      glDepthFunc (GL_LEQUAL);
      glPolygonOffset (-2, -2);
      glEnable (GL_POLYGON_OFFSET_FILL);
      g_list_foreach (pool_table.mark_list, 
		      (GFunc) pool_feed_GL_with_marks, 
		      data);
      glDisable (GL_POLYGON_OFFSET_FILL);
      glDepthFunc (GL_LESS);
    }

  if (draw_shadows && draw_cue)
    {
      pool_fill_cue_shadow_vertices ();
      glNormal3dv (render_up_vector);

      if (properties.real_cue_shadows)
	{
	  glDisable (GL_LIGHT0);

	  pool_feed_GL_with_cue_clips (TRUE);

	  glPolygonOffset (-2,-2);
	  glEnable (GL_POLYGON_OFFSET_FILL);

	  if (data->draw_adepts & PV_Draw_Table)
	    table_render (data, FALSE);

	  ball = a->balls + nob * frame;
	  for (i = 0; i < nob; i++, ball++)
	    feed_GL_with_sphere (data, ball, data -> lod);

	  glDisable (GL_POLYGON_OFFSET_FILL);
	  pool_feed_GL_with_cue_clips (FALSE);
	  glEnable (GL_LIGHT0);
	}
    }


  if (data->with_lights)
    glDisable (GL_LIGHTING); 

  if (draw_shadows)
    {
      glColor3ub (47, 79, 79);		/* dark slate gray */

      if ((! properties.real_cue_shadows) && draw_cue)
	pool_feed_GL_with_simple_cue_shadow (data);

      glPolygonOffset (-2,-2);
      glEnable (GL_POLYGON_OFFSET_FILL);

      ball = a->balls + nob * frame;
      for (i = 0; i < nob; i++, ball++)
	pool_feed_GL_with_sphere_shadow (data, ball);

      glDisable (GL_POLYGON_OFFSET_FILL);
    }

  if (data -> show_hints)
    pool_feed_GL_with_hints (data, a);
}

static void
pool_feed_GL_select_spheres (PoolViewGL *data)
{
  Animation		*a;
  PoolBallSpaceInfo	*ball;
  
  int	i, frame, nob;

  a = gtulpas->game->animation;
  nob = a->nof_balls;
  frame = floor (a->current_frame);
  ball = a->balls + nob * frame;

  for( i = 0; i < nob; i++, ball++)
    {
      glPushName (i);
        feed_GL_with_sphere (data, ball, 0);
      glPopName ();
    }
}

static void
pool_feed_GL_select_hole (PoolViewGL *data)
{
  int	i;

  for (i = 0; i < pool_table.nof_holes; i++)
    {
      glPushName (i);
        table_render_fake_hole (data, pool_table.holes + i, 0);
      glPopName ();
    }
}

/*
 *
 *  main rendering functions...
 */

static void
pool_feed_GL_blink (PoolViewGL *data)
{
  if (! data -> needs_blink)
    return;

  switch (data -> blink)
    {
    case Blink_None:
      glClearColor (0,0,0,1);
      break;
    case Blink_Good:
      glClearColor (0,0,1,1);
      break;
    case Blink_Bad:
      glClearColor (1,0,0,1);
      break;
    }
}

static void
pool_view_setup_GL (PoolViewGL *data)
{
  glEnableClientState (GL_VERTEX_ARRAY);
  glEnable (GL_DITHER);

  glEnable (GL_CULL_FACE);
  glCullFace (GL_BACK);			 /* to ensure default value */

  glEnable (GL_DEPTH_TEST);
  glDepthFunc (GL_LESS);

  if (data -> smooth_shading)
    glShadeModel (GL_SMOOTH);
  else
    glShadeModel (GL_FLAT);

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

  glColorMaterial (GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
  glEnable (GL_COLOR_MATERIAL);

  glPolygonMode (GL_FRONT_AND_BACK, (data -> shading) ? GL_FILL : GL_LINE);
}

static void 
pool_setup_lights (PoolViewGL *data)
{
  if (data->with_lights)
    {
      GLfloat	light_amb [4] = {0.4, 0.4, 0.4, 1};
      GLfloat	light_pos [4] = {0, 0, 1,0 /* 0 if direction. light */};

      glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
      glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
      glLightModelfv (GL_LIGHT_MODEL_AMBIENT, light_amb);

      glEnable (GL_LIGHT0);
      /* glEnable (GL_LIGHTING); -disabled/enabled operationally */

      /* TODO: parametrise following values (in funture) */
      glLightfv (GL_LIGHT0, GL_POSITION, light_pos);
    }
  else
    {
      glDisable (GL_LIGHT0);
      glDisable (GL_LIGHTING);
    }
}

static void
pool_view_setup_camera (PoolViewGL *viewdata, dbl *pick)
{
  if ((viewdata -> camera_change) || (pick))
    {
      glMatrixMode (GL_PROJECTION);
      glLoadIdentity ();

      if (pick)
	{
	  GLint		viewport [4];

	  viewport [0] = 0;
	  viewport [1] = 0;
	  viewport [2] = viewdata -> win_width;
	  viewport [3] = viewdata -> win_height;

	  gluPickMatrix (pick [0], pick [1], pick [2], pick [3], viewport);
	}

      gluPerspective (viewdata->fovy, 
		      ((dbl) (viewdata->win_width)) / ((dbl) (viewdata->win_height)), 
		      16.0, 4096.0);

      glMatrixMode (GL_MODELVIEW);
      viewdata -> camera_change =  (pick != NULL);
    }

  glLoadIdentity ();
  gluLookAt (viewdata->eye [X], viewdata->eye [Y], viewdata->eye [Z],
	     viewdata->focus [X], viewdata->focus [Y], viewdata->focus [Z], 
	     viewdata->up [X], viewdata->up [Y], viewdata->up [Z]);
}

int pool_view_do_select (PoolViewGL *viewdata, int x, int y, QueryType qdest)
{
#if 0
#define DEBUG_SELECT
#endif
  int	saved_lod;
  int 	count, maxim, minimum, i, result;
  dbl   pick_matrix [4];
  GLint *sbuffer, *ptr;
  
  pick_matrix [0] = x;
  pick_matrix [1] = viewdata -> win_height - y;
  pick_matrix [2] = POOL_PICK_WIDTH;
  pick_matrix [3] = POOL_PICK_HEIGHT;

  saved_lod = viewdata -> lod;
  viewdata -> lod = 0;

  pool_view_make_current (viewdata);
  pool_view_setup_camera (viewdata, pick_matrix);

  maxim = 4 * MAX (gtulpas->game->animation->nof_balls, pool_table.nof_holes);
  sbuffer = g_new (GLint, maxim);

  glSelectBuffer (maxim, sbuffer);  
#ifndef DEBUG_SELECT
  glRenderMode (GL_SELECT);
  glInitNames ();
#else
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
#endif

    if (qdest == Query_Ball) 
      pool_feed_GL_select_spheres (viewdata);
    else
      pool_feed_GL_select_hole (viewdata);

#ifndef DEBUG_SELECT
  count = glRenderMode (GL_RENDER);
#else
  gtk_gl_area_swapbuffers (viewdata->area);
  count = 0;
#endif
  viewdata -> lod = saved_lod;

  g_assert (count >= 0 && "Bad estimate of select buffer size");

  if ((count--) > 0)
    {
      /* there are four values in sbuffer:
       *  [0] num of names
       *  [1] min depth
       *  [2] max depth
       *  [3] name
       */

      minimum = sbuffer [1];
      result = sbuffer [3];	/* only one name */

      for (ptr = sbuffer + 4; count; count--, ptr += 4)
	if (minimum > ptr [1])
	  {
	    minimum = ptr [1];
	    result = ptr [3];
	  }
    }
  else
    result = -1;

  g_free (sbuffer);
  return result;
}

static int has_buggy_glx_swap_buffers = 0;

static int
my_fake_x_error_hanler (Display *disp, XErrorEvent *err)
{
  has_buggy_glx_swap_buffers = 1;
#ifdef DEBUG
  g_message ("Buggy GLX - ignored X error ...");
#endif
  return 0;
}

void
render_score_ball (Game *g, int ball_id, int width, int height,
		   GdkPixmap **pixmap, GdkBitmap **bitmap)
{
  static int		attrib [] = { GDK_GL_USE_GL,
				      GDK_GL_RGBA,
				      GDK_GL_NONE};
  static GLfloat	light_pos [4] = { 0,1,1, 0};

#define M 10.0

  int		depth;
  dbl		radius;

  GdkWindow	*win;
  GdkColormap   *cmap;
  GdkVisual	*vis;

  GdkGLPixmap	*glpix;
  GdkGLContext  *ctx;
  GdkPixmap	*tmpixmap;

  GdkImage	*image;
  char		*data, *dptr;

  GdkGC		*gc;
  GdkGCValues	 gc_values;


  dbl		a1, a2;

  PoolViewGL		fake_data;
  PoolBall	       *ball;

  void		       *ptr;

  vis = gdk_gl_choose_visual (attrib);
  depth = vis -> depth;
  
  fake_data.shading = 1;
  fake_data.smooth_shading = 1;
  fake_data.with_lights = 1;

  ball = g -> balls + ball_id;
  radius = ball -> ball_type -> radius;

  /* 
   *	PIXMAP
   */

  ctx = gdk_gl_context_new (vis);
  *pixmap = gdk_pixmap_new (NULL, width, height, depth);

  glpix = gdk_gl_pixmap_new (vis, *pixmap);

  if (! gdk_gl_pixmap_make_current (glpix, ctx))
    {
      g_warning ("Can not render ball %d", ball_id);
      *pixmap = NULL;
      *bitmap = NULL;
      return ;
    }

  pool_view_setup_GL (&fake_data);
  glDisable (GL_DEPTH_TEST);

  pool_setup_lights (&fake_data);
  glEnable (GL_LIGHTING);

  render_material (NULL);

  glClearColor (0,0,0,1);
  glClear (GL_COLOR_BUFFER_BIT);

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  gluPerspective (12.5,  /* 2 * atan2 (1,10) */
		  1,  /* aspect */
		  8. * radius,
		  12 * radius);
  gluLookAt ( 0, M*radius,0, /* eye */
	      0,0,0,	  /* center */
	      0,0,1);
  glMatrixMode (GL_MODELVIEW);
  
  glLightfv (GL_LIGHT0, GL_POSITION, light_pos);
		  
  glRotated ( a1 = (((((dbl) random ()) / ((dbl) RAND_MAX)) - 0.5) * 45.0),
	      1,0,0);
  glRotated ( a2 - ((((dbl) random ()) / ((dbl) RAND_MAX)) * 360.0),
	      0,1,0);

  feed_GL_with_raw_sphere (ball, 2, TRUE);

  /* Quick hack - overcome some mesa bux... 
   * We face two big problems:
   *  1) some Mesa implementations have problems with rendering
   *     to pixmaps and glXSwapBuffers must be always used... :(
   *  2) one sw glx implementation ignore the fact, that 
   *     glXSwapBuffers should not generate X error even buffer
   *     does not have double buffers... urmgh... 
   *  here is quick & dirty hack:
   */

  glFinish ();
  if (! has_buggy_glx_swap_buffers)
    {
      ptr = XSetErrorHandler (my_fake_x_error_hanler);
      gdk_gl_swap_buffers (*pixmap); 
      XSync (gdk_display, False);
      XSetErrorHandler (ptr);
    }
  XSync (gdk_display, False);

  gdk_gl_context_unref (ctx);
  gdk_gl_pixmap_unref (glpix);



  /*
   *
   *	BITMAP
   */

  ctx = gdk_gl_context_new (vis);
  tmpixmap = gdk_pixmap_new (NULL, width, height, depth);

  glpix = gdk_gl_pixmap_new (vis, tmpixmap);

  if (! gdk_gl_pixmap_make_current (glpix, ctx))
    {
      g_warning ("Can not render ball %d", ball_id);
      *pixmap = NULL;
      *bitmap = NULL;
      return;
    }

  pool_view_setup_GL (&fake_data);
  glDisable (GL_DEPTH_TEST);
  glShadeModel (GL_FLAT);
  glDisable (GL_COLOR_MATERIAL);
  glDisable (GL_DITHER);

  glClearColor (0,0,0,1);
  glClear (GL_COLOR_BUFFER_BIT);
  gluPerspective (13, 1, 8. * radius,12.*radius);
  gluLookAt (0, M*radius,0,
	     0,0,0,
	     0,0,1);
		  
  glRotated ( a1, 1,0,0);
  glRotated ( a2, 0,1,0);

  glColor3d (1,1,1);
  feed_GL_with_raw_sphere (ball, 2, FALSE);

  glFinish ();
  if (! has_buggy_glx_swap_buffers)
    gdk_gl_swap_buffers (tmpixmap);
  XSync (gdk_display, False);

  gdk_gl_context_unref (ctx);
  gdk_gl_pixmap_unref (glpix);

  gc_values.function = GDK_COPY;
  gc = gdk_gc_new_with_values (tmpixmap, &gc_values, GDK_GC_FUNCTION);

  image = gdk_image_get (tmpixmap, 0,0,width, height);

  data = (char *) calloc ((width / 8 + 1) * height, 1);
  
  {
    int i,j, s;

    for (dptr = data, i = 0; i < height; i++)
      {
	for (j = 0, s = 1; j < width; j++)
	  {
	    *dptr |= gdk_image_get_pixel (image, j,i) ? s : 0;
	    if (s == 128)
	      { s = 1; dptr ++; }
	    else
	      s <<= 1;
	  }
	if (s != 1)
	  dptr ++;
      }
  }
  
  *bitmap = gdk_bitmap_create_from_data (*pixmap, data, width, height);
  
  free (data);
  gdk_image_destroy (image);

#if 0
  { /* TEST */
    GtkWidget	*w;
    GtkWidget	*p;
    
    gtk_widget_push_colormap (gdk_colormap_new (vis, FALSE));
    gtk_widget_push_visual (vis);
    
    w = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    p = gtk_pixmap_new (*pixmap, NULL /* *bitmap */);
    
    gtk_container_add (GTK_CONTAINER (w), p);
    gtk_widget_show_all (w);
    
    gtk_widget_pop_visual ();
    gtk_widget_pop_colormap ();
  }
#endif
#undef M
}

void pool_view_do_redraw (PoolViewGL *viewdata)
{
  if (! pool_view_make_current (viewdata))
    return;
  pool_view_setup_camera (viewdata, NULL);
  
  if (viewdata -> dirty_flag)
    {
      pool_view_setup_GL (viewdata);
      viewdata -> dirty_flag = 0;
    }

  pool_setup_lights (viewdata);
  pool_feed_GL_blink (viewdata);

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  pool_feed_GL (viewdata);
  gtk_gl_area_swapbuffers (viewdata->area);
}

void pool_view_GL_finish (PoolViewGL *viewdata)
{
  if (! pool_view_make_current (viewdata))
    return;

  glFinish ();
}

void
pool_view2table_coords (PoolViewGL *data, PoolBallSpaceInfo *binfo,
			int win_x, int win_y)
{
  GLdouble	 model_matrix [16];
  GLdouble	 projection_matrix [16];
  GLint		 viewport [4];

  dbl		 p1 [3];
  dbl		 p2 [3];
  dbl		 v [3];
  dbl		 res [3];
  dbl		 delta, ha, radius;
  int		 i;
  
  PoolBall	*ball;
  PoolBallSpaceInfo tmp;

  if (! pool_view_make_current (data))
    return;

  glGetDoublev (GL_PROJECTION_MATRIX, projection_matrix);
  glGetDoublev (GL_MODELVIEW_MATRIX, model_matrix);
  glGetIntegerv (GL_VIEWPORT, viewport);

  win_y = viewport [3] - win_y;
  ball = binfo -> ball;
  radius = ball -> ball_type -> radius;

  gluUnProject ( win_x, win_y, 1,
		 model_matrix, projection_matrix, viewport,
		 p1 + X, p1 + Y, p1 + Z);
  gluUnProject ( win_x, win_y, 0,
		 model_matrix, projection_matrix, viewport,
		 p2 + X, p2 + Y, p2 + Z);

  delta = p1 [Z] - p2 [Z];

  if (delta >  EPSILON)	/* parallel to table or pointing to the sky */
    return;

  ha = (p1 [Z] - radius) / delta;
  vec_sub (v, p2, p1);
  vec_mul_c (res, v, ha);
  vec_add (res, res, p1);

  binfo -> p [X] = res [X];
  binfo -> p [Y] = res [Y];
  binfo -> p [Z] = radius;

  return;
}

