#include "glib.h"

#include <gdk/gdkkeysyms.h>
#include <gdk/gdktypes.h>

#include "poolkeys.h"

static GList* pool_key_list = NULL;

void pool_key_init_default (void)
{
  pool_key_add (GDK_Tab, 0, PKB_NONE);

  pool_key_add (GDK_plus, 0,   PKB_ZOOM_IN);
  pool_key_add (GDK_equal, 0,   PKB_ZOOM_IN);
  pool_key_add (GDK_plus, GDK_SHIFT_MASK, PKB_ZOOM_IN);
  pool_key_add (GDK_KP_Add, 0, PKB_ZOOM_IN);

  pool_key_add (GDK_minus, 0,  PKB_ZOOM_OUT);
  pool_key_add (GDK_KP_Subtract, 0,  PKB_ZOOM_OUT);  

  pool_key_add (GDK_Left,    GDK_MOD1_MASK, PKB_LEFT);
  pool_key_add (GDK_KP_Left, GDK_MOD1_MASK, PKB_LEFT);

  pool_key_add (GDK_Right,    GDK_MOD1_MASK, PKB_RIGHT);
  pool_key_add (GDK_KP_Right, GDK_MOD1_MASK, PKB_RIGHT);

  pool_key_add (GDK_Left,     0, PKB_ROTATE_LEFT);
  pool_key_add (GDK_KP_Left,  0, PKB_ROTATE_LEFT);

  pool_key_add (GDK_Right,    0, PKB_ROTATE_RIGHT);
  pool_key_add (GDK_KP_Right, 0, PKB_ROTATE_RIGHT);

  pool_key_add (GDK_Up,    GDK_MOD1_MASK, PKB_UP);
  pool_key_add (GDK_KP_Up, GDK_MOD1_MASK, PKB_UP);

  pool_key_add (GDK_Down,    GDK_MOD1_MASK, PKB_DOWN);
  pool_key_add (GDK_KP_Down, GDK_MOD1_MASK, PKB_DOWN);

  pool_key_add (GDK_Up,     0, PKB_ROTATE_UP);
  pool_key_add (GDK_KP_Up,  0, PKB_ROTATE_UP);

  pool_key_add (GDK_Down,    0, PKB_ROTATE_DOWN);
  pool_key_add (GDK_KP_Down, 0, PKB_ROTATE_DOWN);

  pool_key_add ('H', 0, PKB_TOGGLE_HINT);

  pool_key_add ('S', 0, PKB_MOD_SPIN);
  pool_key_add ('S', GDK_CONTROL_MASK, PKB_STOP_SPIN);
  pool_key_add ('R', 0, PKB_MOD_ROTATE);

  pool_key_add ('Z', 0, PKB_MOD_ZOOM);
  pool_key_add ('P', 0, PKB_MOD_PAN);

  pool_key_add ('C', 0, PKB_MOD_CUE_MOTION);
  pool_key_add ('V', 0, PKB_MOD_CUE_ROTATE);

  pool_key_add ('M', 0, PKB_MOD_MOVE_BALL);

  /* views */

  pool_key_add (GDK_KP_0, 0, PKB_VIEW_BIND_CUE);

  pool_key_add (GDK_KP_1, 0, PKB_VIEW_1);
  pool_key_add (GDK_KP_2, 0, PKB_VIEW_2);
  pool_key_add (GDK_KP_3, 0, PKB_VIEW_3);
  pool_key_add (GDK_KP_4, 0, PKB_VIEW_4);
  pool_key_add (GDK_KP_5, 0, PKB_VIEW_5);
  pool_key_add (GDK_KP_6, 0, PKB_VIEW_6);
  pool_key_add (GDK_KP_7, 0, PKB_VIEW_7);
  pool_key_add (GDK_KP_8, 0, PKB_VIEW_8);
  pool_key_add (GDK_KP_9, 0, PKB_VIEW_9);


  pool_key_add (GDK_KP_Insert, 0, PKB_VIEW_BIND_CUE);

  pool_key_add (GDK_KP_End, 0, PKB_VIEW_1);
  pool_key_add (GDK_KP_Down, 0, PKB_VIEW_2);
  pool_key_add (GDK_KP_Page_Down, 0, PKB_VIEW_3);
  pool_key_add (GDK_KP_Left, 0, PKB_VIEW_4);
  pool_key_add (GDK_KP_Begin, 0, PKB_VIEW_5); /* ??? */
  pool_key_add (GDK_KP_Right, 0, PKB_VIEW_6);
  pool_key_add (GDK_KP_Home, 0, PKB_VIEW_7);
  pool_key_add (GDK_KP_Up, 0, PKB_VIEW_8);
  pool_key_add (GDK_KP_Page_Up, 0, PKB_VIEW_9);
}

void pool_key_add (guint keyval, guint state, PoolKeyAction action)
{
  PoolKey	*pk;

  pk = g_new (PoolKey, 1);

  pk -> keyval = keyval;
  pk -> state = state;
  pk -> action = action;

  pool_key_list = g_list_prepend (pool_key_list, pk);
}

PoolKey*
pool_key_lookup (guint keyval, guint state)
{
  GList		*ptr;

  if (('a' <= keyval) && (keyval <= 'z'))
      keyval += (int ) 'A' -  (int)'a';

  if (('A' <= keyval) && (keyval <= 'Z'))
    state &= ~(GDK_SHIFT_MASK | GDK_LOCK_MASK);

  for (ptr = pool_key_list; ptr != NULL; ptr = ptr->next)
    if ((((PoolKey *)(ptr->data)) -> keyval == keyval) &&
	(((PoolKey *)(ptr->data)) -> state == (state & POOL_KEY_MASK)))
      return (PoolKey *)(ptr->data);

  return NULL;
}

void pool_key_remove (PoolKey *pk)
{
  pool_key_list = g_list_remove (pool_key_list, pk);
}

void pool_key_remove_all (void)
{
  g_list_free (pool_key_list);
  pool_key_list = NULL;
}
