#include "debug.h"

#include <ctype.h>
#include <stdio.h>

#include <gtk/gtk.h>
#include <gdk/gdkprivate.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include <gtkgl/gtkglarea.h>

#include <math.h>
#include "mathutils.h"

#include "gtulpas.h"

#include "poolview.h"
#include "poolfeedgl.h"

#include "cue.h"
#include "poolkeys.h"

static char*  pool_view_data_key = "pool-view-GL-data";

static int pool_view_attr_list [] = {
  GDK_GL_USE_GL,
  GDK_GL_RGBA,
  GDK_GL_DOUBLEBUFFER,
  GDK_GL_DEPTH_SIZE, 1,
  GDK_GL_NONE
};

GList*		pool_view_list = NULL;

PoolModAction pool_current_action = PMod_NOTHING;

	/* ball in hand */
static int	pool_bih_ok = FALSE;


#define POOLVIEW_EVENT_MASK	  GDK_EXPOSURE_MASK		\
				| GDK_ENTER_NOTIFY_MASK		\
				| GDK_LEAVE_NOTIFY_MASK		\
				| GDK_BUTTON_PRESS_MASK		\
				| GDK_BUTTON_RELEASE_MASK	\
				| GDK_POINTER_MOTION_MASK	\
/* 				| GDK_POINTER_MOTION_HINT_MASK*/\
				| GDK_KEY_PRESS_MASK		\
				| GDK_KEY_RELEASE_MASK

#define POOLVIEW_EVENT_GRAB_MASK  GDK_ENTER_NOTIFY_MASK		\
				| GDK_LEAVE_NOTIFY_MASK  	\
				| GDK_BUTTON_PRESS_MASK 	\
				| GDK_BUTTON_RELEASE_MASK 	\
								\
				| GDK_POINTER_MOTION_MASK	\
/* 				| GDK_POINTER_MOTION_HINT_MASK  */

	/* mouse speed buffer */

#define MOUSE_BUFFER_SIZE 4 /* 16 */

static dbl	mbuffer [MOUSE_BUFFER_SIZE] = {0,0,0,0 /* , 0,0,0,0, 0,0,0,0, 0,0,0,0*/};
static dbl	mbuffer_sum = 0;
static int      mbuffer_cur = 0;



void views_update_position_from_cue (void);


void pool_view_redraw (PoolViewGL *view)
{
  view ->needs_redraw = 1;
}

int
pool_view_make_current (PoolViewGL *view)
{
  if (! (GTK_WIDGET_VISIBLE (GTK_WIDGET (view->area)) && GTK_WIDGET_MAPPED (view -> area)))
	  return FALSE;

#ifdef HAVE_GTK_GL_AREA_MAKE_CURRENT
  if (! gtk_gl_area_make_current (view -> area))
#else
	  if (! gtk_gl_area_begingl (view -> area))
#endif
    {
      g_error ("glXMakeCurrent was not successfull!");
      exit (1);
    }

  return TRUE;
}

void
pool_view_draw_fps (PoolViewGL *view, dbl render_time)
{
  static GdkGC  *gc = NULL;
  static GdkFont *font = NULL;
  static GdkColor color;
  GdkWindow *window;

  char buffer [64];
  int length, i, num;
  dbl sum, *ptr;

  window = ((GtkWidget *) (view->area)) -> window;
  
  if (gc == NULL)
    {
      GdkColormap	*cmap;

      gc = gdk_gc_new (window);
      cmap = ((GdkWindowPrivate *) window) -> colormap;
      gdk_color_white (cmap, &color);

      gdk_gc_set_foreground (gc, &color);
      font = gdk_font_load ("9x15bold");
    }

  view -> fps [(int) (view -> fps_pos)] = 1.0 / render_time;
  view -> fps_pos = (view -> fps_pos + 1) & (POOL_FPS_CACHE_SIZE - 1);

  sum = 0.0;
  for (num = i = 0, ptr = view -> fps; i < POOL_FPS_CACHE_SIZE; i++, ptr++)
    if (*ptr != 0.0)
      { sum += *ptr; num++; }

  length = snprintf (buffer, 64, "fps (%3.1f), cue (%3.1f)", 
		     sum / (dbl) num,
		     pool_cue_mouse_speed * mbuffer_sum / (dbl) MOUSE_BUFFER_SIZE);
  gdk_draw_text ((GdkDrawable *) window, font, gc, 10, 10,
		 buffer, length);
}

static void
view_ball_in_hand (PoolViewGL *view, Game *game,
		   int win_x, int win_y)
{
  int			bid;
  BallInHand	       *bih_info;
  PoolBallSpaceInfo    *bsi;
  PoolBall	       *ball;

  bih_info = (BallInHand *) (game -> balls_in_hand_list -> data);
  bid = bih_info -> ball_id;
  bsi = ANIMATION_LAST_FRAME (game->animation);
  ball = bsi [bid] . ball;

  pool_view2table_coords (view, bsi + bid, win_x, win_y);
  pool_bih_ok = pool_place_ball_is_correct (bid, bih_info, bsi + bid, bsi);

  ball -> is_invalid = ! pool_bih_ok;
  pool_view_redraw_mark_all ();
}

static void
view_ball_move (PoolViewGL *view, Game *g, dbl delta_x, dbl delta_y)
{
  PoolBallSpaceInfo	tmp, *frame;
  BallInHand		bih;

  dbl			vx [3], vy [3];
  int			id;

  if (g -> ball_move_id == -1)
    return;
  
  id = g->ball_move_id;
  frame = ANIMATION_LAST_FRAME (g->animation);
  
  vec_cpy (vx, view -> right);
  vy [X] = - vx [Y];
  vy [Y] =   vx [X];
  vy [Z] = 0.0;

  vec_mul_c (vx, vx, delta_x);
  vec_mul_c (vy, vy, delta_y);
  vec_add (vx,vx,vy);

  tmp = *(frame + id);
  vec_add (tmp.p, tmp.p, vx);

  bih.ball_id = id;
  bih.validate = g->ball_move_verify_func;

  if (pool_place_ball_is_correct (id, &bih, &tmp, frame))
    {
      *(frame + id) = tmp;
      vec_cpy (pool_cue.center_of_rotation, tmp.p);
      cue_update_position (&pool_cue);

      views_update_position_from_cue ();
      pool_view_redraw_mark_all ();
    }
}

void
view_update_orientation (PoolViewGL *view)
{
  angles_to_vectors (view -> vp . alfa, view -> vp . beta, 
		     view -> orientation, view -> right, view -> up);
}


void
view_update_position (PoolViewGL *view)
{
  vec_mul_c (view -> eye, view -> orientation, - (view -> vp . eye_center_distance));
  vec_mul_c (view -> focus, view -> orientation, view -> vp .eye_focus_distance);
  vec_add (view -> eye, view -> eye, view -> vp . center_of_rotation);
  vec_add (view -> focus, view -> focus, view -> eye);
}

static void					/* start           end */
view_interpolate (ViewPosition *dest, dbl t2, ViewPosition *a1, ViewPosition *a2)
{
  dbl t1;

  t1 = 1.0 - t2;

  dest -> alfa = fmod (t1 * a1->alfa + t2 * a2->alfa, 2.0 * PI);
  dest -> beta = fmod (t1 * a1->beta + t2 * a2->beta, 2.0 * PI);

  dest -> eye_focus_distance = t1 * a1->eye_focus_distance + t2 * a2->eye_focus_distance;
  dest -> eye_center_distance = t1 * a1->eye_center_distance + t2 * a2->eye_center_distance;

  dest -> center_of_rotation [X] = t1 * a1 -> center_of_rotation [X] + t2 * a2 -> center_of_rotation [X];
  dest -> center_of_rotation [Y] = t1 * a1 -> center_of_rotation [Y] + t2 * a2 -> center_of_rotation [Y];
  dest -> center_of_rotation [Z] = t1 * a1 -> center_of_rotation [Z] + t2 * a2 -> center_of_rotation [Z];
}

static void 
interpolate_adjust_angles (dbl *start, dbl *end, dbl *delta)
{
  *start = fmod (*start, 2. * PI);
  *end = fmod (*end, 2. * PI);

  *delta = *end - *start;

  if (*delta > PI)
    *delta = - (2. * PI - *delta);

  if (*delta < -PI)
    *delta = 2. * PI + *delta;

  *end = *start + *delta;
  *delta = fabs (*delta);
}

static void
view_start_interpolate (PoolViewGL* view, ViewPosition *end)
{
  ViewPosition	*start;
  dbl		 l;
  dbl		 delta_alfa;
  dbl		 delta_beta;

  if (view -> interp_timer == NULL)
    view -> interp_timer = g_timer_new ();
  else
    g_timer_reset (view -> interp_timer);

  view -> needs_interp = TRUE;
  view -> interp.start = view -> vp;
  view -> interp.end = *end;

  start = & (view -> interp.start);
  l = euclid (view -> vp . center_of_rotation, end -> center_of_rotation) / POOL_DEFAULT_INTERP_SPEED;
  
  interpolate_adjust_angles (&(start -> alfa), &(view -> interp.end.alfa), &delta_alfa);
  interpolate_adjust_angles (&(start -> beta), &(view -> interp.end.beta), &delta_beta);

  view -> interp . time = 
    MAX (
	 MAX (l,
	      MAX ( delta_beta / POOL_DEFAULT_INTERP_ANGLE_SPEED,
		    delta_alfa / POOL_DEFAULT_INTERP_ANGLE_SPEED)),
	 MAX (fabs (start->eye_focus_distance - end->eye_focus_distance) / POOL_DEFAULT_INTERP_SPEED,
	      fabs (start->eye_center_distance - end->eye_center_distance) / POOL_DEFAULT_INTERP_SPEED));
}

static void
view_set_from_game (PoolViewGL *view, int id)
{
  ViewPosition *vp;

  vp = gtulpas -> game -> view_positions + id;
  if (vp -> eye_focus_distance < 0)
    return;

  view -> bind_cue = FALSE;
  view_start_interpolate (view, vp);
}

static void 
view_toggle_bind_to_cue (PoolViewGL *view, PoolCue *cue)
{
  if (! view -> bind_cue)
    {
      ViewPosition vp;

      vp . alfa = cue -> alfa;
      vp . beta = cue -> beta + POOL_DEFAULT_EYE_CUE_ANGLE;
      vp . eye_center_distance = POOL_DEFAULT_EYE_TOP_CUE_DISTANCE;
      vp . eye_focus_distance =  POOL_DEFAULT_EYE_TOP_CUE_DISTANCE;
      
      vec_cpy (vp . center_of_rotation, cue -> center_of_rotation);

      view_start_interpolate (view, &vp);
    }

  view -> bind_cue = ! (view -> bind_cue);
}

static void
views_rebind_to_cue_cb (PoolViewGL *view)
{
  if (! view -> bind_cue)
    return;

  view -> bind_cue = FALSE;
  view_toggle_bind_to_cue (view, &pool_cue);
}

void
views_rebind_to_cue (void)
{
  g_list_foreach (pool_view_list, (GFunc) views_rebind_to_cue_cb, NULL);
}

static void 
pool_view_zoom (PoolViewGL *view, dbl zoom_coeff)
{
  if (view -> needs_interp)
    return;

  view -> vp . eye_center_distance += zoom_coeff;
  view_update_position (view);

  view -> camera_change = TRUE;
  pool_view_redraw (view);
}

static void
pool_eye_sidewalk (PoolViewGL *data, dbl h_change, dbl v_change)
{
  ViewPosition *view;
  dbl		hv [3];
  dbl   	vv [3];

  if (data -> needs_interp)
    return;

  if (data -> bind_cue && animation_finished_p ())
      return;

  view = &(data -> vp);

  vec_mul_c (hv, data -> right, h_change);
  vec_mul_c (vv, data -> up, v_change);

  vec_add (hv, hv, vv);

  vec_add (view -> center_of_rotation, view -> center_of_rotation, hv);
  vec_add (data -> eye, data -> eye, hv);
  vec_add (data -> focus, data -> focus, hv);

  data -> camera_change = TRUE;
  pool_view_redraw (data);
}

static void
views_update_position_from_cue_cb (PoolViewGL *view)
{
  if (! view -> bind_cue)
    return;

  vec_cpy (view -> vp . center_of_rotation,pool_cue.center_of_rotation);

  view_update_orientation (view);
  view_update_position (view);
  
  view -> camera_change = TRUE;
  pool_view_redraw (view);
}

void
views_update_position_from_cue (void)
{
  g_list_foreach (pool_view_list, (GFunc) views_update_position_from_cue_cb, NULL);
}

static void
pool_rotate_binded (PoolViewGL *view, PoolViewGL *m)
{
  if ((view == m) || (! view -> bind_cue))
    return;

  
  view -> vp . beta = m -> vp . beta;
  view -> vp . alfa = m -> vp . alfa;

  view_update_orientation (view);
  view_update_position (view);

  view -> camera_change = TRUE;
  pool_view_redraw (view);
}

static void
pool_eye_rotate (PoolViewGL *view, dbl h_change, dbl v_change)
{
  if (view -> needs_interp)
    return;

  if (view -> bind_cue && animation_finished_p ())
    {
      if (view -> vp . beta > (POOL_DEFAULT_EYE_CUE_ANGLE + EPSILON))
	{
	  cue_rotate (&pool_cue, - h_change, - v_change);
	  view -> vp . beta = pool_cue . beta + POOL_DEFAULT_EYE_CUE_ANGLE;
	}
      else
	{
	  cue_rotate (&pool_cue, - h_change, 0.0);
	  view -> vp . beta = fmod (view -> vp . beta + v_change + PI, 2.0 * PI) - PI;  
	}

      view -> vp . alfa = pool_cue . alfa;
      if (view -> vp . beta < DEG2RAD (7.5))
	view -> vp . beta = DEG2RAD (7.5);

      g_list_foreach (pool_view_list, (GFunc) pool_rotate_binded, (gpointer) view);
    }
  else
    {
      view -> vp . alfa = fmod(view -> vp . alfa + h_change, 2.0 * PI);
      view -> vp . beta = fmod(view -> vp . beta + v_change + PI, 2.0 * PI) - PI;

      if (view -> vp . beta < 0)
	view -> vp . beta = 0;
    }

  view_update_orientation (view);
  view_update_position (view);

  view -> camera_change = TRUE;
  pool_view_redraw (view);
}

void
pool_views_mover (void)
{
  GList		*lptr;
  PoolViewGL	*view;
  dbl		 delta_t;

  delta_t = g_timer_elapsed (pool_anim_spin_timer, NULL);
  g_timer_reset (pool_anim_spin_timer);

  for (lptr = pool_view_list; lptr; lptr = lptr -> next)
    {
      view = (PoolViewGL *) lptr -> data;

      if (view -> needs_interp)
	{
	  delta_t = g_timer_elapsed (view -> interp_timer, NULL);

	  if (delta_t <= view->interp.time)
	    view_interpolate (&(view -> vp), delta_t / view -> interp.time,
			      &(view -> interp.start), &(view -> interp.end));
	  else
	    {
	      view -> vp = view -> interp . end;
	      view -> needs_interp = FALSE;
	    }

	  view_update_orientation (view);
	  view_update_position (view);

	  view -> camera_change = TRUE;
	  pool_view_redraw (view);
	}

      if (view -> spin_speed == 0.0)
	continue;

      if (view-> spin_done == 0)
	pool_eye_rotate (view, POOL_ROTATE_MOUSE_COEFF * delta_t * (view -> spin_speed), 0);

      view -> spin_done = 0;
    }
}

PoolViewGL*
pool_view_get_data (GtkWidget *widget)
{
  PoolViewGL *ret;

  ret = gtk_object_get_data (GTK_OBJECT (widget), pool_view_data_key);
  if (! ret)
    g_error ("pool_view_get_data: widget has not pvGL-data!");

  return ret;
}

gint poolview_expose (GtkWidget *widget, GdkEventExpose *event, PoolViewGL *viewdata)
{
  /* draw only last expose */

  if (event->count == 0) 
    viewdata -> needs_redraw = 1;

  return TRUE;
}

gint poolview_destroy (GtkWidget *widget, PoolViewGL *viewdata)
{
  pool_view_list = g_list_remove (pool_view_list, viewdata);

  return TRUE;
}

static void
view_change_size (PoolViewGL *viewdata, gint width, gint height)
{
  if (! pool_view_make_current (viewdata))
    return;

  viewdata->win_height = height;
  viewdata->win_width = width;

  glViewport (0, 0, width, height);
  glMatrixMode (GL_MODELVIEW);

  viewdata->camera_change = TRUE;
  pool_view_redraw (viewdata);

}

int
views_force_size (void)
{
  GList	     *ptr;
  PoolViewGL *data;
  GtkWidget  *widget;

  for (ptr = pool_view_list; ptr; ptr = ptr -> next)
    {
      data = (PoolViewGL *) (ptr->data);
      widget = (GtkWidget *) data -> area;

      view_change_size (data, widget->allocation.width, widget->allocation.height); 
    }

  /* sometimes it view_change_size is called from idle fuctions -=> we should return FALSE */
  return FALSE;
}

gint poolview_configure (GtkWidget *widget, GdkEventConfigure *event, PoolViewGL *viewdata)
{
  view_change_size (viewdata, event->width, event->height);
  return TRUE;
}

gint poolview_enter_notify (GtkWidget *widget)
{ 
  gtk_widget_grab_focus (widget);
  return TRUE;
}

gint poolview_motion_notify (GtkWidget *widget, GdkEventMotion *event, PoolViewGL *viewdata)
{ 
  static int	last_x, last_y;
  static guint32 last_time = 0;
  int 		x,y;
  dbl		delta_x, delta_y, delta_t, speed;
  PoolModAction action;
  Game		*game;

  GdkModifierType state;
 
  x = (int) event -> x;
  y= (int) event->y;
  state = (GdkModifierType) (event -> state);

  game = gtulpas -> game;
  action = PMod_NOTHING;

  if (state & GDK_BUTTON1_MASK)
    {
      if (pool_current_action == PMod_NOTHING)
	action = PMod_ROTATE;
      else
	action = pool_current_action;
    }
  if (state & GDK_BUTTON3_MASK)
    action = PMod_CUE_MOTION;

  if (state & GDK_BUTTON2_MASK)
    action = PMod_PAN;

  if ((action == PMod_CUE_MOTION) && (game->view_answer || game->balls_in_hand_list))
    action = PMod_NOTHING;

  if ((state & (GDK_BUTTON1_MASK | GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) == 0
      && game -> balls_in_hand_list)
    action = PMod_BALL_IN_HAND;

  if (last_time == event->time)
    last_time--; /* to prevent delta_t == 0 */

  delta_x = ((dbl) (x - last_x));
  delta_y = ((dbl) (y - last_y));
  delta_t = (dbl) (event->time - last_time) / 1000.0;

  if (state & GDK_CONTROL_MASK)
    { delta_x /= 20.0 ; delta_y /= 20.; }

  last_x = x;
  last_y = y;
  last_time = event -> time;

  speed = fabs (delta_y) / delta_t;

  mbuffer_sum -= mbuffer [mbuffer_cur];
  mbuffer_sum += mbuffer [mbuffer_cur++] = speed;

  if (mbuffer_cur == MOUSE_BUFFER_SIZE)
    mbuffer_cur = 0;

#if 0
  g_message ("delty (%s %.6f %.6f %.6f)", (event->is_hint ? "hint" : "event"), delta_x, delta_y, delta_t);
#endif

  switch (action)
    {
    case PMod_SPIN:
      viewdata -> spin_speed = - delta_x / delta_t;
      /* NO BREAK !!! */

    case PMod_ROTATE:
      viewdata -> spin_done = 1;
      pool_eye_rotate (viewdata, 
		       - delta_x * POOL_ROTATE_MOUSE_COEFF,
		       delta_y * POOL_ROTATE_MOUSE_COEFF);
      break;


    case PMod_ZOOM:
      pool_view_zoom (viewdata, delta_y / 2.0);
      break;

    case PMod_PAN:
      pool_eye_sidewalk (viewdata, 
			 - delta_x * POOL_SIDEWALK_MOUSE_COEFF,
			 delta_y * POOL_SIDEWALK_MOUSE_COEFF);
      break;
	
    case PMod_CUE_MOTION:
      cue_motion (&pool_cue, delta_y, mbuffer_sum / (dbl) MOUSE_BUFFER_SIZE);
      break;

    case PMod_CUE_ROTATE:
      cue_rotate (&pool_cue,
		  - delta_x * POOL_ROTATE_MOUSE_COEFF,
		  (state & GDK_SHIFT_MASK) 
		  ? 0
		  : delta_y * POOL_ROTATE_MOUSE_COEFF);
      break;

    case PMod_MOVE_BALL:
      view_ball_move (viewdata, game, delta_x, - delta_y);
      break;

    case PMod_BALL_IN_HAND:
      view_ball_in_hand (viewdata, game, x, y);
      break;

    default:
      break;
    }

  return TRUE;
}


gint poolview_button_press (GtkWidget *widget, GdkEventButton *event, PoolViewGL *viewdata)
{
  if (viewdata -> has_pointer == 0)
    {
#ifdef DEBUG
      g_message("Grabbing");
#endif
  
      viewdata->has_pointer = 1;
      gdk_pointer_grab (widget->window, TRUE, POOLVIEW_EVENT_GRAB_MASK, 
			NULL, NULL, event->time);
      gtk_grab_add (widget);
    }

  if ((pool_current_action == PMod_NOTHING) 
      && (event -> button == 1))
    {
      Game	*g;

      g = gtulpas->game;

      if (g->balls_in_hand_list)
	{
	  if (pool_bih_ok)
	    game_ball_in_hand_answer (g);
	  view_blink_start (viewdata, pool_bih_ok ? Blink_Good : Blink_Bad);
	}
      else if (g->view_answer)
	game_answer_query (gtulpas->game, viewdata, event->x, event->y);
    }
  return TRUE;
}

gint poolview_button_release (GtkWidget *widget, GdkEventButton *event, PoolViewGL *viewdata)
{
  if (viewdata->has_pointer)
    {
#ifdef DEBUG
      g_message ("Ungrabbing");
#endif

      gdk_pointer_ungrab (event->time);
      gtk_grab_remove (widget);
      viewdata->has_pointer = 0;
    }
  
  return TRUE;
}

gint poolview_key_press (GtkWidget *widget, GdkEventKey *event, PoolViewGL* viewdata)
{
  PoolKey	*pk;

  pk = pool_key_lookup (event->keyval, event->state);
  g_message ("Press");

  if (pk == NULL)
    { 
#ifdef DEBUG
      /* quick'n'dirty hack... */
      if (((event->state & POOL_KEY_MASK) == 0) &&
	  (isalnum ((char) event->keyval)))
	{
	  if (('1' <= event->keyval) && (event->keyval <= '7'))
	    {
	      viewdata->lod = event->keyval - '1';
#ifdef DEBUG
	      g_message (">> Setting lod: %d", viewdata->lod);
#endif
	      pool_view_redraw (viewdata);
	      return TRUE;
	    }

	  switch (event->keyval)
	    {
	      /* dirty hack - as any other entry here - but this is more hackish :) */
	    case 'n':
	      viewdata -> show_fps ^= 1;
	      pool_view_redraw (viewdata);
	      return TRUE;
	    case 'w':
	      viewdata->with_lights = 0;
	      viewdata->shading = 0; 
	      viewdata->smooth_shading = 0;
	      viewdata->dirty_flag = 1;
	      pool_view_redraw (viewdata);
	      return TRUE;
	    case 'f':
	      viewdata->with_lights = 1;
	      viewdata->shading = 1; viewdata->smooth_shading = 0;
	      viewdata->dirty_flag = 1;
	      pool_view_redraw (viewdata);
	      return TRUE;
	    case 'g':
	      viewdata->with_lights = 1;
	      viewdata->dirty_flag = 1;
	      viewdata->shading = 1; viewdata->smooth_shading = 1;
	      pool_view_redraw (viewdata);
	      return TRUE;
	    case 'l':
	      viewdata->with_lights = ! viewdata->with_lights;
	      viewdata->dirty_flag = 1;
	      pool_view_redraw (viewdata);
	      return TRUE;
	    default:
	      break;
	    }
	}
      g_message ("Nic"); 
#endif /* DEBUG */
      return TRUE; 
    }

  switch (pk->action)
    {
    case PKB_NONE:
      break;
    case PKB_ZOOM_IN:
    case PKB_ZOOM_OUT:
      pool_view_zoom (viewdata, ((pk->action == PKB_ZOOM_OUT) ? 1 : (-1)) * 5);
      break;
    case PKB_LEFT:
      pool_eye_sidewalk (viewdata, -POOL_WALK_LENGTH, 0);
      break;
    case PKB_RIGHT:
      pool_eye_sidewalk (viewdata, +POOL_WALK_LENGTH, 0);
      break;
    case PKB_ROTATE_LEFT:
      pool_eye_rotate (viewdata, -POOL_ROTATE_ANGLE, 0);
      break;
    case PKB_ROTATE_RIGHT:
      pool_eye_rotate (viewdata,  POOL_ROTATE_ANGLE, 0);
      break;

    case PKB_UP:
      pool_eye_sidewalk (viewdata, 0, +POOL_WALK_LENGTH);
      break;
    case PKB_DOWN:
      pool_eye_sidewalk (viewdata, 0, -POOL_WALK_LENGTH);
      break;
    case PKB_ROTATE_UP:
      pool_eye_rotate (viewdata, 0,  POOL_ROTATE_ANGLE);
      break;
    case PKB_ROTATE_DOWN:
      pool_eye_rotate (viewdata, 0, -POOL_ROTATE_ANGLE);
      break;

    case PKB_TOGGLE_HINT:
      viewdata -> show_hints ^= 1;
      cue_recalc_hints (viewdata);
      break;
    case PKB_STOP_SPIN:
      viewdata -> spin_speed = 0.0;
      break;

    case PKB_MOD_ROTATE:
      pool_current_action = PMod_ROTATE;
      break; 	
    case PKB_MOD_SPIN:
      pool_current_action = PMod_SPIN;
      break;
    case PKB_MOD_ZOOM:
      pool_current_action = PMod_ZOOM;
      break;
    case PKB_MOD_PAN:
      pool_current_action = PMod_PAN;
      break;
    case PKB_MOD_CUE_MOTION:
      pool_current_action = PMod_CUE_MOTION;
      break;
    case PKB_MOD_CUE_ROTATE:
      pool_current_action = PMod_CUE_ROTATE;
      break;
    case PKB_MOD_MOVE_BALL:
      pool_current_action = PMod_MOVE_BALL;
      break;

    case PKB_VIEW_BIND_CUE:
      view_toggle_bind_to_cue (viewdata, &pool_cue);
      break;

    case PKB_VIEW_1: case PKB_VIEW_2: case PKB_VIEW_3: case PKB_VIEW_4:
    case PKB_VIEW_5: case PKB_VIEW_6: case PKB_VIEW_7: case PKB_VIEW_8:
    case PKB_VIEW_9: 
      view_set_from_game (viewdata, pk->action - PKB_VIEW_1);
      break;

    default:
      g_assert_not_reached ();
      break;
    }

  gtk_signal_emit_stop_by_name (GTK_OBJECT (widget), "key_press_event");
  return TRUE;
}

gint
poolview_key_release (GtkWidget *widget, GdkEventKey *event, PoolViewGL* viewdata)
{
  PoolKey	*pk;

  g_message ("Release");

  pk = pool_key_lookup (event->keyval, event->state);
  if (pk == NULL)
    return TRUE;

  switch (pk->action)
    {
    case PKB_MOD_ROTATE:
      pool_current_action &= ~PMod_ROTATE;
      break; 	
    case PKB_MOD_SPIN:
      pool_current_action &= ~PMod_SPIN;
      break;
    case PKB_MOD_ZOOM:
      pool_current_action &= ~PMod_ZOOM;
      break;
    case PKB_MOD_PAN:
      pool_current_action &= ~PMod_PAN;
      break;
    case PKB_MOD_CUE_MOTION:
      pool_current_action &= ~PMod_CUE_MOTION;
      break;
    case PKB_MOD_CUE_ROTATE:
      pool_current_action &= ~PMod_CUE_ROTATE;
      break;
    case PKB_MOD_MOVE_BALL:
      pool_current_action &= ~PMod_MOVE_BALL;
      break;
    default:
      break;
    }

  return TRUE;
}

PoolViewGL*
pool_view_new (GtkWidget *to_clone)
{
  static int	pool_view_id = 1;

  GtkWidget	*area;
  int		i;
  
  PoolViewGL	*newview;
  PoolViewGL	*shareview;


  newview = g_new (PoolViewGL, 1);

  if (to_clone != NULL)
    {
      shareview = pool_view_get_data (to_clone);
      *newview = *shareview;
    }
  else
    {
      if (pool_view_list == NULL)
	shareview = NULL;
      else
	shareview = (PoolViewGL *) (pool_view_list -> data);

      newview -> vp . alfa = POOL_INIT_ALFA;
      newview -> vp . beta = POOL_INIT_BETA;

      newview -> vp . eye_focus_distance = POOL_INIT_EYE_FOCUS;
      newview -> vp . eye_center_distance = POOL_INIT_EYE_CENTER;

      newview -> vp . center_of_rotation [X] = POOL_INIT_CENTER_X;
      newview -> vp . center_of_rotation [Y] = POOL_INIT_CENTER_Y;
      newview -> vp . center_of_rotation [Z] = POOL_INIT_CENTER_Z;

      view_update_orientation (newview);
      view_update_position (newview);

      newview -> fovy = POOL_INIT_FOVY;

      newview -> with_lights = POOL_DEFAULT_LIGHTS;
      newview -> shading = POOL_DEFAULT_SHADING;
      newview -> smooth_shading = POOL_DEFAULT_SMOOTH;

      newview -> draw_adepts = PV_DEFAULT_ADEPTS;

      /* DIRTY HACK */
      newview -> lod = 0;

      newview -> show_hints = 0;
      newview -> show_fps = 0;
    }

  newview -> has_pointer = 0;
  newview -> camera_change = 1;
  newview -> dirty_flag = 1;
  newview -> pv_id = pool_view_id++;

  newview -> spin_speed = 0;
  newview -> spin_done = 0;

  newview -> needs_blink = TRUE;
  newview -> blink_timer = NULL;

  newview -> needs_interp = FALSE;
  newview -> interp_timer = NULL;

  pool_view_list = g_list_prepend (pool_view_list, (gpointer) newview);

  area = gtk_gl_area_share_new (pool_view_attr_list, 
				shareview == NULL ? NULL : shareview -> area);

  newview -> area = (GtkGLArea *) area;

  gtk_quit_add_destroy (1, GTK_OBJECT (area));
  gtk_object_set_data (GTK_OBJECT (area), pool_view_data_key, (gpointer) newview);

  GTK_WIDGET_SET_FLAGS(area, GTK_CAN_FOCUS | GTK_CAN_DEFAULT);
  gtk_widget_add_events(area, POOLVIEW_EVENT_MASK);

  gtk_signal_connect (GTK_OBJECT(area), "expose_event",
                      GTK_SIGNAL_FUNC(poolview_expose), (gpointer) newview);
  gtk_signal_connect (GTK_OBJECT(area), "configure_event",
                      GTK_SIGNAL_FUNC(poolview_configure), (gpointer) newview);
  gtk_signal_connect (GTK_OBJECT(area), "destroy",
		      GTK_SIGNAL_FUNC (poolview_destroy), (gpointer) newview); 

  gtk_signal_connect (GTK_OBJECT(area), "motion_notify_event",
                      GTK_SIGNAL_FUNC(poolview_motion_notify), (gpointer) newview);

  gtk_signal_connect (GTK_OBJECT(area), "enter_notify_event",
                      GTK_SIGNAL_FUNC(poolview_enter_notify), (gpointer) newview);

  gtk_signal_connect (GTK_OBJECT(area), "button_press_event",
                      GTK_SIGNAL_FUNC(poolview_button_press), (gpointer) newview);

  gtk_signal_connect (GTK_OBJECT(area), "button_release_event",
                      GTK_SIGNAL_FUNC(poolview_button_release), (gpointer) newview);

  gtk_signal_connect (GTK_OBJECT(area), "key_press_event",
		      GTK_SIGNAL_FUNC(poolview_key_press), (gpointer) newview);
  gtk_signal_connect (GTK_OBJECT(area), "key_release_event",
		      GTK_SIGNAL_FUNC(poolview_key_release), (gpointer) newview);


  /* init fps cache */

  for (i = 0; i < POOL_FPS_CACHE_SIZE; i++)
    newview -> fps [i] = 0.0;

  newview -> fps_pos = 0;;

  return newview;
}


/*
 *
 *		blinking
 */

static dbl	 view_blink_duration = 0.2;

void
view_blink_set_duration (dbl duration)
{
  view_blink_duration = duration;
}

void
view_blink_start (PoolViewGL *view, ViewBlinkType blink)
{
  if (blink == view -> blink)
    return;

  view -> blink = blink;
  view -> needs_blink = TRUE;
  view -> needs_redraw = TRUE;

  if (blink != Blink_None)
    {
      if (view -> blink_timer == NULL)
	view -> blink_timer = g_timer_new ();

      g_timer_start (view -> blink_timer);
    }
}

static void
view_blink_stop (PoolViewGL *data)
{
  if (data -> blink == Blink_None)
    return;

  data -> blink = Blink_None;

  data -> needs_redraw = TRUE;
  data -> needs_blink = TRUE;
}

void
view_blink_process (PoolViewGL *data)
{
  if (data -> blink_timer == NULL || data -> blink == Blink_None)
    return;

  if (g_timer_elapsed (data -> blink_timer, NULL) > view_blink_duration)
    view_blink_stop (data);
}

