#ifndef __POOLVIEW_H__
#define __POOLVIEW_H__

#include <gtk/gtk.h>
#include <gtkgl/gtkglarea.h>
#include "mathutils.h"

#define POOL_WALK_LENGTH	 10.0
#define POOL_SIDEWALK_MOUSE_COEFF 1.0

#define POOL_ZOOM_STEP		  5

#define POOL_ROTATE_ANGLE	0.15
#define POOL_ROTATE_MOUSE_COEFF (PI / 180.0)

#define POOL_INIT_CENTER_X	0
#define POOL_INIT_CENTER_Y	0
#define POOL_INIT_CENTER_Z	0

#define POOL_INIT_ALFA		(PI  * 5.0 / 4.0)
#define POOL_INIT_BETA		(PI / 4.0)

#define POOL_INIT_EYE_CENTER	100
#define POOL_INIT_EYE_FOCUS	100

#define POOL_INIT_FOVY		30

#define POOL_DEFAULT_EYE_CUE_ANGLE		 (DEG2RAD(20.0))
#define POOL_DEFAULT_EYE_TOP_CUE_DISTANCE	 125
#define POOL_DEFAULT_INTERP_SPEED		 300.0 
#define POOL_DEFAULT_INTERP_ANGLE_SPEED		 (PI * 2.0)

#define POOL_DEFAULT_LIGHTS	1
#define POOL_DEFAULT_SHADING	1
#define POOL_DEFAULT_SMOOTH	1
#define POOL_DEFAULT_UMBRA	1

#define POOL_FPS_CACHE_SIZE    16

#define POOL_DEFAULT_REQUIRED_FPS	30
#define POOL_DEFAULT_BIG_MOVER_FPS	60

#define POOL_SHADOW_SIZE_COEFF	0.8

typedef enum {
  PV_Draw_Balls			= (1 << 0),
  PV_Draw_Umbrae		= (1 << 1),
  PV_Draw_Cue			= (1 << 2),
  PV_Draw_Marks			= (1 << 3),
  PV_Draw_Table			= (1 << 4),
} PoolView_DrawAdepts;

typedef enum 
{
  Blink_None,
  Blink_Good,
  Blink_Bad
} ViewBlinkType;

#define PV_DEFAULT_ADEPTS (PV_Draw_Balls | PV_Draw_Umbrae | PV_Draw_Cue | PV_Draw_Marks | PV_Draw_Table)

typedef struct _PoolViewGL   PoolViewGL;
typedef struct _ViewPosition ViewPosition;

struct _ViewPosition 
{
	  /* orientation around Z axe */
  dbl			alfa;
	 /* declination */
  dbl			beta;

  dbl			eye_focus_distance;
  dbl			eye_center_distance;

  dbl			center_of_rotation [3];
};

struct _PoolViewGL
{
  GtkGLArea		*area;

  /* position and orientation of camera */
  
  ViewPosition		vp;

  dbl			fovy;

  dbl			eye [3];			/* c */
  dbl			focus [3];			/* c */

  dbl			up [3];				/* c */
  dbl			right [3];			/* c */
  dbl			orientation [3];		/* c */

  guint			win_width, win_height;

  guint16		pv_id;
  PoolView_DrawAdepts	draw_adepts;

  ViewBlinkType		blink;

	/* timers */
  GTimer	       *blink_timer;
  GTimer	       *interp_timer;

    /* GL drawing options */
  guint			with_lights	: 1;
  guint			shading		: 1;
  guint			smooth_shading	: 1;

    /* show various infomation */
  guint			show_hints	: 1;
  guint			show_fps	: 1;

    /* dirty flags. Set -=> recompute something*/
  guint			camera_change	: 1;
  guint			dirty_flag	: 1;


   /* Do we have to redraw ? */

  guint			needs_redraw	: 1;
  guint			redrawed	: 1;
  guint			needs_blink	: 1;

   /* GDK for flags */
  guint			has_pointer	: 1;

   /* motion of viewport */

  guint			bind_cue	: 1;
  guint			spin_done	: 1;
  guint			needs_interp	: 1;
  gint			spin_speed;


  struct {
    ViewPosition	start;
    ViewPosition	end;

    dbl			time;
  } interp;

    /* FPS cache */
  dbl			fps [POOL_FPS_CACHE_SIZE];
  char			fps_pos;

    /* DIRTY HACKS */  

  guint			lod;  
};


PoolViewGL*	 pool_view_new (GtkWidget *to_clone);
extern GList	*pool_view_list;

void pool_answer_querry (PoolViewGL *, int x, int y);
int  pool_view_should_answer ();

void pool_view_draw_fps (PoolViewGL *view, dbl render_time);

void views_rebind_to_cue (void);



/*
 *
 *	poolanim - plays pool animation 
 */

void poolanim_fps_set (int newfps);
void poolanim_setup (void);
void poolanim_shutdown (void);

extern GTimer		*pool_anim_spin_timer;

void pool_views_mover (void);
void pool_view_redraw_mark_all (void);

void pool_play_current_animation (void);
void pool_replay_current_animation (void);

/*
 *	dangerous functions
 */
int  views_force_size (void);

/*
 *
 *	view_blink - PoolViewGL blinking
 */

void view_blink_set_duration (dbl duration);
void view_blink_start (PoolViewGL *view, ViewBlinkType blink);
void view_blink_process (PoolViewGL *data);

void view_update_position    (PoolViewGL *view);
void view_update_orientation (PoolViewGL *view);

#endif
