/* properties and properties dialog2 1999/11/27 16:10:11
   $Id: properties.c,v 1.8 2000/01/18 20:50:06 0rfelyus Exp $

   Copyright (C) 1999 Radek Doulik <rodo@pha.eridan.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "debug.h"
#include <gnome.h>

#include "properties.h"
#include "gtulpas.h"

#include "poolview.h"
#include "pool.h"

#include "table.h"

GTulpasProperties properties;
static GTulpasProperties temp_prop;

/********* properties dialog ***********/

static GtkWidget *win = NULL;

static void
properties_cancel (void)
{
	debug ("properties cancel\n");

	win  = NULL;
}

#define CH(x, flag)  if (properties.## x != temp_prop.## x) flag = 1;

static void
properties_apply ()
{
	gint layout = 0;
	gint anim = 0;
	gint table = 0;
	gint redraw = 0;

	debug ("properties apply\n");

	CH (show_mini_views, layout);
	CH (show_game_status, layout);
	CH (blink_duration, anim);
	
	CH (glu_tesselator_ok, table);
	CH (funny_pockets, table);
	CH (render_table_outline, table);

	CH (real_cue_shadows, redraw);	
	CH (draw_shadows, redraw);

	properties = temp_prop;

	if (layout)
		gtulpas_layout_change (gtulpas);

	if (anim)
		view_blink_set_duration (properties.blink_duration);

	if (redraw)
	  pool_view_redraw_mark_all ();

	if (table)
	  {
	    pool_table.update_view_info = 1;
	    pool_view_redraw_mark_all ();
	  }
}

void
properties_load (GTulpasProperties *prop, gchar *prefix)
{
	gnome_config_push_prefix (prefix);

	prop->show_mini_views  = gnome_config_get_int ("show_mini_views=1");
	prop->show_game_status = gnome_config_get_int ("show_game_status=1");
	prop->blink_duration = gnome_config_get_float ("blink_duration=0.1");

	prop -> glu_tesselator_ok = gnome_config_get_bool ("GLU_tesselator_OK=false");
	prop -> funny_pockets = gnome_config_get_bool ("funny_pockets=false");
	prop -> render_table_outline = gnome_config_get_bool ("table_outline=true");

	prop -> real_cue_shadows = gnome_config_get_bool ("real_cue_shadows=false");
	prop -> draw_shadows = gnome_config_get_bool ("draw_shadows=true");


	prop->shading = gnome_config_get_int ("shading=2");
	prop->lod     = gnome_config_get_int ("level_of_detail=2");

	view_blink_set_duration (prop -> blink_duration);

	gnome_config_pop_prefix ();
}

static gint
properties_check_cb (GtkWidget *widget, gpointer data)
{
	guint		*loc;
	guint		newval;

	newval = (0 != GTK_TOGGLE_BUTTON (widget) -> active);
	loc = (guint *) data;

	if (*loc != newval) {
		*loc = newval;
		debug ("changed\n");
		gnome_property_box_changed (GNOME_PROPERTY_BOX (win));
	}

	return TRUE;
}

GtkWidget*
properties_check (char* label, guint *loc)
{
	GtkWidget	*w;

	w = gtk_check_button_new_with_label (label);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w), *loc);

	gtk_signal_connect (GTK_OBJECT (w), "toggled",
			    (GtkSignalFunc) properties_check_cb, (gpointer) loc);

	return w;
}

static gint
properties_radio_cb (GtkWidget *widget, gpointer data)
{
	guint		*loc;
	guint		myval;
	guint		newval;

	debug ("properties_radio_cb\n");

	newval = (0 != GTK_TOGGLE_BUTTON (widget) -> active);
	if (! newval)
		return TRUE;

	loc = (guint *) data;
	myval = (guint) gtk_object_get_user_data (GTK_OBJECT (widget));

	if (*loc != myval) {
		*loc = myval;
		debug ("changed\n");
		gnome_property_box_changed (GNOME_PROPERTY_BOX (win));
	}

	return TRUE;
}

GtkWidget*
properties_radio (GtkWidget *gwidget, char* label, guint val, guint *loc)
{
	GtkWidget*	w;

	w = (gwidget == NULL) 
		? gtk_radio_button_new_with_label (NULL, label)
		: gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (gwidget), label);
	gtk_object_set_user_data (GTK_OBJECT (w), (gpointer) val);

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w), val == *loc);

	gtk_signal_connect (GTK_OBJECT (w), "toggled",
			    (GtkSignalFunc) properties_radio_cb, (gpointer) loc);

	return w;
}

#define ADDC(name, prop)          check = properties_check (_(name), &temp_prop.## prop); \
                                  gtk_box_pack_start (GTK_BOX (vbox), check, FALSE, FALSE, 0);

#define ADDR(name, val, prop)     radio = properties_radio (radio, _(name), val, &temp_prop.## prop); \
                                  gtk_box_pack_start (GTK_BOX (vbox), radio, FALSE, FALSE, 0);

void
properties_save (GTulpasProperties *prop, gchar *prefix)
{
	gnome_config_push_prefix (prefix);

	gnome_config_set_int ("show_mini_views", prop->show_mini_views);
	gnome_config_set_int ("show_game_status", prop->show_game_status);
	gnome_config_set_float ("blink_duration", prop->blink_duration);

	gnome_config_set_int ("shading", prop->shading);
	gnome_config_set_int ("level_of_detail", prop->lod);

	gnome_config_set_bool ("GLU_tesselator_OK", prop->glu_tesselator_ok);
	gnome_config_set_bool ("funny_pockets", prop->funny_pockets);
	gnome_config_set_bool ("table_outline", prop->render_table_outline);

	gnome_config_set_bool ("real_cue_shadows", prop->real_cue_shadows);
	gnome_config_set_bool ("draw_shadows", prop->draw_shadows);

	gnome_config_pop_prefix ();
}

static void
create_tulpas_prop ()
{
	GtkWidget *frame;
	GtkWidget *vbox;
	GtkWidget *check;

	frame = gtk_frame_new (_("Layout"));
	gtk_container_set_border_width (GTK_CONTAINER (frame), 2);
	vbox  = gtk_vbox_new (FALSE, 2);

#if 0
	ADDC ("show mini views",  show_mini_views);
#endif
	ADDC ("show game status", show_game_status);

	gtk_container_add (GTK_CONTAINER (frame), vbox);

	gnome_property_box_append_page (GNOME_PROPERTY_BOX (win), frame, gtk_label_new (_("General")));
}

static void
create_graphics_prop ()
{
	GtkWidget *frame;
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *radio;
	GtkWidget *check;
	GSList    *group;

	hbox  = gtk_hbox_new (FALSE, 2);

	frame = gtk_frame_new (_("Table rendering"));
	gtk_container_set_border_width (GTK_CONTAINER (frame), 2);

	vbox = gtk_vbox_new (FALSE, 2);
	ADDC (_("Real pockets"), glu_tesselator_ok);
	ADDC (_("Shaded pockets"), funny_pockets);
	ADDC (_("Table outlines"), render_table_outline);
	
	gtk_container_add (GTK_CONTAINER (frame), vbox);
	gtk_box_pack_start_defaults (GTK_BOX (hbox), frame);

	frame = gtk_frame_new (_("Shadows"));
	gtk_container_set_border_width (GTK_CONTAINER (frame), 2);

	vbox = gtk_vbox_new (FALSE, 2);			       
	ADDC (_("Draw shadows"), draw_shadows);
	ADDC (_("Realistic cue shadow"), real_cue_shadows);

	gtk_container_add (GTK_CONTAINER (frame), vbox);
	gtk_box_pack_start_defaults (GTK_BOX (hbox), frame);

#if 0
	/* shading */
	frame = gtk_frame_new (_("Shading"));
	gtk_container_set_border_width (GTK_CONTAINER (frame), 2);

	vbox  = gtk_vbox_new (FALSE, 2);

	radio = NULL;
	ADDR ("none",   SH_NONE,   shading);
	ADDR ("linear", SH_LINEAR, shading);
	ADDR ("Goraud", SH_GORAUD, shading);

	gtk_container_add (GTK_CONTAINER (frame), vbox);

	gtk_box_pack_start_defaults (GTK_BOX (hbox), frame);

	/* shading */
	frame = gtk_frame_new (_("Level of detail"));
	gtk_container_set_border_width (GTK_CONTAINER (frame), 2);
	vbox  = gtk_vbox_new (FALSE, 2);

	radio = NULL;
	ADDR ("high",   LOD_HIGH,   lod);
	ADDR ("normal", LOD_NORMAL, lod);
	ADDR ("low",    LOD_LOW,    lod);

	gtk_container_add (GTK_CONTAINER (frame), vbox);

	gtk_box_pack_start_defaults (GTK_BOX (hbox), frame);
#endif
	gnome_property_box_append_page (GNOME_PROPERTY_BOX (win), hbox, gtk_label_new (_("Graphics")));
}

void
properties_show ()
{
	if (win) {
		gdk_window_raise (win->window);
		return;
	}

	temp_prop = properties;

	win = gnome_property_box_new ();
	gtk_window_set_title (GTK_WINDOW (win), _("Tulpas preferences"));

	create_tulpas_prop ();
	create_graphics_prop ();

	gtk_signal_connect (GTK_OBJECT (win), "apply",
			    GTK_SIGNAL_FUNC (properties_apply), NULL);

	gtk_signal_connect (GTK_OBJECT (win), "destroy",
			    GTK_SIGNAL_FUNC (properties_cancel), NULL);

	gtk_widget_show_all (win);
}
