/* queries2 1999/11/14 19:01:36 0rfelyus
   $Id: query.c,v 1.7 2000/02/26 10:21:23 0rfelyus Exp $

   Copyright (C) 1999 Daniel Skarda <0rfelyus@suse.cz>
   Copyright (C) 1999 Radek Doulik <rodo@suse.cz>

   This file is part of gmap.

   gmap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   gmap is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with gmap; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <glib.h>

#include "query.h"
#include "game.h"
#include "gtulpas.h"
#include <guile/gh.h>
#include "guileglue.h"

int
pool_view_should_answer (void)
{
  return FALSE;
  /* fix return (pool_current_query_list != NULL); */
}

gint
query_needs_answer (Query *query, gint round, gint player_id)
{
  SCM          r;

  if (query -> type == Query_Player)
    {
      query -> response = player_id;
      return FALSE;
    }

  query->response = QUERY_DO_CALL;

  if (query -> when_func == SCM_BOOL_T)
    return  TRUE;
  
  r = glue_safe_call3 (query -> when_func, 
		       SCM_MAKINUM (round), 
		       SCM_MAKINUM (player_id),
		       GAME_STATE (gtulpas->game));

  if (r == SCM_BOOL_F) 
    {
      query->response = QUERY_DO_IGNORE;
      return FALSE;
    }

  return TRUE;
}

gint
query_is_valid (Query *query, gint round, gint player_id, int answer)
{
  SCM		  validation;

  if (query -> valid_func == SCM_BOOL_T)
    {
      query -> response = answer;
      return TRUE;
    }

  validation = glue_safe_call4 (query -> valid_func, 
				SCM_MAKINUM (round),
				SCM_MAKINUM (player_id),
				GAME_STATE (gtulpas->game),
				SCM_MAKINUM (answer));


  if (SCM_NIMP (validation) && SCM_STRINGP (validation)) {
    /* not valid answer */
    char		*err;

    err = gh_scm2newstr (validation, NULL);
    query->error = g_strdup (err);      /* fix - possibly memory leak */
    scm_must_free (err);

    return FALSE;
  }

  /* we have valid answer, so do next */
  query -> response = answer;

  return TRUE;
}

#define QUERY_ITEMS_NOF	3

glue_enum_item query_items_sbv [QUERY_ITEMS_NOF];
glue_enum_item query_items_sbs [QUERY_ITEMS_NOF] = {
  { SCM_UNDEFINED, "ball", Query_Ball},
  { SCM_UNDEFINED, "pocket", Query_Pocket},
  { SCM_UNDEFINED, "player", Query_Player}
};


glue_enum_info query_enum_info = {
  query_items_sbs,
  query_items_sbv,
  QUERY_ITEMS_NOF
};

GList *
query_scm2list (SCM obj, int argn, const char *procname)
{
  static int	need_init = TRUE;

#define MASSERT(P,X)	if (! (P)) ERR (X)
#define ERR(X)		do { errobj = X; goto error; } while (0)

  GList			*retlist = NULL;
  Query			*query;
  SCM			errobj, dlist;
  SCM			when_f, valid_f, s_name, s_prompt;
  int			qtype;
  char			*name, *prompt;


  if (need_init)
    { need_init = FALSE; glue_enum_info_init (&query_enum_info); }

  
  while (obj != SCM_EOL)
    {
      MASSERT (SCM_CONSP (obj), obj);
      
      dlist = SCM_CAR (obj);
      obj = SCM_CDR (obj);

      MASSERT (SCM_CONSP (dlist), dlist);

      qtype = glue_enum_scm2value (SCM_CAR (dlist), &query_enum_info);
      MASSERT (qtype != -1, SCM_CAR (dlist));

      dlist = SCM_CDR (dlist);

      
      switch (qtype)
	{
	case Query_Player:
	  MASSERT (dlist == SCM_EOL, dlist);

	  name = g_strdup ("Query player");
	  prompt = g_strdup ("Player?");
	  when_f = SCM_BOOL_T;
	  valid_f = SCM_BOOL_T;
	  break;

	case Query_Ball:
	case Query_Pocket:
	  MASSERT (scm_list_p (dlist) == SCM_BOOL_T && gh_length (dlist) == 4, dlist);

	  s_name = SCM_CAR (dlist);
	  s_prompt = SCM_CADR (dlist);
	  when_f = SCM_CADDR (dlist);
	  valid_f = SCM_CADDDR (dlist);

	  MASSERT (SCM_NIMP (s_name) && SCM_STRINGP (s_name), s_name);
	  MASSERT (SCM_NIMP (s_prompt) && SCM_STRINGP (s_prompt), s_prompt);
	  MASSERT (when_f == SCM_BOOL_T || scm_procedure_p (when_f) == SCM_BOOL_T, when_f);
	  MASSERT (valid_f == SCM_BOOL_T || scm_procedure_p (valid_f) == SCM_BOOL_T, valid_f);
	  
	  name = GLUE_STRDUP (s_name);
	  prompt = GLUE_STRDUP (s_prompt);
	  break;

	default:
	  g_assert_not_reached ();
	}

      query = g_new (Query, 1);
      query -> type = qtype;
      query -> name = name;
      query -> prompt = prompt;
      query -> when_func = when_f;
      query -> valid_func = valid_f;

      if (SCM_NIMP (when_f))
	scm_protect_object (when_f);

      if (SCM_NIMP (valid_f))
	scm_protect_object (valid_f);

      retlist = g_list_append (retlist, query);
    }

  return retlist;

 error:
  SCM_ASSERT (0, errobj, argn, procname);
  return NULL;
}


void
query_free (GList *qlist)
{
  g_list_foreach (qlist, (GFunc) g_free, NULL);
  g_list_free (qlist);
}
