#include <glib.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "mathutils.h"
#include "poolview.h"
#include "poolfeedgl.h"
#include "render.h"

dbl    render_up_vector [3] = {0,0,1};

		/* LIDs = List IDentifierS */
static LOD_array render_disk_LIDs = {
  -1, -1, -1, -1, -1, -1
};

static GLUtesselator *tesselator = NULL;
static GList	     *tess_garbage_list = NULL;

static GLUquadricObj *quadric [2] = {NULL, NULL};


/*
 *
 *		Level of detail
 */

void
LOD_array_init (LOD_array array)
{
  int	i;

  for (i = 0; i < RENDER_MAX_LOD; i++)
    array [i] = -1;
}

void
LOD_array_GL_free (LOD_array array)
{
  int  i;

  for (i = 0; i < RENDER_MAX_LOD; i++)
    if (array [i] != -1)
      { glDeleteLists (array [i],1); array [i] = -1; }
}

/* 
 *
 *	various usefull primitives
 */

void
render_disk (PoolViewGL *data)
{
  int  *iptr;
  
  iptr = render_disk_LIDs + data -> lod;

  if (*iptr == -1)
    {
      int	i, slices = 8;
      GLUquadricObj *qptr;

      for (i = data->lod; i ; i--, slices = (slices * 3) / 2);

      qptr = render_quadric_init (RENDER_QUADRIC_NO_NORMALS);
      *iptr = glGenLists (1);

      glNewList (*iptr, GL_COMPILE_AND_EXECUTE);
        gluDisk (qptr, 0.0, 1.0 , slices, 1);
      glEndList ();

      render_quadric_finish (qptr);
    }
  else
    glCallList (*iptr);
}


/*
 *	Tesselator & callbacks
 */


static void
tess_begin_cb(GLenum which)
{
   glBegin(which);
}

static void
tess_end_cb(void)
{
  glEnd();
}

static void
tess_vertex_cb (GLdouble *p)
{
  glVertex3dv (p);
}

static void 
tess_combine_cb (GLdouble coords[3], void *vertex_data,
		 GLfloat wieght[4], void **outData)
{
  dbl	*tmp;

  tmp = g_new (dbl, 3);

  tess_garbage_list = g_list_prepend (tess_garbage_list, tmp);

  tmp [0] = coords [0];
  tmp [1] = coords [1];
  tmp [2] = coords [2];

  *outData = tmp;
}


static void 
tess_error_cb (GLenum errorCode)
{
   const GLubyte *estring;

   estring = gluErrorString(errorCode);
   g_error ("Tessellation Error (%d, %x): %s\n", errorCode, errorCode, estring);
}


void
render_tesselator_finish (void *ptr)
{
  g_list_foreach (tess_garbage_list, (GFunc) g_free, NULL);
  g_list_free (tess_garbage_list);
  tess_garbage_list = NULL;
}

void*
render_tesselator_init (void)
{
  if (! tesselator)
    tesselator = gluNewTess ();

  gluTessProperty (tesselator, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_POSITIVE);
  gluTessNormal (tesselator, 0,0, 1);

  gluTessCallback (tesselator, GLU_TESS_BEGIN, 
				 (GLvoid (*) ()) &tess_begin_cb);
  gluTessCallback (tesselator, GLU_TESS_END, 
				 (GLvoid (*) ()) &tess_end_cb);
  gluTessCallback (tesselator, GLU_TESS_VERTEX, 
				 (GLvoid (*) ()) &tess_vertex_cb);
  gluTessCallback (tesselator, GLU_TESS_ERROR, 
				  (GLvoid (*) ()) &tess_error_cb);
  gluTessCallback (tesselator, GLU_TESS_COMBINE, 
				  (GLvoid (*) ()) &tess_combine_cb);

  return tesselator;
}

void*
render_quadric_init (int quality)
{
  if (quadric [0] == NULL)
    {
      quadric [RENDER_QUADRIC_NO_NORMALS] = gluNewQuadric ();
      quadric [RENDER_QUADRIC_WITH_NORMALS] = gluNewQuadric ();

      gluQuadricDrawStyle (quadric [RENDER_QUADRIC_NO_NORMALS], (GLenum) GLU_FILL);
      gluQuadricDrawStyle (quadric [RENDER_QUADRIC_WITH_NORMALS], (GLenum) GLU_FILL);
      gluQuadricNormals   (quadric [RENDER_QUADRIC_WITH_NORMALS], (GLenum) GLU_SMOOTH);
    }

  g_return_val_if_fail ((0<=quality) && (quality < 2), quadric [0]);
  return quadric [quality];
}

void
render_quadric_finish (void *ptr)
{
  /* EMPTY */
}
