#include <math.h>
#include <GL/gl.h>
#include <glib.h>

#include "mathutils.h"


GLdouble	*ts_vertices = NULL;
GLdouble	*ts_normals = NULL;
GLdouble        *ts_tex_coord = NULL;
GLint		*ts_indices = NULL;

#define		STORE_INDEX(INDEX,VALUE)	*(INDEX)++ = (VALUE)


#define TS_FREE(X)	do { if (X) g_free (X); (X) = NULL; } while (0)

void
ts_free (void)
{
  TS_FREE (ts_vertices);
  TS_FREE (ts_normals);  
  TS_FREE (ts_indices);
  TS_FREE (ts_tex_coord);
}

void 
ts_sphere_create (dbl radius, int xy, int z)
{
  int		i,j, z1, z2;
  dbl		deltaz, deltaxy;
  dbl		sx,sy,sz;
  dbl		alfa, beta;

  GLint	       *iptr;
  GLdouble     *nptr;
  GLdouble     *vptr;

  ts_free ();

  vptr = ts_vertices = g_new (GLdouble, 3 * ((xy /* + 1 */) * (z + 1) + 2));
  nptr = ts_normals = g_new (GLdouble, 3 * ((xy /* + 1 */) * (z + 1) + 2));
  iptr = ts_indices = g_new (GLint, z * (2 * (xy + 1) + 2));

  deltaz = PI / (dbl) z;
  deltaxy = 2. * PI / (dbl) xy;

  for (i = 1, beta = - PI / 2.0 + deltaz; 
       i < z ;
       i++, beta += deltaz)
    for (j = 0, alfa = 0.0;
	 j < xy; 
	 j++, alfa += deltaxy)
      {
        sz = cos (beta);
	sx = cos (alfa) * sz;
	sy = sin (alfa) * sz;
	sz = sin (beta);

	*vptr++ = radius * (*nptr++ = sx);
	*vptr++ = radius * (*nptr++ = sy);
	*vptr++ = radius * (*nptr++ = sz);
      }

  *vptr++ = radius * (*nptr++ = 0);
  *vptr++ = radius * (*nptr++ = 0);
  *vptr++ = radius * (*nptr++ = 1);

  *vptr++ = radius * (*nptr++ =  0);
  *vptr++ = radius * (*nptr++ =  0);
  *vptr++ = radius * (*nptr++ = -1);

  STORE_INDEX (iptr, GL_TRIANGLE_FAN);
  STORE_INDEX (iptr, 1 + 1 + xy);
  STORE_INDEX (iptr, xy * (z - 1) + 1);		/* center */
  STORE_INDEX (iptr, 0);

  for (i = xy - 1 ; i >= 0 ; i--)
    STORE_INDEX (iptr, i);

  STORE_INDEX (iptr, GL_TRIANGLE_FAN);
  STORE_INDEX (iptr, 1 + 1 + xy);
  STORE_INDEX (iptr, xy * (z - 1));

  STORE_INDEX (iptr, j = xy * (z - 1) - 1);

  for (i = j - xy + 1;
       i <= j;
       i++)
    STORE_INDEX (iptr, i);

  for (j = 0; j < z - 2; j++)
    {
      STORE_INDEX (iptr, GL_QUAD_STRIP);
      STORE_INDEX (iptr, 2 * xy + 2);
      STORE_INDEX (iptr, z1 = xy * (j + 1));
      STORE_INDEX (iptr, z2 = xy * j);

      for (i = 1; i < xy; i++)
	{
	  STORE_INDEX (iptr, *(iptr - 2) + 1);
	  STORE_INDEX (iptr, *(iptr - 2) + 1);
	}
      STORE_INDEX (iptr, z1);
      STORE_INDEX (iptr, z2);
    }

  STORE_INDEX (iptr, 0);
  STORE_INDEX (iptr, 0);
}

void
ts_sphere_cap (dbl radius, dbl s_width, int xy, int z)
{
  int		i,j, realz, z1,z2;
  GLint		*iptr;
  GLdouble	*vptr, *nptr;

  dbl		sx,sy,sz;
  dbl		start,alfa, beta;
  dbl		delta_alfa, delta_beta;

  g_assert (s_width > 0.0 && s_width < 1.0);

  start = asin (s_width);
  realz = ceil (((dbl) z) * (PI/2.0 - start) /  PI);

  ts_free ();

  vptr = ts_vertices = g_new (GLdouble, 3 * (realz * xy + 1));
  nptr = ts_normals = g_new (GLdouble, 3 * (realz * xy + 1));
  iptr = ts_indices = g_new (GLint, 2 + realz * (2 + 2 * (xy + 1)) + (2 + 2 + xy));

  delta_alfa = 2. * PI / (dbl) xy;
  delta_beta = (PI/2. - start) / (dbl) realz;

  for (i = 0, beta = start;
       i < realz;
       i++, beta += delta_beta)
    for (j = 0, alfa = 0.0;
	 j < xy;
	 j++, alfa += delta_alfa)
      {
        sz = cos (beta);
	sx = cos (alfa) * sz;
	sy = sin (alfa) * sz;
	sz = sin (beta);

	*vptr++ = radius * (*nptr++ = sx);
	*vptr++ = radius * (*nptr++ = sy);
	*vptr++ = radius * (*nptr++ = sz);
      }

  *vptr++ = radius * (*nptr++ = 0);
  *vptr++ = radius * (*nptr++ = 0);
  *vptr++ = radius * (*nptr++ = 1);

  STORE_INDEX (iptr, GL_TRIANGLE_FAN);
  STORE_INDEX (iptr, 1 + 1 + xy);
  STORE_INDEX (iptr, realz * xy);
  STORE_INDEX (iptr, j = xy * realz - 1);

  for (i = j - xy + 1; i <= j; i++)
    STORE_INDEX (iptr, i);

  for (j = 0; j < realz - 1; j++)
    {
      STORE_INDEX (iptr, GL_QUAD_STRIP);
      STORE_INDEX (iptr, 2 * xy + 2);
      STORE_INDEX (iptr, z1 = xy * (j + 1));
      STORE_INDEX (iptr, z2 = xy * j);

      for (i = 1; i < xy; i++)
	{
	  STORE_INDEX (iptr, *(iptr - 2) + 1);
	  STORE_INDEX (iptr, *(iptr - 2) + 1);
	}
      STORE_INDEX (iptr, z1);
      STORE_INDEX (iptr, z2);
    }

  STORE_INDEX (iptr, 0);
  STORE_INDEX (iptr, 0);
}

void
ts_sphere_strip (dbl radius, dbl s_width, int xy, int z)
{
  int		i,j, realz, z1,z2;
  GLint		*iptr;
  GLdouble	*vptr, *nptr;

  dbl		sx,sy,sz;
  dbl		start,alfa, beta;
  dbl		delta_alfa, delta_beta;

  g_assert (s_width > 0.0 && s_width < 1.0);

  start = asin (s_width);
  realz = ceil (((dbl) z) * (2. * start) /  PI);

  ts_free ();

  vptr = ts_vertices = g_new (GLdouble, 3 * (realz + 1) * xy);
  nptr = ts_normals = g_new (GLdouble, 3 * (realz + 1) * xy);
  iptr = ts_indices = g_new (GLint, 2 + realz * (2 + (xy+1) * 2));

  delta_alfa = 2. * PI / (dbl) xy;
  delta_beta = 2. * start / (dbl) realz;

  for (i = 0, beta = - start;
       i <= realz;
       i++, beta += delta_beta)
    for (j = 0, alfa = 0.0;
	 j < xy;
	 j++, alfa += delta_alfa)
      {
        sz = cos (beta);
	sx = cos (alfa) * sz;
	sy = sin (alfa) * sz;
	sz = sin (beta);

	*vptr++ = radius * (*nptr++ = sx);
	*vptr++ = radius * (*nptr++ = sy);
	*vptr++ = radius * (*nptr++ = sz);
      }

  for (j = 0; j < realz; j++)
    {
      STORE_INDEX (iptr, GL_QUAD_STRIP);
      STORE_INDEX (iptr, 2 * xy + 2);
      STORE_INDEX (iptr, z1 = xy * (j + 1));
      STORE_INDEX (iptr, z2 = xy * j);

      for (i = 1; i < xy; i++)
	{
	  STORE_INDEX (iptr, *(iptr - 2) + 1);
	  STORE_INDEX (iptr, *(iptr - 2) + 1);
	}
      STORE_INDEX (iptr, z1);
      STORE_INDEX (iptr, z2);
    }

  STORE_INDEX (iptr, 0);
  STORE_INDEX (iptr, 0);
}

void
ts_sphere_tstrip (dbl radius, dbl s_width, int xy, int z)
{
  int		i,j, realz, z1,z2, tmp;
  GLint		*iptr;
  GLdouble	*vptr, *nptr, *tptr;

  dbl		sx,sy,sz;
  dbl		start,alfa, beta;
  dbl		delta_alfa, delta_beta;

  g_assert (s_width > 0.0 && s_width < 1.0);

  start = asin (s_width);
  realz = ceil (((dbl) z) * (2. * start) /  PI);

  ts_free ();

  tmp = (realz + 1) * xy + realz + 1;

  vptr = ts_vertices = g_new (GLdouble, 3 * tmp);
  tptr = ts_tex_coord = g_new (GLdouble, 2 * tmp);
  nptr = ts_normals = g_new (GLdouble, 3 * tmp);
  iptr = ts_indices = g_new (GLint, 2 + realz * (2 + (xy+1) * 2) + realz);

  delta_alfa = 2. * PI / (dbl) xy;
  delta_beta = 2. * start / (dbl) realz;

  for (i = 0, beta = - start;
       i <= realz;
       i++, beta += delta_beta)
    {
      for (j = 0, alfa = 0.0;
	   j < xy;
	   j++, alfa += delta_alfa)
	{
	  sz = cos (beta);
	  sx = cos (alfa) * sz;
	  sy = sin (alfa) * sz;
	  sz = sin (beta);
	  
	  *vptr++ = radius * (*nptr++ = sx);
	  *vptr++ = radius * (*nptr++ = sy);
	  *vptr++ = radius * (*nptr++ = sz);

	  *tptr++ = alfa / (2. * PI);
	  *tptr++ = beta / (2. * PI);
	}
      
      sz = cos (beta);
      sx = cos (alfa + delta_alfa) * sz;
      sy = sin (alfa + delta_alfa) * sz;
      sz = sin (beta);
      
      *vptr++ = radius * (*nptr++ = sx);
      *vptr++ = radius * (*nptr++ = sy);
      *vptr++ = radius * (*nptr++ = sz);
      *tptr++ = 1.0;
      *tptr++ = beta / (2. * PI);
    }

  for (j = 0; j < realz; j++)
    {
      STORE_INDEX (iptr, GL_QUAD_STRIP);
      STORE_INDEX (iptr, 2 * xy + 2);
      STORE_INDEX (iptr, z1 = xy * (j + 1));
      STORE_INDEX (iptr, z2 = xy * j);

      for (i = 1; i < xy; i++)
	{
	  STORE_INDEX (iptr, *(iptr - 2) + 1);
	  STORE_INDEX (iptr, *(iptr - 2) + 1);
	}
      STORE_INDEX (iptr, z1);
      STORE_INDEX (iptr, z2);
    }

  STORE_INDEX (iptr, 0);
  STORE_INDEX (iptr, 0);
}

void
ts_sphere_mark (dbl radius, dbl m_radius, int xy, int z)
{
  int		i;
  GLint		*iptr;
  GLdouble	*vptr, *nptr;
  GLdouble	top [3];
  dbl		sz,sx,sy;
  dbl		alfa, beta;
  dbl		delta_alfa;

  ts_free ();

  vptr = ts_vertices = g_new (GLdouble, 3 * (xy + 1));
  nptr = ts_normals = g_new (GLdouble, 3 * (xy + 1));
  iptr = ts_indices = g_new (GLint, xy + 2 + 2 + 2);

  delta_alfa = 2. * PI / (dbl) xy;
  beta = (PI / 2.0) -  PI / (dbl) z;

  top [X] = 0;
  top [Y] = 0;
  top [Z] = radius;

  for (i = 0, alfa = 0.0;
       i < xy;
       i++, alfa += delta_alfa)
    {
        sz = cos (beta);
	sx = cos (alfa) * sz;
	sy = sin (alfa) * sz;
	sz = sin (beta);

	*(vptr + X) = radius * (*nptr++ = sx);
	*(vptr + Y) = radius * (*nptr++ = sy);
	*(vptr + Z) = radius * (*nptr++ = sz);

	vec_sub (vptr, vptr, top);
	vec_norm (vptr, vptr);
	vec_mul_c (vptr, vptr, m_radius);
	vec_add (vptr, vptr, top);

	vptr += 3;
    }

  *vptr++ = radius * (*nptr++ = 0);
  *vptr++ = radius * (*nptr++ = 0);
  *vptr++ = radius * (*nptr++ = 1);

  STORE_INDEX (iptr, GL_TRIANGLE_FAN);
  STORE_INDEX (iptr, xy + 2);
  STORE_INDEX (iptr, xy);
  STORE_INDEX (iptr, xy - 1);

  for (i =0; i < xy; i++)
    STORE_INDEX (iptr, i);

  STORE_INDEX (iptr, 0);
  STORE_INDEX (iptr, 0);
}

