#include <glib.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "mathutils.h"
#include "pool.h"

#include "poolview.h"
#include "poolfeedgl.h"
#include "table.h"
#include "fyjik.h"

#include "properties.h"

static LID table_LID = -1;
static LID outline_LID = -1;
static LID outline_sides_LID = -1;

static LOD_array *pocket_borders_LIDs = NULL;
static int	  nof_pocket_borders = 0;

void
table_render_fake_hole (PoolViewGL *data, TableHole *hole, dbl z)
{
  dbl		 radius;

  glPushMatrix ();
  glTranslated (hole->p[X], hole->p[Y], z);
  glScaled (hole->radius, hole->radius, hole->radius);

  render_disk (data);

  glPopMatrix ();
}

static void
render_cushions (void)
{
  int   i, j, k;
  dbl   n [3];

  glFrontFace (GL_CW);

  glBegin( GL_QUADS );
  for( k = i = 0; i < nof_edges; i ++ )
    {
      planeNormal( table_sides + k, n );
      vec_neg( n );
      n [Z] = 0.5; /* CHEATER !!! :) */
      vec_norm (n,n);
      glNormal3dv( n );
      for( j = 0; j < counts [i]; j++, k++ )
        glVertex3dv( table_sides [k] );
    }

  glEnd();

  glFrontFace (GL_CCW);
}

static void
clear_table_LIDs (void)
{
  int i;

  if (table_LID != -1)
    { glDeleteLists (table_LID, 1); table_LID = -1; }

  if (outline_LID != -1)
    { glDeleteLists (outline_LID, 1); outline_LID = -1; }

  if (outline_sides_LID != -1)
    { glDeleteLists (outline_sides_LID, 1); outline_sides_LID = -1; }

  if (pocket_borders_LIDs == NULL)
    return;

  for (i = 0; i <  nof_pocket_borders; i++)
    LOD_array_GL_free (pocket_borders_LIDs [i]);

  g_free (pocket_borders_LIDs);
  pocket_borders_LIDs = NULL;
  nof_pocket_borders = 0;
}

#define HXHY(ANGLE)	nx = cos (ANGLE); ny = sin (ANGLE);	\
			hx = (nx*(hole->radius)+hole->p[X]);	\
			hy = (ny*(hole->radius)+hole->p[Y]);    \
			norm[X] = -nx; norm [Y] = -ny; norm [Z] = 0.5 /* CHEAT!! */ ; \
			vec_norm (norm, norm)

#define FEED		glNormal3dv (norm);				\
			if (funny)					\
			  {						\
			    glColor3dv (col2);				\
			    glVertex3d (hx,hy, hole_depth);		\
			    glColor3dv (col1);				\
			    glVertex3d (hx,hy, hole_height);		\
			  }						\
			else						\
			  {						\
			    glVertex3d (hx,hy, hole_depth);		\
			    glVertex3d (hx,hy, hole_height);		\
			  }

static void
render_part_of_pocket_border (PoolViewGL *view, TableHole *hole, dbl start, dbl end, 
			      dbl hole_depth, dbl hole_height,
			      dbl *ucolor, dbl *col1, dbl *col2)
{
  int slices;
  dbl delta, hx,hy, nx, ny, angle;
  dbl norm [3];
  int funny,i;

  funny = properties.funny_pockets;

  angle = end - start;

  slices = 8;
  for (i = view->lod; i; i--, slices = (slices * 3) / 2);
  slices = ceil (((dbl) slices) * fabs (angle) / (2. * PI));
  delta = angle / ((dbl) slices);

  if (! funny)
    glColor3dv (ucolor);

  glBegin (GL_QUAD_STRIP);
    HXHY (start);
    FEED;

    for (angle = start + delta; 
	 slices; 
	 slices--, angle += delta)
      {
	HXHY (angle);
	FEED;
      }
  glEnd ();
}
#undef HXHY
#undef FEED

static void
render_pocket_border (PoolViewGL *view, int hole_id)
{
  int	i, *iptr;


  if (pocket_borders_LIDs == NULL)
    {
      nof_pocket_borders = pool_table.nof_holes;
      pocket_borders_LIDs = g_new (LOD_array, nof_pocket_borders);

      for (i = 0; i < nof_pocket_borders; i++)
	LOD_array_init (pocket_borders_LIDs [i]);
    }

  iptr = &(pocket_borders_LIDs [hole_id][view->lod]);

  if (*iptr == -1)
    {
      int	slices;
      dbl	start, end, tmp;
      TableHole *hole;
      Skin	*skin;

      dbl	c1[3], c2[3], *c3;

      start = hole_cuts [2 * hole_id];
      end  = hole_cuts [2 * hole_id + 1];
      if (end < start) end += 2. * PI;

      hole = pool_table.holes + hole_id;
      skin = hole->skin;

      tmp = ((dbl) pool_table.table_height) / (dbl) (pool_table.table_height - POOL_TABLE_HOLE_DEPTH);
      vec_mul_c (c1, skin -> color2, tmp);
      vec_mul_c (c2, skin -> color1, (1.0 - tmp));
      vec_add (c1, c1, c2);
      c3 = c1;

      *iptr = glGenLists (1);
      glNewList (*iptr, GL_COMPILE_AND_EXECUTE);

        if (end - start > EPSILON)
	  render_part_of_pocket_border (view, hole, 
					start, end, 
					POOL_TABLE_HOLE_DEPTH, pool_table.table_height,
					skin -> color1, skin -> color1, skin -> color2);
	else
	  c3 = skin -> color1;

	render_part_of_pocket_border (view, hole, 
				      end, start + 2. * PI,
				      POOL_TABLE_HOLE_DEPTH, 0,
				      skin -> color1, c3, skin -> color2);
      glEndList ();
    }
  else
    glCallList (*iptr);
}


static void
render_pocket_borders (PoolViewGL *view)
{
  int	i;
  int   both_sides;

  both_sides =  ((! properties.render_table_outline) ||
	          pool_table.outer_vertices == NULL);

  if (both_sides)
    glDisable (GL_CULL_FACE);

  for (i = 0; i < pool_table.nof_holes; i++)
    render_pocket_border (view, i);

  if (both_sides)
    glEnable (GL_CULL_FACE);
}

static void
render_holes_pre_1_2 (PoolViewGL *data, dbl z)
{
  int i;

  glColorMask (GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

      /* it could seem useless to set color but some buggy opengl
	 implentaions ignore glColorMask - so we have at least
	 black holes ... :) */

  glColor3d (0,0,0);

  glPolygonOffset (-3, -3);
  glEnable (GL_POLYGON_OFFSET_FILL);
  if (data -> with_lights)
    glDisable (GL_LIGHTING);
      

  for (i = 0; i < pool_table.nof_holes; i++)
    table_render_fake_hole (data, pool_table.holes + i,z);

  if (data -> with_lights)
    glEnable (GL_LIGHTING);
  glDisable (GL_POLYGON_OFFSET_FILL);
  glColorMask (GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
}

static void
render_table_outline_sides (PoolViewGL *data)
{
  if (outline_sides_LID != -1)
    glCallList (outline_sides_LID);
  else
    {
      int		i;
      dbl		*ct, *cp;
      dbl		 vt [3], vp [3], norm [3], sub [3];

#define VEC_LOAD(V,DV2)		do { (V)[X] = (DV2)[X]; (V)[Y] = (DV2)[Y]; (V)[Z] = 0; } while (0)

      ct = pool_table.outer_vertices;
      cp = ct + 2 * (pool_table.nof_outer_vertices - 1);

      outline_sides_LID = glGenLists (1);
      glNewList (outline_sides_LID, GL_COMPILE_AND_EXECUTE);

      glColor3dv (pool_table.outer_skin->color1);
      glBegin (GL_QUAD_STRIP);

      VEC_LOAD (vt, ct);
      VEC_LOAD (vp, cp);

      vec_sub (sub, vt, vp);
      vec_cross (norm, sub, render_up_vector);
      vec_norm (norm, norm);

      glNormal3dv (norm);
      vp [Z] = pool_table.table_height; glVertex3dv (vp);
      vp [Z] = DEFAULT_TABLE_DEPTH; glVertex3dv (vp);

      for (i = 0; i < pool_table.nof_outer_vertices; i++)
	{
	  VEC_LOAD (vt, ct);
	  VEC_LOAD (vp, cp);
	    
	  vec_sub (sub, vt, vp);
	  vec_cross (norm, sub, render_up_vector);
	  vec_norm (norm, norm);

	  glNormal3dv (norm);

	  vt [Z] = pool_table.table_height; glVertex3dv (vt);
	  vt [Z] = DEFAULT_TABLE_DEPTH; glVertex3dv (vt);

	  cp = ct;
	  ct += 2;
	}

      glEnd ();
      glEndList ();
    }
#undef VEC_LOAD
}

static void
render_table_outline (PoolViewGL *data)
{
  if (outline_LID != -1)
    glCallList (outline_LID);
  else
    {
      GLUtesselator	*tess;
      dbl		*vertex, *vptr, *dptr;
      int		i;

      outline_LID = glGenLists (1);

      tess = render_tesselator_init ();
      vertex = vptr = g_new (dbl, 3 * (pool_table.nof_vertices + pool_table.nof_outer_vertices));

      glNewList (outline_LID, GL_COMPILE_AND_EXECUTE);

      glColor3dv(pool_table.outer_skin->color1);
      glNormal3dv (render_up_vector);

      gluTessBeginPolygon (tess, NULL);
       gluTessBeginContour (tess);
       for (i = 0, dptr = pool_table.outer_vertices; 
	    i < pool_table.nof_outer_vertices;
	    i++, dptr +=2, vptr += 3)
	 {
	   vptr [X] = dptr [X];
	   vptr [Y] = dptr [Y];
	   vptr [Z] = pool_table.table_height;

	   gluTessVertex (tess, vptr, vptr);
	 }
       gluTessEndContour (tess);
       gluTessBeginContour (tess);
       for (i = 0, dptr = (pool_table.vertices + 2 * (pool_table.nof_vertices - 1));
	    i < pool_table.nof_vertices;
	    i++, dptr -= 2, vptr += 3)
	 {
	   vptr [X] = dptr [X];
	   vptr [Y] = dptr [Y];
	   vptr [Z] = pool_table.table_height;

	   gluTessVertex (tess, vptr, vptr);
	 }
       gluTessEndContour (tess);
      gluTessEndPolygon (tess);

      render_tesselator_finish (tess);
      g_free (vertex);

      glEndList ();

    }
}

static void
render_table_ground (PoolViewGL *data)
{
  GLUtesselator *tess;

  dbl		*dptr;
  dbl		*vertex, *vptr;
  dbl		*holevert = NULL;
  TableHole	*hole;
  int		i;

  tess = render_tesselator_init ();
  vertex = g_new (dbl, 3 * pool_table.nof_vertices);

  glNormal3dv (render_up_vector);
  gluTessBeginPolygon (tess, NULL);
  gluTessBeginContour (tess);

  for (i = pool_table.nof_vertices, dptr = pool_table.vertices, vptr = vertex;
       i;
       i--, dptr += 2, vptr += 3)
    {
      vptr [X] = dptr [0];
      vptr [Y] = dptr [1];
      vptr [Z] = 0;
	    
      gluTessVertex (tess, vptr, vptr);
    }

  gluTessEndContour (tess);

  if (pool_table.nof_holes != 0 && properties.glu_tesselator_ok)
    {
      vptr = holevert = g_new (dbl, 4 * 3 * pool_table.nof_holes);

      for (i = 0, hole = pool_table.holes; 
	   i < pool_table.nof_holes; 
	   i++, hole++)
	{
	  gluTessBeginContour (tess);

	  vptr [X] = hole->p[X] + hole->radius;
	  vptr [Y] = hole->p[Y] + hole->radius;
	  vptr [Z] = 0.0;
	  gluTessVertex (tess, vptr, vptr);
	  vptr += 3;

	  vptr [X] = hole->p[X] + hole->radius;
	  vptr [Y] = hole->p[Y] - hole->radius;
	  vptr [Z] = 0;
	  gluTessVertex (tess, vptr, vptr);
	  vptr += 3;

	  vptr [X] = hole->p[X] - hole->radius;
	  vptr [Y] = hole->p[Y] - hole->radius;
	  vptr [Z] = 0;
	  gluTessVertex (tess, vptr, vptr);
	  vptr += 3;
		    
	  vptr [X] = hole->p[X] - hole->radius;
	  vptr [Y] = hole->p[Y] + hole->radius;
	  vptr [Z] = 0;
	  gluTessVertex (tess, vptr, vptr);
	  vptr += 3;

	  gluTessEndContour (tess);
	}
    }

  gluTessEndPolygon (tess);

  render_tesselator_finish (tess);
  g_free (vertex);

  if (holevert) 
    g_free (holevert);
}


void
table_render (PoolViewGL *data, int allow_fake_holes)
{
  int	draw_outlines;

  draw_outlines = properties.render_table_outline &&
    pool_table.outer_vertices;

  if (pool_table.update_view_info == 1)
    {
      clear_table_LIDs ();
      generate_table();

      table_LID = glGenLists (1);
      glNewList (table_LID, GL_COMPILE);

        glColor3dv (default_table_skin -> color1);
        render_cushions ();

	render_table_ground (data);
      glEndList (); 
      pool_table.update_view_info = 0;
    }

  render_pocket_borders (data);
  if (draw_outlines)
    render_table_outline_sides (data);

  if (allow_fake_holes && ! properties.glu_tesselator_ok)
    render_holes_pre_1_2 (data, 0);

  glCallList (table_LID);

  if (draw_outlines)
    {
      if (allow_fake_holes && ! properties.glu_tesselator_ok)
	render_holes_pre_1_2 (data, pool_table.table_height);
      render_table_outline (data);
    }
}
