#include <libguile.h>
#include <gtk/gtk.h>
#include <gnome.h>

#include "gtulpas.h"
#include "game.h"

#include "guileglue.h"
#include "variant.h"


#define VARIANT_EVENT_ITEMS_NOF		4

glue_enum_item variant_event_items_sbv [VARIANT_EVENT_ITEMS_NOF];
glue_enum_item variant_event_items_sbs [VARIANT_EVENT_ITEMS_NOF] = {
  { SCM_UNDEFINED, "yes/no", Variant_YesNo },
  { SCM_UNDEFINED, "one-of", Variant_OneOf },
  { SCM_UNDEFINED, "integer", Variant_Integer },
  { SCM_UNDEFINED, "group", Variant_Group }
};

glue_enum_info variant_enum_info = {
  variant_event_items_sbs,
  variant_event_items_sbv,
  VARIANT_EVENT_ITEMS_NOF
};

GList *
variant_scm2vlist (SCM list, SCM *error )
{
#define ERROR(S)	{ *error = (S); return NULL; }

  static int	need_init = TRUE;

  SCM		dlist, tmp;
  int		sid;

  GList	       *vlist = NULL;
  Variant      *var;

  if (need_init)
    { need_init = FALSE; glue_enum_info_init (&variant_enum_info); }


  while (list != SCM_EOL)
    {
      if (! SCM_CONSP (list))
	ERROR (list);

      dlist = SCM_CAR (list);
      if (! SCM_CONSP (dlist))
	ERROR (dlist);

      sid = glue_enum_scm2value (SCM_CAR (dlist), &variant_enum_info);
      if (sid == -1) ERROR (SCM_CAR (dlist));

      dlist = SCM_CDR (dlist);
      
      var = g_new (Variant, 1);
      vlist = g_list_prepend (vlist, var);

      var -> type = sid;
      var -> data = NULL;
      var -> name = NULL;

      if (sid != Variant_OneOf)
	{
	  if (! SCM_CONSP (dlist))
	    ERROR (dlist);

	  tmp = SCM_CAR (dlist);
	  dlist = SCM_CDR (dlist);

	  if (sid == Variant_Group && tmp == SCM_BOOL_F)
	    var -> name = NULL;
	  else
	    {
	      if (! (SCM_NIMP (tmp) && SCM_STRINGP (tmp)))
		ERROR (tmp);

	      var -> name = GLUE_STRDUP (tmp);
	    }
	}

      switch (sid)
	{
	case Variant_YesNo:
	  if (scm_ilength (dlist) != 1)
	    ERROR (dlist);

	  var -> value = (SCM_CAR (dlist) == SCM_BOOL_F) ? FALSE : TRUE;
	  break;

	case Variant_OneOf: 
	  if (scm_ilength (dlist) < 3)
	    ERROR (dlist);

	  tmp = SCM_CAR (dlist);
	  dlist = SCM_CDR (dlist);

	  if (! SCM_INUMP (tmp))
	    ERROR (tmp);

	  var -> value = SCM_INUM (tmp);

	  while (dlist != SCM_EOL)
	    {
	      if (SCM_NCONSP (dlist))
		ERROR (dlist);

	      tmp = SCM_CAR (dlist);
	      if (! (SCM_NIMP (tmp) && SCM_STRINGP (tmp)))
		ERROR (tmp);

	      var -> data = g_list_prepend (var->data, GLUE_STRDUP (tmp));
	      dlist = SCM_CDR (dlist);
	    }

	  var -> data = g_list_reverse (var->data);
	  break;

	case Variant_Integer:
	  if (scm_ilength (dlist) != 3)
	    ERROR (dlist);

	  if (! SCM_INUMP (SCM_CAR (dlist)))
	    ERROR (SCM_CAR (dlist));

	  var -> value = SCM_INUM (SCM_CAR (dlist));
	  dlist = SCM_CDR (dlist);

	  if (! SCM_INUMP (SCM_CAR (dlist)))
	    ERROR (SCM_CAR (dlist));

	  var -> lo = SCM_INUM (SCM_CAR (dlist));
	  dlist = SCM_CDR (dlist);

	  if (! SCM_INUMP (SCM_CAR (dlist)))
	    ERROR (SCM_CAR (dlist));

	  var -> up = SCM_INUM (SCM_CAR (dlist));

	  break;

	case Variant_Group:
	  tmp = SCM_UNSPECIFIED;
	  var -> data = variant_scm2vlist (dlist, &tmp);
	  
	  if (var -> data == NULL && tmp != SCM_UNSPECIFIED)
	    ERROR (tmp);

	  break;

	default:
	  g_assert_not_reached ();
	}

      list = SCM_CDR (list);
    }

  return g_list_reverse (vlist);
#undef ERROR
}

SCM*
v2s_internal (GList *vlist, SCM *ptr)
{
  SCM first,val;
  Variant *var;

  while (vlist != NULL)
    {
      var = (Variant *) (vlist -> data);
      switch (var -> type)
	{
	case Variant_YesNo:
	  val = (var->value) ? SCM_BOOL_T : SCM_BOOL_F;
	  break;

	case Variant_OneOf:
	case Variant_Integer:
	  val = SCM_MAKINUM (var->value);
	  break;

	case Variant_Group:
	  ptr = v2s_internal (var->data, ptr);
	  vlist = vlist -> next;
	  continue; /* no break !!! */
	default:
	  g_assert_not_reached ();
	}

      *ptr = scm_cons (val, SCM_EOL);
      ptr = SCM_CDRLOC (*ptr);

      vlist = vlist -> next;
    }

  return ptr;
}


SCM
variant_vlistval2scm (GList *vlist)
{
  SCM		ptr = SCM_EOL;

  v2s_internal (vlist,&ptr);
  return ptr;
}

/*
 *
 *		GUI
 */

static void
v2dialog_check_cb (GtkWidget *widget, Variant *var)
{
  var -> tmp_value = GTK_TOGGLE_BUTTON (widget) -> active;
}

static void
v2dialog_radio_cb (GtkWidget *widget, Variant *var)
{
  if (GTK_TOGGLE_BUTTON (widget) -> active)
    var -> tmp_value = (int) gtk_object_get_user_data (GTK_OBJECT (widget));
}

static void
v2dialog_adjust_cb (GtkAdjustment *adj, Variant *var)
{
  var -> tmp_value = adj -> value;
}

static GtkWidget*
v2dialog_internal (GList *vlist)
{
  Variant	*var;
  GtkWidget	*vbox;
  GtkWidget	*widget;
  GtkAdjustment *adj;

  if (! vlist) 
    return NULL;

  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 5);

  while (vlist)
    {
      var = (Variant *) (vlist -> data);
      var -> tmp_value = var -> value;

      switch (var -> type)
	{
	case Variant_YesNo:
	  widget = gtk_check_button_new_with_label (var -> name);
	  gtk_signal_connect (GTK_OBJECT (widget), "toggled",
			      (GtkSignalFunc) v2dialog_check_cb, var);
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget),
					var -> value ? TRUE : FALSE);
	  break;
	case Variant_OneOf:
	  {
	    GList	*ptr;
	    char	*vname;
	    int		id;

	    widget = NULL;
	    for (ptr = var -> data, id = 0; ptr != NULL; ptr = ptr -> next, id++)
	      {
		vname = (char *) (ptr -> data);
		widget = 
		  gtk_radio_button_new_with_label_from_widget ((GtkRadioButton *) widget, vname);

		if (var->value == id)
		  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);

		gtk_object_set_user_data (GTK_OBJECT (widget), (gpointer) id);
		gtk_signal_connect (GTK_OBJECT (widget), "toggled",
				    (GtkSignalFunc) v2dialog_radio_cb, var);

		gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 0);
	      }
	    widget = NULL;
	    break;
	  }
	case Variant_Integer:
	  {
	    GtkWidget   *label;
	    GtkWidget   *spin;

	    widget = gtk_hbox_new (FALSE, 10);

	    label = gtk_label_new (var -> name);
	    gtk_box_pack_start (GTK_BOX (widget), label, FALSE, FALSE, 0);
	    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

	    adj = (GtkAdjustment *) gtk_adjustment_new (var->value, var->lo, var->up, 1, 1, 1);
	    spin = gtk_spin_button_new (adj, 1, 0);
	    gtk_box_pack_end (GTK_BOX (widget), spin, FALSE, FALSE, 0);

	    gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
				(GtkSignalFunc) v2dialog_adjust_cb, var);
	    }
	  break;
	case Variant_Group:
	  {
	    GtkWidget	*iw;

	    widget = gtk_frame_new (var->name);
	    gtk_frame_set_shadow_type (GTK_FRAME (widget), GTK_SHADOW_ETCHED_IN);
	    gtk_container_set_border_width (GTK_CONTAINER (widget), 5);

	    iw = v2dialog_internal (var -> data);
	    gtk_container_add (GTK_CONTAINER (widget), iw);
	    break;
	  }

	default:
	  g_assert_not_reached ();
	}

      if (widget)
	gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE,0);
      vlist  = vlist -> next;
    }
  return vbox;
}

static void
variant_list_apply (GList *list)
{
  Variant	*var;

  while (list)
    {
      var = (Variant *) (list -> data);

      if (var -> type == Variant_Group)
	variant_list_apply (var -> data);
      else
	var -> value = var -> tmp_value;

      list = list -> next;
    }
}

static void
v2dialog_should_restart (Game *game)
{
  static GtkWidget *dialog = NULL;

  if (! dialog)
    dialog = gnome_message_box_new (_("Should I restart current game with new settings?"),
				    GNOME_MESSAGE_BOX_WARNING,
				    GNOME_STOCK_BUTTON_YES,
				    GNOME_STOCK_BUTTON_NO,
				    NULL);

  gnome_dialog_close_hides (GNOME_DIALOG (dialog), TRUE);
  gnome_dialog_set_default (GNOME_DIALOG (dialog), 1);

  switch (gnome_dialog_run_and_close (GNOME_DIALOG (dialog)))
    {
    case 0:
#ifdef DEBUG
      g_message ("Restart?: Yes");
#endif
      gtulpas_select_game (gtulpas, game);
      game_add_event_type (gtulpas->game, NEW_GAME);
      break;

    default:
#ifdef DEBUG
      g_message ("Restart?: No");
#endif
      break;
    }
}

static GtkWidget *v2dialog_dialog = NULL;
static Game *v2dialog_game = NULL;
static GtkWidget *v2dialog_vbox = NULL;
static GtkWidget *v2dialog_aarea = NULL;

static void
v2dialog_destroy_cb (void)
{
  v2dialog_dialog = NULL;
  v2dialog_game = NULL;  
  v2dialog_vbox = NULL;  
  v2dialog_aarea = NULL; 
}

static void
v2dialog_clicked_cb (GnomeDialog *dialog, int bid)
{
  Game	*tmp;

  tmp = v2dialog_game;
  switch (bid)
    {
    case 0:
#ifdef DEBUG
      g_message ("var:OK");
#endif
      gtk_widget_destroy (v2dialog_dialog);
      /* no break */
    case 1:
#ifdef DEBUG
      g_message ("var:Apply");
#endif
      variant_list_apply (tmp -> variant_list);
      v2dialog_should_restart (tmp);
      break;
    case 2:
#ifdef DEBUG
      g_message ("var:Cancel");
#endif
      gtk_widget_destroy (v2dialog_dialog);
      break;
    default:
#ifdef DEBUG
      g_message ("var:Huh? %d", bid);
#endif
    }
}

void
variant_dialog_new (Game *game, int change_only)
{
  GtkWidget	*bbox;
  GtkWidget	*button;
  char		*ptr;

  if (game -> variant_list == NULL)
    game = NULL;

  if (game == NULL)
    {
      if (v2dialog_dialog)
	gtk_widget_destroy (v2dialog_dialog);

      v2dialog_dialog = NULL;
      v2dialog_game = NULL;
      v2dialog_vbox = NULL;

      return;
    }

  if (v2dialog_dialog == NULL && change_only)
    return;

  if (game != v2dialog_game && v2dialog_vbox)
    {
      gtk_widget_destroy (v2dialog_vbox);
      v2dialog_vbox = NULL;
    }

  if (v2dialog_dialog == NULL)
    {
      v2dialog_dialog = gnome_dialog_new ("",
					  GNOME_STOCK_BUTTON_OK,
					  GNOME_STOCK_BUTTON_APPLY,
					  GNOME_STOCK_BUTTON_CANCEL,
					  NULL);
      gnome_dialog_close_hides (GNOME_DIALOG (v2dialog_dialog), FALSE);
      gnome_dialog_set_default (GNOME_DIALOG (v2dialog_dialog), 0);

      gtk_signal_connect (GTK_OBJECT (v2dialog_dialog), "destroy",
			  (GtkSignalFunc) v2dialog_destroy_cb, NULL);

      gtk_signal_connect (GTK_OBJECT (v2dialog_dialog), "clicked",
			  (GtkSignalFunc) v2dialog_clicked_cb, NULL);

    }

  ptr = g_strdup_printf (_("Game options - %s"), game -> name);
  gtk_window_set_title (GTK_WINDOW (v2dialog_dialog), ptr);
  g_free (ptr);

  if (! v2dialog_vbox)
    {
      v2dialog_vbox = v2dialog_internal (game -> variant_list);
      gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (v2dialog_dialog) -> vbox), v2dialog_vbox, 
			  FALSE, FALSE, 0);
    }

  v2dialog_game = game;
  gtk_widget_show_all (v2dialog_dialog);
}
